#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

/* 
 * A utility for processing texinfo files and generating math images
 * in HTML output.
 * The program reads standard input and substitutes the mathematical
 * text inside @m{} or @mm{} macros with texinfo macros for mathematical
 * expressions in TeX mode and images in HTML mode.
 * It generates unique IDs for all mathematical expressions.
 *
 * Usage:
 * procmath < input > output prefix list
 *
 * where 'list' is the generated script that generates the mathematical
 * images to be included in the HTML code;
 * 'prefix' is a unique prefix for the image filenames (the rest is
 * the image number within the given file).
 *
 * The 'list' is appended to enable processing of several files. You
 * may want to delete its contents before start.
 */

struct pair
{
	char a;
	char * b;
};

/* Characters that texinfo cannot process directly.
 * Add new pairs as you wish. */
struct pair transl_pairs[] =
{
/*	{ '\\', "@" },
	{ '^', "@sup " },
	{ '_', "@sub " },*/
	{ 'd', "d" },	//dummy
};
int trans_pair_cnt = sizeof(transl_pairs)/sizeof(struct pair);

int id_counter;

enum mode
{
	M_NORMAL,
	M_MATH,
	M_DISPLAY_MATH,
	M_AT,
	M_M,
	M_MM,
} mode;

int backslash;
int line_counter = 1;

char * MATH_START = "@c\n@iftex\n@tex\n$";
char * MMATH_START = "@c\n@iftex\n@tex\n$$";
char * MATH_END = "$%\n@end tex\n@end iftex\n@ifhtml\n@image{math/%s%d,,,math%s%d,png}@c\n@end ifhtml\n@nothing{}";
char * MMATH_END = "$$\n@end tex\n@end iftex\n@ifhtml\n@center @image{math/%s%d,,,math%s%d,png}@c\n@end ifhtml\n@nothing{}";
char * LIST_MATH_START = "./tex2png '$ ";
char * LIST_MMATH_START = "./tex2png '$$ ";
char * LIST_MATH_END = " $' math/%s%d.png\n";
char * LIST_MMATH_END = " $$' math/%s%d.png\n";

#define OUT(_c) \
{ \
	printf( "%c", _c ); \
	backslash = (_c == '\\'); \
	line_counter += (_c == '\n'); \
}

char buf[5];
int pos;

int depth;

void flush_buf(void)
{
	int i = 0;
	for ( ; i < pos; i++ )
		OUT(buf[i]);
	pos = 0;
}

int main( int argc, char ** argv )
{
	FILE * list;
	char c;
	int translated, i;

	if (argc < 3)
	{
		fprintf( stderr, "procmath: error: not enough arguments.\nUsage: procmath < input > output prefix list\n" );
		return 1;
	}

	list = fopen(argv[2], "a");
	if (!list)
	{
		fprintf( stderr, "procmath: error: could not open file %s for writing: %s\n", argv[2], strerror(errno) );
		return 1;
	}

	while ( scanf("%c", &c) == 1 )
	{
		switch (mode)
		{
		case M_NORMAL:
			if (c == '@')
			{ mode = M_AT; buf[pos++] = '@'; }
			else
				OUT(c);
			break;
		case M_AT:
			if (c == 'm')
			{ mode = M_M; buf[pos++] = 'm'; }
			else
			{ flush_buf(); OUT(c); mode = M_NORMAL; }
			break;
		case M_M:
			switch (c)
			{
			case 'm':
				mode = M_MM;
				buf[pos++] = 'm';
				break;
			case '{':
				mode = M_MATH; //start of math mode
				depth = 1;
				pos = 0;
				printf( MATH_START );
				fprintf( list, LIST_MATH_START );
				break;
			default:
				mode = M_NORMAL;
				flush_buf(); OUT(c);
			}
			break;
		case M_MM:
			if (c == '{')
			{
				mode = M_DISPLAY_MATH; // start of diplay math mode
				depth = 1;
				pos = 0;
				printf( MMATH_START );
				fprintf( list, LIST_MMATH_START );
			} else
			{ flush_buf(); OUT(c); mode = M_NORMAL; }
			break;
		case M_MATH:
		case M_DISPLAY_MATH:;
			translated = 0;
			if (c == '}' && !backslash)
				depth--;
			if (c == '{' && !backslash)
				depth++;
			if (depth == 0)
			{
				if (mode == M_MATH)
				{
					// end of math mode
					printf(MATH_END, argv[1], id_counter, argv[1], id_counter);
					fprintf( list, LIST_MATH_END, argv[1], id_counter );
				} else {
					// end of display math mode
					printf( MMATH_END, argv[1], id_counter, argv[1], id_counter );
					fprintf( list, LIST_MMATH_END, argv[1], id_counter );
				}
				mode = M_NORMAL;
				id_counter++;
			} else {
				for (i = 0; i < trans_pair_cnt; i++)
					if ( c == transl_pairs[i].a )
					{
						printf("%s", transl_pairs[i].b);
						fprintf( list, "%s", transl_pairs[i].b);
						translated = 1;
					}
				if (!translated)
				{
					OUT(c);
					fprintf(list, "%c", c);
				}
			}
			backslash = (c == '\\');
			break;
		}
	}

	fclose(list);
	if ( depth != 0 )
	{
		fprintf(stderr, "procmath: error: not enough '}' at end of file :-(\n");
		return 1;
	}

	return 0;
}
