/*
 * FONTLIB - Conversions among font formats PFA and PFB, TT and T42
 *
 * Copyright (c) 2004-2005,2007    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG

#include "lib/lib.h"
#include "font/internal.h"
#include <time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <alloca.h>

#define HEX(b) ((b)<10 ? (b)+'0' : (b)-10+'a')
#define OUT(...) fprintf(out, __VA_ARGS__)

int font_pfb_to_pfa(int fnt_id, FILE *out)
{
	byte buf[64*1024];
	size_t s;
	u32 len;
	byte hdr[6];
	uns i;
	const char *name;
	FILE *in;

	name = font_filename(fnt_id);
	if (!name)
		return -1;

	in = fopen(name, "r");
	if (!in)
		return -1;

	s = fread(hdr, 1, 6, in);
	if (s>0 && hdr[0] != 0x80) {
		DBG("Hmm... looks like PFA.\n");
		/* Let's hope it is PFA file, just rewrite it */
		if (fwrite(hdr, 1, s, out) != s)
			goto write_err;
		while ((s = fread(buf, 1, sizeof(buf), in))>0)
			if (s != fwrite(buf, 1, s, out))
				goto write_err;
	} else
	for (;;) { /* We have PFB file, so we must recode it into PFA */

		/* The 'EOF' segment, header could be shorter */
		if (s >=2 && hdr[0] == 0x80 && hdr[1] == 3)
			break;

		if (s != 6 || hdr[0] != 0x80 || !(hdr[1] & 3)) {
			err("Invalid PFB header at position %ld.\n", ftell(in));
			goto end;
		}
		len = hdr[2] | (hdr[3] << 8) | (hdr[4] << 16) | (hdr[5] << 24);

		/* ASCII segment */
		if (hdr[1] == 1) {
			while (len && (s = fread(buf, 1, MIN(sizeof(buf), len), in))>0) {
				/* We need to recode the CR characters into LF */
				for (i=0; i<s; i++)
					if (buf[i] == 0x0d)
						buf[i] = 0x0a;

				if (fwrite(buf, 1, s, out) != s)
					goto write_err;
				len -= s;
			}
		}

		/* Binary segment */
		if (hdr[1] == 2) {
			char hexbuf[65];
			uns l;

			while (len && (s = fread(buf, 1, MIN(sizeof(buf), len), in))>0) {

				/* Now we encode the whole buffer into hexadecimal */
				l = 0;
				while (l < s) {
					for (i=0; i<MIN(s-l, 32); i++) {
						hexbuf[2*i] = HEX(buf[i+l] >> 4);
						hexbuf[2*i+1] = HEX(buf[i+l] & 15);
					}
					l += i;
					hexbuf[2*i] = '\n';

					if (fwrite(hexbuf, 1, 2*i+1, out) != 2*i+1)
						goto write_err;
				}
				len -= s;
			}
		}
		if (len) {
			err("Premature EOF of PFB file (or invalid header).\n");
			goto end;
		}
		s = fread(hdr, 1, 6, in);
	}
	fputc('\n', out);
	fclose(in);
	return 0;
write_err:
	err("Error writing to PS file.\n");
end:
	fclose(in);
	return -1;
}

static u32 compute_chksum(u32 *buf, uns bytelen)
{
	u32 r = 0;
	u32 n = (bytelen + 3) / 4;

	while (n-- > 0) {
		r += (*buf >> 24) + ((*buf >> 8) & 0xff00) + ((*buf & 0xff00) << 8) + (*buf << 24);
		buf++;
	}
	return r;
}

static void write_table(FILE *out, byte *buf, uns len, uns align_bytes)
{
	byte charbuf[100];
	uns i, l;

	OUT("<");
	l = 0;
	while (l < len) {
		for (i=0; i<MIN(len-l, 32); i++) {
			charbuf[2*i] = HEX(buf[i+l] >> 4);
			charbuf[2*i+1] = HEX(buf[i+l] & 15);
		}
		l += i;

		OUT("\n");
		fwrite(charbuf, 1, 2*i, out);
	}
	OUT("\n");

	while (align_bytes--)
		OUT("00");
	/* Backwards compatibility padding */
	OUT("00>\n");
}

static void write_glyftable(FILE *out, byte *glyf, uns glyf_len, byte *loca, uns loca_len, int loca_long)
{
	uns glyphs;

	if (loca_long)
		glyphs = loca_len / 4 - 1;
	else
		glyphs = loca_len / 2 - 1;

	DBG("using %s version of loca table\n", (loca_long ? "long" : "short"));
	DBG("loca table has %d entries\n", glyphs);

	u32 off, loff;
	if (loca_long)
		loff = (loca[0] << 24) + (loca[1] << 16) + (loca[2] << 8) + loca[3];
	else
		loff = ((loca[0] << 8) + loca[1]) << 1;

	uns written = 0;
	for (uns i=1; i <= glyphs; i++) {
		if (loca_long)
			off = (loca[4*i] << 24) + (loca[4*i+1] << 16) + (loca[4*i+2] << 8) + loca[4*i+3];
		else
			off = ((loca[2*i] << 8) + loca[2*i+1]) << 1;

		DBG("glyph %d offset is %d, len is %d\n", i-1, loff, off-loff);
		OUT("%% glyph %d, len %d\n", i-1, off-loff);

		written += off-loff;
		if (off-loff) {		// omit empty glyphs
			if (i == glyphs) {	// treat last glyphs specially because of padding
				/*
				 * Many TTF fonts have broken loca table and
				 * are missing few bytes now to finish the glyf
				 * table.  Here is the fix.
				 */
				uns n = ((glyf_len+3) & ~3) - written;
				write_table(out, glyf+loff, off-loff, n);
			}
			else
				write_table(out, glyf+loff, off-loff, ((off-loff+1) & ~1) - (off-loff));
		}
		loff = off;
	}
}

static void write_sfnts(FILE *out, FT_Face face)
{
	uns i, j;
	FT_ULong len;
	u32 tables[11], table_size[11];
	byte *table_buf[11];
	uns table_cnt = 0;

	/* We want to load these tables, if present */
	tables[0] = TTAG_cvt;
	tables[1] = TTAG_fpgm;
	tables[2] = TTAG_glyf;
	tables[3] = TTAG_head;
	tables[4] = TTAG_hhea;
	tables[5] = TTAG_hmtx;
	tables[6] = TTAG_loca;
	tables[7] = TTAG_maxp;
	tables[8] = TTAG_prep;
	tables[9] = TTAG_vhea;
	tables[10] = TTAG_vmtx;

	/* Scan entire font file and load suitable tables */
	for (uns t=0; t < 11; t++) {
		len = 0;
		if (FT_Load_Sfnt_Table(face, tables[t], 0, NULL, &len)<0 || !len) {
			DBG("Omitting table number %d, not found or zero size\n", t);
			table_size[t] = 0;
			continue;
		}
		table_cnt++;
		table_size[t] = len;
		DBG("Table %d needs %d bytes, allocating...\n", t, (uns)len);
		table_buf[t] = xmalloc(len);
		FT_Load_Sfnt_Table(face, tables[t], 0, table_buf[t], &len);
	}

	/* Write the font directory */

	/* The offset subtable */
	OUT("<00010000");			// scaler type, M$ WIN
	OUT("000%c", HEX(table_cnt));		// number of tables

	i = 1;
	j = 0;
	while (i <= table_cnt) {
		i <<= 1;
		j++;
	}
	i >>= 1;
	j--;
	OUT("00%c0", HEX(i));			// searchRange (maximum power of 2 <= numTables)*16
	OUT("000%c", HEX(j));			// entrySelector log2(maximum power of 2 <= numTables)
	OUT("00%c0\n", HEX(table_cnt - i));	// rangeShift numTables*16-searchRange

	/* Checksum of the beginning */
	u32 total_chksum = 0x10000 + (table_cnt << 16) + (i << 4) + (j << 16) + ((table_cnt-i) << 4);

	/* Zero the chksumAdj field of head table */
	table_buf[3][8] = table_buf[3][9] = table_buf[3][10] = table_buf[3][11] = 0;

	/* The table directory */
	u32 offset = 12 + 16*table_cnt;	// for font directory
	for (uns t=0; t < 11; t++)
		if (table_size[t]) {
			OUT("%08x", tables[t]);			// 4-byte identifier
			u32 chksum = compute_chksum((u32 *) table_buf[t], table_size[t]);
			OUT("%08x", chksum);			// checksum
			OUT("%08x", offset);			// table position in sfnts
			OUT("%08x\n", table_size[t]);
			total_chksum += tables[t] + chksum + offset + table_size[t];

			/* Each table must be aligned to 4 byte boundary */
			offset += (table_size[t]+3) & ~3;
		}
	OUT(">\n");

	/* Store the chksumAdj into head table */
	total_chksum = 0xb1b0afba - total_chksum;
	table_buf[3][8] = total_chksum >> 24;
	table_buf[3][9] = (total_chksum >> 16) & 0xff;
	table_buf[3][10] = (total_chksum >> 8) & 0xff;
	table_buf[3][11] = total_chksum & 0xff;

	for (uns t=0; t < 11; t++) {
		if (!table_size[t])
			continue;
		OUT("%% table %d, len %d\n", t, table_size[t]);
		if (t == 2) {	// glyf table
			/* Detect whether the loca table is of long format (offset 51, head table) */
			write_glyftable(out, table_buf[t], table_size[t], table_buf[6], table_size[6], table_buf[3][51]);
		}
		else
			write_table(out, table_buf[t], table_size[t], ((table_size[t]+3) & ~3) - table_size[t]);
		xfree(table_buf[t]);
	}
}

static void write_string(FILE *out, byte *buf, uns len)
{
	uns i;
	int p = 0;

	OUT("(");
	for (i=0; i < len; i++) {
		if (buf[i] == '(')
			p++;
		else if (buf[i] == ')')
			if (--p < 0)
				break;
	}

	if (p == 0)
		fwrite(buf, 1, len, out);
	else {
		for (i=0; i < len; i++) {
			if (buf[i] == '(' || buf[i] == ')')
				putc('\\', out);
			putc(buf[i], out);
		}
	}
	OUT(")");
}

static void chew_nametable(FILE *out, FT_Face face)
{
	uns i;
	uns sfnt_name_cnt = FT_Get_Sfnt_Name_Count(face);
	int copy_set = 0, family_set = 0, full_set = 0, version_set = 0, subfam_set = 0;

	FT_SfntName sfnt_name;
	DBG("Parsing name table\n");
	for (i = 0; i < sfnt_name_cnt; i++) {
		if (FT_Get_Sfnt_Name(face, i, &sfnt_name))
			continue;
		DBG("Got platform=%d encoding=%d language=%d name=%d\n", sfnt_name.platform_id, sfnt_name.encoding_id,
				sfnt_name.language_id, sfnt_name.name_id);
		if (sfnt_name.platform_id == 0 || sfnt_name.encoding_id != 0) {
			DBG("rejecting\n");
			continue;
		}

		switch (sfnt_name.name_id) {
		case 0:		// copyright
			if (copy_set)
				break;
			copy_set = 1;
			OUT("   /notice ");
			write_string(out, sfnt_name.string, sfnt_name.string_len);
			OUT(" def\n");
			break;
		case 1:		// font family
			if (family_set)
				break;
			family_set = 1;
			OUT("   /FamilyName ");
			write_string(out, sfnt_name.string, sfnt_name.string_len);
			OUT(" def\n");
			break;
		case 2:		// font subfamily
			if (subfam_set)
				break;
			subfam_set = 1;
			OUT("   /Weight ");
			write_string(out, sfnt_name.string, sfnt_name.string_len);
			OUT(" def\n");
			break;
		case 4:		// full name
			if (full_set)
				break;
			full_set = 1;
			OUT("   /FullName ");
			write_string(out, sfnt_name.string, sfnt_name.string_len);
			OUT(" def\n");
			break;
		case 5:		// version
			if (version_set)
				break;
			version_set = 1;
			OUT("   /version ");
			write_string(out, sfnt_name.string, sfnt_name.string_len);
			OUT(" def\n");
			break;
		default:
			break;
		}
	}
}


int font_tt_to_type42(int fontid, const char *fontname, FILE *out)
{
	FT_Face face;
	uns i;

	if (FTC_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)fontid, &face))
		return -1;

	if (!fontname)
		fontname = FT_Get_Postscript_Name(face);

	TT_Header *tthdr = FT_Get_Sfnt_Table(face, ft_sfnt_head);
	TT_Postscript *ttps_tbl = FT_Get_Sfnt_Table(face, ft_sfnt_post);

	if (tthdr)
		OUT("%%!PS-TrueTypeFont-%d.%d-%d.%d\n", (uns)(tthdr->Table_Version >> 16), (int)(tthdr->Table_Version & 0xffff),
				(uns)(tthdr->Font_Revision >> 16), (int)(tthdr->Font_Revision & 0xffff));
	else
		OUT("%%!PS-TrueTypeFont-1-1\n");

	OUT("%%%%Title: %s\n", fontname);
	if (ttps_tbl)
		OUT("%%%%VMusage: %d %d\n", (uns)ttps_tbl->minMemType42, (uns)ttps_tbl->maxMemType42);
	OUT("%%%%Creator: VRR  (http://vrr.ucw.cz/)\n");
	time_t tm = time(NULL);
	OUT("%%%%CreationDate: %s\n", ctime(&tm));
	OUT("15 dict begin\n");
	OUT("/FontType 42 def\n");
	OUT("/FontMatrix [1 0 0 1 0 0] def\n");
	OUT("/FontName /%s def\n", fontname);

#if 0
	/* Build our own encoding */
	OUT("/Encoding 256 array\n0 1 255 { 1 index exch /.notdef put } for\n");
	load_charmap(face, FONT_CHARMAP_NONE);
	for (i=0; i<256; i++) {
		char name_buf[100];
		int gi;

		gi = FT_Get_Char_Index(face, i);

		if (FT_Get_Glyph_Name(face, i, name_buf, sizeof(name_buf)))
			OUT("   dup %d /.notdef put\n", i);
		else {
			for (uns j=0; j < strlen(name_buf); j++)
				if (!isgraph(name_buf[j]))
					name_buf[j] = '_';

			OUT("   dup %d /%s put\n", i, name_buf);
		}
	}
	OUT("readonly def\n");
#else
	/* Use standard PS encoding */
	OUT("/Encoding StandardEncoding def\n");
#endif

#if 0
	/* Use Freetype to compute fontbbox */
	OUT("/FontBBox [%d %d %d %d] def\n", (int)face->bbox.xMin, (int)face->bbox.yMin, (int)face->bbox.xMax, (int)face->bbox.yMax);
#else
	/* Get fontbbox from head table */
	if (tthdr)
		OUT("/FontBBox [%d %d %d %d] def\n", tthdr->xMin, tthdr->yMin, tthdr->xMax, tthdr->yMax);
	else
		OUT("/FontBBox [%d %d %d %d] def\n", (int)face->bbox.xMin, (int)face->bbox.yMin, (int)face->bbox.xMax, (int)face->bbox.yMax);
#endif

	/* Omitted UniqueID and XUID */

	/* filled glyphs */
	OUT("/PaintType 0 def\n");
	OUT("/CIDMap 0 def\n");

	/* Omitted StrokeWidth, Metrics, Metrics2, CDevProc */

	/* Build the list of glyphnames using freetype */
	OUT("/CharStrings %d dict dup begin\n", (int)face->num_glyphs);
	for (i=0; i < (uns)face->num_glyphs; i++) {
		char name_buf[50];
		if (FT_Get_Glyph_Name(face, i, name_buf, sizeof(name_buf)))
			OUT("   /.notdef %d def\n %% !!!", i);
		else {
			for (uns j=0; j < strlen(name_buf); j++)
				if (!isgraph(name_buf[j]))
					name_buf[j] = '_';

			OUT("   /%s %d def\n", name_buf, i);
		}
	}
	OUT("end readonly def\n");

	/* Build the FontInfo dictionary */
	OUT("/FontInfo 10 dict dup begin\n");
	if (ttps_tbl) {
		OUT("   /ItalicAngle %f def\n", ttps_tbl->italicAngle / (real)0x10000);
		OUT("   /isFixedPitch %s def\n", (ttps_tbl->isFixedPitch ? "true" : "false"));
		OUT("   /UnderlinePosition %d def\n", ttps_tbl->underlinePosition);
		OUT("   /UnderlineThickness %d def\n", ttps_tbl->underlineThickness);
	}
	/* Parse the nametable for the rest */
	chew_nametable(out, face);
	OUT("end readonly def\n");

	/* Omitted GDBytes, GlyphDirectory, MetricsCount */

	OUT("/sfnts [\n");
	write_sfnts(out, face);
	OUT("] def\n");

	OUT("FontName currentdict end definefont pop\n");

	return 0;
}

#undef OUT

int font_pfb_sections(byte *buf, uns size, uns *length, uns *clear_text, uns *encrypted, uns *fixed)
{
	uns total, txt, enc, fix;
	uns len;
	byte *obuf;

	if (buf[0] != 0x80)
		return -1;

	obuf = buf;
	total = txt = enc = fix = 0;
	for (;;) {
		/*
		 * A typical Type1 font program, encoded in binary (a.k.a. PFB)
		 * consists of three sections, ASCII, binary, and ASCII again.
		 * However, Type1 reference doesn't forbid nonstandard font programs,
		 * with nonstandard sections and nonstandard section order.
		 * Moreover, Downloadable font reference is very unclear,
		 * whether there cannot be multiple data segments. That means
		 * if the ASCII section or binary sections is divided into
		 * multiple consecutive ASCII or binary segments.
		 *
		 * Let's hope nobody produces such a font and assume the form
		 *
		 *   ASCII header - size: clear_text
		 *   Binary body - size: encrypted
		 *   ASCII trailer - size: fixed
		 */

		/* EOF segment, header could be shorter */
		if (size >= 2 && buf[0] == 0x80 && buf[1] == 3) {
			DBG("EOF at %d\n", buf - obuf);
			break;
		}
		if (size < 6 || buf[0] != 0x80 || !(buf[1] & 3)) {
			DBG("Invalid PFB header at offset %d\n", buf - obuf);
			return -1;
		}
		len = buf[2] | (buf[3] << 8) | (buf[4] << 16) | (buf[5] << 24);
		if (size < len + 6) {
			DBG("Segment at %d too long (want %d, have %d)\n", buf - obuf, len, size - 6);
			return -1;
		}
		total += len;

		/* ASCII segment */
		if (buf[1] == 1) {
			DBG("ASCII segment length %d\n", len);
			if (!txt)
				txt = len;
			else
				fix = len;
		}
		/* Binary segment */
		else if (buf[1] == 2) {
			enc += len;
			DBG("Binary segment length %d, total %d binary\n", len, enc);
		}
		else {
			DBG("Invalid segment ID (%d)\n", buf[1]);
			return -1;
		}

		buf += len + 6;
		size -= len + 6;
	}

	if (length)
		*length = total;
	if (clear_text)
		*clear_text = txt;
	if (encrypted)
		*encrypted = enc;
	if (fixed)
		*fixed = fix;

	return 0;
}

int font_pfb_data_store(byte *buf, uns size, byte *outbuf)
{
	byte *obuf;
	int total = 0;
	uns len;

	obuf = buf;
	for (;;) {
		/*
		 * We just follow the segment structure and dump all data
		 * into output buffer. Let's hope the file structure is OK, as mentioned
		 * in font_pfb_pdf_sections().
		 */

		/* EOF segment, header could be shorter */
		if (size >= 2 && buf[0] == 0x80 && buf[1] == 3) {
			DBG("EOF at %d\n", buf - obuf);
			break;
		}
		if (size < 6 || buf[0] != 0x80 || !(buf[1] & 3)) {
			DBG("Invalid PFB header at offset %d\n", buf - obuf);
			return -1;
		}
		len = buf[2] | (buf[3] << 8) | (buf[4] << 16) | (buf[5] << 24);
		if (size < len + 6) {
			DBG("Segment at %d too long (want %d, have %d)\n", buf - obuf, len, size - 6);
			return -1;
		}

		DBG("%s segment length %d\n", (buf[1] == 1 ? "ASCII" : "Binary"), len);
		memcpy(outbuf, buf+6, len);
		outbuf += len;
		total += len;
		buf += len + 6;
		size -= len + 6;
	}
	return total;
}

static int encode_body(byte *s, uns size, byte *outbuf)
{
	int zeroes = 0;
	byte *os = s;
	byte *first_zero;
	byte *oob;
	int v;
	uns last;
	uns c;

	byte *body_hdr = outbuf;
	outbuf += 6;
	oob = outbuf;
	last = 0;

	DBG("Munching eexec string\n");
	for (c=0; *s; s++) {
		if (*s == ' ' || *s == '\t' || *s == '\n' || *s == '\r')
			continue;

		v = toupper(*s);
		if (v >= '0' && v <= '9')
			v -= '0';
		else if (v >= 'A' && v <= 'F')
			v = v - 'A' + 10;
		else {
			DBG("Invalid character '%c' in hexstring\n", *s);
			return -1;
		}

		if (c & 1)
			*outbuf++ = (last << 4) + v;
		else
			last = v;
		c++;

		if (*s == '0') {
			if (!zeroes)
				first_zero = s;
			zeroes++;
		}
		else
			zeroes = 0;

		if (zeroes == 512)
			break;
	}
	/* Last tuple can be incomplete, assume zero */
	if (!(c & 1)) {
		DBG("Wow, there is one tuple incomplete ;-)\n");
		*outbuf++ = last << 4;
	}

	if (*s) {
		DBG("There is ASCII trailer\n");
		outbuf -= 256;
	}

	u32 len = outbuf-oob;
	DBG("Updating body header, len=%d\n", len);
	body_hdr[0] = 0x80;
	body_hdr[1] = 2;
	body_hdr[2] = len & 0xff;
	body_hdr[3] = (len & 0xffff) >> 8;
	body_hdr[4] = (len & 0xffffff) >> 16;
	body_hdr[5] = len >> 24;

	if (*s) {	/* We have the ASCII trailer */
		u32 l;

		s = first_zero;
		size -= s-os;
		l = size;

		DBG("Storing trailer, len=%d\n", l);
		outbuf[0] = 0x80;
		outbuf[1] = 1;
		outbuf[2] = l & 0xff;
		outbuf[3] = (l & 0xffff) >> 8;
		outbuf[4] = (l & 0xffffff) >> 16;
		outbuf[5] = l >> 24;
		memcpy(outbuf+6, s, l);
		outbuf += l+6;
	}
	return outbuf - oob + 6;
}

int font_pfa_to_pfb_buf(byte *buf, uns size, byte *outbuf)
{
	byte *obuf = buf;
	byte *oob = outbuf;
	byte *b = buf;
	byte *s;
	char *hdr1 = "%!FontType1-";
	char *hdr2 = "%!PS-AdobeFont-1.0:";
	u32 len;

	/* Skip blanks */
	while (*b == ' ' || *b == '\n' || *b == '\t' || *b == '\r')
		b++;

	/* Search for one of allowed headers */
	DBG("Searching for headers\n");
	s = b;
	while (*hdr1 && *hdr1 == *s) {
		hdr1++;
		s++;
	}
	s = buf;
	while (*hdr2 && *hdr2 == *s) {
		hdr2++;
		s++;
	}

	if (*hdr1 && *hdr2) {
		DBG("Not PFA file. Just copying it.\n");
		memcpy(outbuf, obuf, size);
		return size;
	}
	DBG("Found, it's PFA\n");

	DBG("Searching for eexec\n");
	int lev = 0;
	while (*s) {
		/* Search for eexec, avoid strings */
		if (lev == 0 && *s == 'e'
				&& (*(s-1) == ' ' || *(s-1) == '\n' || *(s-1) == '\t' || *(s-1) == '\r')
				&& size - (s-b) >= 6 && strncmp((char *) s, "eexec", 5) == 0)
			break;
		if (*s == '(' && *(s-1) != '\\')
			lev++;
		if (*s == ')' && *(s-1) != '\\' && lev > 0)
			lev--;
		s++;
	}
	if (!*s) {
		DBG("eexec not found!\n");
		return -1;
	}

	s += 5;
	while (*s == ' ' || *s == '\n' || *s == '\t' || *s == '\r')
		s++;

	/* Store the first segment - ASCII header */
	len = s-b;
	DBG("Storing the ASCII header, len=%d\n", len);
	outbuf[0] = 0x80;
	outbuf[1] = 1;
	outbuf[2] = len & 0xff;
	outbuf[3] = (len & 0xffff) >> 8;
	outbuf[4] = (len & 0xffffff) >> 16;
	outbuf[5] = len >> 24;
	memcpy(outbuf+6, b, s-b);
	outbuf += s-b + 6;

	int wr = encode_body(s, size - (s-obuf), outbuf);
	if (wr < 0)
		return -1;
	outbuf += wr;

	/* EOF segment */
	outbuf[0] = 0x80;
	outbuf[1] = 3;
	outbuf += 2;

	return outbuf - oob;
}

int font_pfa_to_pfb(int fntid, FILE *out)
{
	const char *filename = font_filename(fntid);
	byte *buf, *outbuf;
	int fd;
	struct stat stat;
	int r;

	if (!filename)
		return -1;
	fd = open(filename, O_RDONLY);
	if (fd < 0)
		return -1;
	if (fstat(fd, &stat)<0)
		goto fail;

	buf = alloca(stat.st_size);
	outbuf = alloca(stat.st_size + 20);
	if (read(fd, buf, stat.st_size) != stat.st_size)
		goto fail;

	if ((r = font_pfa_to_pfb_buf(buf, stat.st_size, outbuf)) < 0)
		goto fail;
	if (fwrite(outbuf, 1, r, out) != (uns)r)
		goto fail;

	return 0;
fail:
	close(fd);
	return -1;
}
