/*
 * VRR FONTLIB main module
 *
 * Copyright (c) 2004-2005,2007    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG

#include "lib/lib.h"
#include "font/internal.h"
#include "build/path.h"
#include "lib/garr.h"
#include <fontconfig/fontconfig.h>

#define string kpathsea_string
#include <kpathsea/kpathsea.h>
#undef string


FT_Library font_lib;
FTC_Manager font_mgr;
FTC_ImageCache font_img_cache;
FTC_CMapCache font_cmap_cache;

struct fntid {
	uns id;
	char fname[1];
};

#define HASH_PREFIX(x) fntid_##x
#define HASH_NODE struct fntid
#define HASH_KEY_ENDSTRING fname
#define HASH_WANT_CLEANUP
#define HASH_WANT_NEW
#define HASH_WANT_FIND
#include "lib/sht.h"

static struct garr *fnttbl;
static uns max_fntid;

#define GARR_PREFIX(x) fnttbl_##x
#define GARR_TYPE char *
#include "lib/garr.h"


static char *xstrncpy(char *dest, const char *src, size_t n)
{
	if(src)
		return strncpy(dest, src, n);
	dest[0] = 0;
	return NULL;
}

static FT_Error face_request(FTC_FaceID face_id, FT_Library library, FT_Pointer request_data UNUSED, FT_Face *aface)
{
	if (!(addr_int_t)face_id)
		return 1;
	char *fname = *fnttbl_get(fnttbl, (addr_int_t)face_id);
	FT_Error r;

	r = FT_New_Face(library, fname, 0, aface);
	return r;
}

void font_init(const char* vrr_name)
{
	if (FT_Init_FreeType(&font_lib))
		die("FreeType library initialization error\n");
	if (FTC_Manager_New(font_lib, 0, 0, 0, face_request, NULL, &font_mgr))
		die("FreeType font caching subsystem initialization error\n");
	if (FcInit() != FcTrue)
		die("Fontconfig initialization error\n");

	fntid_init();
	fnttbl = fnttbl_new(32);
	kpse_set_program_name(vrr_name, NULL);
}

void font_finish(void)
{
	fnttbl_destroy(fnttbl);
	fntid_cleanup();
	FTC_Manager_Done(font_mgr);
	FT_Done_FreeType(font_lib);
}

int font_id_valid(int fnt_id)
{
	FT_Face face;

	return fnt_id>0 && !FTC_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)fnt_id, &face);
}

const char *font_filename(int fnt_id)
{
	if (!fnt_id)
		return NULL;
	return *fnttbl_get(fnttbl, fnt_id);
}

int font_load_file(const char *fname)
{
	FT_Face face;

	struct fntid *f = fntid_find((char*)fname);

	if (!f) {
		/* Try to load the font to check if it is usable by Freetype */
		if (FT_New_Face(font_lib, fname, 0, &face))
			return 0;	// it isn't...
		FT_Done_Face(face);
			
		f = fntid_new((char*)fname);
		f->id = ++max_fntid;
		*fnttbl_get(fnttbl, f->id) = f->fname;
	}
	return f->id;
}

#define POINT2MM (25.4/72.0)

int font_get_glyph_advances(int font_id, struct garr *width)
{
	FT_Face face;
	int i;
	int w;

	if (FTC_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)font_id, &face))
		return -1;
	
	DBG("Scanning widths of font %s\n", font_filename(font_id));
	for (i=0; i < face->num_glyphs; i++) {
		if (FT_Load_Glyph(face, i, FT_LOAD_NO_SCALE)) {
			DBG("Error loading glyph %d!\n", i);
			w = 0;
		}
		else {
			w = face->glyph->metrics.width;
			DBG("glyph %d metrics: width=%d hbx=%d hby=%d ha=%d lha=%d ax=%d\n", i, (int)face->glyph->metrics.width,
					(int)face->glyph->metrics.horiBearingX, (int)face->glyph->metrics.horiBearingY,
					(int)face->glyph->metrics.horiAdvance, (int)face->glyph->linearHoriAdvance,
					(int)face->glyph->advance.x);
		}
		*(real *)garr_get(width, i) = POINT2MM * w;
	}
	return face->num_glyphs;
}

void font_info_dump(int fnt_id)
{
	FT_Face face;
	uns flg;
	char buf[200];
	int i;
	struct font_t1_info fi;

	if (FTC_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)fnt_id, &face)) {
		msg("Font %d not found\n", fnt_id);
		return;
	}

	msg("Font %d, family '%s', style '%s':\n", fnt_id,
		face->family_name ? face->family_name : "", face->style_name ? face->style_name : "");
	msg("%d glyphs, %d charmaps, %d fixed sizes\n", (int)face->num_glyphs, face->num_charmaps, face->num_fixed_sizes);
	const char *name = FT_Get_Postscript_Name(face);
	if (name)
		msg("PS name: '%s'\n", name);
	if (font_t1_info(fnt_id, &fi)) {
		msg("T1 info:\n");
		msg("family '%s'\n", fi.family);
		msg("style '%s'\n", fi.style);
		msg("full_name '%s'\n", fi.full_name);
		msg("family_name '%s'\n", fi.family_name);
	}

	flg = face->face_flags;
	msg("Flags: %s%s%s%s%s%s%s%s%s\n",
		flg & FT_FACE_FLAG_SCALABLE ? "scalable " : "",
		flg & FT_FACE_FLAG_FIXED_SIZES ? "fixed_sizes " : "",
		flg & FT_FACE_FLAG_FIXED_WIDTH ? "fixed_width " : "",
		flg & FT_FACE_FLAG_SFNT ? "sfnt " : "",
		flg & FT_FACE_FLAG_HORIZONTAL ? "horizontal " : "",
		flg & FT_FACE_FLAG_VERTICAL ? "vertical " : "",
		flg & FT_FACE_FLAG_KERNING ? "kerning " : "",
		flg & FT_FACE_FLAG_MULTIPLE_MASTERS ? "multimasters " : "",
		flg & FT_FACE_FLAG_GLYPH_NAMES ? "glyph_names " : "");

	msg("Style: %s%s\n",
		face->style_flags & FT_STYLE_FLAG_ITALIC ? "italic " : "",
		face->style_flags & FT_STYLE_FLAG_BOLD ? "bold " : "");
	for (i=0; i < face->num_charmaps; i++) {
		u32 e = face->charmaps[i]->encoding;
		msg("Charmap %d: encoding=%c%c%c%c, platform_id=%d encoding_id=%d\n", i, e >> 24, (e >> 16) & 255, (e >> 8) & 255, e & 255,
				face->charmaps[i]->platform_id, face->charmaps[i]->encoding_id);
	}

	if (FT_HAS_GLYPH_NAMES(face)) {
		msg("Glyph names:\n");
		for (i=0; i < face->num_glyphs; i++) {
			FT_Get_Glyph_Name(face, i, buf, sizeof(buf));
			msg("%d: %s\n", i, buf);
		}
	}
}

static int detect_type(FT_Face face)
{
	PS_FontInfoRec psfi;

	if (!FT_Get_PS_Font_Info(face, &psfi))
		return FONT_TYPE_TYPE1;

	if (FT_Get_Sfnt_Table(face, ft_sfnt_head))
		return FONT_TYPE_TRUETYPE;

	return FONT_TYPE_UNKNOWN;
}

int font_get_type(int fnt_id)
{
	FT_Face face;

	if (FTC_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)fnt_id, &face))
		return FONT_TYPE_INVALID;
	
	return detect_type(face);
}

static struct font_info *fnt_info(FT_Face face, struct font_info *fnt_info)
{
	struct font_info fi;

	if (!(face->face_flags & FT_FACE_FLAG_SCALABLE))
		return NULL;
	fi.type = detect_type(face);
	fi.italic = face->style_flags & FT_STYLE_FLAG_ITALIC;
	fi.bold = face->style_flags & FT_STYLE_FLAG_BOLD;
	fi.fixed_width = face->face_flags & FT_FACE_FLAG_FIXED_WIDTH;
	fi.glyphs = face->num_glyphs;
	fi.charmaps = face->num_charmaps;

	if (face->family_name) {
		strncpy(fi.family, face->family_name, sizeof(fi.family)-1);
		fi.family[sizeof(fi.family)-1] = 0;
	} else
		fi.family[0] = 0;

	if (face->style_name) {
		strncpy(fi.style, face->style_name, sizeof(fi.style)-1);
		fi.style[sizeof(fi.style)-1] = 0;
	} else
		fi.style[0] = 0;

	fi.fontid = -1;

	const char *name = FT_Get_Postscript_Name(face);
	if (!name)
		goto end;
	strncpy(fi.name, name, sizeof(fi.name)-1);
	fi.name[sizeof(fi.name)-1] = 0;

	fi.bbox = (struct geom_rectangle){ { face->bbox.xMin, face->bbox.yMin }, { face->bbox.xMax, face->bbox.yMax } };
	fi.ascender = face->ascender;
	fi.descender = face->descender;
	
	*fnt_info = fi;
	return fnt_info;
end:
	return NULL;
}

struct font_info *font_info_temporary(const char *fontpath, struct font_info *fi)
{
	FT_Face face;

	ASSERT(fi);

	if (FT_New_Face(font_lib, fontpath, 0, &face))
		return NULL;
	fnt_info(face, fi);
	FT_Done_Face(face);
	return fi;
}

struct font_info *font_info(int fntid, struct font_info *fi)
{
	FT_Face face;

	ASSERT(fi);

	if (FTC_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)fntid, &face))
		return NULL;
	return fnt_info(face, fi);
}

const char *font_get_ps_name(int fnt_id)
{
	FT_Face face;

	if (FTC_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)fnt_id, &face))
		return NULL;

	return FT_Get_Postscript_Name(face);
}

struct font_t1_info *font_t1_info(int fnt_id, struct font_t1_info *fi)
{
	FT_Face face;
	PS_FontInfoRec psfi;

	ASSERT(fi);

	if (FTC_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)fnt_id, &face) || FT_Get_PS_Font_Info(face, &psfi))
		return NULL;

	if (!(face->face_flags & FT_FACE_FLAG_SCALABLE))
		return NULL;
	fi->italic = face->style_flags & FT_STYLE_FLAG_ITALIC;
	fi->bold = face->style_flags & FT_STYLE_FLAG_BOLD;
	fi->glyphs = face->num_glyphs;
	fi->charmaps = face->num_charmaps;

	if (face->family_name) {
		strncpy(fi->family, face->family_name, sizeof(fi->family)-1);
		fi->family[sizeof(fi->family)-1] = 0;
	} else
		fi->family[0] = 0;

	if (face->style_name) {
		strncpy(fi->style, face->style_name, sizeof(fi->style)-1);
		fi->style[sizeof(fi->style)-1] = 0;
	} else
		fi->style[0] = 0;
	
	/* FIXME: check whether the bbox is 16/16 or 24/8 format */
	fi->bbox.ll.x = face->bbox.xMin / (double) 0x10000;
	fi->bbox.ll.y = face->bbox.yMin / (double) 0x10000;
	fi->bbox.ll.x = face->bbox.xMax / (double) 0x10000;
	fi->bbox.ll.y = face->bbox.xMax / (double) 0x10000;

	fi->units_per_em = face->units_per_EM;
	fi->ascender = face->ascender;
	fi->descender = face->descender;

	xstrncpy(fi->full_name, psfi.full_name, sizeof(fi->full_name)-1);
	xstrncpy(fi->family_name, psfi.family_name, sizeof(fi->family_name)-1);
	xstrncpy(fi->version, psfi.version, sizeof(fi->version)-1);
	xstrncpy(fi->notice, psfi.notice, sizeof(fi->notice)-1);
	xstrncpy(fi->weight, psfi.weight, sizeof(fi->weight)-1);
	fi->italic_angle = psfi.italic_angle;
	fi->is_fixed_pitch = psfi.is_fixed_pitch;
	fi->underline_position = psfi.underline_position;
	fi->underline_thickness = psfi.underline_thickness;

	return fi;
}

struct font_truetype_info *font_truetype_info(int fnt_id, struct font_truetype_info *tt)
{
	FT_Face face;
	FT_Sfnt_Tag tag = ft_sfnt_head;

	ASSERT(tt);

	if (FTC_Manager_LookupFace(font_mgr, (FTC_FaceID)(addr_int_t)fnt_id, &face) || !FT_Get_Sfnt_Table(face, tag))
		return NULL;

	return tt;
}

uns font_get_max_fontid(void)
{
	return max_fntid;
}

char *font_search(const char *family, const char *style, char *outbuf, uns bufsize)
{
	FcPattern *patt, *fnt;
	FcResult res;
	FcChar8 *rv = NULL;
	
	patt = FcPatternCreate();
	if (!patt)
		return NULL;

	FcPatternAddBool(patt, FC_SCALABLE, FcTrue);
	if (family)
		FcPatternAddString(patt, FC_FAMILY, (FcChar8 *)family);
	if (style)
		FcPatternAddString(patt, FC_STYLE, (FcChar8 *)style);

	FcConfigSubstitute(NULL, patt, FcMatchFont);
	FcDefaultSubstitute(patt);
	fnt = FcFontMatch(NULL, patt, &res);
	/* It seems that fontconfig ignores res. How am I supposed to test the result? */

	FcPatternGetString(fnt, FC_FILE, 0, &rv);
	strncpy(outbuf, (char *)rv, bufsize-1);
	uns l = strlen((char *)rv);
	if (l >= bufsize)
		outbuf[l-1] = 0;
	rv = (FcChar8 *)outbuf;

	FcPatternDestroy(patt);
	FcPatternDestroy(fnt);
	return (char *)rv;
}

