/*
 * GEOMLIB - Bernstein polynomials
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#define GEOM_SOURCE
#include "geomlib/base.h"
#include "geomlib/polynomial.h"
#include "geomlib/bernstein.h"


/* geom_bernstein_valid */

int geom_bernstein_valid(uns degree, double coef[])
{
	uns i;

	if (degree > GEOM_MAX_DEGREE || !coef)
		return 0;

	for (i = 0; i <= degree; i++)
		if (!isfinite(coef[i]))
			return 0;

	return 1;
}


/* geom_power_to_bernstein */

int geom_bernstein_to_power(uns degree, double coef[], double result[])
{
	uns i, j;

	BASSERT(geom_bernstein_valid(degree, coef) && result);

	for (i = 0; i <= degree; i++) {
		result[i] = 0;

		for (j = 0; j <= i; j++)
			result[i] +=
			    (double) geom_binomial(degree, i) *
			    (double) geom_binomial(i, j) * (((i - j) % 2 == 1) ? -1 : 1) * coef[j];

		if (!isfinite(result[i]))
			return -1;
	}

	return degree;
}

/* geom_bernstein_to_power */

int geom_power_to_bernstein(uns degree, double coef[], double result[])
{
	uns i, j;

	BASSERT(geom_bernstein_valid(degree, coef) && result);

	for (i = 0; i <= degree; i++) {
		result[i] = 0;

		for (j = 0; j <= i; j++)
			result[i] +=
			    (double) geom_binomial(i, j) / (double) geom_binomial(degree, j) * coef[j];

		if (!isfinite(result[i]))
			return -1;
	}

	return degree;
}


/* geom_bernstein_evaluate
  TODO:
  - rychlejsi je Hornerova metoda O(n), ted O(n^2)
*/

double geom_bernstein_evaluate(uns degree, double coef[], double t)
{
	uns i, j;
	double a[GEOM_MAX_DEGREE + 1];

	BASSERT(geom_bernstein_valid(degree, coef) && isfinite(t));

	if (t < 0)
		return coef[0];

	else if (t > 1)
		return coef[degree];

	for (i = 0; i <= degree; i++)
		a[i] = coef[i];
	for (j = 1; j <= degree; j++)
		for (i = 0; i <= degree - j; i++)
			a[i] = a[i] + t * (a[i + 1] - a[i]);

	if (!isfinite(a[0]))
		return NAN;
	return a[0];
}


/* geom_bernstein_add */

int geom_bernstein_add(uns degree, double coef1[], double coef2[], double result[])
{
	uns i;

	BASSERT(geom_bernstein_valid(degree, coef1) && geom_bernstein_valid(degree, coef2) && result);

	for (i = 0; i <= degree; i++)
		if (!isfinite(result[i] = coef1[i] + coef2[i]))
			return -1;

	return degree;
}


/* geom_bernstein_sub */

int geom_bernstein_sub(uns degree, double coef1[], double coef2[], double result[])
{
	uns i;

	BASSERT(geom_bernstein_valid(degree, coef1) && geom_bernstein_valid(degree, coef2) && result);

	for (i = 0; i <= degree; i++)
		if (!isfinite(result[i] = coef1[i] - coef2[i]))
			return -1;

	return degree;
}


/* geom_bernstein_mul */

int geom_bernstein_mul(uns degree1, double coef1[], uns degree2, double coef2[], double result[])
{
	int i, j, k, n;

	BASSERT(degree1 + degree2 <= GEOM_MAX_DEGREE &&
	       geom_bernstein_valid(degree1, coef1) && geom_bernstein_valid(degree2, coef2) && result);

	n = degree1 + degree2;

	for (i = 0; i <= n; i++) {
		result[i] = 0;

		for (j = MAX(0, (int)(i - degree2)); j <= MIN(i, (int)degree1); j++) {
			k = i - j;
			result[i] +=
			    (double) geom_binomial(degree1, j) *
			    (double) geom_binomial(degree2, k) /
			    (double) geom_binomial(n, i) * coef1[j] * coef2[k];
		}

		if (!isfinite(result[i]))
			return -1;
	}

	return n;
}


/* geom_bernstein_derivate */

int geom_bernstein_derivate(uns degree, double coef[], double result[])
{
	uns i;

	BASSERT(degree > 0);
	BASSERT(geom_bernstein_valid(degree, coef));
	BASSERT(result);

	for (i = 0; i < degree; i++)
		if (!isfinite(result[i] = degree * (coef[i + 1] - coef[i])))
			return -1;

	return degree - 1;
}


/* geom_bernstein_normalize */

#if 0
static void geom_bernstein_normalize(uns degree, double coef[], double result[])
{
	int i;
	double m;

	if (!degree)
		return;

	m = fabs(coef[0]);
	for (i = 1; i <= degree; i++)
		if (m < fabs(coef[i]))
			m = fabs(coef[i]);

	if (m > 0)
		for (i = 0; i <= degree; i++)
			result[i] = fminmax(coef[i] / m, -1, 1);
	else
		for (i = 0; i <= degree; i++)
			result[i] = 0;
}
#endif


/* geom_bernstein_solve
   TODO:
   - dodelat algoritmus BCOM pro hledani korenu v <0,1>... nesmi nikdy vracet -1
   - aktualni algoritmus sice funguje, ale je v nekterych pripadech pomaly
     a neumi se vyporadat s GEOM_SOLVE_MULTIPLICITY
*/

//static int geom_bernstein_solve_slow(int degree, double coef[], int flags, double result[]);

int geom_bernstein_solve(uns degree, double coef[], uns flags, double result[])
{
	double /*norm[GEOM_MAX_DEGREE + 1], */power[GEOM_MAX_DEGREE + 1];
	int n;

	BASSERT(geom_bernstein_valid(degree, coef) && result);

	//geom_bernstein_normalize(degree, coef, norm);
	if (geom_bernstein_to_power(degree, coef, power) < 0) /*{
		if (flags & GEOM_SOLVE_UNIT_INTERVAL)
			return geom_bernstein_solve_slow(degree, coef, flags, result);
		else
			return -1;
	}*/
		return 0; // FIXME
	/*for (int i = 0; i <= degree; i++)
		msg("%f ", norm[i]);
	msg("\n");
	for (int i = 0; i <= degree; i++)
		msg("%f ", power[i]);
	msg("\n");*/

	n = geom_polynomial_solve(degree, power, flags, result);
	if (n >= 0 /*|| !(flags & GEOM_SOLVE_UNIT_INTERVAL)*/)
		return n;

	//return geom_bernstein_solve_slow(degree, coef, flags, result);
	return 0; // FIXME
}


/*-----------------------------------------------------------------------------------------------------*/


#if 0
/* TODO: rozdelane resitko polynomu v bernsteinove bazi */


struct geom_bcom_poly {
	int d;                           /* degree of bernstein polynomial */
	double a, b;                     /* domain range, have to be a<b */
	double c[GEOM_MAX_DEGREE + 1];   /* coefficients */
	double e[GEOM_MAX_DEGREE + 1];   /* errors in coefficients */
	double tol;                      /* global error bound */
};


#ifdef LOCAL_DEBUG

static int geom_bcom_dbg_index = 0;
static char geom_bcom_dbg_buf[4][1000];

/* geom_bcom_dbg
 * Debug string for geom_bcom_poly structure. */

static char *geom_bcom_dbg(struct geom_bcom_poly *p)
{
	char *s, *t;
	int i;

	/* ! nekontroluju preteceni bufferu, ale jde jen o debugovaci vypisy */
	t = s = geom_bcom_dbg_buf[geom_bcom_dbg_index++];
	geom_bcom_dbg_index &= 3;
	s += sprintf(s, "B(%p d=%d a=%.3f b=%.3f)(%.3f", p, p->d, p->a, p->b, p->c[0]);
	for (i = 1; i <= p->d; i++)
		s += sprintf(s, " %.3f", p->c[i]);
	s += sprintf(s, ")");
	return t;
}

#endif


/* geom_bcom_copy
 * Copies bernstein polynomial. */

static void geom_bcom_copy(struct geom_bcom_poly *p, struct geom_bcom_poly *d)
{
	int i;

	DBG("copying %s to %p\n", geom_bcom_dbg(p), d);

	d->d = p->d;
	d->a = p->a;
	d->b = p->b;
	d->tol = p->tol;
	for (i = 0; i <= p->d; i++) {
		d->c[i] = p->c[i];
		d->e[i] = p->e[i];
	}
}


/* geom_bcom_evaluate
 * Evaluates bernstein polynomial at given parameter.
 * TODO: pocitani chyby */

static double geom_bcom_evaluate(struct geom_bcom_poly *p, double t, double *err)
{
	double scale, s, v, r, e;
	int i;

	s = 1 - t;
	scale = 1;
	e = 0;

	if (t < 0.5) {
		v = t / s;
		r = p->c[p->d];
		for (i = p->d - 1; i >= 0; i--) {
			r = (double)(p->d - i) / (i + 1) * v * r + p->c[i];
			scale *= s;
		}
	}

	else {
		v = s / t;
		r = p->c[0];
		for (i = 1; i <= p->d; i++) {
			r = (double)i / (p->d - i + 1) * v * r + p->c[i];
			scale *= t;
		}
	}

	r *= scale;
	*err = e;

	DBG("evaluating %s at %f -> %f, e=%f\n", geom_bcom_dbg(p), t, r, e);
	return r;
}


/* geom_bcom_normalize
 * Normalize bernstein coefficients to avoid floating-point overflow and underflow.
 * TODO: mozna pouzit logaritmicke scalovani podobne jako v RPOLY */

static void geom_bcom_normalize(struct geom_bcom_poly *p)
{
	double s;
	int i;

	/* find the largest coefficient */
	s = 0;
	for (i = 0; i <= p->d; i++)
		if (s < fabs(p->c[i]))
			s = fabs(p->c[i]);

	/* scale */
	if (s > 0) {
		for (i = 0; i <= p->d; i++) {
			p->c[i] /= s;
			p->e[i] /= s;
		}
		p->tol /= s;
	}

	DBG("normalizing %s\n", geom_bcom_dbg(p));
}


/* geom_bcom_deflate */

/*static void geom_bcom_deflate(struct geom_bcom_poly *p, double t)
{
	int i;
	double u, tmp1, tmp2;

	DBG("deflating %f from %s\n", t, geom_bcom_dbg(p));

	if (t < 0.5) {
		u = t / (1 - t);
		tmp1 = p->c[p->d - 1];
		p->c[p->d - 1] = p->c[p->d];
		for (i = 1; i < p->d; i++) {
			tmp2 = p->c[p->d - i - 1];
			p->c[p->d - i - 1] = (p->d * tmp1 + i * p->c[p->d - i] * u) / (p->d - i);
			tmp1 = tmp2;
		}
	}

	else {
		u = (1 - t) / t;
		for (i = 1; i < p->d; i++)
			p->c[i] = (p->d * p->c[i] + i * p->c[i - 1] * u) / (p->d - i);
	}

	p->d--;
}*/


/* geom_bcom_deflate_left */

static void geom_bcom_deflate_left(struct geom_bcom_poly *p)
{
	int i;

	DBG("deflating left from %s\n", geom_bcom_dbg(p));

	for (i = 0; i < p->d; i++)
		p->c[i] = p->c[i + 1] * p->d / (i + 1);

	p->d--;
}


/* geom_bcom_deflate_right */

static void geom_bcom_deflate_right(struct geom_bcom_poly *p)
{
	int i;

	DBG("deflating right from %s\n", geom_bcom_dbg(p));

	for (i = 0; i < p->d; i++)
		p->c[i] = p->c[i] * p->d / (p->d - i);

	p->d--;
}


/* geom_bcom_end_left */

static int geom_bcom_end_left(struct geom_bcom_poly *p)
{
	int i;

	DBG("deflating left roots from %s\n", geom_bcom_dbg(p));

	for (i = 0; p->d > 0 && IS_ZERO(p->c[0]); i++) {
		geom_bcom_deflate_left(p);
		geom_bcom_normalize(p);
	}

	return i;
}


/* geom_bcom_end_right */

static int geom_bcom_end_right(struct geom_bcom_poly *p)
{
	int i;

	DBG("deflating right roots from %s\n", geom_bcom_dbg(p));

	for (i = 0; p->d > 0 && IS_ZERO(p->c[p->d]); i++) {
		geom_bcom_deflate_right(p);
		geom_bcom_normalize(p);
	}

	return i;
}


/* geom_bcom_derivate */

static void geom_bcom_derivate(struct geom_bcom_poly *p, struct geom_bcom_poly *d)
{
	int i;

	d->d = p->d - 1;
	d->a = p->a;
	d->b = p->b;

	for (i = 0; i < p->d; i++)
		d->c[i] = p->c[i + 1] - p->c[i];

	DBG("derivating %s to %s\n", geom_bcom_dbg(p), geom_bcom_dbg(d));
}


/* geom_bcom_subdivide_left */

static void geom_bcom_subdivide_left(struct geom_bcom_poly *p, struct geom_bcom_poly *q, double t)
{
	int i, j;
	double a[GEOM_MAX_DEGREE + 1][GEOM_MAX_DEGREE + 1];

	DBG("subdividing %s at %f, right part writing to %p\n", geom_bcom_dbg(p), t, q);

	q->d = p->d;
	q->a = p->a + t * (p->b - p->a);
	q->b = p->b;
	p->b = q->a;

	for (i = 0; i <= p->d; i++)
		a[0][i] = p->c[i];
	for (j = 1; j <= p->d; j++)
		for (i = 0; i <= p->d - j; i++)
			a[j][i] = a[j - 1][i] + t * (a[j- 1][i + 1] - a[j - 1][i]);
	for (i = 0; i <= p->d; i++) {
		p->c[i] = a[i][0];
		q->c[i] = a[p->d - i][i];
	}
}


/* geom_bcom_refine
 * Function returns a root in given interval of bernstein polynomial.
 * The endpoints of p must have different signs.
 * TODO: pocitani s chybami p->e */

static double geom_bcom_refine(struct geom_bcom_poly *p, double lt, double rt)
{
	double lv, rv, cv, ct, err;
	int i, ls, rs, cs, ps;

	DBG("refining %s in (%f,%f)\n", geom_bcom_dbg(p), lt, rt);

	lv = geom_bcom_evaluate(p, lt, &err);
	ls = lv > 0 ? 0 : 1;

	rv = geom_bcom_evaluate(p, rt, &err);
	rs = rv > 0 ? 0 : 1;

	ASSERT(ls != rs);

	cs = ls;

	for (i = 0; ; i++) {
		ct = lt + (rt - lt) * (lv / (lv - rv));

		if (i > 25)
			break;

		cv = geom_bcom_evaluate(p, ct, &err);
		if (IS_ZERO(cv))
			break;

		ps = cs;
		cs = cv > 0 ? 0 : 1;

		if (cs == ls) {
			lt = ct;
			lv = cv;
			if (ps == cs)
				rv *= 0.5;
		}

		else {
			rt = ct;
			rv = cv;
			if (ps == cs)
				lv *= 0.5;
		}
	}

	DBG("refined to %f in %d steps\n", ct, i + 1);

	return ct;
}


/* geom_bcom_sign_changes */

static int geom_bcom_sign_changes(struct geom_bcom_poly *p, int *first, int *second)
{
	int i, n = 0;

	DBG("looking for sign changes in %s\n", geom_bcom_dbg(p));

	for (i = 0; i < p->d; i++) {
		if ((p->c[i] < 0 && p->c[i + 1] >= 0) || (p->c[i] >= 0 && p->c[i + 1] < 0)) {
			if (!(n++))
				*first = i;
			else {
				*second = i;
				return 2;
			}
		}
	}

	return n;
}


/* geom_bcom_record */

static void geom_bcom_record(double t UNUSED, int m)
{
	if (!m)
		return;

	DBG("****************** root %f with multiplicity %d found\n", t, m);
}


/* geom_bcom
 * TODO: zatim zcela nefunkcni */

int geom_bcom(int degree, double coef[])
{
	int i, m;

	int stack_top;
	struct geom_bcom_poly stack[GEOM_MAX_DEGREE + 1];

	int level;
	int level_max;

	int deriv_pos[GEOM_MAX_DEGREE + 1];
	int deriv_sub[GEOM_MAX_DEGREE + 1];
	int deriv_max_sub;

	struct geom_bcom_poly master[GEOM_MAX_DEGREE + 1];

	struct geom_bcom_poly local[GEOM_MAX_DEGREE + 1];

	int roots_n;
	int roots_right;

	int cross_n, cross_i = 0, cross_j = 0;

	double r, s, t, err;
	double t_left, t_right, t_root, t_eval;



	DBG("starting BCOM algorithm\n");

	/* initialize */
	stack_top = 0;
	level = 0;
	level_max = 0;
	deriv_pos[0] = 0;
	deriv_sub[0] = 0;
	roots_n = 0;
	roots_right = 0;
	deriv_max_sub = 2;

	/* assign the master poly structure */
	master[0].d = degree;
	master[0].a = 0;
	master[0].b = 1;
	for (i = 0; i <= degree; i++) {
		master[0].c[i] = coef[i];
		master[0].e[i] = 0;
	}
	DBG("input polynomial %s\n", geom_bcom_dbg(&master[0]));

	/* normalize master[0] */
	geom_bcom_normalize(&master[0]);

	/* test roots at right interval limit */
	DBG("looking for roots at right interval limit\n");
	roots_right = geom_bcom_end_right(&master[0]);
	if (roots_right > 0)
		DBG("found %d such roots\n", roots_right);

	/* test roots at left interval limit */
	DBG("looking for roots at left interval limit\n");
	geom_bcom_record(0, geom_bcom_end_left(&master[0]));

	/* assign the local poly structure */
	geom_bcom_copy(&master[0], &local[0]);



	/* real roots isolation loop steps */
	DBG("roots isolation for degree %d derivation %s\n", level, geom_bcom_dbg(&local[level]));

BCOM_2_1:
	/* test for roots at left interval limit */
	if (level > 0)
		while (IS_ZERO(local[level].c[0]))
			geom_bcom_deflate_left(&local[level]);

BCOM_2_2:
	/* look for sign changes in coefficients to estimate number of roots in local[level] */
	DBG("looking for sign changes in degree %d derivation %s\n", level, geom_bcom_dbg(&local[level]));
	cross_n = geom_bcom_sign_changes(&local[level], &cross_i, &cross_j);

	/* multiple sign changes -> there can be any number of roots */
	if (cross_n > 1) {
		DBG("multiple sign changes found\n");

		r = (cross_i + cross_j + 1.5) / (2 * local[level].d);
		s = local[level].a + r * (local[level].b - local[level].a);

		t = geom_bcom_evaluate(&local[level], s, &err);

		if (deriv_sub[level] > deriv_max_sub || IS_ZERO(t)) {
			DBG("using degree %d derivation to isolate roots\n", level + 1);

			deriv_pos[level] = stack_top;
			deriv_sub[level] = 0;
			deriv_sub[++level] = 0;

			if (level > level_max) {
				geom_bcom_derivate(&master[level - 1], &master[level]);
				level_max++;
			}

			geom_bcom_derivate(&local[level - 1], &local[level]);
		}

		else {
			DBG("using subdivision to isolate roots\n");

			deriv_sub[level]++;
			geom_bcom_subdivide_left(&local[level], &stack[stack_top], r);
			stack_top++;

			goto BCOM_2_2;
		}
	}

	/* single sign change -> there is exactly one root */
	else if (cross_n == 1) {
		DBG("single sign change -> one root in this interval\n");

		t_left = local[level].a;
		t_right = local[level].b;
		goto BCOM_3_1;
	}

	/* no sign change -> there are no roots */
	else {
		DBG("no sign change -> no root in this interval\n");
	}

BCOM_2_4_3:
	/* if stack empty */
	if (stack_top == 0) {
		DBG("no more polynomials on stack\n");
		if (level == 0)
			goto BCOM_2_5;
		else
			level--;
	}

	/* pop next polynomial from stack */
	else {
		DBG("pop next polynomial from stack\n");
		geom_bcom_copy(&stack[--stack_top], &local[level]);
		goto BCOM_2_1;
	}


BCOM_2_5:
	DBG("ending algorithm\n");
	geom_bcom_record(1, roots_right);
	return 0;


BCOM_3_1:
	/* refine t_root in (t_left, t_right) with multiplicity 1 */
	m = 1;
	t_root = geom_bcom_refine(&local[level], t_left, t_right);

BCOM_3_2:
	/* refined root of original polynomial */
	if (level == 0) {
		geom_bcom_record(t_root, m);
		goto BCOM_2_4_3;
	}

	/* refined root of a derivation */
	else {
		DBG("degree %d derivation root(s) found\n", level);

		level--;
		stack_top = deriv_pos[level];

		t_eval = geom_bcom_evaluate(&local[level], t_root, &err);

		if (IS_ZERO(t_eval)) {
			m++;
			goto BCOM_3_2;

		}

		else {
			m = 0;

			if (local[level].c[0] * t_eval < 0) {
				t_left = local[level].a;
				t_right = t_root;
				goto BCOM_3_1;
			}

			else {
				deriv_sub[level] = 0;
				t = (t_root - local[level].a) / (local[level].b - local[level].a);
				geom_bcom_subdivide_left(&local[level], &stack[stack_top], t);
				stack_top++;
				goto BCOM_2_1;
			}
		}
	}



	return 0;
}



/* geom_bernstein_solve_slow */

#define GEOM_BERNSTEIN_MAX_SUBDIV 25

int geom_bernstein_solve_slow(int degree, double coef[], int flags, double result[])
{
	int i, j, n, m;
	int stack_top, stack_d[GEOM_MAX_DEGREE + 1];
	struct geom_bcom_poly stack[GEOM_BERNSTEIN_MAX_SUBDIV + 1], *p;

	stack_top = 0;
	stack_d[0] = 0;
	stack[0].d = degree;
	stack[0].a = 0;
	stack[0].b = 1;
	for (i = 0; i <= degree; i++)
		stack[0].c[i] = coef[i];
	geom_bcom_normalize(&stack[0]);
	n = 0;

	while (stack_top >= 0) {
		p = &stack[stack_top];
		m = geom_bcom_sign_changes(p, &i, &j);
		if (m > 1 && stack_d[stack_top] < GEOM_BERNSTEIN_MAX_SUBDIV) {
			geom_bcom_copy(p, &stack[stack_top + 1]);
			geom_bcom_subdivide_left(&stack[stack_top + 1], p, 0.5);
			stack_d[stack_top + 1] = ++stack_d[stack_top];
			stack_top++;
		}
		else if (m == 0)
			stack_top--;
		else {
			if (m == 1) {
				result[n] = p->a + (p->b - p->a) * geom_bcom_refine(p, 0, 1);
				result[n] = MIN(result[n], p->b);
				n++;
			}
			else if (m > 1 && p->c[0] * p->c[p->d] <= 0) {
				result[n] = p->a + 0.5 * (p->b - p->a);
				result[n] = MIN(result[n], p->b);
				n++;
			}
			if (n == 1 && flags & GEOM_SOLVE_LEFT_ONLY)
				return 1;
			stack_top--;
		}
	}

	return n;
}


/* geom_bcom_test */

void geom_bcom_test(void)
{
	struct geom_bcom_poly p;
	double r[10], s[10];
	int i, n, m;

	p.a = 0;
	p.b = 1;
	p.d = 5;
	p.c[0] = -1;
	p.c[1] = 2;
	p.c[2] = 10;
	p.c[3] = -20;
	p.c[4] = 20;
	p.c[5] = -10;

	/*geom_bcom(p.d, p.c);*/

	m = geom_bernstein_solve_slow(p.d, p.c, 0, r);
	n = geom_bernstein_solve(p.d, p.c, GEOM_SOLVE_UNIT_INTERVAL, s);
	for (i = 0; i < m; i++)
		DBG("slow = %f\n", r[i]);
	for (i = 0; i < n; i++)
		DBG("norm = %f\n", s[i]);
}

#endif
