/*
 * GEOMLIB - Circles and arcs
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#define GEOM_SOURCE
#include "geomlib/vector.h"
#include "geomlib/transform.h"
#include "geomlib/fpath.h"
#include "geomlib/bezier.h"
#include "geomlib/conic.h"
#include "geomlib/circle.h"


/*----------------------------------------------*
 *                                              *
 *              INITIALIZATION                  *
 *                                              *
 *----------------------------------------------*/


/* geom_arc_set */

void geom_arc_set(struct geom_arc *arc, struct geom_point *center, real radius, real start, real end)
{
	FUNC;
	INSTANCE_IS_ASSERT(arc, arc);
	GASSERT(point, center);
	RASSERT(radius);
	RASSERT(start);
	RASSERT(end);

	arc->center = *center;
	arc->radius = radius;
	arc->start = start;
	arc->end = end;
	
	ITEM_ASSERT(arc, arc);
	RET;
}


/**
 * Fits starting angle to interval [0, 2*PI].
 *
 * @aio arc  Arc to normalize.
 */
static inline void geom_arc_normalize_angles(struct geom_arc *arc UNUSED)
{
	double dif;

	FUNC;
	RASSERT(arc->start);
	RASSERT(arc->end);

	dif = arc->end - arc->start;
	arc->start = fmodfloorr(arc->start, 2 * M_PI);
	arc->end = inf2finr(arc->start + dif);

	RASSERT(arc->start);
	RASSERT(arc->end);
	RET;
}


/* geom_arc_set_endpoints */

int geom_arc_set_endpoints(struct geom_arc *arc, struct geom_point *start, struct geom_point *end, struct geom_point *mid, uns flags)
{
	double t1, t2, t3;
	int swap;

	FUNC;

	ITEM_ASSERT(arc, arc);

	if (flags == GEOM_CONIC_ARC_ENTIRE) {
		arc->start = 0;
		arc->end = INFINITY;
		geom_arc_after_change(arc);
		goto EXIT;
	}

	GASSERT(point, start);
	GASSERT(point, end);

	t1 = nfinset(atan2(start->y - arc->center.y, start->x - arc->center.x), 0);
	t2 = nfinset(atan2(end->y - arc->center.y, end->x - arc->center.x), 0);

	switch (flags & ~1) {
		case GEOM_CONIC_ARC_CCW:
			swap = t1 > t2;
			break;

		case GEOM_CONIC_ARC_SMALLER:
			swap = fabs(t2 - t1) > M_PI;
			break;

		case GEOM_CONIC_ARC_MIDDLE:
			GASSERT(point, mid);
			t3 = nfinset(atan2(mid->y - arc->center.y, mid->x - arc->center.x), 0);
			swap = (t1 < t3) ^ (t3 < t2);
			break;

		default:
			BUG();
	}

	if (swap ^ (flags & 1)) {
		if (t1 < t2)
			t1 += 2 * M_PI;
		else
			t2 += 2 * M_PI;
	}

	RASSERTX(arc->start = t1);
	RASSERTX(arc->end = t2);
	geom_arc_normalize_angles(arc);
	geom_arc_after_change(arc);

EXIT:
	ERET(0);
}


/* geom_arc_from_center_point */

int geom_arc_from_center_point(struct geom_arc *arc, struct geom_point *center, struct geom_point *point)
{
	FUNC;
	INSTANCE_IS_ASSERT(arc, arc);
	ITEM_ASSERT(arc, arc);
	GASSERT(point, center);
	GASSERT(point, point);

	arc->center = *center;
	RTEST_NUM(arc->radius = hypot(center->x - point->x, center->y - point->y));
	arc->start = 0;
	arc->end = INFINITY;
	geom_arc_after_change(arc);

	ERET(0);
}


/* geom_arc_from_2_points_thales */

int geom_arc_from_2_points_thales(struct geom_arc *arc, struct geom_point *point1, struct geom_point *point2)
{
	FUNC;
	INSTANCE_IS_ASSERT(arc, arc);
	GASSERT(point, point1);
	GASSERT(point, point2);

	arc->center.x = fmidfinr(point1->x, point2->x);
	arc->center.y = fmidfinr(point1->y, point2->y);
	RTEST_NUM(arc->radius = hypot(point1->x - arc->center.x, point1->y - arc->center.y));
	arc->start = 0;
	arc->end = INFINITY;
	geom_arc_after_change(arc);

	ERET(0);
}


/* geom_arc_from_3_points */

int geom_arc_from_3_points(struct geom_arc *arc, struct geom_point *point1, struct geom_point *point2, struct geom_point *point3)
{
	double ax, ay, bx, by, cx, cy, abx, aby, bcx, bcy, t, ab, bc, det, t1, t2;
	real minx, maxx, miny, maxy;

	FUNC;
	INSTANCE_IS_ASSERT(arc, arc);
	GASSERT(point, point1);
	GASSERT(point, point2);
	GASSERT(point, point3);

	ax = point1->x; ay = point1->y;
	bx = point2->x; by = point2->y;
	cx = point3->x; cy = point3->y;

	t = SQR(bx) + SQR(by);
	ab = 0.5 * (SQR(ax) + SQR(ay) - t);
	bc = 0.5 * (t - SQR(cx) - SQR(cy));
	abx = ax - bx; aby = ay - by;
	bcx = bx - cx; bcy = by - cy;

	t1 = abx * bcy;
	t2 = bcx * aby;
	det = t1 - t2;
	if (unlikely(IS_ZERO(det / t1)))
		goto COLLINEAR;
	det = 1 / det;
	if (!FTEST(det))
		goto COLLINEAR;

	arc->center.x = (ab * bcy - bc * aby) * det;
	arc->center.y = (abx * bc - bcx * ab) * det;
	if (!RTEST(arc->center.x) || !RTEST(arc->center.y)) {
COLLINEAR:
		minx = MIN(point1->x, point2->x); minx = MIN(minx, point3->x);
		maxx = MAX(point1->x, point2->x); maxx = MAX(maxx, point3->x);
		miny = MIN(point1->y, point2->y); miny = MIN(miny, point3->y);
		maxy = MAX(point1->y, point2->y); maxy = MAX(maxy, point3->y);
		arc->center.x = fmidfinr(minx, maxx);
		arc->center.y = fmidfinr(miny, maxy);
	}
	RTEST_NUM(arc->radius = hypot(arc->center.x - point1->x, arc->center.y - point1->y));
	arc->start = 0;
	arc->end = INFINITY;
	geom_arc_after_change(arc);

	ERET(0);
}


/*----------------------------------------------*
 *                                              *
 *             PARAMETRIZATIONS                 *
 *                                              *
 *----------------------------------------------*/


/* geom_arc_alength */

real geom_arc_alength(struct geom_arc *arc)
{
	FUNC;
	ITEM_ASSERT(arc, arc);

	if (geom_item_is_closed(arc))
		RRET(2 * M_PI * arc->radius);
	else
		RRET(fabs(arc->start - arc->end) * arc->radius);
}


/* geom_arc_time_to_atime */

real geom_arc_time_to_atime(struct geom_arc *arc, real time)
{
	double l;

	FUNC;
	ITEM_ASSERT(arc, arc);
	RASSERT(time);

	if (time <= 0)
		RRET(0);
	if (geom_item_is_closed(arc))
		l = 2 * M_PI * arc->radius;
	else
		l = fabs(arc->start - arc->end) * arc->radius;
	RRET(time >= 1 ? l : l * time);
}


/* geom_arc_atime_to_time */

real geom_arc_atime_to_time(struct geom_arc *arc, real atime)
{
	double l;

	FUNC;
	ITEM_ASSERT(arc, arc);
	RASSERT(atime);

	if (atime <= 0)
		RRET(0);
	if (geom_item_is_closed(arc))
		l = 2 * M_PI * arc->radius;
	else
		l = fabs(arc->start - arc->end) * arc->radius;
	RRET((double)atime >= l ? 1 : (double)atime / l);
}


/* geom_arc_time_to_btime */

real geom_arc_time_to_btime(struct geom_arc *arc, struct geom_expansion *expansion, real time)
{
	FUNC;
	ITEM_ASSERT(arc, arc);
	GASSERT(expansion, expansion);
	RASSERT(time);

	if (time <= 0)
		RRET(0);
	if (time >= 1)
		RRET(geom_group_count(&expansion->path));
	else {
		double t = time * geom_group_count(&expansion->path), index = floor(t), btime, alpha, beta, abs_beta, s;
		if (!geom_item_is_closed(arc))
			alpha = 0.25 * fabs(arc->end - arc->start) / geom_group_count(&expansion->path);
		else
			alpha = 0.5 * M_PI / 3;
		beta = 2 * (t - index - 0.5) * alpha;
		abs_beta = fabs(beta);
		t = sin(alpha + abs_beta);
		s = sin(alpha - abs_beta);
		t += s;
		s *= 2;
		if (s <= 0)
			btime = beta < 0 ? index : index + 1;
		else if (s >= t)
			btime = index + 0.5;
		else {
			t = 0.5 * s / t;
			btime = index + (beta < 0 ? t : 1 - t);
		}
		RRET(btime);
	}
}


/* geom_arc_btime_to_time */

real geom_arc_btime_to_time(struct geom_arc *arc, struct geom_expansion *expansion, real btime)
{
	FUNC;
	ITEM_ASSERT(arc, arc);
	GASSERT(expansion, expansion);
	RASSERT(btime);

	if (btime <= 0)
		RRET(0);
	else if (btime >= geom_group_count(&expansion->path))
		RRET(1);
	else {
		double index = floorr(btime), t = btime - index, alpha, half, beta, y, x;
		if (!geom_item_is_closed(arc)) {
			alpha = (arc->end - arc->start) / (geom_group_count(&expansion->path));
			half = 0.5 * alpha;
			y = sin(half) * (0.5 - t);
			t = t - SQR(t);
			x = cos(half) * (0.5 - t) + t;
			beta = atan2(y, x);
			RRET((index + 0.5 - beta / alpha) / (geom_group_count(&expansion->path)));
		}
		else {
			y = sqrt(3) * (0.5 - t);
			x = 0.5 + t - SQR(t);
			beta = atan2(y, x);
			RRET((index + 0.5 - beta / (2 * M_PI / 3)) / 3);
		}
	}
}


/*----------------------------------------------*
 *                                              *
 *                  CONVERSION                  *
 *                                              *
 *----------------------------------------------*/


/* geom_arc_expand_to_tail */

/*int geom_arc_expand_to_tail(struct geom_arc *arc, struct geom_path *path)
{
	uns i, n;
	double w, dif, abs_dif, mid, end, a, b, c;
	real x1, x2;
	struct geom_bezier bezier;

	FUNC;

	GEOM_ARC_ASSERT(arc);
	GEOM_PATH_ASSERT(path);

	bezier.degree = GEOM_BEZIER_QUADRATIC;
	bezier.flags = 0;
	bezier.pt[0].w = bezier.pt[2].w = 1;

	if (geom_arc_is_closed(arc)) {
		bezier.pt[1].w = 0.5;
		a = sqrt(3) * arc->radius;
		b = 0.5 * arc->radius;
		c = 0.5 * a;
		RTEST_NUM(x1 = arc->radius + arc->center.x);
		RTEST_NUM(x2 = -b + arc->center.x);

		bezier.pt[0].x = x1;
		bezier.pt[0].y = arc->center.y;
		bezier.pt[1].x = x1;
		RTEST_NUM(bezier.pt[1].y = a + arc->center.y);
		bezier.pt[2].x = x2;
		RTEST_NUM(bezier.pt[2].y = c + arc->center.y);
		GASSERTX(e,geom_path_insert_tail(path, GEOM_CURVE_BEZIER, &bezier));

		bezier.pt[0] = bezier.pt[2];
		RTEST_NUM(bezier.pt[1].x = -2 * arc->radius + arc->center.x);
		bezier.pt[1].y = arc->center.y;
		RTEST_NUM(bezier.pt[2].y = -c + arc->center.y);
		GASSERTX(e,geom_path_insert_tail(path, GEOM_CURVE_BEZIER, &bezier));

		bezier.pt[0].y = bezier.pt[2].y;
		bezier.pt[1].x = x1;
		RTEST_NUM(bezier.pt[1].y = -a + arc->center.y);
		bezier.pt[2].x = x1;
		bezier.pt[2].y = arc->center.y;
		GASSERTX(e,geom_path_insert_tail(path, GEOM_CURVE_BEZIER, &bezier));

		ERET(0);
	}

	FTEST_NUM(dif = arc->end - arc->start);
	TEST_ERR((abs_dif = fabs(dif)) <= 5 * M_PI, CONIC_TOO_LONG);
	n = 1 + (int)(abs_dif * (4.0 / 3 / M_PI));
	dif /= n;
	end = arc->start;
	mid = end - 0.5 * dif;

	bezier.pt[1].w = w = cos(0.5 * dif);
	TEST_NUM(bezier.pt[1].w > 0);
	a = arc->radius / w;

	RTEST_NUM(bezier.pt[2].x = cos(arc->start) * arc->radius + arc->center.x);
	RTEST_NUM(bezier.pt[2].y = sin(arc->start) * arc->radius + arc->center.y);

	for (i = 1; i <= n; i++) {
		end += dif;
		mid += dif;

		bezier.pt[0] = bezier.pt[2];
		RTEST_NUM(bezier.pt[1].x = cos(mid) * a + arc->center.x);
		RTEST_NUM(bezier.pt[1].y = sin(mid) * a + arc->center.y);
		RTEST_NUM(bezier.pt[2].x = cos(end) * arc->radius + arc->center.x);
		RTEST_NUM(bezier.pt[2].y = sin(end) * arc->radius + arc->center.y);
		GASSERTX(e,geom_path_insert_tail(path, GEOM_CURVE_BEZIER, &bezier));
	}
	ERET(0);
}*/


/* geom_arc_to_bezier */

/*int geom_arc_to_bezier(struct geom_arc *arc, struct geom_bezier *bezier)
{
	double w, mid, dif, coef;

	FUNC;

	GEOM_ARC_ASSERT(arc);
	GEOM_BEZIER_INIT_ASSERT(bezier);

	dif = fabs(arc->end - arc->start);
	TEST_ERR(dif < M_PI, CONIC_TOO_LONG);

	bezier->degree = GEOM_BEZIER_QUADRATIC;
	bezier->flags = 0;
	bezier->pt[0].w = bezier->pt[2].w = 1;

	mid = fmidfin(arc->start, arc->end);

	RTEST_NUM(bezier->pt[0].x = cos(arc->start) * arc->radius + arc->center.x);
	RTEST_NUM(bezier->pt[0].y = sin(arc->start) * arc->radius + arc->center.y);
	RTEST_NUM(bezier->pt[2].x = cos(arc->end) * arc->radius + arc->center.x);
	RTEST_NUM(bezier->pt[2].y = sin(arc->end) * arc->radius + arc->center.y);

	bezier->pt[1].w = w = cos(0.5 * dif);
	TEST_NUM(bezier->pt[1].w > 0);
	coef = arc->radius / w;
	RTEST_NUM(bezier->pt[1].x = cos(mid) * coef + arc->center.x);
	RTEST_NUM(bezier->pt[1].y = sin(mid) * coef + arc->center.y);

	ERET(0);
}*/


/* geom_arc_to_poly */

int geom_arc_to_poly(struct geom_arc *arc, real poly[6])
{
	FUNC;
	ITEM_ASSERT(arc, arc);
	PASSERT(poly);

	poly[0] = 1;
	poly[1] = 0;
	poly[2] = 1;
	RTEST_NUM(poly[3] = -2 * arc->center.x);
	RTEST_NUM(poly[4] = -2 * arc->center.y);
	RTEST_NUM(poly[5] = SQR(arc->center.x) + SQR(arc->center.y) - SQR(arc->radius));

	ERET(0);
}


/*----------------------------------------------*
 *                                              *
 *               TRANSFORMATIONS                *
 *                                              *
 *----------------------------------------------*/


/* geom_arc_transform */

int geom_arc_transform(struct geom_arc *arc, struct geom_transform *transform, struct geom_arc *result)
{
	FUNC;
	ITEM_ASSERT(arc, arc);
	GASSERT(transform, transform);
	INSTANCE_IS_ASSERT(arc, result);

	TEST_ERR(transform->flags & GEOM_TRANSFORM_SIMILAR, TRANSFORM_NOT_SIMILAR);
	ETEST_RET(geom_point_transform(&arc->center, transform, &result->center));

	/* TODO: transformace polomeru a koncovych uhlu */
	result->radius = arc->radius;
	result->start = arc->start;
	result->end = arc->end;

	ERET(0);
}


/* geom_arc_normalize */

int geom_arc_normalize(struct geom_arc *arc)
{
	FUNC;
	ITEM_ASSERT(arc, arc);

	if (!geom_item_is_closed(arc))
		geom_arc_normalize_angles(arc);

	ERET(0);
}


/*----------------------------------------------*
 *                                              *
 *             GEOMETRIC PROPERTIES             *
 *                                              *
 *----------------------------------------------*/


/* geom_arc_point_at_time */

int geom_arc_point_at_time(struct geom_arc *arc, real time, struct geom_point *result)
{
	double t;

	FUNC;
	ITEM_ASSERT(arc, arc);
	RASSERT(time);
	PASSERT(result);

	t = fcoefrng(arc->start, arc->end, time);
	RTEST_NUM(result->x = cos(t) * arc->radius + arc->center.x);
	RTEST_NUM(result->y = sin(t) * arc->radius + arc->center.y);
	ERET(0);
}


/* geom_arc_derivation_at_time */

int geom_arc_derivation_at_time(struct geom_arc *arc, real time, struct geom_vector *result)
{
	double t;

	FUNC;
	ITEM_ASSERT(arc, arc);
	RASSERT(time);
	PASSERT(result);

	t = fcoefrng(arc->start, arc->end, time);
	RTEST_NUM(result->dx = -sin(t) * arc->radius);
	RTEST_NUM(result->dy = cos(t) * arc->radius);
	ERET(0);
}


