#ifndef _VRR_GEOMLIB_CIRCLE_H_
#define _VRR_GEOMLIB_CIRCLE_H_

/** @file
 *
 * GEOMLIB - Circles and arcs.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */


#include "geomlib/curve.h"
#include "geomlib/vector.h"


/* FIXME: vetsina veci asi nebude fungovat */


/*----------------------------------------------*
 *                                              *
 *                   ARC CLASS                  *
 *                                              *
 *----------------------------------------------*/


#define geom_arc_VMT    geom_curve_VMT
#define geom_arc_INIT   geom_curve_INIT \
	\
	.alength = (real (*)(void *)) &geom_arc_alength, \
	.time_to_atime = (real (*)(void *, real)) &geom_arc_time_to_atime, \
	.atime_to_time = (real (*)(void *, real)) &geom_arc_atime_to_time, \
	.time_to_btime = (real (*)(void *, struct geom_expansion *, real)) &geom_arc_time_to_btime, \
	.btime_to_time = (real (*)(void *, struct geom_expansion *, real)) &geom_arc_btime_to_time, \
	\
	.transform = (int (*)(void *, struct geom_transform *, void *)) &geom_arc_transform, \
	\
	.point_at_time = (int (*)(void *, real, struct geom_point *)) &geom_arc_point_at_time, \
	.derivation_at_time = (int (*)(void *, real, struct geom_vector *)) &geom_arc_derivation_at_time, \

GEOM_CLASS_HEAD(arc, curve);


/*----------------------------------------------*
 *                                              *
 *              INITIALIZATION                  *
 *                                              *
 *----------------------------------------------*/


static inline void geom_arc_create(struct geom_arc *arc)
{ geom_curve_create(&arc->curve); }

static inline void geom_arc_destroy(struct geom_arc *arc)
{ geom_curve_destroy(&arc->curve); }


/**
 * Sets arc from given center point, radius and endpoints angles.
 *
 * @ao  arc     Resulting arc (initialized).
 * @ai  center  Center point.
 * @ai  radius  Radius (>=0).
 * @ai  start   Angle of starting point.
 * @ai  end     Angle of ending point.
 *
 * @rv  zero  Always.
 */
void geom_arc_set(struct geom_arc *arc, struct geom_point *center, real radius, real start, real end);


/**
 * Sets entire circle from given center point, and radius.
 *
 * @ao  arc     Resulting arc (initialized).
 * @ai  center  Center point.
 * @ai  radius  Radius (>=0).
 *
 * @rv  zero  Always.
 */
static inline void geom_arc_set_closed(struct geom_arc *arc, struct geom_point *center, real radius)
{ geom_arc_set(arc, center, radius, 0, INFINITY); }


/**
 * Sets new endpoints to circle or circular arc.
 *
 * @aio arc    Circle or circular arc.
 * @ai  start  Point very near to ellipse (or NULL).
 * @ai  end    Point very near to ellipse (or NULL).
 * @ai  mid    Point very near to ellipse (or NULL).
 * @ai  flags  Flags describing the arc to select (see GEOM_CONIC_ARC_...).
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_arc_set_endpoints(struct geom_arc *arc, struct geom_point *start, struct geom_point *end, struct geom_point *mid, uns flags);


/**
 * Sets the circle with given center point going through another point.
 *
 * @ao  arc     Resulting circle (initialized).
 * @ai  center  Center point.
 * @ai  point   Point on the circle.
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_arc_from_center_point(struct geom_arc *arc, struct geom_point *center, struct geom_point *point);


/**
 * Sets the Thales circle for two given points.
 *
 * @ao  arc     Resulting circle (initialized).
 * @ai  point1  First point.
 * @ai  point2  Second point.
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_arc_from_2_points_thales(struct geom_arc *arc, struct geom_point *point1, struct geom_point *point2);


/**
 * Sets the circle going through given three points.
 *
 * If points are collinear, largest Thales circle of two points is returned.
 *
 * @ao  arc     Resulting circle (initialized).
 * @ai  point1  First point.
 * @ai  point2  Second point.
 * @ai  point3  Third point.
 */
int geom_arc_from_3_points(struct geom_arc *arc, struct geom_point *point1, struct geom_point *point2, struct geom_point *point3);


/*----------------------------------------------*
 *                                              *
 *           OVERRIDED VIRTUAL METHODS          *
 *                                              *
 *----------------------------------------------*/


static inline void geom_arc_after_change(struct geom_arc *arc)
{ geom_curve_after_change(&arc->curve); }


/* Parametrizations */

real geom_arc_alength(struct geom_arc *arc);

real geom_arc_time_to_atime(struct geom_arc *arc, real time);
real geom_arc_atime_to_time(struct geom_arc *arc, real atime);

real geom_arc_time_to_btime(struct geom_arc *arc, struct geom_expansion *expansion, real time);
real geom_arc_btime_to_time(struct geom_arc *arc, struct geom_expansion *expansion, real btime);


/* Transformation */

int geom_arc_transform(struct geom_arc *arc, struct geom_transform *transform, struct geom_arc *result);


/* Geometric properties */

int geom_arc_point_at_time(struct geom_arc *arc, real time, struct geom_point *result);
int geom_arc_derivation_at_time(struct geom_arc *arc, real time, struct geom_vector *result);


/*----------------------------------------------*
 *                                              *
 *                  CONVERSION                  *
 *                                              *
 *----------------------------------------------*/

/**
 * Converts circle to form Ax^2 + Bxy + Cy^2 + Dx + Ey + F = 0.
 *
 * @ai  arc   Arc to convert.
 * @ao  poly  Resulting coefficents.
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_arc_to_poly(struct geom_arc *arc, real poly[6]);


/*----------------------------------------------*
 *                                              *
 *                TRANSFORMATIONS               *
 *                                              *
 *----------------------------------------------*/


/**
 * Normalize arc so that starting point angle is in interval [0, 2 * PI].
 *
 * @aio  arc  Arc to normalize.
 *
 * @rv  zero  Always.
 */
int geom_arc_normalize(struct geom_arc *arc);


/*----------------------------------------------*
 *                                              *
 *             GEOMETRIC PROPERTIES             *
 *                                              *
 *----------------------------------------------*/


/**
 * Returns whether the arc is singular (zero radius).
 *
 * @ai  arc  Arc.
 *
 * @rv  nonzero  Arc is singular.
 * @rv  zero     Arc is not singular.
 */
static inline int geom_arc_is_singular(struct geom_arc *arc)
{ return arc->radius == 0; }



#endif

