/*
 * GEOMLIB - Elementary curves ancestor.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#define GEOM_SOURCE
#include "geomlib/fpath.h"
#include "geomlib/bezier.h"


/*----------------------------------------------*
 *                                              *
 *           PARAMETRIZATIONS LENGTHS           *
 *                                              *
 *----------------------------------------------*/


/* geom_item_alength__from_atime */

real geom_item_alength__from_atime(void *item)
{
	FUNC;
	RRET(geom_item_time_to_atime(item, REAL_MAX));
}


/* geom_item_blength__expand */

int geom_item_blength__expand(void *item)
{
	struct geom_expansion *e;
	uns tlen;

	FUNC;
	PTEST_ERRNO(e = geom_item_expand(item));
	tlen = geom_group_count(&e->path);
	geom_expansion_unref(e);
	ERET(tlen);
}


/*----------------------------------------------*
 *                                              *
 *       PARAMETRIZATIONS:  TIME <-> ATIME      *
 *                                              *
 *----------------------------------------------*/


/* geom_item_time_to_atime__expand */

real geom_item_time_to_atime__expand(void *item, real time)
{
	struct geom_expansion *e;
	real t;

	FUNC;
	PTEST_NAN(e = geom_item_expand(item));
	if (RTEST(t = geom_item_time_to_btime(item, e, time)))
		t = geom_path_time_to_atime(&e->path.path, t);
	geom_expansion_unref(e);
	RRET(t);
}


/* geom_item_times_to_atimes__cycle */

int geom_item_times_to_atimes__cycle(void *item, uns count, real *times, real *atimes)
{
	uns i;
	real t = 0;

	FUNC;
	for (i = 0; i < count; i++) {
		RTEST_NUM(atimes[i] = geom_item_time_to_atime(item, times[i]));
		if (likely(atimes[i] >= t)) t = atimes[i];
		else atimes[i] = t;
	}
	GEOM_REALS_SORTED_ASSERT(count, atimes);
	ERET(0);
}


/* geom_item_atime_to_time__expand */

real geom_item_atime_to_time__expand(void *item, real atime)
{
	struct geom_expansion *e;
	real t;

	FUNC;
	PTEST_NAN(e = geom_item_expand(item));
	if (RTEST(t = geom_path_atime_to_time(&e->path.path, atime)))
		t = geom_item_btime_to_time(item, e, t);
	geom_expansion_unref(e);
	RRET(t);
}


/* geom_item_atimes_to_times__cycle */

int geom_item_atimes_to_times__cycle(void *item, uns count, real *atimes, real *times)
{
	uns i;
	real t = 0;

	FUNC;
	for (i = 0; i < count; i++) {
		RTEST_NUM(times[i] = geom_item_atime_to_time(item, atimes[i]));
		if (likely(times[i] >= t)) t = times[i];
		else times[i] = t;
	}
	GEOM_REALS_SORTED_ASSERT(count, times);
	ERET(0);
}


/*----------------------------------------------*
 *                                              *
 *       PARAMETRIZATIONS:  TIME <-> RATIME     *
 *                                              *
 *----------------------------------------------*/


/* geom_item_time_to_ratime__from_atime */

real geom_item_time_to_ratime__from_atime(void *item, real time)
{
	real t, l;

	FUNC;
	if (time <= 0)
		RRET(0);
	if (time >= 1)
		RRET(1);

	RTEST_NAN(t = geom_item_time_to_atime(item, time));
	if (t <= 0)
		RRET(0);
	RTEST_NAN(l = geom_item_alength(item));
	if (t >= l)
		RRET(1);
	RRET(RASSERTX(t / l));
}


/* geom_item_times_to_ratimes__cycle */

int geom_item_times_to_ratimes__cycle(void *item, uns count, real *times, real *ratimes)
{
	uns i;
	real t = 0;

	FUNC;
	for (i = 0; i < count; i++) {
		RTEST_NUM(ratimes[i] = geom_item_time_to_ratime(item, times[i]));
		if (likely(ratimes[i] >= t)) t = ratimes[i];
		else ratimes[i] = t;
	}
	GEOM_REALS_SORTED_ASSERT(count, ratimes);
	ERET(0);
}


/* geom_item_ratime_to_time__from_atime */

real geom_item_ratime_to_time__from_atime(void *item, real ratime)
{
	real t;

	FUNC;
	if (ratime <= 0)
		RRET(0);
	if (ratime >= 1)
		RRET(1);
	RTEST_NAN(t = ratime * geom_item_alength(item));
	RRET(geom_item_atime_to_time(item, t));
}


/* geom_item_ratimes_to_times__cycle */

int geom_item_ratimes_to_times__cycle(void *item, uns count, real *ratimes, real *times)
{
	uns i;
	real t = 0;

	FUNC;
	for (i = 0; i < count; i++) {
		RTEST_NUM(times[i] = geom_item_ratime_to_time(item, ratimes[i]));
		if (likely(times[i] >= t)) t = times[i];
		else times[i] = t;
	}
	GEOM_REALS_SORTED_ASSERT(count, times);
	ERET(0);
}


/*----------------------------------------------*
 *                                              *
 *       PARAMETRIZATIONS:  TIME <-> BTIME      *
 *                                              *
 *----------------------------------------------*/


/* geom_item_time_to_btime__from_atime */

real geom_item_time_to_btime__from_atime(void *item, struct geom_expansion *expansion, real time)
{
	real t;

	FUNC;
	RTEST_NAN(t = geom_item_time_to_atime(item, time));
	RRET(geom_path_atime_to_time(&expansion->path.path, t));
}


/* geom_item_times_to_btimes__copy */

int geom_item_times_to_btimes__copy(void UNUSED *item, struct geom_expansion UNUSED *expansion, uns count, real *times, real *btimes)
{
	FUNC;
	if (times != btimes)
		memcpy(btimes, times, count * sizeof(real));
	ERET(0);
}


/* geom_item_times_to_btimes__cycle */

int geom_item_times_to_btimes__cycle(void *item, struct geom_expansion *expansion, uns count, real *times, real *btimes)
{
	uns i;
	real t = 0;

	FUNC;
	for (i = 0; i < count; i++) {
		RTEST_NUM(btimes[i] = geom_item_time_to_btime(item, expansion, times[i]));
		if (likely(btimes[i] >= t)) t = btimes[i];
		else btimes[i] = t;
	}
	GEOM_REALS_SORTED_ASSERT(count, btimes);
	ERET(0);
}


/* geom_item_times_to_btimes__from_atimes */

int geom_item_times_to_btimes__from_atimes(void *item, struct geom_expansion UNUSED *expansion, uns count, real *times, real *btimes)
{
	FUNC;
	ETEST_RET(geom_item_times_to_atimes(item, count, times, btimes));
	ETEST_RET(geom_path_atimes_to_times(&expansion->path.path, count, btimes, btimes));
	GEOM_REALS_SORTED_ASSERT(count, btimes);
	ERET(0);
}


/* geom_item_btime_to_time__from_atime */

real geom_item_btime_to_time__from_atime(void *item, struct geom_expansion UNUSED *expansion, real btime)
{
	real t;

	FUNC;
	RTEST_NAN(t = geom_path_time_to_atime(&expansion->path.path, btime));
	RRET(geom_item_atime_to_time(item, t));
}


/* geom_item_btimes_to_times__cycle */

int geom_item_btimes_to_times__cycle(void *item, struct geom_expansion *expansion, uns count, real *btimes, real *times)
{
	uns i;
	real t = 0;

	FUNC;
	for (i = 0; i < count; i++) {
		RTEST_NUM(times[i] = geom_item_btime_to_time(item, expansion, btimes[i]));
		if (likely(times[i] >= t)) t = times[i];
		else times[i] = t;
	}
	GEOM_REALS_SORTED_ASSERT(count, times);
	ERET(0);
}


/* geom_item_btimes_to_times__copy */

int geom_item_btimes_to_times__copy(void UNUSED *item, struct geom_expansion UNUSED *expansion, uns count, real *btimes, real *times)
{
	FUNC;
	if (btimes != times)
		memcpy(times, btimes, count * sizeof(real));
	ERET(0);
}


/* geom_item_btimes_to_times__from_atimes */

int geom_item_btimes_to_times__from_atimes(void *item, struct geom_expansion UNUSED *expansion, uns count, real *btimes, real *times)
{
	FUNC;
	ETEST_RET(geom_path_times_to_atimes(&expansion->path.path, count, btimes, times));
	ETEST_RET(geom_item_atimes_to_times(item, count, times, times));
	GEOM_REALS_SORTED_ASSERT(count, times);
	ERET(0);
}


/*----------------------------------------------*
 *                                              *
 *                   EXPANSIONS                 *
 *                                              *
 *----------------------------------------------*/


 #if 0
/* geom_item_expand */

struct geom_expansion *geom_item_expand(void *item)
{
	int err;
	struct geom_expansion *expansion;

	FUNC;
	ITEM_ASSERT(item, item);

	expansion = xmalloc_zero(sizeof(struct geom_expansion));
	GEOM_INSTANCE_ZERO_CREATE(&expansion->path, fpath, geom_fpath_create);
	if (ETEST(err = geom_item_expansion_append(item, &expansion->path))) {
		geom_fpath_finish(&expansion->path);
		expansion->ref = 1;
		PRET(expansion);
	}
	else {
		GEOM_INSTANCE_DESTROY2(&expansion->path, fpath);
		xfree(expansion);
		PRET(NULL);
	}
}


/* geom_expansion_unref */

void geom_expansion_unref(struct geom_expansion *expansion)
{
	FUNC;
	GASSERT(expansion, expansion);
	BASSERT(expansion->ref > 0);

	expansion->ref--;
	if (expansion->ref == 0) {
		GEOM_INSTANCE_DESTROY2(&expansion->path, fpath);
		xfree(expansion);
	}
	RET;
}
#endif

/* geom_old_expand */

int geom_old_expand(void *item, struct geom_path *path)
{
	struct geom_expansion *e;
	if (!GEOM_PTEST(e = geom_item_expand(item)))
		return geom_errno;
	geom_group_free_all(path);
	geom_item_copy(path, &e->path);
	geom_expansion_unref(e);
	return 0;
}


/*----------------------------------------------*
 *                                              *
 *     CONVERSION TO OTHER TYPES OF CURVES      *
 *                                              *
 *----------------------------------------------*/


/* geom_item_to_bezier__expand */

int geom_item_to_bezier__expand(void *item, struct geom_bezier *bezier)
{
	struct geom_expansion *e;
	int err;

	FUNC;
	PTEST_ERRNO(e = geom_item_expand(item));
	if (geom_group_count(&e->path) == 1) {
		geom_bezier_copy(bezier, (struct geom_bezier *)geom_group_head((struct geom_group *)&e->path));
		err = 0;
	}
	else
		err = GEOM_ERR_CONVERSION_IMPOSSIBLE;
	geom_expansion_unref(e);
	ERET(err);
}


/* geom_item_conversion_append__expand */

int geom_item_conversion_append__expand(void *item, uns flags, struct geom_fpath *result)
{
	struct geom_expansion *e;
	int err;
	
	FUNC;
	PTEST_ERRNO(e = geom_item_expand(item));
	err = geom_path_conversion_append(&e->path.path, flags, result);
	geom_expansion_unref(e);
	ERET(err);
}


/*----------------------------------------------*
 *                                              *
 *        CURVE POINTS AND DERIVATIONS          *
 *                                              *
 *----------------------------------------------*/


/* geom_item_point_start__from_time */

int geom_item_point_start__from_time(void *item, struct geom_point *result)
{
	FUNC;
	ITEM_ASSERT(item, item);
	PASSERT(result);

	ERET(geom_item_point_at_time(item, 0, result));
}


/* geom_item_point_end__from_time */

int geom_item_point_end__from_time(void *item, struct geom_point *result)
{
	FUNC;
	ERET(geom_item_point_at_time(item, REAL_MAX, result));
}


/* geom_item_point_at_time__expand */

int geom_item_point_at_time__expand(void *item, real time, struct geom_point *result)
{
	struct geom_expansion *e;
	real t;
	int err;

	FUNC;
	PTEST_ERRNO(e = geom_item_expand(item));
	if (!RTEST(t = geom_item_time_to_btime(item, e, time)))
		err = GEOM_ERR_NUMERIC;
	else
		err = geom_path_point_at_time(&e->path.path, t, result);
	geom_expansion_unref(e);
	ERET(err);
}


/* geom_item_derivation_start__from_time */

int geom_item_derivation_start__from_time(void *item, struct geom_vector *result)
{
	FUNC;
	ERET(geom_item_derivation_at_time(item, 0, result));
}


/* geom_item_derivation_end__from_time */

int geom_item_derivation_end__from_time(void *item, struct geom_vector *result)
{
	FUNC;
	ERET(geom_item_derivation_at_time(item, REAL_MAX, result));
}


/* geom_item_derivation_at_time__expand */

int geom_item_derivation_at_time__expand(void *item, real time, struct geom_vector *result)
{
	struct geom_expansion *e;
	real t;
	int err;

	FUNC;
	PTEST_ERRNO(e = geom_item_expand(item));
	if (!RTEST(t = geom_item_time_to_btime(item, e, time)))
		err = GEOM_ERR_NUMERIC;
	else
		err = geom_path_derivation_at_time(&e->path.path, t, result);
	geom_expansion_unref(e);
	ERET(err);
}


/*----------------------------------------------*
 *                                              *
 *         OTHER GEOMETRIC PROPERTIES           *
 *                                              *
 *----------------------------------------------*/


/* geom_item_bbox__expand */

int geom_item_bbox__expand(void *item, struct geom_rectangle *result)
{
	struct geom_expansion *e;
	int err;

	FUNC;
	PTEST_ERRNO(e = geom_item_expand(item));
	err = geom_path_bbox(&e->path.path, result);
	geom_expansion_unref(e);
	ERET(err);
}


/* geom_item_nearest_to_point__expand */

int geom_item_nearest_to_point__expand(void *item, struct geom_point *point, uns flags, struct geom_nearest *nearest)
{
	struct geom_expansion *e;
	int err;

	FUNC;
	PTEST_ERRNO(e = geom_item_expand(item));
	if (ETEST(err = geom_path_nearest_to_point(&e->path.path, point, flags, nearest)) && (flags & GEOM_NEAREST_TIME))
		if (!RTEST(nearest->time = geom_item_btime_to_time(item, e, nearest->time)))
			err = GEOM_ERR_NUMERIC;
	geom_expansion_unref(e);
	ERET(err);
}


/* geom_item_direction_times__expand */

int geom_item_direction_times__expand(void *item, struct geom_vector *dir, struct garr *result)
{
	struct geom_expansion *e;
	real *times;
	int err, n;

	FUNC;
	PTEST_ERRNO(e = geom_item_expand(item));
	if ((n = geom_path_direction_times(&e->path.path, dir, result)) > 0) {
		times = (real *)garr_get(result, 0);
		if (!ETEST(err = geom_item_btimes_to_times(item, e, n, times, times)))
			n = err;
	}
	geom_expansion_unref(e);
	ERET(n);
}


/* geom_item_distance_times__expand */

int geom_item_distance_times__expand(void *item, struct geom_point *point, real dist, struct garr *result)
{
	struct geom_expansion *e;
	real *times;
	int err, n;

	FUNC;
	PTEST_ERRNO(e = geom_item_expand(item));
	if ((n = geom_path_distance_times(&e->path.path, point, dist, result)) > 0) {
		times = (real *)garr_get(result, 0);
		if (!ETEST(err = geom_item_btimes_to_times(item, e, n, times, times)))
			n = err;
	}
	geom_expansion_unref(e);
	ERET(n);
}


/*----------------------------------------------*
 *                                              *
 *                  INTERSECTIONS               *
 *                                              *
 *----------------------------------------------*/


/* geom_items_intersections */

#define ASORT_PREFIX(x) geom_items_intersections_sort
#define ASORT_KEY_TYPE int
#define ASORT_ELT(i) (perm[(i)])
#define ASORT_LT(x, y) (times[x] < times[y])
#define ASORT_EXTRA_ARGS , uns *perm, real *times
#include "lib/arraysort.h"

int geom_items_intersections(void *item1, void *item2, uns flags, struct garr *result)
{
	struct geom_expansion *expansion1 = NULL, *expansion2 = NULL;
	struct geom_path *path1, *path2;
	int convert1, convert2;
	uns i, n;
	struct geom_intersection *inter;

	FUNC;
	ITEM_ASSERT(item, item1);
	ITEM_ASSERT(item, item2);
	BASSERT(!(flags & ~GEOM_INTERSECTION_ALL));
	PASSERT(result);

	/* Special cases */
	if (geom_instance_class(item1) == CLASS(bezier) && geom_instance_class(item2) == CLASS(bezier))
		ERET(geom_beziers_intersections(item1, item2, flags, result));
	if ((ITEM_FLAGS(item1) & GEOM_ITEM_PATH) && (ITEM_FLAGS(item2) & GEOM_ITEM_PATH))
		ERET(geom_paths_intersections(item1, item2, flags, result));

	/* Expand if necessary */
	if (ITEM_FLAGS(item1) & GEOM_ITEM_PATH)
		path1 = item1;
	else {
		expansion1 = geom_item_expand(item1);
		path1 = &expansion1->path.path;
	}
	if (ITEM_FLAGS(item2) & GEOM_ITEM_PATH)
		path2 = item2;
	else {
		expansion2 = geom_item_expand(item2);
		path2 = &expansion2->path.path;
	}

	/* Main code */
	if ((geom_errno = geom_paths_intersections(path1, path2, flags, result)) > 0) {
		convert1 = flags & GEOM_INTERSECTION_TIME1 && expansion1 && geom_item_time_dif_btime(item1);
		convert2 = flags & GEOM_INTERSECTION_TIME2 && expansion2 && geom_item_time_dif_btime(item2);

		/* BTIME -> TIME correction needed */
		if (convert1 || convert2)
		{
			if ((n = geom_errno) == 1) {
				inter = (struct geom_intersection *)garr_get(result, 0);
				if (convert1)
					if (!RTEST(inter->time1 = geom_item_btime_to_time(item1, expansion1, inter->time1))) {
						geom_errno = GEOM_ERR_NUMERIC; goto EXIT; }
				if (convert2)
					if (!RTEST(inter->time2 = geom_item_btime_to_time(item2, expansion2, inter->time2))) {
						geom_errno = GEOM_ERR_NUMERIC; goto EXIT; }
			}
			else {
				real times[n];
				if (convert1) {
					for (i = 0; i < n; i++)
						times[i] = ((struct geom_intersection *)garr_get(result, i))->time1;
					if (!ETEST(geom_errno = geom_item_btimes_to_times(item1, expansion1, n, times, times)))
						goto EXIT;
					for (i = 0; i < n; i++)
						((struct geom_intersection *)garr_get(result, i))->time1 = times[i];
				}
				if (convert2) {
					uns perm[n];
					for (i = 0; i < n; i++) {
						times[i] = ((struct geom_intersection *)garr_get(result, i))->time2;
						perm[i] = i;
					}
					geom_items_intersections_sort(n, perm, times);
					for (i = 0; i < n; i++)
						times[i] = ((struct geom_intersection *)garr_get(result, perm[i]))->time2;
					if (!ETEST(geom_errno = geom_item_btimes_to_times(item2, expansion2, n, times, times)))
						goto EXIT;
					for (i = 0; i < n; i++)
						((struct geom_intersection *)garr_get(result, perm[i]))->time2 = times[i];
				}
				geom_errno = n;
			}
		}
	}
EXIT:
	if (expansion1)
		geom_expansion_unref(expansion1);
	if (expansion2)
		geom_expansion_unref(expansion2);
	RET_ERRNO;
}


/*----------------------------------------------*
 *                                              *
 *             GEOMETRIC PROPERTIES             *
 *                                              *
 *----------------------------------------------*/


/* geom_curve_solve_y */

/*int geom_curve_solve_y(uns type, void *ptr, real min_time, real max_time, real y, uns flags, struct geom_curve_point *result)
{

	switch (type) {
		case GEOM_CURVE_PATH: return geom_path_solve_y(ptr, min_time, max_time, y, flags, result);
		case GEOM_CURVE_BEZIER: return geom_bezier_solve_y(ptr, min_time, max_time, y, flags, result);

		default: ;
			struct geom_path expansion;
			int err;
			real time1, time2;

			FUNC;

			GASSERTX(e,geom_path_init(&expansion));
			if (ETEST(err = geom_curve_expand(type, ptr, 0, &expansion))) {
				if (!RTEST(time1 = geom_curve_time_to_btime(type, ptr, &expansion, min_time)) ||
					!RTEST(time2 = geom_curve_time_to_btime(type, ptr, &expansion, max_time)))
					err = GEOM_ERR_NUMERIC;
				else if (ETEST(err = geom_path_solve_y(&expansion, time1, time2, y, flags, result))) {
					if (flags & GEOM_CURVE_POINT_TIME)
						if (!RTEST(result->time = geom_curve_btime_to_time(type, ptr, &expansion, result->time)))
							err = GEOM_ERR_NUMERIC;
				}
			}
			GASSERTX(e,geom_path_destroy(&expansion));
			ERET(err);
	}
}


*/
