#ifndef _VRR_GEOMLIB_H_
#define _VRR_GEOMLIB_H_

/** @file
 *
 * GEOMLIB - Main geomlib header files.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "geomlib/base.h"
#include "geomlib/item.h"
#include "geomlib/group.h"
#include "geomlib/numeric.h"
#include "geomlib/polynomial.h"
#include "geomlib/bernstein.h"
#include "geomlib/transform.h"
#include "geomlib/transform2.h"
#include "geomlib/bezier.h"
#include "geomlib/vector.h"
#include "geomlib/rectangle.h"
#include "geomlib/rtree.h"
#include "geomlib/approx.h"
#include "geomlib/matrix.h"
#include "geomlib/segment.h"
#include "geomlib/circle.h"
#include "geomlib/ellipse.h"
#include "geomlib/nurbs.h"
#include "geomlib/conic.h"
#include "geomlib/curve.h"
#include "geomlib/path.h"
#include "geomlib/fpath.h"
#include "geomlib/pathset.h"
#include "geomlib/special.h"
#include "geomlib/cache.h"
#include "geomlib/pattern.h"


/**
 * @defgroup geomlib Geomlib
 *
 * Geometric library.
 */

/**
 * @defgroup geomlib_conventions Calling Conventions
 * @ingroup geomlib
 *
 * <h2>Errors</h2>
 *
 * Many functions in \ref geomlib "geomlib" can fail because of
 * floating point error or different reason.
 *
 * There are two groups of possibly failing routines:
 * -# If a function returns integer type, zero or possitive result means success, while
 *    negative result means one of defined \ref geomlib_errors "error codes".
 * -# If a function returns floating point type, it success, when result is
 *    a finite number, or it fails otherwise.
 *
 * <h2>Input and ouput</h2>
 *
 * If nothing else is said in description, all parameters
 * must contain valid data (floating point types must contain finite numbers,
 * pointers must not be NULL and structures must agree with their definitions).
 *
 * Successfuly ended functions always returns valid data or specially defined results.
 * Pointers to resulting structures should not overlap with any other parameters.
 * Resulting data of failed function is undefined.
 */

/**
 * @defgroup geomlib_errors Error Codes
 * @ingroup geomlib
 *
 * List of error codes:
 */

/**
 * @defgroup geomlib_curves Curves
 * @ingroup geomlib
 */

/**
 * @defgroup geomlib_curves_path General Paths
 * @ingroup geomlib_curves
 * @sa struct geom_path2
 */

/**
 * @defgroup geomlib_curves_segment Segments
 * @ingroup geomlib_curves
 * @sa struct geom_segment
 */

/**
 * @defgroup geomlib_curves_bezier Rational Bezier Curves
 * @ingroup geomlib_curves
 *
 * Rational Bezier curves are the most universal curves in geomlib
 * and any other defined curves can be converted to them.
 *
 * <h2>Definition</h2>
 *
 * General rational Bezier curve is represented by
 * @f[P(t) = {\sum_{i=0}^n w_i B_{i,n}(t) P_i \over \sum_{i=0}^n w_i B_{i,n}(t)}.@f]
 *
 * Terminology:
 * - The number <i>n</i> is called <i>curve degree</i> (values up to 3 are supported).
 * - @f$B_{i,n}@f$ (<i>i</i> = 0,...,<i>n</i>) are <i>Bernstein polynomials</i> of degree <i>n</i> and are defined by
 *   @f[B_{i,n}(t) = {n! \over i!(n-i)!} (1-t)^{n-i} t^i.@f]
 * - The points @f$P_i@f$ are called <i>control points</i> of the Bezier curve.
 * - The scalars @f$w_i@f$ are called the <i>weights</i> (only positive weights are supported).
 *
 * <h2>Implementation</h2>
 *
 * @sa struct geom_bezier
 *
 */

/**
 * @defgroup geomlib_curves_arc Circular Arcs
 * @ingroup geomlib_curves
 *
 * @sa struct geom_arc
 */

/**
 * @defgroup geomlib_curves_elliptic_arc Elliptic Arcs
 * @ingroup geomlib_curves
 *
 * @sa struct geom_elliptic_arc
 */

/**
 * @defgroup geomlib_curves_parabolic_arc Parabolic Arcs
 * @ingroup geomlib_curves
 *
 * @sa struct geom_parabolic_arc
 */

/**
 * @defgroup geomlib_curves_hyperbolic_arc Hyperbolic Arcs
 * @ingroup geomlib_curves
 *
 * @sa struct geom_hyperbolic_arc
 */

/**
 * @defgroup geomlib_curves_nurbs NURBS
 * @ingroup geomlib_curves
 *
 * \todo Not implemented, yet :)
 */

/*----------------------------------------------*/

/**
 * @defgroup geomlib_param Parametrizations
 * @ingroup geomlib
 */

/**
 * @defgroup geomlib_param_time TIME
 * @ingroup geomlib_param
 *
 * Primary curves parametrization.
 *
 * It is defined for each type of curve to allow
 * fast internal calculations.
 * All curves excepts paths should be parametrized
 * by interval [0, 1].
 */

/**
 * @defgroup geomlib_param_btime BTIME
 * @ingroup geomlib_param
 *
 * Bezier expansion parametrization.
 *
 * Equals to TIME parametrization of
 * Bezier expansion of the curve.
 */

/**
 * @defgroup geomlib_param_atime ATIME
 * @ingroup geomlib_param
 *
 * Euclidean arc length parametrization.
 *
 * Parametrization by arc length from starting point of the curve.
 * Corresponds to interval [0, alength], where alength is Euclidean
 * arc length of the curve.
 */

/**
 * @defgroup geomlib_param_ratime RATIME
 * @ingroup geomlib_param
 *
 * Relative Euclidean arc length parametrization.
 *
 * Parametrization by arc length from starting point
 * divided by arc length of entire curve.
 * Corresponds to interval [0, 1].
 */


#endif
