/** @file
 *
 * GEOMLIB - Geomlib functionality testing.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */


#include "lib/lib.h"
#include "geomlib/geomlib.h"
#include "lib/cache.h"
#include <stdlib.h>


#define COUNT 20


/*----------------------------------------------*
 *                                              *
 *                  DEFINITIONS                 *
 *                                              *
 *----------------------------------------------*/


#define SMALL_EPS 0.0001
#define MED_EPS 0.001
#define LARGE_EPS 0.1
double EPS = MED_EPS;

#define ENTER(x) do{double _save_eps=EPS;test_enter(x);test_depth++;
#define LEAVE test_depth--;EPS=_save_eps;}while(0)

static uns test_depth = 0;

static void indent(void)
{
	for (uns i = 0; i < test_depth; i++)
		msg("  ");
}

static void test_enter(const char *s)
{
	indent();
	msg("Testing `%s'\n", s);
}

#define WARN(...) do{ indent(); msg("Warning: "); msg(__VA_ARGS__); msg("\n"); }while(0)

static void warn_test(int line, const char *s)
{
	WARN("Test `%s' failed at line %d", s, line);
}

static void warn_eps(int line, const char *s, real eps)
{
	WARN("%f error in `%s' at line %d", eps, s, line);
}

static void warn_eps2(int line, const char *s, real eps1, real eps2)
{
	WARN("(%f, %f) error in `%s' at line %d", eps1, eps2, s, line);
}

static void test(int line, const char *s, int x)
{
	if (!x)
		warn_test(line, s);
}

static void etest(int line, const char *s, int x)
{
	if (!GEOM_ETEST(x))
		warn_test(line, s);
}

UNUSED static void test_zero(int line, const char *s, real x)
{
	if (!(x > -EPS && x < EPS))
		warn_eps(line, s, x);
}

static void test_eq(int line, const char *s, real x, real y)
{
	y = x - y;
	if (!(y > -EPS && y < EPS))
		warn_eps(line, s, y);
}

static void test_range(int line, const char *s, real x, real min, real max)
{
	if (!(x > min - EPS))
		warn_eps(line, s, x - min);
	else if (!(x < max + EPS))
		warn_eps(line, s, x - max);
}

static void test_pt_eq(int line, const char *s, struct geom_point x, struct geom_point y)
{
	y.x = x.x - y.x;
	y.y = x.y - y.y;
	if (!(y.x > -EPS && y.x < EPS && y.y > -EPS && y.y < EPS))
		warn_eps2(line, s, y.x, y.y);
}

static void test_vec_eq(int line, const char *s, struct geom_vector x, struct geom_vector y)
{
	y.dx = x.dx - y.dx;
	y.dy = x.dy - y.dy;
	if (!(y.dx > -EPS && y.dx < EPS && y.dy > -EPS && y.dy < EPS))
		warn_eps2(line, s, y.dx, y.dy);
}

#define TEST(x) do{test(__LINE__,#x,(x));}while(0)
#define ETEST(x) do{etest(__LINE__,#x,(x));}while(0)
#define TEST_ZERO(x) do{test_zero(__LINE__,"["#x"] = 0",x);}while(0)
#define TEST_EQ(x,y) do{test_eq(__LINE__,"["#x"] = ["#y"]",x,y);}while(0)
#define TEST_RANGE(x,min,max) do{test_range(__LINE__,"["#x"] in (["#min"],["#max"])",x,min,max);}while(0)
#define TEST_PT_EQ(x,y) do{test_pt_eq(__LINE__,"["#x"] = ["#y"]",x,y);}while(0)
#define TEST_VEC_EQ(x,y) do{test_vec_eq(__LINE__,"["#x"] = ["#y"]",x,y);}while(0)

static real rrand(real min, real max)
{ return fcoefrngr(min, max, (rand() % 1000000001) / 1000000000.0); }

static real rand_coord(void)
{ return rrand(-100, 100); }

static struct geom_point rand_point(void)
{ return (struct geom_point){.x = rand_coord(), .y = rand_coord()}; }

static struct geom_point_w rand_point_w(void)
{ return (struct geom_point_w){.x = rand_coord(), .y = rand_coord(), .w = rrand(0.3, 5)}; }


/*----------------------------------------------*
 *                                              *
 *                    BASE                      *
 *                                              *
 *----------------------------------------------*/


static void test_base_v(void) { GEOM_FUNC; GEOM_IN(); GEOM_OUT(); GEOM_RET; }
static int test_base_e(void) { GEOM_FUNC; test_base_v(); GEOM_ERET(12); }
static real test_base_r(void) { GEOM_FUNC; test_base_e(); GEOM_RRET(REAL_MAX); }
static double test_base_f(void) { GEOM_FUNC; test_base_r(); GEOM_FRET(DBL_MAX); }

static void test_base(void)
{
	ENTER("Basic definitions") {

		ENTER("Message macros") {

#undef GEOM_DEBUG_TRACE
#define GEOM_DEBUG_TRACE 0
			GEOM_MSG("This should not display: int %d, ptr %p.\n", 1, (void *)&test_base);
			GEOM_TRACE("This should not display: Hello, world.");
			GEOM_WPT;
			GEOM_DUMP_ENTER(0, "This should not display: Hello, world.");
			GEOM_DUMP("This should not display: Hello, world.");
			GEOM_DUMP("This should not display: int %d, ptr %p.", 1, (void *)&test_base);
			GEOM_DUMP_ENTER(0, "This should not display: int %d, ptr %p.", 1, (void *)&test_base);
			GEOM_DUMP("This should not display: Hello, world.");
			GEOM_DUMP("This should not display: int %d, ptr %p.", 1, (void *)&test_base);
			GEOM_DUMP_LEAVE;
			GEOM_DUMP_GROUP;
			GEOM_DUMP("This should not display: Hello, world.");
			GEOM_DUMP("This should not display: int %d, ptr %p.", 1, (void *)&test_base);
			GEOM_DUMP_LEAVE;
			GEOM_DUMP_SEP;
			GEOM_SDUMP("This should not display: Hello, world.");
			GEOM_SDUMP("This should not display: int %d, ptr %p.", 1, (void *)&test_base);
			GEOM_DUMP_LEAVE;

		} LEAVE;

		ENTER("GEOM_*TEST macros") {
			TEST(GEOM_TEST((double)1 == (double)1));
			TEST(GEOM_TEST((void *)&test_base));
			TEST(!GEOM_TEST((double)1 == (double)2));
			TEST(GEOM_PTEST(&test_base));
			TEST(!GEOM_PTEST(NULL));
			TEST(GEOM_RTEST(REAL_MAX));
			TEST(!GEOM_RTEST(2 * (double)REAL_MAX));
			TEST(!GEOM_RTEST(NAN));
			TEST(!GEOM_RTEST(REAL_INF));
			TEST(GEOM_FTEST(DBL_MAX));
			TEST(!GEOM_FTEST(2 * DBL_MAX));
			TEST(!GEOM_FTEST(NAN));
			TEST(!GEOM_FTEST(INFINITY));
			TEST(GEOM_ETEST(0));
			TEST(GEOM_ETEST(1234));
			TEST(!GEOM_ETEST(GEOM_ERR_NUMERIC));
			TEST(GEOM_RTEST_MINMAX(123, 122, 124));
			TEST(GEOM_RTEST_MINMAX(REAL_MAX, REAL_MAX, REAL_MAX));
			TEST(GEOM_RTEST_MINMAX(REAL_INF, REAL_INF, REAL_INF));
			TEST(GEOM_RTEST_MINMAX(0, -REAL_INF, REAL_INF));
			TEST(!GEOM_RTEST_MINMAX(121, 122, 124));
			TEST(!GEOM_RTEST_MINMAX(125, 122, 124));
			TEST(!GEOM_RTEST_MINMAX(2 * REAL_MAX, 122, 124));
			TEST(!GEOM_RTEST_MINMAX(0, REAL_INF, REAL_INF));
			TEST(GEOM_FTEST_MINMAX(123, 122, 124));
			TEST(GEOM_FTEST_MINMAX(DBL_MAX, DBL_MAX, DBL_MAX));
			TEST(GEOM_FTEST_MINMAX(DBL_INF, DBL_INF, DBL_INF));
			TEST(GEOM_FTEST_MINMAX(0, -REAL_INF, REAL_INF));
			TEST(!GEOM_FTEST_MINMAX(121, 122, 124));
			TEST(!GEOM_FTEST_MINMAX(125, 122, 124));
			TEST(!GEOM_FTEST_MINMAX(2 * REAL_MAX, 122, 124));
			TEST(!GEOM_FTEST_MINMAX(0, DBL_INF, DBL_INF));

		} LEAVE;

		ENTER("Assertion macros") {

			GEOM_ASSERT(b, (double)2 == (double)2);
			GEOM_BASSERT((double)2 == (double)2);
			GEOM_BASSERT(3);
			GEOM_BASSERT(&test_base);
			GEOM_ASSERT(p, (void *)&test_base);
			GEOM_PASSERT(&test_base);
			GEOM_ASSERT(r, REAL_MAX);
			GEOM_ASSERT(f, DBL_MAX);
			GEOM_ASSERT(e, 0);
			GEOM_ASSERT(e, 1234);
			GEOM_ASSERT2(r_minmax, 123, 122, 124);
			GEOM_ASSERT2(r_minmax, REAL_MAX, REAL_MAX, REAL_MAX);
			GEOM_ASSERT2(r_minmax, REAL_INF, REAL_INF, REAL_INF);
			GEOM_ASSERT2(r_minmax, 0, -REAL_INF, REAL_INF);
			GEOM_ASSERT2(f_minmax, 123, 122, 124);
			GEOM_ASSERT2(f_minmax, DBL_MAX, DBL_MAX, DBL_MAX);
			GEOM_ASSERT2(f_minmax, DBL_INF, DBL_INF, DBL_INF);
			GEOM_ASSERT2(f_minmax, 0, -DBL_INF, DBL_INF);

			TEST(GEOM_ASSERTX(b, (double)2 == (double)2) == 1);
			TEST(GEOM_BASSERTX((double)2 == (double)2) == 1);
			TEST(GEOM_BASSERTX(3) == 1);
			TEST(GEOM_BASSERTX(&test_base) == 1);
			TEST(GEOM_ASSERTX(p, (void *)&test_base) == (void *)&test_base);
			TEST(GEOM_PASSERTX(&test_base) == (void *)&test_base);
			TEST(GEOM_ASSERTX(r, REAL_MAX) == REAL_MAX);
			TEST(GEOM_ASSERTX(f, DBL_MAX) == DBL_MAX);
			TEST(GEOM_ASSERTX(e, 0) == 0);
			TEST(GEOM_ASSERTX(e, 1234) == 1234);
			TEST(GEOM_ASSERTX2(r_minmax, 123, 122, 124) == 123);
			TEST(GEOM_ASSERTX2(r_minmax, REAL_MAX, REAL_MAX, REAL_MAX) == REAL_MAX);
			TEST(GEOM_ASSERTX2(r_minmax, REAL_INF, REAL_INF, REAL_INF) == REAL_INF);
			TEST(GEOM_ASSERTX2(r_minmax, 0, -REAL_INF, REAL_INF) == 0);
			TEST(GEOM_ASSERTX2(f_minmax, 123, 122, 124) == 123);
			TEST(GEOM_ASSERTX2(f_minmax, DBL_MAX, DBL_MAX, DBL_MAX) == DBL_MAX);
			TEST(GEOM_ASSERTX2(f_minmax, DBL_INF, DBL_INF, DBL_INF) == DBL_INF);
			TEST(GEOM_ASSERTX2(f_minmax, 0, -DBL_INF, DBL_INF) == 0);

		} LEAVE;

		ENTER("Entering and leaving functions") {

			TEST(test_base_f() == DBL_MAX);
			TEST(test_base_r() == REAL_MAX);
			TEST(test_base_e() == 12);
			test_base_v();

		} LEAVE;

	} LEAVE;
}


/*----------------------------------------------*
 *                                              *
 *                   CLASSES                    *
 *                                              *
 *----------------------------------------------*/


static uns class_state;
#define A_CONSTRUCTOR 0x1
#define A_DESTRUCTOR 0x2

struct geom_class_a { struct geom_o o; int i; };

#define geom_class_a_VMT  geom_o_VMT int virtual, virtual2, virtual3;
#define geom_class_a_INIT geom_o_INIT	\
	.destroy = (void (*)(void *)) &geom_class_a_destroy, \
	.virtual = 1, \
	.virtual2 = 3,
GEOM_CLASS_HEAD(class_a, o);

static void geom_class_a_create(struct geom_class_a *a)
{
	TEST(geom_instance_class(a) == GEOM_CLASS(class_a));
	TEST(a->i == 0);
	class_state |= A_CONSTRUCTOR;
}

static void geom_class_a_create2(struct geom_class_a *a, int param)
{
	geom_class_a_create(a);
	TEST(param == 123);
}

static void geom_class_a_destroy(struct geom_class_a *a)
{
	TEST(geom_instance_class(a) == GEOM_CLASS(class_a));
	class_state |= A_DESTRUCTOR;
}

struct geom_class_b { struct geom_class_a a; int b; };

#define geom_class_b_VMT  geom_class_a_VMT
#define geom_class_b_INIT geom_class_a_INIT \
	.destroy = (void (*)(void *))&geom_o_destroy, \
	.virtual = 2,
GEOM_CLASS_HEAD(class_b, class_a);

GEOM_CLASS_DEF(class_a);
GEOM_CLASS_DEF(class_b);


static void test_classes(void)
{
	ENTER("Classes emulation") {

		struct geom_class_a a, *pa;
		struct geom_class_b b;
		struct geom_o o;

		ENTER("VMT initialization") {
			GEOM_CLASS_INIT(class_a);
			GEOM_CLASS_INIT(class_b);
			TEST(GEOM_CLASS(class_a)->class_id != GEOM_CLASS(class_b)->class_id);
			TEST(GEOM_CLASS(o)->class_parent == NULL);
			TEST(GEOM_CLASS(class_a)->class_parent == GEOM_CLASS(o));
			TEST(GEOM_CLASS(class_b)->class_parent == GEOM_CLASS(class_a));
			TEST(GEOM_CLASS_VMT(class_a, virtual3) == 0);
		} LEAVE;

		ENTER("Constructors and destructors") {

			memset(&a, 0xff, sizeof(a));
			class_state = 0; GEOM_INSTANCE_INIT(&a, class_a); geom_class_a_create(&a); TEST(class_state == A_CONSTRUCTOR);
			class_state = 0; GEOM_INSTANCE_DESTROY(&a); TEST(class_state == A_DESTRUCTOR);

			memset(&a, 0xff, sizeof(a));
			class_state = 0; GEOM_INSTANCE_CREATE(&a, class_a, geom_class_a_create); TEST(class_state == A_CONSTRUCTOR);
			class_state = 0; GEOM_INSTANCE_DESTROY2(&a, class_a); TEST(class_state == A_DESTRUCTOR);

			memset(&a, 0xff, sizeof(a));
			class_state = 0; GEOM_INSTANCE_CREATE2(&a, class_a, geom_class_a_create2, 123); TEST(class_state == A_CONSTRUCTOR);
			class_state = 0; GEOM_INSTANCE_DESTROY(&a); TEST(class_state == A_DESTRUCTOR);

			memset(&a, 0, sizeof(a));
			class_state = 0; GEOM_INSTANCE_ZERO_INIT(&a, class_a); geom_class_a_create(&a); TEST(class_state == A_CONSTRUCTOR);
			class_state = 0; GEOM_INSTANCE_DESTROY(&a); TEST(class_state == A_DESTRUCTOR);

			memset(&a, 0, sizeof(a));
			class_state = 0; GEOM_INSTANCE_ZERO_CREATE(&a, class_a, geom_class_a_create); TEST(class_state == A_CONSTRUCTOR);
			class_state = 0; GEOM_INSTANCE_DESTROY(&a); TEST(class_state == A_DESTRUCTOR);

			memset(&a, 0, sizeof(a));
			class_state = 0; GEOM_INSTANCE_ZERO_CREATE2(&a, class_a, geom_class_a_create2, 123); TEST(class_state == A_CONSTRUCTOR);
			class_state = 0; GEOM_INSTANCE_DESTROY(&a); TEST(class_state == A_DESTRUCTOR);

			class_state = 0; pa = GEOM_INSTANCE_NEW(class_a); geom_class_a_create(pa); TEST(class_state == A_CONSTRUCTOR);
			class_state = 0; GEOM_INSTANCE_FREE(pa); TEST(class_state == A_DESTRUCTOR);

			class_state = 0; pa = GEOM_INSTANCE_NEW(class_a); geom_class_a_create(pa); TEST(class_state == A_CONSTRUCTOR);
			class_state = 0; GEOM_INSTANCE_FREE(pa); TEST(class_state == A_DESTRUCTOR);

			class_state = 0; GEOM_INSTANCE_NEW_CREATE(pa, class_a, geom_class_a_create); TEST(class_state == A_CONSTRUCTOR);
			class_state = 0; GEOM_INSTANCE_FREE(pa); TEST(class_state == A_DESTRUCTOR);

			class_state = 0; GEOM_INSTANCE_NEW_CREATE2(pa, class_a, geom_class_a_create2, 123); TEST(class_state == A_CONSTRUCTOR);
			class_state = 0; GEOM_INSTANCE_FREE2(pa, class_a); TEST(class_state == A_DESTRUCTOR);

		} LEAVE;

		ENTER("Virtual data and methods") {

			GEOM_INSTANCE_CREATE(&a, class_a, geom_class_a_create);
			TEST(GEOM_INSTANCE_VMT(&a, class_a, virtual) == 1);
			TEST(GEOM_INSTANCE_VMT(&a, class_a, virtual2) == 3);
			TEST(geom_instance_size(&a) == sizeof(struct geom_class_a));
			TEST(geom_instance_class(&a) == GEOM_CLASS(class_a));
			TEST(strcmp(geom_instance_class_name(&a), "class_a") == 0);
			TEST(geom_instance_is(&a, GEOM_CLASS(o)));
			TEST(geom_instance_is(&a, GEOM_CLASS(class_a)));
			TEST(!geom_instance_is(&a, GEOM_CLASS(class_b)));
			GEOM_INSTANCE_DESTROY(&a);

			GEOM_INSTANCE_CREATE(&b, class_b, geom_o_create);
			TEST(GEOM_INSTANCE_VMT(&b, class_a, virtual) == 2);
			TEST(GEOM_INSTANCE_VMT(&b, class_a, virtual2) == 3);
			TEST(geom_instance_size(&b) == sizeof(struct geom_class_b));
			TEST(geom_instance_class(&b) == GEOM_CLASS(class_b));
			TEST(strcmp(geom_instance_class_name(&b), "class_b") == 0);
			TEST(geom_instance_is(&b, GEOM_CLASS(o)));
			TEST(geom_instance_is(&b, GEOM_CLASS(class_a)));
			TEST(geom_instance_is(&b, GEOM_CLASS(class_b)));
			GEOM_INSTANCE_DESTROY(&b);

		} LEAVE;

		ENTER("Assertions") {

			GEOM_INSTANCE_CREATE(&a, class_a, geom_class_a_create);
			GEOM_ASSERT(class, geom_instance_class(&a));
			TEST(GEOM_ASSERTX(class, geom_instance_class(&a)) == GEOM_CLASS(class_a));
			GEOM_ASSERT(instance, &a);
			GEOM_INSTANCE_IS_ASSERT(class_a, &a);
			GEOM_INSTANCE_IS_ASSERT(o, &a);
			TEST(GEOM_INSTANCE_IS_ASSERTX(class_a, &a) == &a);
			TEST(GEOM_INSTANCE_IS_ASSERTX(o, &a) == (struct geom_o *)&a);
			TEST(GEOM_ASSERTX(instance, &a) == &a);

			GEOM_INSTANCE_DESTROY(&a);

		} LEAVE;

		ENTER("Class o") {

			GEOM_INSTANCE_CREATE(&o, o, geom_o_create);
			TEST(geom_instance_size(&o) == sizeof(struct geom_o));
			TEST(geom_instance_class(&o) == GEOM_CLASS(o));
			TEST(strcmp(geom_instance_class_name(&o), "o") == 0);
			TEST(geom_instance_is(&o, GEOM_CLASS(o)));
			TEST(!geom_instance_is(&o, GEOM_CLASS(class_a)));
			TEST(!geom_instance_is(&o, GEOM_CLASS(class_b)));
			GEOM_INSTANCE_DESTROY(&o);

		} LEAVE;

	} LEAVE;
}


/*----------------------------------------------*
 *                                              *
 *                   NUMERIC                    *
 *                                              *
 *----------------------------------------------*/


static void test_numeric(void)
{
	ENTER("Numeric routines") {

		ENTER("Basic macros") {

			TEST(IS_EQUAL(12.123, 12.123));
			TEST(IS_EQUAL(INFINITY, INFINITY));
			TEST(!IS_EQUAL(1.0,2.0));
			TEST(!IS_EQUAL(-INFINITY,INFINITY));
			TEST(IS_ZERO(0.0));
			TEST(!IS_ZERO(1.0));
			TEST(!IS_ZERO(INFINITY));
			TEST(!IS_ZERO(NAN));

			TEST(IS_EQUAL_L(12.123, 12.123));
			TEST(IS_EQUAL_L(INFINITY, INFINITY));
			TEST(!IS_EQUAL_L(1.0,2.0));
			TEST(!IS_EQUAL_L(-INFINITY,INFINITY));
			TEST(IS_ZERO_L(0.0));
			TEST(!IS_ZERO_L(1.0));
			TEST(!IS_ZERO_L(INFINITY));
			TEST(!IS_ZERO_L(NAN));

		} LEAVE;

	} LEAVE;
}


/*----------------------------------------------*
 *                                              *
 *               ITEMS AND GROUPS               *
 *                                              *
 *----------------------------------------------*/


/* test_items */

static void test_items(void)
{
	ENTER("Items and groups") {

		ENTER("Class item") {
			struct geom_item i;
			GEOM_INSTANCE_CREATE(&i, item, geom_item_create);
			TEST(geom_item_valid(&i));
			TEST(!geom_item_has_owner(&i));
			TEST(geom_item_owner(&i) == NULL);
			GEOM_ITEM_ASSERT(item, &i);
			TEST(GEOM_ITEM_ASSERTX(item, &i) == &i);
			GEOM_INSTANCE_DESTROY(&i);
		} LEAVE;

		ENTER("Class group") {
			struct geom_group g;
			GEOM_INSTANCE_CREATE(&g, group, geom_group_create);
			TEST(geom_item_valid(&g));
			TEST(!geom_item_has_owner(&g.item));
			TEST(geom_group_empty(&g));
			GEOM_ITEM_ASSERT(group, &g);
			TEST(GEOM_ITEM_ASSERTX(group, &g) == &g);
			GEOM_ITEM_ASSERT(item, &g);
			TEST(GEOM_ITEM_ASSERTX(item, &g) == &g);
			GEOM_INSTANCE_DESTROY(&g);
		} LEAVE;

		ENTER("Group items manipulation") {

			struct geom_group g;
			struct geom_item item[2500], *p, *q;
			uns i, n = ARRAY_SIZE(item);

			GEOM_INSTANCE_CREATE(&g, group, geom_group_create);
			for (i = 0; i < n; i++)
				GEOM_INSTANCE_CREATE(&item[i], item, geom_item_create);

			ENTER("Head access") {
				for (i = 0; i < n; i++) {
					geom_group_add_head(&g, &item[i]);
					TEST(geom_group_head(&g) == &item[i]);
				}
				for (i = 0, p = geom_group_head(&g); i < n; i++, p = geom_group_next(p))
					TEST(p == &item[n - i - 1]);
				for (i = 0; i < n; i++) {
					TEST(geom_group_head(&g) == &item[n - i - 1]);
					geom_group_remove_head(&g);
				}
				TEST(geom_group_empty(&g));
			} LEAVE;

			ENTER("Tail access") {
				for (i = 0; i < n; i++) {
					geom_group_add_tail(&g, &item[i]);
					TEST(geom_group_tail(&g) == &item[i]);
				}
				for (i = 0, p = geom_group_tail(&g); i < n; i++, p = geom_group_prev(p))
					TEST(p == &item[n - i - 1]);
				for (i = 0; i < n; i++) {
					TEST(geom_group_tail(&g) == &item[n - i - 1]);
					geom_group_remove_tail(&g);
				}
				TEST(geom_group_empty(&g));
			} LEAVE;

			ENTER("Random access") {
				uns j, k, l, d, used[n];
				for (i = 0; i < n; i++)
					used[i] = 0;
				for (i = 0; i < 5000; i++) {
					if (geom_group_count(&g) < n / 5) j = 0;
					else if (geom_group_count(&g) > n / 3) j = 1;
					else j = rand() % 2;

					/* random insert */
					if (j == 0) {
						do { j = rand() % n; } while (used[j]);
						used[j] = 1;
						p = &item[j];
						if (geom_group_empty(&g))
							geom_group_add_head(&g, p);
						else {
							for (k = j + 1; k < n && !used[k]; k++);
							for (l = j - 1; l < n && !used[l]; l--);
							if (k >= n) d = 0;
							else if (l >= n) d = 1;
							else d = rand() % 2;
							if (d) geom_group_insert_before(p, &item[k]);
							else geom_group_insert_after(p, &item[l]);
						}
					}

					/* random delete */
					else {
						j = rand() % geom_group_count(&g);
						p = geom_group_index(&g, j);
						geom_group_remove(p);
						k = (((byte *)p) - ((byte *)&item)) / sizeof(item[0]);
						TEST(p == &item[k]);
						used[k] = 0;
						for (l = 0; l < k; l++) if (used[l]) j--;
						TEST(j == 0);
					}

					/* test order */
					q = NULL;
					GEOM_GROUP_WALK(&g, p) {
						if (q)
							TEST(((byte *)q) < ((byte *)p));
						q = p;
					}
				}
				geom_group_remove_all(&g);
				TEST(geom_group_empty(&g));

			} LEAVE;

			for (i = 0; i < n; i++)
				GEOM_INSTANCE_DESTROY(&item[i]);
			GEOM_INSTANCE_DESTROY(&g);

		} LEAVE;

		ENTER("Items dumping") {
			struct geom_group group[2];
			struct geom_item item[1];
			uns i;

			for (i = 0; i < 2; i++)
				GEOM_INSTANCE_CREATE(&group[i], group, geom_group_create);
			for (i = 0; i < 1; i++)
				GEOM_INSTANCE_CREATE(&item[i], item, geom_item_create);

			geom_group_add_tail(&group[0], &item[0]);
			geom_group_add_tail(&group[0], &group[1].item);
			indent(); geom_item_dump(NULL, GEOM_DUMP_ALL);
			indent(); geom_item_dump(&item[0], GEOM_DUMP_ALL);
			indent(); geom_item_dump(&group[0], GEOM_DUMP_ALL ^ GEOM_DUMP_OWNER ^ GEOM_DUMP_RECURSIVE);
			indent(); geom_item_dump(&group[0], GEOM_DUMP_ALL ^ GEOM_DUMP_OWNER);

			for (i = 0; i < 2; i++) {
				geom_group_remove_all(&group[i]);
				GEOM_INSTANCE_DESTROY(&group[i]);
			}
			for (i = 0; i < 1; i++)
				GEOM_INSTANCE_DESTROY(&item[i]);
		} LEAVE;

	} LEAVE;
}


/*----------------------------------------------*
 *                                              *
 *                     CURVES                   *
 *                                              *
 *----------------------------------------------*/

#define TEST_CURVE_ELEMENTARY 0x1


static void test_curve(void *item, uns flags)
{
	real tlen, alen, time[10], time2[10], time3[10];
	int i, j, n = (int)ARRAY_SIZE(time);

	{
		TEST((tlen = geom_item_tlength(item)) >= 0);
		TEST(geom_item_valid(item));
		GEOM_ITEM_ASSERT(item, item);

		if (flags & TEST_CURVE_ELEMENTARY) TEST(tlen == 1);

		/* Point evalutaion */
		{
			struct geom_point pt[3];
			ETEST(geom_item_point_start(item, &pt[0]));
			ETEST(geom_item_point_end(item, &pt[1]));
			for (i = 0; i < COUNT; i++)
				ETEST(geom_item_point_at_time(item, rrand(0, tlen), &pt[2]));
			ETEST(geom_item_point_at_time(item, 0, &pt[2]));         TEST_PT_EQ(pt[0], pt[2]);
			ETEST(geom_item_point_at_time(item, -REAL_MAX, &pt[2])); TEST_PT_EQ(pt[0], pt[2]);
			ETEST(geom_item_point_at_time(item, tlen, &pt[2]));      TEST_PT_EQ(pt[1], pt[2]);
			ETEST(geom_item_point_at_time(item, REAL_MAX, &pt[2]));  TEST_PT_EQ(pt[1], pt[2]);
		}

		/* Derivations */
		{
			struct geom_vector vec[3];
			ETEST(geom_item_derivation_start(item, &vec[0]));
			ETEST(geom_item_derivation_end(item, &vec[1]));
			for (i = 0; i < COUNT; i++)
				ETEST(geom_item_derivation_at_time(item, rrand(0, tlen), &vec[2]));
			ETEST(geom_item_derivation_at_time(item, 0, &vec[2]));          TEST_VEC_EQ(vec[0], vec[2]);
			ETEST(geom_item_derivation_at_time(item, -REAL_MAX, &vec[2])); TEST_VEC_EQ(vec[0], vec[2]);
			ETEST(geom_item_derivation_at_time(item, tlen, &vec[2]));       TEST_VEC_EQ(vec[1], vec[2]);
			ETEST(geom_item_derivation_at_time(item, REAL_MAX, &vec[2]));   TEST_VEC_EQ(vec[1], vec[2]);
		}

		/* TIME <-> ATIME */
		{
			TEST(GEOM_RTEST_MINMAX(alen = geom_item_alength(item), 0, REAL_MAX));

			TEST_EQ(geom_item_time_to_atime(item, 0), 0);
			TEST_EQ(geom_item_time_to_atime(item, -REAL_MAX), 0);
			TEST_EQ(geom_item_time_to_atime(item, tlen), alen);
			TEST_EQ(geom_item_time_to_atime(item, REAL_MAX), alen);
			for (i = 0; i < COUNT; i++)
				TEST_RANGE(geom_item_time_to_atime(item, rrand(0, tlen)), 0, alen);

			TEST_EQ(geom_item_atime_to_time(item, 0), 0);
			TEST_EQ(geom_item_atime_to_time(item, -REAL_MAX), 0);
			for (i = 0; i < COUNT; i++)
				TEST_RANGE(geom_item_atime_to_time(item, rrand(0, alen)), 0, alen);

			if (alen > EPS) {
				TEST_EQ(geom_item_atime_to_time(item, alen), tlen);
				TEST_EQ(geom_item_atime_to_time(item, REAL_MAX), tlen);
				for (i = 0; time[0] = rrand(0, tlen), i < 10; i++)
					TEST_EQ(geom_item_atime_to_time(item, geom_item_time_to_atime(item, time[0])), time[0]);

				for (j = 0; j < 5; j++) {
					for (i = 0; i < n; i++) time[i] = rrand(0, tlen);
					geom_reals_sort(n, time);
					ETEST(geom_item_times_to_atimes(item, n, time, time2));
					ETEST(geom_item_atimes_to_times(item, n, time2, time3));
					/*for (i = 0; i < n; i++) {
						EQ_TEST(geom_item_time_to_atime(item, time[i]), time2[i]);
						EQ_TEST(time[i], time2[i]);
					}*/

				}
			}
		}

		/* Bounding box */
		{
			struct geom_rectangle bbox;
			struct geom_point pt;
			ETEST(geom_item_bbox(item, &bbox));
			for (int i = 0; i < COUNT; i++) {
				ETEST(geom_item_point_at_time(item, rrand(0, tlen), &pt));
				TEST_RANGE(pt.x, bbox.ll.x, bbox.ur.x);
				TEST_RANGE(pt.y, bbox.ll.y, bbox.ur.y);
			}
		}

		/* Nearest point */
		{
			for (int i = 0; i < COUNT; i++) {
				struct geom_point point = rand_point();
				struct geom_nearest nearest;
				ETEST(geom_item_nearest_to_point(item, &point, GEOM_NEAREST_ALL, &nearest));
				TEST(GEOM_RTEST(nearest.time));
				TEST(geom_point_valid(&nearest.point));
				TEST(GEOM_RTEST_MINMAX(nearest.dist, 0, REAL_MAX));
				for (int i = 0; i < COUNT; i++) {
					struct geom_point pt;
					ETEST(geom_item_point_at_time(item, rrand(0, tlen), &pt));
					TEST_RANGE(hypot(pt.x - point.x, pt.y - point.y), hypot(nearest.point.x - point.x, nearest.point.y - point.y), REAL_MAX);
				}
			}
		}

		/* Instance duplicate */
		{
			struct geom_item *copy;
			copy = geom_item_duplicate(item);
			TEST(geom_item_valid(copy));
			for (int i = 0; i < COUNT; i++) {
				struct geom_point pt1, pt2;
				real t = rrand(0, tlen);
				ETEST(geom_item_point_at_time(item, t, &pt1));
				ETEST(geom_item_point_at_time(copy, t, &pt2));
				TEST_PT_EQ(pt1, pt2);
			}
			GEOM_INSTANCE_FREE(copy);
		}

	}
}


/*----------------------------------------------*
 *                                              *
 *                     SEGMENT                  *
 *                                              *
 *----------------------------------------------*/


static void test_segment(struct geom_segment *s)
{
	struct geom_point pt;

	test_curve(s, TEST_CURVE_ELEMENTARY);

	ETEST(geom_item_point_start(s, &pt)); TEST_PT_EQ(pt, s->a);
	ETEST(geom_item_point_end(s, &pt)); TEST_PT_EQ(pt, s->b);
	for (int i = 0; i < 10; i++) {
		real t = rrand(0, 1);
		ETEST(geom_item_point_at_time(s, t, &pt));
	}
}

static struct geom_segment *new_segment(struct geom_point a, struct geom_point b)
{
	struct geom_segment *s;
	GEOM_INSTANCE_NEW_CREATE2(s, segment, geom_segment_set, &a, &b);
	return s;
}

static struct geom_segment *rand_segment(void)
{ return new_segment(rand_point(), rand_point()); }

static void test_segments(void)
{
	ENTER("Segments") {
		struct geom_segment *s;

		ENTER("Random segments") {
			for (int i = 0; i < 10 * COUNT; i++) {
				s = rand_segment();
				test_segment(s);
				GEOM_INSTANCE_FREE(s);
			}
		} LEAVE;

		ENTER("Zero segment") {
			s = new_segment(geom_point_zero, geom_point_zero);
			test_segment(s);
			GEOM_INSTANCE_FREE(s);
		} LEAVE;

		ENTER("Point segments") {
			struct geom_point pt = rand_point();
			s = new_segment(pt, pt);
			test_segment(s);
			GEOM_INSTANCE_FREE(s);
		} LEAVE;

	} LEAVE;
}


/*----------------------------------------------*
 *                                              *
 *                     BEZIER                   *
 *                                              *
 *----------------------------------------------*/


static void test_bezier(struct geom_bezier *b)
{
	test_curve(b, TEST_CURVE_ELEMENTARY);
}

static struct geom_bezier *rand_bezier_ln(void)
{
	struct geom_bezier *b;
	struct geom_point pt0 = rand_point(), pt1 = rand_point();
	GEOM_INSTANCE_NEW_CREATE2(b, bezier, geom_bezier_set_linear_nonrational, &pt0, &pt1);
	return b;
}

static struct geom_bezier *rand_bezier_qn(void)
{
	struct geom_bezier *b;
	struct geom_point pt0 = rand_point(), pt1 = rand_point(), pt2 = rand_point();
	GEOM_INSTANCE_NEW_CREATE2(b, bezier, geom_bezier_set_quadratic_nonrational, &pt0, &pt1, &pt2);
	return b;
}

static struct geom_bezier *rand_bezier_cn(void)
{
	struct geom_bezier *b;
	struct geom_point pt0 = rand_point(), pt1 = rand_point(), pt2 = rand_point(), pt3 = rand_point();
	GEOM_INSTANCE_NEW_CREATE2(b, bezier, geom_bezier_set_cubic_nonrational, &pt0, &pt1, &pt2, &pt3);
	return b;
}

static struct geom_bezier *rand_bezier_lr(void)
{
	struct geom_bezier *b;
	struct geom_point_w pt0 = rand_point_w(), pt1 = rand_point_w();
	GEOM_INSTANCE_NEW_CREATE2(b, bezier, geom_bezier_set_linear_rational, &pt0, &pt1);
	return b;
}

static struct geom_bezier *rand_bezier_qr(void)
{
	struct geom_bezier *b;
	struct geom_point_w pt0 = rand_point_w(), pt1 = rand_point_w(), pt2 = rand_point_w();
	GEOM_INSTANCE_NEW_CREATE2(b, bezier, geom_bezier_set_quadratic_rational, &pt0, &pt1, &pt2);
	return b;
}

static struct geom_bezier *rand_bezier_cr(void)
{
	struct geom_bezier *b;
	struct geom_point_w pt0 = rand_point_w(), pt1 = rand_point_w(), pt2 = rand_point_w(), pt3 = rand_point_w();
	GEOM_INSTANCE_NEW_CREATE2(b, bezier, geom_bezier_set_cubic_rational, &pt0, &pt1, &pt2, &pt3);
	return b;
}

static struct geom_bezier *rand_bezier(void)
{
	switch (rand() % 6) {
	case 0: return rand_bezier_ln();
	case 1: return rand_bezier_qn();
	case 2: return rand_bezier_cn();
	case 3: return rand_bezier_lr();
	case 4: return rand_bezier_qr();
	case 5: return rand_bezier_cr();
	default: BUG();
	}
}

static void test_beziers(void)
{
	ENTER("Bezier curves") {
		ENTER("Linear nonrational") {
			for (int i = 0; i < COUNT; i++) {
				struct geom_bezier *b = rand_bezier_ln();
				test_bezier(b);
				GEOM_INSTANCE_FREE(b);
			}
		} LEAVE;

		ENTER("Quadratic nonrational") {
			for (int i = 0; i < COUNT; i++) {
				struct geom_bezier *b = rand_bezier_qn();
				test_bezier(b);
				GEOM_INSTANCE_FREE(b);
			}
		} LEAVE;

		ENTER("Cubic nonrational") {
			for (int i = 0; i < COUNT; i++) {
				struct geom_bezier *b = rand_bezier_cn();
				test_bezier(b);
				GEOM_INSTANCE_FREE(b);
			}
		} LEAVE;

		ENTER("Linear rational") {
			for (int i = 0; i < COUNT; i++) {
				struct geom_bezier *b = rand_bezier_lr();
				test_bezier(b);
				GEOM_INSTANCE_FREE(b);
			}
		} LEAVE;

		ENTER("Quadratic rational") {
			for (int i = 0; i < COUNT; i++) {
				struct geom_bezier *b = rand_bezier_qr();
				test_bezier(b);
				GEOM_INSTANCE_FREE(b);
			}
		} LEAVE;

		ENTER("Cubic rational") {
			for (int i = 0; i < COUNT; i++) {
				struct geom_bezier *b = rand_bezier_cr();
				test_bezier(b);
				GEOM_INSTANCE_FREE(b);
			}
		} LEAVE;

	} LEAVE;
}


/*----------------------------------------------*
 *                                              *
 *                ELLIPTIC_ARC                  *
 *                                              *
 *----------------------------------------------*/


static void test_elliptic_arc(struct geom_elliptic_arc *a)
{
	test_curve(a, TEST_CURVE_ELEMENTARY);
}

static struct geom_elliptic_arc *rand_elliptic_arc(void)
{
	struct geom_point center = rand_point();
	struct geom_elliptic_arc *e;

	GEOM_INSTANCE_NEW_CREATE2(e, elliptic_arc, geom_elliptic_arc_set,
		&center, rrand(-100, 100), rrand(0, 100), rrand(0, 100), rrand(-100, 100), rrand(0, R_2PI));
	return e;
}

static void test_elliptic_arcs(void)
{
	ENTER("Elliptic arcs") {
		struct geom_elliptic_arc *e;

		ENTER("Random elliptic arcs") {
			for (int i = 0; i < COUNT; i++) {
				e = rand_elliptic_arc();
				test_elliptic_arc(e);
				GEOM_INSTANCE_FREE(e);
			}
		} LEAVE;

	} LEAVE;
}


/*----------------------------------------------*
 *                                              *
 *                      PATHS                   *
 *                                              *
 *----------------------------------------------*/


static void *rand_curve(void)
{
	switch (rand() % 3) {
	case 0: return rand_segment();
	case 1: return rand_elliptic_arc();
	case 2: return rand_bezier();
	default: BUG();
	}
}

static struct geom_path *rand_path(void)
{
	struct geom_path *p;
	GEOM_INSTANCE_NEW_CREATE(p, path, geom_path_create);
	for (int i = 1 + rand() % 10; i; i--)
		geom_group_add_tail(p, rand_curve());
	return p;
}

static void test_path(struct geom_path *p)
{
	test_curve(p, 0);
}

static void test_paths(void)
{
	ENTER("Paths") {

		ENTER("Empty path") {
			struct geom_path *p;
			GEOM_INSTANCE_NEW_CREATE(p, path, geom_path_create);
			TEST(geom_group_empty(p));
			GEOM_INSTANCE_FREE(p);
		} LEAVE;

		ENTER("Random paths") {
			struct geom_path *p;
			for (int i = 0; i < COUNT; i++) {
				p = rand_path();
				test_path(p);
				GEOM_INSTANCE_FREE(p);
			}
		} LEAVE;

	} LEAVE;
}


/*----------------------------------------------*
 *                                              *
 *                     FPATHS                   *
 *                                              *
 *----------------------------------------------*/


static struct geom_fpath *rand_fpath(void)
{
	struct geom_fpath *p;
	struct geom_curve *c, *n;
	GEOM_INSTANCE_NEW_CREATE(p, fpath, geom_fpath_create);
	for (int i = 1 + rand() % 10; i; i--) {
		c = rand_curve();
		n = geom_fpath_append_dup(p, c);
		GEOM_INSTANCE_FREE(c);
	}
	geom_fpath_finish(p);
	return p;
}

static void test_fpath(struct geom_fpath *p)
{
	test_curve(p, 0);
}

static void test_fpaths(void)
{
	ENTER("Fast paths") {

		ENTER("Empty path") {
			struct geom_fpath *p;
			GEOM_INSTANCE_NEW_CREATE(p, fpath, geom_fpath_create);
			TEST(geom_group_empty(p));
			GEOM_INSTANCE_FREE(p);
		} LEAVE;

		ENTER("Random paths") {
			struct geom_fpath *p;
			for (int i = 0; i < COUNT; i++) {
				p = rand_fpath();
				test_fpath(p);
				GEOM_INSTANCE_FREE(p);
			}
		} LEAVE;

	} LEAVE;
}


/*----------------------------------------------*
 *                                              *
 *                 EXPANSIONS                   *
 *                                              *
 *----------------------------------------------*/


static void test_expansions(void)
{
	ENTER("Expansions") {

		ENTER("Old interface") {
			struct geom_path *p;
			void *c;
			real alen;

			for (int i = 0; i < COUNT; i++) {
				if (i < 10)
					c = rand_path();
				else
					c = rand_curve();
				GEOM_INSTANCE_NEW_CREATE(p, path, geom_path_create);
				ETEST(geom_old_expand(c, p));
				alen = geom_item_alength(c);
				TEST_EQ(geom_item_alength(p), alen);
				GEOM_INSTANCE_FREE(p);
				GEOM_INSTANCE_FREE(c);
			}

		} LEAVE;

	} LEAVE;
}


/*----------------------------------------------*
 *                                              *
 *                   POINT ITEM                 *
 *                                              *
 *----------------------------------------------*/

static void test_point_items(void)
{
	ENTER("Point items") {

		ENTER("Random points") {
			for (int i = 0; i < 10 * COUNT; i++) {
				struct geom_point_item *p;
				struct geom_point pt = rand_point();
				GEOM_INSTANCE_NEW_CREATE2(p, point_item, geom_point_item_set, &pt);
				test_curve(p, 0);
				GEOM_INSTANCE_FREE(p);
			}
		} LEAVE;

	} LEAVE;
}


/*----------------------------------------------*
 *                                              *
 *           CALLBACK EXPANSION ITEM            *
 *                                              *
 *----------------------------------------------*/

static void *test_callback_org;

static int test_callback_func(struct geom_callback_item UNUSED *item, struct geom_fpath *path)
{
	return geom_item_expansion_append(test_callback_org, path);
}

static void test_callback_items(void)
{
	ENTER("Callback expansion items") {

		ENTER("Random curves") {
			for (int i = 0; i < COUNT; i++) {
				struct geom_callback_item *p;
				test_callback_org = rand_curve();
				GEOM_INSTANCE_NEW_CREATE2(p, callback_item, geom_callback_item_create_set, &test_callback_func);
				test_curve(p, 0);
				GEOM_INSTANCE_FREE(test_callback_org);
				GEOM_INSTANCE_FREE(p);
			}
		} LEAVE;

	} LEAVE;
}


/*----------------------------------------------*
 *                                              *
 *                INTERSECTIONS                 *
 *                                              *
 *----------------------------------------------*/

static int test_items_intersections(void *curve1, void *curve2)
{
	int n;
	struct garr *garr;
	garr = garr_new(10, sizeof(struct geom_intersection));
	ETEST(n = geom_items_intersections(curve1, curve2, GEOM_INTERSECTION_ALL, garr));
	for (int i = 0; i < n; i++) {
		struct geom_intersection *inter = (struct geom_intersection *)garr_get(garr, i);
		struct geom_nearest nearest;

		ETEST(geom_item_nearest_to_point(curve1, &inter->point, GEOM_NEAREST_ALL, &nearest));
		TEST_PT_EQ(inter->point, nearest.point);
		TEST_EQ(inter->time1, nearest.time);

		ETEST(geom_item_nearest_to_point(curve2, &inter->point, GEOM_NEAREST_ALL, &nearest));
		TEST_PT_EQ(inter->point, nearest.point);
		TEST_EQ(inter->time2, nearest.time);
	}
	garr_destroy(garr);
	return MAX(0, n);
}

static void test_intersections(void)
{
	ENTER("Intersections") {

#define TEST_INTERSECTIONS_COMBO(name, count, eps, first, second) do{\
		ENTER(name) { EPS = eps; \
			int n = 0; \
			for (int i = 0; i < count; i++) { \
				void *p1 = rand_##first(), *p2 = rand_##second(); \
				n += test_items_intersections(p1, p2); \
				GEOM_INSTANCE_FREE(p1); GEOM_INSTANCE_FREE(p2); } \
			indent(); msg("%d intersections found in %d tests\n", n, count); \
		} LEAVE;}while(0)

		TEST_INTERSECTIONS_COMBO("Random segments", 100 * COUNT, SMALL_EPS, segment, segment);
		TEST_INTERSECTIONS_COMBO("Random Bezier curves", 10 * COUNT, LARGE_EPS, bezier, bezier);
		TEST_INTERSECTIONS_COMBO("Random elliptic arcs", 10 * COUNT, LARGE_EPS, elliptic_arc, elliptic_arc);
		TEST_INTERSECTIONS_COMBO("Random curves combinations", 10 * COUNT, LARGE_EPS, curve, curve);
		TEST_INTERSECTIONS_COMBO("Random paths", COUNT, LARGE_EPS, path, path);

	} LEAVE;
}

/*----------------------------------------------*
 *                                              *
 *                  MAIN CODE                   *
 *                                              *
 *----------------------------------------------*/


/* main */

int main(void)
{
	//srandom(129);

	ENTER("Geomlib") {

		ENTER("Geomlib initialization")	{ cache_init(&global_cache); geomlib_init(); } LEAVE;

		test_base();
		test_classes();
		test_numeric();
		test_items();
		test_segments();
		test_beziers();
		test_elliptic_arcs();
		test_paths();
		test_fpaths();
		test_expansions();
		test_point_items();
		test_callback_items();
		test_intersections();

		ENTER("Geomlib finalization") { geomlib_finish(); cache_destroy(&global_cache); } LEAVE;

	} LEAVE;

	msg("All tests passed.\n");
	return 0;
}

