/*
 * GEOMLIB - Nonuniform rational Bezier splines.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#define GEOM_SOURCE
#include "geomlib/vector.h"
#include "geomlib/transform.h"
#include "geomlib/bezier.h"
#include "geomlib/fpath.h"
#include "geomlib/nurbs.h"


static inline struct geom_nurbs_knot *geom_nurbs_knot_alloc(struct geom_nurbs UNUSED *nurbs)
{ return xmalloc(sizeof(struct geom_nurbs_knot)); }

static inline void geom_nurbs_knot_free(struct geom_nurbs UNUSED *nurbs, struct geom_nurbs_knot *knot)
{ xfree(knot); }

static inline struct geom_nurbs_knot *geom_nurbs_knot_first(struct geom_nurbs *nurbs)
{ return SKIP_BACK(struct geom_nurbs_knot, node, nurbs->knots.head.next); }

static inline struct geom_nurbs_knot *geom_nurbs_knot_last(struct geom_nurbs *nurbs)
{ return SKIP_BACK(struct geom_nurbs_knot, node, nurbs->knots.head.prev); }

static inline struct geom_nurbs_knot *geom_nurbs_knot_prev(struct geom_nurbs_knot *knot)
{ return SKIP_BACK(struct geom_nurbs_knot, node, knot->node.prev); }

static inline struct geom_nurbs_knot *geom_nurbs_knot_next(struct geom_nurbs_knot *knot)
{ return SKIP_BACK(struct geom_nurbs_knot, node, knot->node.next); }

static struct geom_nurbs_knot *geom_nurbs_knot_search(struct geom_nurbs *nurbs, real time)
{
	struct geom_nurbs_knot *knot;
	for (knot = geom_nurbs_knot_first(nurbs); knot->time <= time; knot = geom_nurbs_knot_next(knot));
	return knot;
}


/*----------------------------------------------*
 *                                              *
 *                INITIALIZATION                *
 *                                              *
 *----------------------------------------------*/


/* geom_nurbs_clear */

static void geom_nurbs_clear(struct geom_nurbs *nurbs)
{
	struct geom_nurbs_knot *knot;
	if (nurbs->degree) {
		nurbs->degree = 0;
		while (knot = cclist_head(&nurbs->knots)) {
			cclist_remove(&nurbs->knots, &knot->node);
			geom_nurbs_knot_free(nurbs, knot);
		}
	}
}


/* geom_nurbs_destroy */

void geom_nurbs_destroy(struct geom_nurbs *nurbs)
{
	FUNC;
	geom_nurbs_clear(nurbs);
	geom_curve_destroy(&nurbs->curve);
	RET;
}


/* geom_nurbs_copy */

void geom_nurbs_copy(struct geom_nurbs *nurbs, struct geom_nurbs *src)
{
	uns i;
	struct geom_nurbs_knot *knot, *nknot;

	FUNC;
	INSTANCE_IS_ASSERT(nurbs, nurbs);
	ITEM_ASSERT(nurbs, src);

	geom_nurbs_clear(nurbs);
	nurbs->degree = src->degree;
	knot = geom_nurbs_knot_first(src);
	for (i = 0; i < src->knots.count; i++) {
		nknot = geom_nurbs_knot_alloc(nurbs);
		nknot->point = knot->point;
		nknot->time = knot->time;
		cclist_add_tail(&nurbs->knots, &nknot->node);
		knot = geom_nurbs_knot_next(knot);
	}

	RET;
}


/* geom_nurbs_from_bezier */

int geom_nurbs_from_bezier(struct geom_nurbs *nurbs, struct geom_bezier *bezier)
{
	int i;
	struct geom_nurbs_knot *knot;

	FUNC;
	INSTANCE_IS_ASSERT(nurbs, nurbs);
	ITEM_ASSERT(bezier, bezier);

	geom_nurbs_clear(nurbs);
	nurbs->degree = BEZIER_DEGREE(bezier);
	for (i = 0; i <= BEZIER_DEGREE(bezier); i++) {
		knot = geom_nurbs_knot_alloc(nurbs);
		knot->time = 0;
		knot->point = bezier->pt[i];
		cclist_add_tail(&nurbs->knots, &knot->node);
	}
	for (i = 0; i <= BEZIER_DEGREE(bezier); i++) {
		knot = geom_nurbs_knot_alloc(nurbs);
		knot->time = 1;
		cclist_add_tail(&nurbs->knots, &knot->node);
	}

	ERET(0);
}


/*----------------------------------------------*
 *                                              *
 *               PARAMETRIZATIONS               *
 *                                              *
 *----------------------------------------------*/


/* geom_nurbs_time_to_btime */

real geom_nurbs_time_to_btime(struct geom_nurbs *nurbs, struct geom_expansion *expansion, real time)
{
	uns i, j;
	real t1, t2;
	struct geom_nurbs_knot *knot;

	FUNC;
	ITEM_ASSERT(nurbs, nurbs);
	ITEM_ASSERT(fpath, expansion);
	RASSERT(time);

	if (time <= 0)
		RRET(0);
	else if (time >= 1)
		RRET(geom_group_count(&expansion->path));
	else {
		knot = geom_nurbs_knot_first(nurbs);
		t1 = 0;
		j = 0;
		for (i = 0; i < nurbs->knots.count; i++) {
			t2 = knot->time;
			if (t1 != t2) {
				if (time <= t2) {
					t1 = fminmaxr(inf2finr((time - t1) / (t2 - t1)), 0, 1);
					RRET(j + t1);
				}
				else
					j++;
				t1 = t2;
			}
			knot = geom_nurbs_knot_next(knot);
		}
		BUG();
	}
}


/* geom_nurbs_btime_to_time */

real geom_nurbs_btime_to_time(struct geom_nurbs *nurbs, struct geom_expansion *expansion, real btime)
{
	uns i, j;
	real t1, t2, t;
	struct geom_nurbs_knot *knot;

	FUNC;
	ITEM_ASSERT(nurbs, nurbs);
	ITEM_ASSERT(fpath, expansion);
	RASSERT(btime);

	if (btime <= 0)
		RRET(0);
	else if (btime >= geom_group_count(&expansion->path))
		RRET(1);
	else {
		j = btime;
		j = MIN(j, geom_group_count(&expansion->path) - 1);
		t = fminmaxr(btime - j, 0, 1);
		knot = geom_nurbs_knot_first(nurbs);
		t1 = 0;
		for (i = 0; i < nurbs->knots.count; i++) {
			t2 = knot->time;
			if (t1 != t2) {
				if (j--) {
					t = fcoefrngr(t1, t2, t);
					RRET(t);
				}
				t1 = t2;
			}
			knot = geom_nurbs_knot_next(knot);
		}
		BUG();
	}
}


/*----------------------------------------------*
 *                                              *
 *                     EDIT                     *
 *                                              *
 *----------------------------------------------*/


/* geom_nurbs_knot_insert */

int geom_nurbs_knot_insert(struct geom_nurbs *nurbs, real time)
{
	int d, i;
	struct geom_point_w p[4], q;
	real a, u[6];
	struct geom_nurbs_knot *knot, *k, *nknot;

	FUNC;
	ITEM_ASSERT(nurbs, nurbs);
	GASSERT2(r_minmax, time, 0, 1);

	nknot = geom_nurbs_knot_alloc(nurbs);
	nknot->time = time;
	if (time == 0) {
		nknot->point = geom_nurbs_knot_first(nurbs)->point;
		cclist_add_head(&nurbs->knots, &nknot->node);
	}
	else if (time == 1) {
		nknot->point = geom_nurbs_knot_last(nurbs)->point;
		cclist_add_tail(&nurbs->knots, &nknot->node);
	}
	else {
		d = nurbs->degree;
		k = knot = geom_nurbs_knot_search(nurbs, time);
		for (i = d; i >= 0; i--) {
			k = geom_nurbs_knot_prev(k);
			p[i] = k->point;
		}
		for (i = 0; i < d + d; i++) {
			k = geom_nurbs_knot_next(k);
			u[i] = k->time;
		}
		cclist_insert_before(&nurbs->knots, &nknot->node, &knot->node);
		nknot->point = p[d];
		knot = geom_nurbs_knot_prev(knot);
		for (i = d - 1; i >= 0; i--) {
			a = fminmaxr(inf2finr((time - u[i]) / (u[i + d] - u[i])), 0, 1);
			q.w = fcoefrngr(p[i].w, p[i + 1].w, a);
			a = fminmaxr(inf2finr(a * p[i + 1].w / q.w), 0, 1);
			q.x = fcoefrngr(p[i].x, p[i + 1].x, a);
			q.y = fcoefrngr(p[i].y, p[i + 1].y, a);
			knot = geom_nurbs_knot_prev(knot);
			knot->point = q;
		}
	}
	ITEM_ASSERT(nurbs, nurbs);
	ERET(0);
}



/*----------------------------------------------*
 *                                              *
 *     CONVERSION TO GIVEN TYPES OF CURVES      *
 *                                              *
 *----------------------------------------------*/


/* geom_nurbs_expansion_append */

int geom_nurbs_expansion_append(struct geom_nurbs *nurbs, struct geom_fpath *expansion)
{
	struct geom_nurbs temp;
	struct geom_nurbs_knot *knot;
	struct geom_bezier *b;
	real time;
	uns i, j;
	struct geom_point_w pt[4];

	// FIXME: slow and no error tests

	FUNC;
	ITEM_ASSERT(nurbs, nurbs);
	ITEM_ASSERT(fpath, expansion);

	GEOM_INSTANCE_CREATE2(&temp, nurbs, geom_nurbs_copy, nurbs);
	knot = geom_nurbs_knot_first(&temp);
	time = 0;
	j = 0;
	for (i = 0; i < temp.knots.count; i++) {
		if (knot->time == time)
			j++;
		else {
			while (j < temp.degree) {
				geom_nurbs_knot_insert(&temp, time);
				j++;
			}
			j = 1;
			time = knot->time;
		}
		knot = geom_nurbs_knot_next(knot);
	}
	
	knot = geom_nurbs_knot_first(&temp);
	j = 0;
	for (i = temp.degree + 1; i < temp.knots.count; i++) {
		pt[j] = knot->point;
		if (++j > temp.degree) {
			switch (temp.degree) {
			case GEOM_BEZIER_LINEAR:
				GEOM_FPATH_APPEND_NEW_CREATE2(b, expansion, bezier, geom_bezier_set_linear_rational, &pt[0], &pt[1]);
				break;
			case GEOM_BEZIER_QUADRATIC:
				GEOM_FPATH_APPEND_NEW_CREATE2(b, expansion, bezier, geom_bezier_set_quadratic_rational, &pt[0], &pt[1], &pt[2]);
				break;
			case GEOM_BEZIER_CUBIC:
				GEOM_FPATH_APPEND_NEW_CREATE2(b, expansion, bezier, geom_bezier_set_cubic_rational, &pt[0], &pt[1], &pt[2], &pt[3]);
				break;
			}
			j = 1;
			pt[0] = pt[temp.degree];
		}
		knot = geom_nurbs_knot_next(knot);
	}
	GEOM_INSTANCE_DESTROY2(&temp, nurbs);

	ERET(0);
}


