/*
 * GEOMLIB - Continous paths from more curves.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#define GEOM_SOURCE
#include "geomlib/vector.h"
#include "geomlib/rectangle.h"
#include "geomlib/transform.h"
#include "geomlib/segment.h"
#include "geomlib/bezier.h"
#include "geomlib/curve.h"
#include "geomlib/path.h"


/* geom_path_copy */

void geom_path_copy(struct geom_path *path, struct geom_path *src)
{
	struct geom_curve *curve;

	BASSERT(geom_instance_class(path) == CLASS(path));
	INSTANCE_IS_ASSERT(path, src);

	if (path->group.list.head.next)
		geom_group_free_all(path);
	else
		geom_path_create(path);
	GEOM_GROUP_WALK(src, curve)
		geom_path_append_dup(path, curve);
}


/*----------------------------------------------*
 *                                              *
 *      PARAMETRIZATIONS:  TIME <-> ATIME       *
 *                                              *
 *----------------------------------------------*/


/*
 * FIXME:
 *   S vyvazenym stromem u grup by melo smysl cachovat vyhledavaci tabulku souctu delek v ceste nebo pridat real do geom_curve.
 */

/** Update cached arc length of given path. */
static void geom_path_do_update_alength(struct geom_path *path)
{
	struct geom_path_item *item;
	real sum;

	sum = 0;
	GEOM_GROUP_WALK(path, item)
		if (!RTEST(sum += geom_item_alength(item)))
			break;
	path->alength = sum;
	INSTANCE_FLAGS(path) |= GEOM_PATH_ALENGTH_VALID;
}


/** If necessary, update chached arc length of given path. */
static inline real geom_path_update_alength(struct geom_path *path)
{
	if (!(INSTANCE_FLAGS(path) & GEOM_PATH_ALENGTH_VALID))
		geom_path_do_update_alength(path);
	return path->alength;
}


/* Path walk structure */
struct geom_path_pos {
	struct geom_curve *curve;
	uns index;
	real time_offset;
	real atime_offset;
	real alen_sum;
};


/**
 * Converts TIME parameter to curve index and offset.
 * For empty path, zeros are returned.
 *
 * @ai  path    Path.
 * @ai  time    TIME parameter (must be finite and in interval <0-epsilon, path->count+epsilon>).
 * @ao  index   Curve index (0 <= index < path->count).
 * @ao  offset  Curve offset (in interval <0-epsilon, 1+epsilon>, always finite).
 */
static void geom_path_time_to_index_offset(struct geom_path *path, real time, struct geom_path_pos *pos)
{
	uns count;

	RASSERT(time);
	count = geom_group_count(path);
	BASSERT(count > 0);
	if (time <= 0) {
		pos->index = 0;
		pos->time_offset = 0;
	}
	else if (time >= count) {
		pos->index = count - 1;
		pos->time_offset = 1;
	}
	else {
		pos->index = time;
		if ((int)pos->index < 0)
			pos->index = 0;
		else if (pos->index >= count)
			pos->index = count - 1;
		pos->time_offset = inf2fin(time - pos->index);
	}
}


/**
 * Converts TIME parameter to curve index and offset.
 * For empty path, NULL is returned.
 *
 * @ai  path    Path.
 * @ai  time    TIME parameter (must be finite and in interval <0-epsilon, path->count+epsilon>).
 * @ao  index   Curve index (0 <= index < path->count).
 * @ao  offset  Curve offset (in interval <0-epsilon, 1+epsilon>, always finite).
 */
static struct geom_curve *geom_path_time_to_curve(struct geom_path *path, real time, struct geom_path_pos *pos)
{
	uns count;

	RASSERT(time);
	count = geom_group_count(path);
	BASSERT(count > 0);
	if (time <= 0) {
		pos->index = 0;
		pos->time_offset = 0;
		pos->curve = geom_group_head(path);
	}
	else if (time >= count) {
		pos->index = count - 1;
		pos->time_offset = 1;
		pos->curve = geom_group_tail(path);
	}
	else {
		pos->index = time;
		if ((int)pos->index < 0)
			pos->index = 0;
		else if (pos->index >= count)
			pos->index = count - 1;
		pos->time_offset = inf2fin(time - pos->index);
		pos->curve = geom_group_index(path, pos->index);
	}
	return pos->curve;
}


/* geom_path_alength */

real geom_path_alength(struct geom_path *path)
{
	FUNC;
	ITEM_ASSERT(path, path);

	geom_path_update_alength(path);
	RRET(path->alength);
}


/**
 * Converts TIME parameter in nonempty path to curve index and offset.
 *
 * @ai  path    Path.
 * @ai  time    TIME parameter (must be finite and in interval <0-epsilon, path->count+epsilon>).
 * @ao  index   Curve index (0 <= index < path->count).
 * @ao  offset  Curve offset (in interval <0-epsilon, 1+epsilon>, always finite).
 */
static void geom_path_time_to_atime_first(struct geom_path *path, real time, struct geom_path_pos *pos)
{
	uns i;
	geom_path_time_to_index_offset(path, time, pos);
	pos->curve = geom_group_head(path);
	pos->alen_sum = 0;
	for (i = 0; i < pos->index; i++) {
		pos->alen_sum += geom_item_alength(pos->curve);
		pos->curve = geom_group_next(pos->curve);
	}
}

static void geom_path_time_to_atime_next(struct geom_path *path, real time, struct geom_path_pos *pos)
{
	uns i = pos->index;
	geom_path_time_to_index_offset(path, time, pos);
	if (unlikely(pos->index < i)) {
		pos->index = i;
		pos->time_offset = 0;
	}
	else
		while (pos->index > i) {
			pos->alen_sum += geom_item_alength(pos->curve);
			pos->curve = geom_group_next(pos->curve);
			i++;
		}
}


/* geom_path_time_to_atime */

real geom_path_time_to_atime(struct geom_path *path, real time)
{
	struct geom_path_pos pos;

	FUNC;
	ITEM_ASSERT(path, path);
	RASSERT(time);

	if (time <= 0)
		RRET(0);
	RTEST_NAN(geom_path_update_alength(path));
	if (time >= geom_group_count(path))
		RRET(path->alength);
	geom_path_time_to_atime_first(path, time, &pos);
	RRET(pos.alen_sum + geom_item_time_to_atime(pos.curve, pos.time_offset));
}


/* geom_path_times_to_atimes */

int geom_path_times_to_atimes(struct geom_path *path, uns count, real *times, real *atimes)
{
	struct geom_path_pos pos;
	struct geom_curve *block_curve;
	uns path_count, i, j, l, r, block_count;
	real t, block_sum;

	FUNC;
	ITEM_ASSERT(path, path);
	GEOM_REALS_SORTED_ASSERT(count, times);
	PASSERT(atimes);

	for (l = 0, r = count; l < r && times[l] <= 0; l++)
		atimes[l] = 0;
	RTEST_NUM(geom_path_update_alength(path));
	path_count = geom_group_count(path);
	while (l < r && times[r - 1] >= path_count)
		atimes[--r] = path->alength;
	if (l == r)
		ERET(0);
	block_curve = NULL;
	block_count = block_sum = 0; /* <- not necessary, warning elimination */
	for (i = l; i <= r; i++) {
		if (i < r) {
			if (i == l)
				geom_path_time_to_atime_first(path, times[i], &pos);
			else
				geom_path_time_to_atime_next(path, times[i], &pos);
			atimes[i] = pos.time_offset;
		}
		if (i == r || pos.curve != block_curve) {
			if (block_curve) {
				if (block_count == 1)
					RTEST_NUM(atimes[i - 1] = block_sum + geom_item_time_to_atime(block_curve, atimes[i - 1]));
				else if (block_count > 1) {
					j = i - block_count;
					ETEST_RET(geom_item_times_to_atimes(block_curve, block_count, atimes + j, atimes + j));
					for (; j < i; j++)
						RTEST_NUM(atimes[j] = block_sum + atimes[j]);
				}
			}
			block_curve = pos.curve;
			block_sum = pos.alen_sum;
			block_count = 1;
		}
		else
			block_count++;
	}
	t = path->alength;
	for (i = r - 1; (int)i >= (int)l; i--) {
		if (atimes[i] < 0)
			atimes[i] = 0;
		else if (times[i] > t)
			atimes[i] = t;
		t = atimes[i];
	}

	GEOM_REALS_SORTED_ASSERT(count, atimes);
	ERET(0);
}


/**
 * Converts ATIME parameter in nonempty path to curve index and offset.
 *
 * @ai  path    Path.
 * @ai  atime   ATIME parameter (must be finite and in interval [0, alength]).
 * @ao  index   Curve index (0 <= index < count).
 * @ao  offset  Curve offset (in interval [0-epsilon, curve alength+epsilon], always finite).
 */
static void geom_path_atime_to_time_first(struct geom_path *path, real atime, struct geom_path_pos *pos)
{
	real alen;
	struct geom_curve *next;

	RASSERT(atime);
	pos->curve = geom_group_head(path);
	pos->alen_sum = 0;
	pos->index = 0;
	while (1) {
		if (atime <= (alen = geom_item_alength(pos->curve)))
			break;
		if (!(next = geom_group_next(pos->curve)))
			break;
		pos->curve = next;
		pos->alen_sum += alen;
		pos->index++;
		atime -= alen;
	}
	pos->atime_offset = atime;
}


/**
 * Converts next ATIME in nonempty path parameter to curve index and offset.
 *
 * @ai  path    Path.
 * @ai  atime   ATIME parameter (must be finite and in interval [previous atime, alength]).
 * @ao  index   Curve index (0 <= index < count).
 * @ao  offset  Curve offset (in interval [0-epsilon, 1+epsilon], always finite).
 */
static inline void geom_path_atime_to_time_next(struct geom_path *path, real atime, struct geom_path_pos *pos)
{ geom_path_atime_to_time_first(path, atime, pos); }


/* geom_path_atime_to_time */

real geom_path_atime_to_time(struct geom_path *path, real atime)
{
	struct geom_path_pos pos;

	FUNC;
	ITEM_ASSERT(path, path);
	RASSERT(atime);

	if (atime <= 0)
		RRET(0);
	RTEST_NAN(geom_path_update_alength(path));
	if (atime >= path->alength)
		RRET(geom_group_count(path));
	geom_path_atime_to_time_first(path, atime, &pos);
	RRET(pos.index + geom_item_atime_to_time(pos.curve, pos.atime_offset));
}


/* geom_path_atimes_to_times */

int geom_path_atimes_to_times(struct geom_path *path, uns count, real *atimes, real *times)
{
	struct geom_path_pos pos;
	struct geom_curve *block_curve;
	uns path_count, i, j, l, r, block_index, block_count;
	real t;

	FUNC;
	ITEM_ASSERT(path, path);
	GEOM_REALS_SORTED_ASSERT(count, atimes);
	PASSERT(times);

	path_count = geom_group_count(path);
	for (l = 0, r = count; l < r && atimes[l] <= 0; l++)
		times[l] = 0;
	RTEST_NUM(geom_path_update_alength(path));
	while (l < r && atimes[r - 1] >= path->alength)
		times[--r] = path_count;
	if (l == r)
		ERET(0);
	block_curve = NULL;
	block_index =block_count = 0; /* <- not necessary, warning elimination */
	for (i = l; i <= r; i++) {
		if (i < r) {
			if (i == l)
				geom_path_atime_to_time_first(path, atimes[i], &pos);
			else
				geom_path_atime_to_time_next(path, atimes[i], &pos);
			times[i] = pos.atime_offset;
		}
		if (i == r || pos.curve != block_curve) {
			if (block_curve) {
				if (block_count == 1)
					RTEST_NUM(times[i - 1] = block_index + geom_item_atime_to_time(block_curve, times[i - 1]));
				else if (block_count > 1) {
					j = i - block_count;
					ETEST_RET(geom_item_atimes_to_times(block_curve, block_count, times + j, times + j));
					for (; j < i; j++)
						RTEST_NUM(times[j] += block_index);
				}
			}
			block_curve = pos.curve;
			block_index = pos.index;
			block_count = 1;
		}
		else
			block_count++;
	}
	t = path_count;
	for (i = r - 1; (int)i >= (int)l; i--) {
		if (times[i] < 0)
			times[i] = 0;
		else if (times[i] > t)
			times[i] = t;
		t = times[i];
	}

	GEOM_REALS_SORTED_ASSERT(count, times);
	ERET(0);
}


/*----------------------------------------------*
 *                                              *
 *      PARAMETRIZATIONS:  TIME <-> RATIME      *
 *                                              *
 *----------------------------------------------*/


/* geom_path_time_to_ratime */

real geom_path_time_to_ratime(struct geom_path *path, real time)
{
	struct geom_path_pos pos;
	real atime;

	FUNC;
	ITEM_ASSERT(path, path);
	RASSERT(time);

	if (time <= 0)
		RRET(0);
	else if (time >= 1)
		RRET(1);
	RTEST_NAN(geom_path_update_alength(path));
	geom_path_time_to_atime_first(path, time, &pos);
	atime = pos.alen_sum + geom_item_time_to_atime(pos.curve, pos.time_offset);
	if (atime <= 0)
		RRET(0);
	else if (atime >= path->alength)
		RRET(1);
	else
		RRET(atime / path->alength);
}


/* geom_path_times_to_ratimes */

int geom_path_times_to_ratimes(struct geom_path *path, uns count, real *times, real *ratimes)
{
	uns i;
	real t;

	FUNC;
	ITEM_ASSERT(path, path);
	GEOM_REALS_SORTED_ASSERT(count, times);
	PASSERT(ratimes);

	if (!count)
		ERET(0);
	RTEST_NUM(geom_path_update_alength(path));
	ETEST_RET(geom_path_times_to_atimes(path, count, times, ratimes));
	if (path->alength > 0) {
		t = 0;
		for (i = 0; i < count; i++) {
			t = MAX(t, ratimes[i] / path->alength);
			ratimes[i] = t;
		}
	}
	GEOM_REALS_SORTED_ASSERT(count, ratimes);
	ERET(0);
}


/* geom_path_ratime_to_time */

real geom_path_ratime_to_time(struct geom_path *path, real ratime)
{
	struct geom_path_pos pos;
	real atime;

	FUNC;
	ITEM_ASSERT(path, path);
	RASSERT(ratime);

	if (ratime <= 0)
		RRET(0);
	else if (ratime >= 1)
		RRET(1);
	geom_path_update_alength(path);
	RTEST_NAN(atime = ratime * path->alength);
	geom_path_atime_to_time_first(path, atime, &pos);
	RRET(pos.index + geom_item_atime_to_time(pos.curve, pos.atime_offset));
}


/* geom_path_ratimes_to_times */

int geom_path_ratimes_to_times(struct geom_path *path, uns count, real *ratimes, real *times)
{
	uns i;
	real t;

	FUNC;
	ITEM_ASSERT(path, path);
	GEOM_REALS_SORTED_ASSERT(count, ratimes);
	PASSERT(times);

	if (!count)
		ERET(0);
	RTEST_NUM(geom_path_update_alength(path));
	if (path->alength > 0) {
		t = 0;
		for (i = 0; i < count; i++) {
			t = MAX(t, ratimes[i] * path->alength);
			times[i] = t;
		}
		GEOM_REALS_SORTED_ASSERT(count, times);
		ERET(geom_path_atimes_to_times(path, count, times, times));
	}
	else
		ERET(geom_path_atimes_to_times(path, count, ratimes, times));
}


/*----------------------------------------------*
 *                                              *
 *                   CONVERSIONS                *
 *                                              *
 *----------------------------------------------*/


 /* geom_path_expansion_append */

 int geom_path_expansion_append(struct geom_path *path, struct geom_fpath *expansion)
 {
 	struct geom_curve *curve;

 	FUNC;
	GEOM_GROUP_WALK(path, curve)
		ETEST_RET(geom_item_expansion_append(curve, expansion));
	ERET(0);
 }


/* geom_path_conversion_append */

int geom_path_conversion_append(struct geom_path *path, uns flags, struct geom_fpath *result)
{
	struct geom_curve *curve;

	ITEM_ASSERT(path, path);
	ITEM_ASSERT(fpath, result);
	
	FUNC;
	GEOM_GROUP_WALK(path, curve)
		ETEST_RET(geom_item_conversion_append(curve, flags, result));
	ERET(0);
}


/*----------------------------------------------*
 *                                              *
 *               TRANSFORMATIONS                *
 *                                              *
 *----------------------------------------------*/


/* geom_path_transform */

int geom_path_transform(struct geom_path *path, struct geom_transform *transform, struct geom_path *result)
{
	struct geom_path_curve *curve;

	FUNC;
	ITEM_ASSERT(path, path);
	GASSERT(transform, transform);
	INSTANCE_IS_ASSERT(path, result);

	if (transform->flags & GEOM_TRANSFORM_IDENTITY) {
		if (path != result)
			geom_item_copy(result, path);
	}
	else {
		if (path != result)
			geom_item_copy(result, path);
		GEOM_GROUP_WALK(result, curve)
			ETEST_RET(geom_item_transform(curve, transform, curve));
		/* FIXME: kdyz transformuju do jine cesty, daly by se spoustet transformce do jinych krivek */
		geom_path_after_change(result);
	}
	ERET(0);
}


/*----------------------------------------------*
 *                                              *
 *            GEOMETRIC PROPERTIES              *
 *                                              *
 *----------------------------------------------*/


/* geom_path_point_start */

int geom_path_point_start(struct geom_path *path, struct geom_point *result)
{
	struct geom_curve *curve;

	FUNC;
	ITEM_ASSERT(path, path);
	PASSERT(result);

	PTEST_ERR(curve = geom_group_head(path), PATH_EMPTY);
	ERET(geom_item_point_start(curve, result));
}


/* geom_path_point_end */

int geom_path_point_end(struct geom_path *path, struct geom_point *result)
{
	struct geom_curve *curve;

	FUNC;
	ITEM_ASSERT(path, path);
	PASSERT(result);

	PTEST_ERR(curve = geom_group_tail(path), PATH_EMPTY);
	ERET(geom_item_point_end(curve, result));
}


/* geom_path_point_at_time */

int geom_path_point_at_time(struct geom_path *path, real time, struct geom_point *result)
{
	struct geom_path_pos pos;

	FUNC;
	ITEM_ASSERT(path, path);
	RASSERT(time);
	PASSERT(result);

	if (time <= 0)
		ERET(geom_path_point_start(path, result));
	else if (time >= geom_group_count(path))
		ERET(geom_path_point_end(path, result));
	PTEST_ERR(geom_path_time_to_curve(path, time, &pos), PATH_EMPTY);
	ERET(geom_item_point_at_time(pos.curve, pos.time_offset, result));
}


/* geom_path_derivation_start */

int geom_path_derivation_start(struct geom_path *path, struct geom_vector *result)
{
	struct geom_curve *curve;

	FUNC;
	ITEM_ASSERT(path, path);
	PASSERT(result);

	PTEST_ERR(curve = geom_group_head(path), PATH_EMPTY);
	ERET(geom_item_derivation_start(curve, result));
}


/* geom_path_derivation_end */

int geom_path_derivation_end(struct geom_path *path, struct geom_vector *result)
{
	struct geom_curve *curve;

	FUNC;
	ITEM_ASSERT(path, path);
	PASSERT(result);

	PTEST_ERR(curve = geom_group_tail(path), PATH_EMPTY);
	ERET(geom_item_derivation_end(curve, result));
}


/* geom_path_derivation_at_time */

int geom_path_derivation_at_time(struct geom_path *path, const real time, struct geom_vector *result)
{
	struct geom_path_pos pos;

	FUNC;
	ITEM_ASSERT(path, path);
	RASSERT(time);
	PASSERT(result);

	if (time <= 0)
		ERET(geom_path_derivation_start(path, result));
	else if (time >= geom_group_count(path))
		ERET(geom_path_derivation_end(path, result));
	PTEST_ERR(geom_path_time_to_curve(path, time, &pos), PATH_EMPTY);
	ERET(geom_item_derivation_at_time(pos.curve, pos.time_offset, result));
}


/* geom_path_bbox */

int geom_path_bbox(struct geom_path *path, struct geom_rectangle *bbox)
{
	struct geom_curve *curve;
	struct geom_rectangle temp;

	FUNC;
	ITEM_ASSERT(path, path);
	PASSERT(bbox);

	PTEST_ERR(curve = geom_group_head(path), PATH_EMPTY);
	ETEST_RET(geom_item_bbox(curve, bbox));
	while (curve = geom_group_next(curve)) {
		ETEST_RET(geom_item_bbox(curve, &temp));
		if (bbox->ll.x > temp.ll.x) bbox->ll.x = temp.ll.x;
		if (bbox->ll.y > temp.ll.y) bbox->ll.y = temp.ll.y;
		if (bbox->ur.x < temp.ur.x) bbox->ur.x = temp.ur.x;
		if (bbox->ur.y < temp.ur.y) bbox->ur.y = temp.ur.y;
	}
	ERET(0);
}


/* geom_path_nearest_to_point */

int geom_path_nearest_to_point(struct geom_path *path, struct geom_point *point, uns flags, struct geom_nearest *nearest)
{
	struct geom_curve *curve;
	struct geom_nearest temp;
	uns index;

	FUNC;
	ITEM_ASSERT(path, path);
	GASSERT(point, point);
	BASSERT(!(flags & ~GEOM_NEAREST_ALL));
	PASSERT(nearest);

	PTEST_ERR(curve = geom_group_head(path), PATH_EMPTY);
	ETEST_RET(geom_item_nearest_to_point(curve, point, flags, nearest));
	for (index = 1; curve = geom_group_next(curve); index++) {
		ETEST_RET(geom_item_nearest_to_point(curve, point, flags, &temp));
		if (temp.dist < nearest->dist) {
			if (flags & GEOM_NEAREST_TIME)
				RTEST_NUM(temp.time += index);
			*nearest = temp;
		}
	}
	ERET(0);
}


/* geom_path_direction_times */

int geom_path_direction_times(struct geom_path *path, struct geom_vector *direction, struct garr *result)
{
	struct geom_curve *curve;
	uns m, i, n = 0, index = 0;
	real time, last = 0;

	FUNC;
	ITEM_ASSERT(path, path);
	GASSERT(vector, direction);
	PASSERT(result);

	GEOM_GROUP_WALK(path, curve) {
		ETEST_RET((int)(m = geom_item_direction_times(curve, direction, result)));
		for (i = 0; i < m; i++) {
			RTEST_NUM(time = *(real *)garr_get(result, i) + index);
			*(real *)garr_get(result, i) = last = MAX(last, time);
		}
		garr_seek(result, m);
		n += m;
		index++;
	}
	garr_seek(result, -n);
	ERET(n);
}


/* geom_path_distance_times */

int geom_path_distance_times(struct geom_path *path, struct geom_point *point, real dist, struct garr *result)
{
	struct geom_curve *curve;
	uns m, i, n = 0, index = 0;
	real time, last = 0;

	FUNC;
	ITEM_ASSERT(path, path);
	RASSERT(dist);
	PASSERT(result);

	GEOM_GROUP_WALK(path, curve) {
		ETEST_RET((int)(m = geom_item_distance_times(curve, point, dist, result)));
		for (i = 0; i < m; i++) {
			RTEST_NUM(time = (*(real *)garr_get(result, i)) + index);
			last = MAX(last, time);
			*(real *)garr_get(result, i) = last;
		}
		garr_seek(result, m);
		n += m;
		index++;
	}
	garr_seek(result, -n);
	ERET(n);
}


/* geom_paths_intersections */

#define ASORT_PREFIX(x) geom_paths_intersections_sort
#define ASORT_KEY_TYPE struct geom_intersection
#define ASORT_ELT(x) (*(struct geom_intersection *)garr_get(garr, (x)))
#define ASORT_LT(x,y) ((x).time1 < (y).time1)
#define ASORT_EXTRA_ARGS , struct garr *garr
#include "lib/arraysort.h"

int geom_paths_intersections(struct geom_path *path1, struct geom_path *path2, uns flags, struct garr *result)
{
	uns index1 = 0, index2, n = 0, i, m;
	struct geom_curve *curve1, *curve2;
	struct geom_intersection *inter;

	FUNC;
	ITEM_ASSERT(path, path1);
	ITEM_ASSERT(path, path2);
	BASSERT(!(flags & ~GEOM_INTERSECTION_ALL));
	PASSERT(result);

	/* FIXME: slo by udelat rychlejsi napr. rekurzinim delenim cest s orezavanim */

	GEOM_GROUP_WALK(path1, curve1) {
		index2 = 0;
		GEOM_GROUP_WALK(path2, curve2) {
			ETEST_RET((int)(m = geom_items_intersections(curve1, curve2, flags, result)));
			for (i = 0; i < m; i++) {
				inter = (struct geom_intersection *)garr_get(result, i);
				if (flags & GEOM_INTERSECTION_TIME1)
					RTEST_NUM(inter->time1 += index1);
				if (flags & GEOM_INTERSECTION_TIME2)
					RTEST_NUM(inter->time2 += index2);
			}
			garr_seek(result, m);
			n += m;
			index2++;
		}
		index1++;
	}
	garr_seek(result, -n);
	if (flags & GEOM_INTERSECTION_TIME1 && n > 1)
		geom_paths_intersections_sort(n, result);
	ERET(n);
}





#if 0

/* geom_path_solve_y */

int geom_path_solve_y(struct geom_path *path, real min_time, real max_time, real y, uns flags, struct geom_curve_point *result)
{
	uns min_index, max_index, index;
	int err;
	struct geom_path_item *item;
	real time1, time2;

	FUNC;

	if (!path->count)
		ERET(0);

	if (min_time < 1)
		min_index = 0;
	else if (min_time >= path->count)
		min_index = path->count - 1;
	else {
		min_index = floor(min_time);
		min_index = MINMAX((int)min_index, 0, (int)(path->count - 1));
	}

	if (max_time <= 1)
		max_index = 0;
	else if (max_time >= path->count)
		max_index = path->count - 1;
	else {
		max_index = ceil(max_time);
		max_index = MINMAX((int)max_index, 0, (int)(path->count - 1));
	}

	if (min_index > max_index)
		ERET(0);

	item = GEOM_PATH_ITEM(path, min_index);
	for (index = min_index; index <= max_index; item = GEOM_PATH_NEXT_ITEM(path, item), index++) {
		if (index == min_index)
			RTEST_NUM(time1 = fmaxr(0, min_time - index));
		else
			time1 = 0;
		if (index == max_index)
			RTEST_NUM(time2 = fminr(1, max_time - index));
		else
			time2 = 1;
		ETEST_RET(err = geom_curve_solve_y(item->flags & GEOM_CURVE_TYPE, item->ptr, time1, time2, y, flags, result));
		if (err > 0) {
			if (flags & GEOM_CURVE_POINT_TIME)
				RTEST_NUM(result->time = result->time + index);
			ERET(1);
		}
	}
	ERET(0);
}

#endif

