#ifndef _VRR_GEOMLIB_POLYNOMIAL_H_
#define _VRR_GEOMLIB_POLYNOMIAL_H_

/** @file
 *
 * GEOMLIB - Polynomial solver.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "geomlib/base.h"


/* geom_polynomial_solve flags */
#define GEOM_SOLVE_MULTIPLICITY   0x00000001
#define GEOM_SOLVE_LEFT_ONLY      0x00000002
#define GEOM_SOLVE_UNIT_INTERVAL  0x00000004


/**
 * Test for validity of given polynomial.
 *
 * @ai  degree  Degree of the polynomial.
 * @ai  coef    Polynomial coefficients (in order of increasing powers).
 *
 * @rv  nonzero  Polynomial is valid.
 * @rv  zero     Polynomial is not valid.
 */
int geom_polynomial_valid(uns degree, double coef[]);


/**
 * Function evaluates a real polynomial P at given parameter t.
 *
 * @ai  degree  Degree of the polynomial.
 * @ai  coef    Polynomial coefficients (in order of increasing powers).
 * @ai  t       Real parameter to evaluate.
 *
 * @rv  finite     P(t).
 * @rv  otherwise  Numeric error.  
 */
double geom_polynomial_evaluate(uns degree, double coef[], double t);

/**
 * Function returns product of two real polynomials P(t) * Q(t).
 *
 * @ai  degree1  Degree of the first polynomial.
 * @ai  coef1    Coefficients of the first polynomial.
 * @ai  degree2  Degree of the second polynomial.
 * @ai  coef2    Coefficients of the second polynomial.
 * @ao  result   Array to store coefficients of resulting polynomial (of size degree1+degree2+1).
 *
 * @rv  degree1+degree2   Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_polynomial_mul(uns degree1, double coef1[], uns degree2, double coef2[], double result[]);

/**
 * Function finds real roots of a real polynomial.
 *
 * @ai  degree  Degree of the polynomial.
 * @ai  coef    Coefficients of the polynomial (in order of increasing powers).
 * @ai  flags   OR combination of GEOM_SOLVE_x flags.
 *              GEOM_SOLVE_MULTIPLICITY - returns multiple roots as multiple array entries.
 *              GEOM_SOLVE_LEFT_ONLY - returns at most one root (leftmost).
 *              GEOM_SOLVE_UNIT_INTERVAL - returns only roots in closed interval [0,1].
 * @ai  result  Array to store resulting roots (degree entries needed or one with GEOM_SOLVE_LEFT_ONLY).
 *
 * @rv  >=0               Number of roots found.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_polynomial_solve(uns degree, double coef[], uns flags, double result[]);


#endif
