/*
 * 	GEOMLIB - Rectangles.
 *
 *	Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 *	This program is distributed under GNU General Public License, see COPYING.
 */

#define GEOM_SOURCE
#include "geomlib/rectangle.h"


/* geom_rectangle_valid */

int geom_rectangle_valid(const struct geom_rectangle *r)
{
	return PTEST(r) && TEST(geom_point_valid(&r->ll)) && TEST(geom_point_valid(&r->ur));
}


/* geom_rectangle_assert */

const struct geom_rectangle *geom_rectangle_assert(GEOM_ASSERT_DEF, const struct geom_rectangle *r)
{
	geom_p_assert(GEOM_ASSERT_COPY, r);
	if (!(TEST(geom_point_valid(&r->ll)) && TEST(geom_point_valid(&r->ur))))
		die(GEOM_ASSERT_FORMAT ": rectangle (%f, %f, %f, %f) at %p is invalid", GEOM_ASSERT_COPY, r->ll.x, r->ll.y, r->ur.x, r->ur.y, (void *)r);
	return r;
}


/* geom_rectangle_dump */

void geom_rectangle_dump(const struct geom_rectangle *r)
{
	DUMP_ENTER(1, "Dumping geom_rectangle %p: ", (void *)r);
	DUMP_PTR(r) {
		SDUMP("%f", r->ll.x);
		SDUMP("%f", r->ll.y);
		SDUMP("%f", r->ur.x);
		SDUMP("%f", r->ur.y);
	}
	DUMP_LEAVE;
}





int
geom_rectangle_is_valid (const struct geom_rectangle *r)
{
  return ((r == NULL) ||
	  (geom_point_valid (&(r -> ll)) &&
	   geom_point_valid (&(r -> ur))));
}

int

geom_rectangle_is_empty (const struct geom_rectangle *r)
{
  return ((r == NULL) ||
	  (r -> ll . x > r -> ur . x ) ||
	  (r -> ll . y > r -> ur . y ));
}

int
geom_rectangle_set_empty (struct geom_rectangle *r)
{
  if (r == NULL) return (0);
  r -> ll . x =  r -> ll . y = 1;
  r -> ur . x =  r -> ur . y = 0;
  return (0);
}

int
geom_rectangle_init_from_points (struct geom_rectangle *r, const struct geom_point *p1, const struct geom_point *p2)
{
  BASSERT (r && p1 && p2);

  r -> ll . x = MIN (p1 -> x, p2 -> x);
  r -> ll . y = MIN (p1 -> y, p2 -> y);
  r -> ur . x = MAX (p1 -> x, p2 -> x);
  r -> ur . y = MAX (p1 -> y, p2 -> y);

  return (geom_rectangle_is_valid (r) ? 0 : -1);
}


int
geom_rectangle_merge (const struct geom_rectangle *r1,
		      const struct geom_rectangle *r2,
		      struct geom_rectangle *result)
{
	if (geom_rectangle_is_empty (r1)) {
		if (r2)
			*result = *r2;
		else
			geom_rectangle_set_empty(result);
		return 0;
	}
	if (geom_rectangle_is_empty (r2)) return (*result = *r1, 0);

	BASSERT (geom_rectangle_is_valid (r1) && geom_rectangle_is_valid (r2) && result);

	result -> ll . x = MIN (r1 -> ll . x, r2 -> ll . x);
	result -> ll . y = MIN (r1 -> ll . y, r2 -> ll . y);
	result -> ur . x = MAX (r1 -> ur . x, r2 -> ur . x);
	result -> ur . y = MAX (r1 -> ur . y, r2 -> ur . y);

	return (0);
}

int
geom_rectangle_merge_point (const struct geom_rectangle *r,
		      const struct geom_point *p,
		      struct geom_rectangle *result)
{
	if (geom_rectangle_is_empty (r)) {
		if (p)
			*result = (struct geom_rectangle) { *p, *p };
		else
			geom_rectangle_set_empty(result);
		return 0;
	}

	BASSERT (geom_rectangle_is_valid (r) && p && result);

	result -> ll . x = MIN (r -> ll . x, p -> x);
	result -> ll . y = MIN (r -> ll . y, p -> y);
	result -> ur . x = MAX (r -> ur . x, p -> x);
	result -> ur . y = MAX (r -> ur . y, p -> y);

	return (0);
}

int
geom_rectangle_enlarge (const struct geom_rectangle *source, real value,
		        struct geom_rectangle *result)
{
	GASSERT(rectangle, source);
	RASSERT(value);
	PASSERT(result);

	if (geom_rectangle_is_empty (source)) {
		geom_rectangle_set_empty (result);
		return 0;
	}

	result -> ll . x = inf2finr(source -> ll . x - value);
	result -> ll . y = inf2finr(source -> ll . y - value);
	result -> ur . x = inf2finr(source -> ur . x + value);
	result -> ur . y = inf2finr(source -> ur . y + value);
	
	return 0;
}

int
geom_rectangle_have_intersection (const struct geom_rectangle *r1,
				  const struct geom_rectangle *r2)
{
  BASSERT (geom_rectangle_is_valid (r1) && geom_rectangle_is_valid (r2));

  if (geom_rectangle_is_empty (r1) ||
      geom_rectangle_is_empty (r2)) return (0);

  return (((r1 -> ll . x > r2 -> ur . x) ||
	   (r2 -> ll . x > r1 -> ur . x) ||
	   (r1 -> ll . y > r2 -> ur . y) ||
	   (r2 -> ll . y > r1 -> ur . y)) ? 0 : 1);
}

int
geom_rectangle_is_in (const struct geom_rectangle *r1, const struct geom_rectangle *r2)
{
  BASSERT (geom_rectangle_is_valid (r1) && geom_rectangle_is_valid (r2));

  if (geom_rectangle_is_empty (r1)) return (1);
  if (geom_rectangle_is_empty (r2)) return (0);

  return ((r2 -> ll . x < r1 -> ll . x) &&
	  (r2 -> ll . y < r1 -> ll . y) &&
	  (r1 -> ur . x < r2 -> ur . x) &&
	  (r1 -> ur . y < r2 -> ur . y));
}

int
geom_rectangle_is_adjoin (const struct geom_rectangle *r1, const struct geom_rectangle *r2)
{
  BASSERT (geom_rectangle_is_valid (r1) && geom_rectangle_is_valid (r2));

  if (geom_rectangle_is_empty (r1) ||
      geom_rectangle_is_empty (r2)) return (0);

  return ((r2 -> ll . x == r1 -> ll . x) ||
	  (r2 -> ll . y == r1 -> ll . y) ||
	  (r1 -> ur . x == r2 -> ur . x) ||
	  (r1 -> ur . y == r2 -> ur . y));
}
