/*
 * GEOMLIB - Segments.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#define GEOM_SOURCE
#include "geomlib/vector.h"
#include "geomlib/transform.h"
#include "geomlib/bezier.h"
#include "geomlib/fpath.h"
#include "geomlib/approx.h"
#include "geomlib/segment.h"


/*----------------------------------------------*
 *                                              *
 *              PARAMETRIZATIONS                *
 *                                              *
 *----------------------------------------------*/


/* geom_segment_alength */

#define GEOM_SEGMENT_ALENGTH(s) (hypot((double)(s)->a.x - (double)(s)->b.x, (double)(s)->a.y - (double)(s)->b.y))

real geom_segment_alength(struct geom_segment *segment)
{
	FUNC;
	ITEM_ASSERT(segment, segment);
	RRET(GEOM_SEGMENT_ALENGTH(segment));
}


/* geom_segment_time_to_atime */

real geom_segment_time_to_atime(struct geom_segment *segment, real time)
{
	double alength;

	FUNC;
	ITEM_ASSERT(segment, segment);
	RASSERT(time);

	if (time <= 0)
		RRET(0);
	alength = GEOM_SEGMENT_ALENGTH(segment);
	RRET((time >= 1) ? alength : time * alength);
}


/* geom_segment_atime_to_time */

real geom_segment_atime_to_time(struct geom_segment *segment, real atime)
{
	double alength;

	FUNC;
	ITEM_ASSERT(segment, segment);
	RASSERT(atime);

	if (atime <= 0)
		RRET(0);
	alength = GEOM_SEGMENT_ALENGTH(segment);
	if (!isfinite(alength))
		RET_NAN;
	if (atime >= alength)
		RRET(1);
	RRET(atime / alength);
}


/*----------------------------------------------*
 *                                              *
 *                   CONVERSIONS                *
 *                                              *
 *----------------------------------------------*/


int geom_segment_expansion_append(struct geom_segment *segment, struct geom_fpath *expansion)
{
	struct geom_bezier *bezier;
	FUNC;
	GEOM_FPATH_APPEND_NEW_CREATE2(bezier, expansion, bezier, geom_bezier_set_linear_nonrational, &segment->a, &segment->b);
	ERET(0);
}

int geom_segment_to_bezier(struct geom_segment *segment, struct geom_bezier *bezier)
{
	FUNC;
	geom_bezier_set_linear_nonrational(bezier, &segment->a, &segment->b);
	ERET(0);
}

int geom_segment_conversion_append(struct geom_segment *segment, uns flags, struct geom_fpath *result)
{
	struct geom_segment *s;

	if (flags & GEOM_APPROX_SEGMENT) {
		GEOM_FPATH_APPEND_NEW_CREATE2(s, result, segment, geom_item_copy, segment);
		return 0;
	}
	else
		return geom_item_conversion_append__expand(segment, flags, result);
}

/*----------------------------------------------*
 *                                              *
 *               TRANSFORMATIONS                *
 *                                              *
 *----------------------------------------------*/


/* geom_segment_transform */

int geom_segment_transform(struct geom_segment *segment, struct geom_transform *transform, struct geom_segment *result)
{
	FUNC;
	ITEM_ASSERT(segment, segment);
	GASSERT(transform, transform);
	INSTANCE_IS_ASSERT(segment, result);

	ETEST_RET(geom_point_transform(&segment->a, transform, &result->a));
	ETEST_RET(geom_point_transform(&segment->b, transform, &result->b));
	geom_segment_after_change(result);
	ERET(0);
}


/*----------------------------------------------*
 *                                              *
 *             GEOMETRIC PROPERTIES             *
 *                                              *
 *----------------------------------------------*/


/* geom_segment_point_at_time */

int geom_segment_point_at_time(struct geom_segment *segment, real time, struct geom_point *result)
{
	FUNC;
	ITEM_ASSERT(segment, segment);
	RASSERT(time);
	PASSERT(result);

	if (time <= 0)
		*result = segment->a;
	else if (time >= 1)
		*result = segment->b;
	else {
		result->x = fcoeffinr(segment->a.x, segment->b.x, time);
		result->y = fcoeffinr(segment->a.y, segment->b.y, time);
	}
	ERET(0);
}


/* geom_segment_derivation */

int geom_segment_derivation(struct geom_segment *segment, struct geom_vector *result)
{
	FUNC;
	ITEM_ASSERT(segment, segment);
	PASSERT(result);

	RTEST_NUM(result->dx = segment->b.x - segment->a.x);
	RTEST_NUM(result->dy = segment->b.y - segment->a.y);
	ERET(0);
}


/* geom_segment_bbox */

int geom_segment_bbox(struct geom_segment *segment, struct geom_rectangle *bbox)
{
	FUNC;
	ITEM_ASSERT(segment, segment);
	PASSERT(bbox);

	bbox->ll.x = MIN(segment->a.x, segment->b.x);
	bbox->ur.x = MAX(segment->a.x, segment->b.x);
	bbox->ll.y = MIN(segment->a.y, segment->b.y);
	bbox->ur.y = MAX(segment->a.y, segment->b.y);
	ERET(0);
}





