/*
 * GEOMLIB - Affine transformations.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#define GEOM_SOURCE
#include "geomlib/base.h"
#include "geomlib/vector.h"
#include "geomlib/rectangle.h"
#include "geomlib/bezier.h"
#include "geomlib/path.h"
#include "geomlib/transform.h"


/* predefined transformation matrices */

const struct geom_transform geom_transform_identity = {
	GEOM_TRANSFORM_IDENTITY | GEOM_TRANSFORM_SIMILAR,
	{
		{ 1, 0, 0 },
		{ 0, 1, 0 }
	}
};

const struct geom_transform geom_transform_hflip = {
	GEOM_TRANSFORM_SIMILAR,
	{
		{ -1, 0, 0},
		{ 0,  1, 0}
	}
};

const struct geom_transform geom_transform_vflip = {
	GEOM_TRANSFORM_SIMILAR,
	{
		{ 1, 0, 0},
		{ 0, -1, 0}
	}
};

const struct geom_transform geom_transform_rot90 = {
	GEOM_TRANSFORM_SIMILAR,
	{
		{ 0, 1, 0},
		{ -1, 0, 0}
	}
};

const struct geom_transform geom_transform_rot180 = {
	GEOM_TRANSFORM_SIMILAR,
	{
		{ -1, 0, 0},
		{ 0, -1, 0}
	}
};

const struct geom_transform geom_transform_rot270 = {
	GEOM_TRANSFORM_SIMILAR,
	{
		{ 0, -1, 0},
		{ 1, 0, 0}
	}
};

/* geom_transform_valid */

int geom_transform_valid(const struct geom_transform *t)
{
	int i, j;

	if (!t || t->flags & ~(GEOM_TRANSFORM_IDENTITY | GEOM_TRANSFORM_SIMILAR))
		return 0;

	for (i = 0; i < 2; i++)
		for (j = 0; j < 3; j++)
			if (!isfinite(t->coef[i][j]))
				return 0;

	if (t->flags & GEOM_TRANSFORM_IDENTITY)
		for (i = 0; i < 2; i++)
			for (j = 0; j < 3; j++)
				if (t->coef[i][j] != ((i == j) ? 1 : 0))
					return 0;

	return 1;
}


/* geom_transform_assert */

const struct geom_transform *geom_transform_assert(GEOM_ASSERT_DEF, const struct geom_transform *t)
{
	geom_p_assert(GEOM_ASSERT_COPY, t);
	if (!TEST(geom_transform_valid(t))) {
		msg(GEOM_ASSERT_FORMAT ": transformation at %p is invalid\n", GEOM_ASSERT_COPY, (void *)t);
		geom_transform_dump(t);
		BUG();
	}
	return t;
}


/* geom_transform_dump */

void geom_transform_dump(const struct geom_transform *t)
{
	DUMP_ENTER(1, "Dumping geom_transform %p: ", (void *)t);
	DUMP_PTR(t) {
		SDUMP("flags=0x%x", t->flags);
		SDUMP("Tx=%fx+%fy+%f", t->coef[0][0], t->coef[0][1], t->coef[0][2]);
		SDUMP("Ty=%fx+%fy+%f", t->coef[1][0], t->coef[1][1], t->coef[1][2]);
	}
	DUMP_LEAVE;
}


/* geom_transform_set */

int geom_transform_set(struct geom_transform *t, real a, real b, real c, real d, real e, real f)
{
	PASSERT(t);
	RASSERT(a);
	RASSERT(b);
	RASSERT(c);
	RASSERT(d);
	RASSERT(e);
	RASSERT(f);

	if (IS_EQUAL(a, 1) && IS_ZERO(b) && IS_ZERO(c) && IS_ZERO(d) && IS_EQUAL(e, 1) && IS_ZERO(f))
		*t = geom_transform_identity;
	else {
		t->flags = 0;
		t->coef[0][0] = a;
		t->coef[0][1] = b;
		t->coef[0][2] = c;
		t->coef[1][0] = d;
		t->coef[1][1] = e;
		t->coef[1][2] = f;
	}
	return 0;
}


/* geom_transform_set_move */

int geom_transform_set_move(struct geom_transform *t, const real mx, const real my)
{
	PASSERT(t);
	RASSERT(mx);
	RASSERT(my);

	if (IS_ZERO(mx) && IS_ZERO(my))
		*t = geom_transform_identity;
	else {
		t->flags = GEOM_TRANSFORM_SIMILAR;
		t->coef[0][0] = t->coef[1][1] = 1;
		t->coef[0][1] = t->coef[1][0] = 0;
		t->coef[0][2] = mx;
		t->coef[1][2] = my;
	}
	return 0;
}


/* geom_transform_set_scale */

int geom_transform_set_scale(struct geom_transform *t, const real sx, const real sy)
{
	PASSERT(t);
	RASSERT(sx);
	RASSERT(sy);

	if (IS_EQUAL(sx, 1) && IS_EQUAL(sy, 1))
		*t = geom_transform_identity;
	else {
		if (IS_EQUAL(sx, sy)) {
			t->flags = GEOM_TRANSFORM_SIMILAR;
			t->coef[0][0] = t->coef[1][1] = MAX(sx, sy);
		}
		else {
			t->flags = 0;
			t->coef[0][0] = sx;
			t->coef[1][1] = sy;
		}
		t->coef[0][1] = t->coef[0][2] = t->coef[1][0] = t->coef[1][2] = 0;
	}

	return 0;
}


/* geom_transform_set_center_scale */

int geom_transform_set_center_scale(struct geom_transform *t, const struct geom_point *center, const real sx, const real sy)
{
	real sx1, sy1;

	PASSERT(t);
	GASSERT(point, center);
	RASSERT(sx);
	RASSERT(sy);

	sx1 = 1 - sx;
	sy1 = 1 - sy;

	if (IS_ZERO(sx1) && IS_ZERO(sy1))
		*t = geom_transform_identity;
	else {
		if (IS_EQUAL(sx, sy)) {
			t->flags = GEOM_TRANSFORM_SIMILAR;
			t->coef[0][0] = t->coef[1][1] = MAX(sx, sy);
		}
		else {
			t->flags = 0;
			t->coef[0][0] = sx;
			t->coef[1][1] = sy;
		}
		t->coef[0][1] = 0;
		t->coef[1][0] = 0;
		t->coef[0][2] = sx1 * center->x;
		t->coef[1][2] = sy1 * center->y;
		if (!isfinite(t->coef[0][2]) || !isfinite(t->coef[1][2]))
			return GEOM_ERR_NUMERIC;
	}
	return 0;
}


/* geom_transform_set_rotation */

int geom_transform_set_rotation(struct geom_transform *t, const real angle)
{
	real cos1, sin1;

	PASSERT(t);
	RASSERT(angle);

	cos1 = cos(angle);
	sin1 = sin(angle);

	if (IS_EQUAL(cos1, 1) && IS_ZERO(sin1))
		*t = geom_transform_identity;
	else {
		t->flags = GEOM_TRANSFORM_SIMILAR;
		t->coef[0][2] = t->coef[1][2] = 0;
		t->coef[0][0] = t->coef[1][1] = cos1;
		t->coef[0][1] = -sin1;
		t->coef[1][0] = -t->coef[0][1];
	}
	return 0;
}


/* geom_transform_set_center_rotation */

int geom_transform_set_center_rotation(struct geom_transform *t, const struct geom_point *center, const real rotation)
{
	real cos1, sin1, cos2;

	PASSERT(t);
	GASSERT(point, center);
	RASSERT(rotation);

	cos1 = cos(rotation);
	sin1 = sin(rotation);
	cos2 = 1 - cos1;

	if (IS_ZERO(cos1) && IS_ZERO(cos2))
		*t = geom_transform_identity;
	else {
		t->flags = GEOM_TRANSFORM_SIMILAR;
		t->coef[0][0] = cos1;
		t->coef[0][1] = -sin1;
		t->coef[1][0] = sin1;
		t->coef[1][1] = cos1;
		t->coef[0][2] = cos2 * center->x + sin1 * center->y;
		t->coef[1][2] = cos2 * center->y - sin1 * center->x;
		if (!isfinite(t->coef[0][2]) || !isfinite(t->coef[1][2]))
			return GEOM_ERR_NUMERIC;
	}
	return 0;
}


/* geom_transform_set_slant */

int geom_transform_set_slant(struct geom_transform *t, const real a, const real b)
{
	PASSERT(t);
	RASSERT(a);
	RASSERT(b);

	if (IS_ZERO(a) && IS_ZERO(b))
		*t = geom_transform_identity;
	else {
		t->flags = 0;
		t->coef[0][0] = t->coef[1][1] = 1;
		t->coef[0][2] = t->coef[1][2] = 0;
		t->coef[0][1] = a;
		t->coef[1][0] = b;
	}
	return 0;
}


/* geom_transform_merge */

int geom_transform_merge(const struct geom_transform *t1, const struct geom_transform *t2, struct geom_transform *result)
{
	uns i, j;
	struct geom_transform tmp;

	GASSERT(transform, t1);
	GASSERT(transform, t2);
	PASSERT(result);

	if (t1->flags & GEOM_TRANSFORM_IDENTITY)
		*result = *t2;
	else if (t2->flags & GEOM_TRANSFORM_IDENTITY)
		*result = *t1;
	else {
		tmp.flags = (t1->flags & t2->flags) & GEOM_TRANSFORM_SIMILAR;
		tmp.coef[0][0] = t1->coef[0][0] * t2->coef[0][0] + t1->coef[1][0] * t2->coef[0][1];
		tmp.coef[0][1] = t1->coef[0][1] * t2->coef[0][0] + t1->coef[1][1] * t2->coef[0][1];
		tmp.coef[0][2] = t1->coef[0][2] * t2->coef[0][0] + t1->coef[1][2] * t2->coef[0][1] + t2->coef[0][2];
		tmp.coef[1][0] = t1->coef[0][0] * t2->coef[1][0] + t1->coef[1][0] * t2->coef[1][1];
		tmp.coef[1][1] = t1->coef[0][1] * t2->coef[1][0] + t1->coef[1][1] * t2->coef[1][1];
		tmp.coef[1][2] = t1->coef[0][2] * t2->coef[1][0] + t1->coef[1][2] * t2->coef[1][1] + t2->coef[1][2];
		for (i = 0; i < 2; i++)
			for (j = 0; j < 3; j++)
				if (!isfinite(tmp.coef[i][j]))
					return GEOM_ERR_NUMERIC;
		*result = tmp;
	}
	return 0;
}


/* geom_transform_invert */

int geom_transform_invert(const struct geom_transform *t, struct geom_transform *result)
{
	uns i, j;
	real det;
	struct geom_transform tmp;

	GASSERT(transform, t);
	PASSERT(result);

	if (!(t->flags & GEOM_TRANSFORM_IDENTITY)) {
		det = t->coef[0][0] * t->coef[1][1] - t->coef[0][1] * t->coef[1][0];
		if (IS_ZERO(det))
			return GEOM_ERR_TRANSFORM_SINGULAR;
		det = 1 / det;
		tmp.flags = t->flags & GEOM_TRANSFORM_SIMILAR;
		tmp.coef[0][0] = t->coef[1][1] * det;
		tmp.coef[0][1] = -t->coef[0][1] * det;
		tmp.coef[0][2] = (t->coef[0][1] * t->coef[1][2] - t->coef[0][2] * t->coef[1][1]) * det;
		tmp.coef[1][0] = -t->coef[1][0] * det;
		tmp.coef[1][1] = t->coef[0][0] * det;
		tmp.coef[1][2] = (t->coef[1][0] * t->coef[0][2] - t->coef[0][0] * t->coef[1][2]) * det;
		for (i = 0; i < 2; i++)
			for (j = 0; j < 3; j++)
				if (!isfinite(tmp.coef[i][j]))
					return GEOM_ERR_NUMERIC;
		*result = tmp;
	}
	else if (t != result)
		*result = geom_transform_identity;
	return 0;
}


/* geom_transform_is_linear */

int geom_transform_is_linear(const struct geom_transform *t)
{
	GASSERT(transform, t);

	return IS_ZERO(t->coef[0][2]) && IS_ZERO(t->coef[1][2]);
}


/* geom_transform_is_regular */

int geom_transform_is_regular(const struct geom_transform *t)
{
	real det;

	GASSERT(transform, t);

	det = t->coef[0][0] * t->coef[1][1] - t->coef[0][1] * t->coef[1][0];
	if (!isfinite(det))
		return GEOM_ERR_NUMERIC;
	return IS_ZERO(det) ? 0 : 1;
}


/* geom_transform_rank */

int geom_transform_rank(const struct geom_transform *t)
{
	real det;

	GASSERT(transform, t);

	det = t->coef[0][0] * t->coef[1][1] - t->coef[0][1] * t->coef[1][0];
	if (!isfinite(det))
		return GEOM_ERR_NUMERIC;
	if (!IS_ZERO(det))
		return 2;
	if (IS_ZERO(t->coef[0][0]) && IS_ZERO(t->coef[0][1]) && IS_ZERO(t->coef[1][0]) && IS_ZERO(t->coef[1][1]))
		return 0;
	return 1;
}


/* geom_point_transform */

int geom_point_transform(const struct geom_point *p, const struct geom_transform *t, struct geom_point *result)
{
	struct geom_point tmp;

	GASSERT(point, p);
	GASSERT(transform, t);
	PASSERT(result);

	if (t->flags & GEOM_TRANSFORM_IDENTITY)
		*result = *p;
	else {
		tmp.x = t->coef[0][0] * p->x + t->coef[0][1] * p->y + t->coef[0][2];
		tmp.y = t->coef[1][0] * p->x + t->coef[1][1] * p->y + t->coef[1][2];
		if (!isfinite(tmp.x) || !isfinite(tmp.y))
			return GEOM_ERR_NUMERIC;
		*result = tmp;
	}
	return 0;
}


/* geom_point_w_transform */

int geom_point_w_transform(struct geom_point_w *p, struct geom_transform *t, struct geom_point_w *result)
{
	struct geom_point tmp;

	GASSERT(point_w, p);
	GASSERT(transform, t);
	PASSERT(result);

	if (t->flags & GEOM_TRANSFORM_IDENTITY)
		*result = *p;
	else {
		tmp.x = t->coef[0][0] * p->x + t->coef[0][1] * p->y + t->coef[0][2];
		tmp.y = t->coef[1][0] * p->x + t->coef[1][1] * p->y + t->coef[1][2];
		if (!isfinite(tmp.x) || !isfinite(tmp.y))
			return GEOM_ERR_NUMERIC;
		result->x = tmp.x;
		result->y = tmp.y;
		result->w = p->w;
	}
	return 0;
}


/* geom_vector_transform */

int geom_vector_transform(const struct geom_vector *p, const struct geom_transform *t, struct geom_vector *result)
{
	struct geom_vector tmp;

	GASSERT(vector, p);
	GASSERT(transform, t);
	PASSERT(result);

	if (t->flags & GEOM_TRANSFORM_IDENTITY)
		*result = *p;
	else {
		tmp.dx = t->coef[0][0] * p->dx + t->coef[0][1] * p->dy;
		tmp.dy = t->coef[1][0] * p->dx + t->coef[1][1] * p->dy;
		if (!isfinite(tmp.dx) || !isfinite(tmp.dy))
			return GEOM_ERR_NUMERIC;
		*result = tmp;
	}
	return 0;
}


/* geom_rectangle_transform */

int geom_rectangle_transform(const struct geom_rectangle *r, const struct geom_transform *t, struct geom_rectangle *result)
{
	struct geom_rectangle tmp;
	real a, b;

	BASSERT(geom_rectangle_is_valid(r));
	GASSERT(transform, t);
	PASSERT(result);

	if (geom_rectangle_is_empty(r))
		return (geom_rectangle_set_empty(result));

	if (!(t->flags & GEOM_TRANSFORM_IDENTITY)) {
		tmp.ll.x = tmp.ur.x = t->coef[0][2];
		tmp.ll.y = tmp.ur.y = t->coef[1][2];
		a = t->coef[0][0] * r->ll.x;
		b = t->coef[0][0] * r->ur.x;
		tmp.ll.x += MIN(a, b);
		tmp.ur.x += MAX(a, b);
		a = t->coef[0][1] * r->ll.y;
		b = t->coef[0][1] * r->ur.y;
		tmp.ll.x += MIN(a, b);
		tmp.ur.x += MAX(a, b);
		a = t->coef[1][0] * r->ll.x;
		b = t->coef[1][0] * r->ur.x;
		tmp.ll.y += MIN(a, b);
		tmp.ur.y += MAX(a, b);
		a = t->coef[1][1] * r->ll.y;
		b = t->coef[1][1] * r->ur.y;
		tmp.ll.y += MIN(a, b);
		tmp.ur.y += MAX(a, b);
		if (!isfinite(tmp.ll.x) || !isfinite(tmp.ll.y) || !isfinite(tmp.ur.x) || !isfinite(tmp.ur.y))
			return GEOM_ERR_NUMERIC;
		*result = tmp;
	}
	else if (result != r)
		*result = *r;
	return 0;
}


int geom_transform_is_point_fixed(struct geom_point *p, struct geom_transform *t)
{
	struct geom_point tp;

	FUNC;
	GASSERT(point, p);
	GASSERT(transform, t);

	ETEST_RET(geom_point_transform(p, t, &tp));

	if (IS_EQUAL_XY_L(*p, tp))
		ERET(1);
	else
		ERET(0);
}


