/** @file
 *
 * 	GEOMLIB - Vector handling routines.
 *
 *	Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 *	This program is distributed under GNU General Public License, see COPYING.
 */

#ifndef _VRR_GEOMLIB_VECTOR_H_
#define _VRR_GEOMLIB_VECTOR_H_

#include "geomlib/base.h"


/* point constants */
extern struct geom_point geom_point_zero;  /**< zero point constant */
extern struct geom_point geom_point_right; /**< (1, 0) point constant */
extern struct geom_point geom_point_left;  /**< (-1, 0) point constant */
extern struct geom_point geom_point_up;    /**< (0, 1) point constant */
extern struct geom_point geom_point_down;  /**< (0, -1) point constant */

/* vector constants */
extern struct geom_vector geom_vector_zero;  /**< zero point constant */
extern struct geom_vector geom_vector_right; /**< (1, 0) point constant */
extern struct geom_vector geom_vector_left;  /**< (-1, 0) point constant */
extern struct geom_vector geom_vector_up;    /**< (0, 1) point constant */
extern struct geom_vector geom_vector_down;  /**< (0, -1) point constant */


const struct geom_point *geom_point_assert(GEOM_ASSERT_DEF, const struct geom_point *p);

static inline int geom_point_valid(const struct geom_point *p)
{ return GEOM_PTEST(p) && GEOM_RTEST(p->x) && GEOM_RTEST(p->y); }

void geom_point_dump(const struct geom_point *p);



const struct geom_point_w *geom_point_w_assert(GEOM_ASSERT_DEF, const struct geom_point_w *p);

static inline int geom_point_w_valid(const struct geom_point_w *p)
{ return GEOM_PTEST(p) && GEOM_RTEST(p->x) && GEOM_RTEST(p->y) && GEOM_RTEST(p->w); }

void geom_point_w_dump(const struct geom_point_w *p);



const struct geom_vector *geom_vector_assert(GEOM_ASSERT_DEF, const struct geom_vector *v);

static inline int geom_vector_valid(const struct geom_vector *v)
{ return GEOM_PTEST(v) && GEOM_RTEST(v->dx) && GEOM_RTEST(v->dy); }

void geom_vector_dump(const struct geom_vector *v);





static inline double geom_vector_length(const struct geom_vector *v)
{
	GEOM_ASSERT(vector, v);
	return hypot(v->dx, v->dy);
}


/**
	\brief Function returns new vector from start and end points.

	\param p1  start point
	\param p2  end point
	\param v  new vector

	\retval -1 error occured
	\retval 0 otherwise
*/
int geom_vector_init_two_points(const struct geom_point *p1, const struct geom_point *p2, struct geom_vector *v);


/**
	\brief Function rotates vector 90 degrees CCW.

	\param in input vector
	\param out new vector

	\retval -1 error occured
	\retval 0 otherwise
*/
static inline int geom_vector_left_normal(const struct geom_vector *in, struct geom_vector *out)
{
	GEOM_ASSERT(vector, in);
	GEOM_PASSERT(out);

	out->dx = -in->dy;
	out->dy = in->dx;

	return ((isfinite(out->dx) && isfinite(out->dy)) ? 0 : -1);
}


/**
	\brief Function rotates vector 90 degrees CW.

	\param in input vector
	\param out new vector

	\retval -1 error occured
	\retval 0 otherwise
*/
static inline int geom_vector_right_normal(const struct geom_vector *in, struct geom_vector *out)
{
	GEOM_ASSERT(vector, in);
	GEOM_BASSERT(out);

	out->dx = in->dy;
	out->dy = -in->dx;

	return ((isfinite(out->dx) && isfinite(out->dy)) ? 0 : -1);
}


/**
	\brief Function returns point in some direction and distance from some point.

	\param in input point
	\param v direction
	\param dist distance to new point
	\param out new point

	\retval -1 error occured
	\retval 0 otherwise
*/
static inline int geom_vector_point_in_direction(const struct geom_point *in, const struct geom_vector *v, real dist, struct geom_point *out)
{
	GEOM_ASSERT(point, in);
	GEOM_ASSERT(vector, v);
	GEOM_ASSERT(r, dist);
	GEOM_PASSERT(out);

	out->x = in->x + v->dx * dist;
	out->y = in->y + v->dy * dist;

	return ((isfinite(out->x) && isfinite(out->y)) ? 0 : -1);
}

/**
	\brief Function returns intersection point of two lines given by point and direction (vector).

	\param p1 point of first line
	\param v1 direction of first line
	\param p2 point of second line
	\param v2 direction of second line
	\param p2 distance to new point
	\param result new point

	\retval -1 error occured (no intersection ?)
	\retval 0 otherwise
*/
int geom_vector_intersection(const struct geom_point *p1, const struct geom_vector *v1, const struct geom_point *p2, const struct geom_vector *v2, struct geom_point *result);

#endif
