/** @file
 *
 * GUI - The common VRR dialogs
 *
 * Copyright (c) 2004-2007    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG
#include "lib/lib.h"

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <gtk/gtk.h>
#ifdef USE_LIBPAPER
#include <paper.h>
#endif

#include "kernel/kernel.h"
#include "gui/main.h"
#include "export/export.h"
#include "import/import.h"
#include "gui/dialogs.h"
#include "gui/properties.h"
#include "build/path.h"
#include "build/config.h"
#include "scheme/glt_kernel.h"

#define POINT2MM	(25.4/72)


/*** EXPORT DIALOGS ****/
#define FALLBACK_A4_WIDTH	210
#define FALLBACK_A4_HEIGHT	297

char last_open_folder[PATH_MAX];
char last_save_folder[PATH_MAX];

#ifdef USE_LIBPAPER
static int last_paper;
#endif

static GtkWidget * create_paper_combobox( GtkSpinButton * spinw, GtkSpinButton * spinh );
static void fill_paper_sizes( GtkComboBox * combobox, GtkSpinButton * spinw );

static GtkWidget * create_paper_combobox( GtkSpinButton * spinw, GtkSpinButton * spinh )
{
	GtkComboBox * cb = GTK_COMBO_BOX( gtk_combo_box_new_text() );
#ifdef USE_LIBPAPER
	const struct paper *p;
	for ( p = paperfirst(); p; p = papernext(p) ) gtk_combo_box_append_text( cb, papername(p) );
	gtk_combo_box_set_active( cb, last_paper );
#else
	gtk_combo_box_append_text( cb, "a4" );
	gtk_combo_box_set_active( cb, 0 );
#endif
	g_object_set_data( G_OBJECT(spinw), "other", spinh );
	g_signal_connect( G_OBJECT(cb), "changed", G_CALLBACK(fill_paper_sizes), spinw );
	fill_paper_sizes( cb, spinw );

	return GTK_WIDGET(cb);
}
#ifdef USE_LIBPAPER
static void fill_paper_sizes( GtkComboBox * combobox, GtkSpinButton * spinw )
{
	GtkSpinButton * spinh = GTK_SPIN_BUTTON( g_object_get_data( G_OBJECT(spinw), "other" ));
	const struct paper * p = paperinfo(gtk_combo_box_get_active_text(combobox));
	gtk_spin_button_set_value( spinw, paperpswidth(p) * POINT2MM );
	gtk_spin_button_set_value( spinh, paperpsheight(p) * POINT2MM );
}
#else
static void fill_paper_sizes( GtkComboBox * combobox UNUSED, GtkSpinButton * spinw )
{
	GtkSpinButton * spinh = GTK_SPIN_BUTTON( g_object_get_data( G_OBJECT(spinw), "other" ));
	gtk_spin_button_set_value( spinw, FALLBACK_A4_WIDTH );
	gtk_spin_button_set_value( spinh, FALLBACK_A4_HEIGHT );
}
#endif

void save_folder( const char * filename, char * dest )
{
	int i = strlen(filename) - 1;
	for ( ; i >= 0 && filename[i] != G_DIR_SEPARATOR; i-- ); //find the last dir separator
	for ( dest[i+1] = '\0'; i >= 0; i-- ) dest[i] = filename[i];
	DBG( "extracted folder path:%s\n", dest );
}

void extract_name_from_filename( const char * filename, char * name )
{
	ASSERT( filename && name );
	char * n = strrchr( filename, G_DIR_SEPARATOR );
	strcpy( name, n ? (n + 1) : filename );
	
	// omit the suffix if present
	char * s = strrchr( name, '.' );
	if (s) *s = '\0';
}

int prompt_overwrite_file( const char * filename )
{
	int result = 0;
	GtkWidget * q = gtk_message_dialog_new( NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE,
		"The file \"%s\" already exists. Do you wish to overwrite it?", filename );
	gtk_dialog_add_buttons( GTK_DIALOG(q), "Yes", GTK_RESPONSE_YES, "No" , GTK_RESPONSE_NO, NULL );
	result = ( gtk_dialog_run(GTK_DIALOG(q)) == GTK_RESPONSE_YES );
	gtk_widget_destroy(GTK_WIDGET(q));
	return result;
}

void on_svg_file_import( void * subj UNUSED )
{
	OPEN_DLG_START( "Import SVG file ...", NULL, NULL )
	{
		volatile int res = 1;
		char err_msg[TRANS_ERR_SIZE];
		char err_msg2[TRANS_ERR_SIZE];
		char name[PATH_MAX];
		extract_name_from_filename( filenames[i], name );
		
		TRANS_BEGIN_MAIN( tlo_universum, "Import file", err_msg2 )
			struct obj_doc * doc = doc_new_named(name);
			struct obj_tlo * tlo = tlo_new_named(name);
			obj_link_tail( &tlo->obj, &doc->obj );
			obj_link_tail( &doc->obj, obj_universum );
			TRANS_BEGIN_MAIN (tlo, "Import SVG image", err_msg)
				import_svg (filenames[i]);
			TRANS_FAILED
				trans_fail(err_msg);
			TRANS_END
			create_view(tlo_get_tlg(tlo));
		TRANS_FAILED
			SHOW_ERROR_MESSAGE( GTK_WINDOW(main_window), "Error importing SVG file %s: %s", filenames[i], err_msg2 );
			res = 0;
		TRANS_END

		if (res)
			statusbar_msg( NULL, "File %s imported.", filenames[i] );
	}
	OPEN_DLG_END;
}

void on_import_ipe5( void * subj UNUSED )
{
	OPEN_DLG_START( "Import Ipe 5 file ...", NULL, NULL )
	{
		volatile int res = 1;
		char err_msg[TRANS_ERR_SIZE];
		char err_msg2[TRANS_ERR_SIZE];
		char name[PATH_MAX];
		extract_name_from_filename( filenames[i], name );

		TRANS_BEGIN_MAIN( tlo_universum, "Import file", err_msg2 )
			struct obj_doc * doc = doc_new_named(name);
			struct obj_tlo * tlo = tlo_new_named(name);
			obj_link_tail( &tlo->obj, &doc->obj );
			obj_link_tail( &doc->obj, obj_universum );
			TRANS_BEGIN_MAIN (tlo, "Import Ipe 5 file", err_msg)
				import_tlo_ipe5 (filenames[i]);
			TRANS_FAILED
				trans_fail(err_msg);
			TRANS_END
			create_view(tlo_get_tlg(tlo));
		TRANS_FAILED
			SHOW_ERROR_MESSAGE( GTK_WINDOW(main_window), "Error importing Ipe 5 file %s: %s", filenames[i], err_msg2 );
			res = 0;
		TRANS_END

		if (res)
			statusbar_msg( NULL, "File %s imported.", filenames[i] );
	}
	OPEN_DLG_END;
}

void on_export_pdf( void * volatile subj )
{
	char buf[TRANS_ERR_SIZE];
	FILE * file;

	GtkWidget * table = gtk_table_new( 2, 2, FALSE );

	GtkWidget * include_fonts = gtk_check_button_new_with_mnemonic( "Include font data" );
	gtk_table_attach( GTK_TABLE(table), include_fonts, 0, 2, 0, 1, GTK_EXPAND | GTK_FILL, 0, 0, 0);
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(include_fonts),
		PROP_STORE_GET_UNS(ps_global, "pdf-export-include-fonts", TRUE ));

	gtk_table_attach (GTK_TABLE (table), gtk_label_new("Compression level:"),
		0, 1, 1, 2, GTK_EXPAND | GTK_FILL, 0, 0, 0);
	GtkComboBox * c_box = GTK_COMBO_BOX(gtk_combo_box_new_text());
	gtk_combo_box_append_text( c_box, "No compression" );
	gtk_combo_box_append_text( c_box, "Fast compression" );
	gtk_combo_box_append_text( c_box, "Best compression" );
	gtk_combo_box_set_active( c_box, PROP_STORE_GET_UNS(ps_global, "pdf-export-compression-level", 0) );
	gtk_table_attach (GTK_TABLE (table), GTK_WIDGET(c_box), 1, 2, 1, 2, GTK_EXPAND | GTK_FILL, 0, 0, 0);

	gtk_widget_show_all( table );

	if ( !subj ) subj = obj_get_first_selected();
	ASSERT_DOC( subj );
	SUGGEST_FILENAME( subj, "pdf", nn );

	SAVE_DLG_START( "Export PDF file ...", table, nn )
	{
		if ( !(file = fopen( filename, "w" )) )
			SHOW_ERROR_MESSAGE( GTK_WINDOW(main_window), "Open file failed: %s", strerror(errno) )
		else
		{
			uns incl_fnt = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(include_fonts));
			uns cl = gtk_combo_box_get_active(c_box);
			TRANS_BEGIN_ANONYMOUS(buf)
				export_pdf( (struct obj_doc*)subj, file, !incl_fnt, cl);
				statusbar_msg(NULL, "File %s written successfully.", filename);
			TRANS_FAILED
				statusbar_err(NULL, "PDF export failed: %s", buf );
				SHOW_ERROR_MESSAGE( GTK_WINDOW(main_window), "PDF export failed: %s", buf );
			TRANS_END

			fclose( file );

			PROP_STORE_SET_UNS(ps_global, "pdf-export-include-fonts", incl_fnt );
			PROP_STORE_SET_UNS(ps_global, "pdf-export-compression-level", cl );
		}
	}
	SAVE_DLG_END;
}

void on_export_ps( void * volatile subj )
{
	FILE * file;
	GtkWidget *table;
	GtkWidget *spin_height, *spin_width, *combo_format;
	GtkWidget *label_height, *label_width, *label_format;
	GtkWidget *check_fit;
	GtkWidget *include_fonts;

	table = gtk_table_new (5, 2, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER (table), 10);

	label_format = gtk_label_new ("Set sizes to standard page format:");
	gtk_table_attach (GTK_TABLE (table), label_format, 0, 1, 0, 1, GTK_FILL, 0, 0, 0);
	gtk_misc_set_alignment (GTK_MISC (label_format), 0, 0.5);

	label_width = gtk_label_new ("Page width (in mm):");
	gtk_table_attach (GTK_TABLE (table), label_width, 0, 1, 1, 2, GTK_FILL, 0, 0, 0);
	gtk_misc_set_alignment (GTK_MISC (label_width), 0, 0.5);

	label_height = gtk_label_new ("Page height (in mm):");
	gtk_table_attach (GTK_TABLE (table), label_height, 0, 1, 2, 3, GTK_FILL, 0, 0, 0);
	gtk_misc_set_alignment (GTK_MISC (label_height), 0, 0.5);

	spin_width = gtk_spin_button_new (GTK_ADJUSTMENT (gtk_adjustment_new (210, 0, 1000, 1, 10, 10)), 1, 0);
	gtk_table_attach (GTK_TABLE (table), spin_width, 1, 2, 1, 2, GTK_FILL, 0, 0, 0);

	spin_height = gtk_spin_button_new (GTK_ADJUSTMENT (gtk_adjustment_new (297, 0, 1000, 1, 10, 10)), 1, 0);
	gtk_table_attach (GTK_TABLE (table), spin_height, 1, 2, 2, 3, GTK_FILL, 0, 0, 0);

	combo_format = create_paper_combobox( GTK_SPIN_BUTTON(spin_height), GTK_SPIN_BUTTON(spin_width) );
	gtk_table_attach (GTK_TABLE (table), combo_format, 1, 2, 0, 1, GTK_FILL, 0, 0, 0);

	check_fit = gtk_check_button_new_with_mnemonic ("Fit to page");
	gtk_table_attach (GTK_TABLE (table), check_fit, 0, 2, 3, 4, GTK_EXPAND | GTK_FILL, 0, 0, 0);

	include_fonts = gtk_check_button_new_with_mnemonic ("Include font data");
	gtk_table_attach (GTK_TABLE (table), include_fonts, 0, 2, 4, 5, GTK_EXPAND | GTK_FILL, 0, 0, 0);
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(include_fonts), TRUE );

	gtk_widget_show_all(table);

	if ( !subj ) subj = obj_get_first_selected();
	ASSERT_DOC( subj );
	SUGGEST_FILENAME( subj, "ps", nn );

	SAVE_DLG_START( "Export PS file ...", table, nn )
	{
		if ( !(file = fopen( filename, "w" )) )
			SHOW_ERROR_MESSAGE( GTK_WINDOW(main_window), "Open file failed: %s", strerror(errno) )
		else {
			int width, height, fit, no_fonts;

			width = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (spin_width));
			height = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (spin_height));
			fit = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (check_fit));
			no_fonts = !gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (include_fonts));

			char buf[TRANS_ERR_SIZE];
			TRANS_BEGIN_ANONYMOUS(buf)
				export_ps( subj, file, width, height, fit, no_fonts );
				statusbar_msg(NULL, "File %s written successfully.", filename);
			TRANS_FAILED
				statusbar_err(NULL, "PS export failed: %s", buf );
				SHOW_ERROR_MESSAGE( GTK_WINDOW(main_window), "PS export failed: %s", buf );
			TRANS_END

			fclose( file );
		}
	}
	SAVE_DLG_END;
}

void on_export_eps( void * volatile subj )
{
	FILE * file;
	GtkWidget *include_fonts;

	include_fonts = gtk_check_button_new_with_mnemonic ("Include font data");
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(include_fonts), TRUE );
	gtk_widget_show (include_fonts);

	if ( !subj ) subj = obj_get_first_selected();
	ASSERT_TLO( subj );
	SUGGEST_FILENAME( subj, "eps", nn );

	SAVE_DLG_START( "Export EPS file ...", include_fonts, nn )
	{
		if ( !(file = fopen( filename, "w" )) )
			SHOW_ERROR_MESSAGE( GTK_WINDOW(main_window), "Open file failed: %s", strerror(errno) )
		else
		{
			int no_fonts UNUSED = !gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (include_fonts));
			char buf[TRANS_ERR_SIZE];
			TRANS_BEGIN_ANONYMOUS(buf)
				export_eps( subj, file, no_fonts );
				statusbar_msg(NULL, "File %s written successfully.", filename);
			TRANS_FAILED
				statusbar_err(NULL, "EPS export failed: %s", buf );
				SHOW_ERROR_MESSAGE( GTK_WINDOW(main_window), "EPS export failed: %s", buf );
			TRANS_END

			fclose( file );
		}
	}
	SAVE_DLG_END;
}

void on_export_svg( void * volatile subj )
{
	if ( !subj ) subj = obj_get_first_selected();
	ASSERT_TLO( subj );
	SUGGEST_FILENAME(subj, "svg", nn );

	SAVE_DLG_START("Export SVG file ...", NULL, nn)
	{
		char buf[TRANS_ERR_SIZE];
		TRANS_BEGIN_ANONYMOUS(buf)
			export_svg( subj, filename );
			statusbar_msg( NULL, "File %s written succesfully.", filename );
		TRANS_FAILED
			statusbar_err( NULL, "SVG export failed: %s", buf );
			SHOW_ERROR_MESSAGE( GTK_WINDOW(main_window), "SVG export failed: %s", buf );
		TRANS_END
	}
	SAVE_DLG_END;
}

/*** OPTIONS WINDOW ****/

#define GLOBAL_SETTINGS_CNT 20
struct wnd_settings wnd_settings;
static void destroy_settings_window( GtkWidget * widget, gpointer data );
static void settings_add_prop( string key, struct prop_item * pi, struct o * o, GtkTable * table,
	int * counter, int max, int * row, char * desc );
static void sync_settings_window(struct o *o, struct obj_hook * h );
static void create_settings_window(void);

void on_show_settings_window( void * subj UNUSED )
{
	if ( wnd_settings.window )
	{
		gtk_widget_destroy( wnd_settings.window );
		wnd_settings.window = 0;
	}
	else
		create_settings_window();
}

static void destroy_settings_window( GtkWidget * widget UNUSED, gpointer data UNUSED )
{
	wnd_settings.window = NULL;
	xfree(wnd_settings.pi);
	wnd_settings.pi = NULL;
	obj_hook_remove(&wnd_settings.oh);
	refresh_gui();
}

static void settings_add_prop( string key, struct prop_item * pi, struct o * o, GtkTable * table,
	int * counter, int max, int * row, char * desc )
{
	ASSERT(*counter < max);
	if (!prop_find(o, key)) return; // FIXME: nemelo by byt potreba, vsechny stylove properties by mely byt nastaveny
	prop_item_init( pi, o, key );

	GtkWidget * label = gtk_label_new(desc);
	g_object_set( G_OBJECT(label), "xalign", 0.0, NULL );
	gtk_table_attach( table, label, 0, 1, *row, *row+1, GTK_FILL, GTK_FILL, 5, 0 );
	if (pi->unit_edit)
	{
		gtk_table_attach( table, pi->value_edit, 1, 2, *row, *row+1, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0 );
		gtk_table_attach( table, pi->unit_edit, 2, 3, *row, *row+1, GTK_FILL, GTK_FILL, 0, 0 );
	} else
		gtk_table_attach( table, pi->value_edit, 1, 3, *row, *row+1, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0 );
	(*counter)++;
	(*row)++;
}

static void sync_settings_window(struct o *o, struct obj_hook * h UNUSED)
{
	if ( o != obj2o(obj_universum) ) return;
	for ( uns counter = 0; counter < GLOBAL_SETTINGS_CNT; counter++ )
		prop_sync( wnd_settings.pi + counter );
}

#define SW_ADD_PROP(_key, _desc) settings_add_prop(_key, wnd_settings.pi + counter, obj2o(obj_universum),\
	GTK_TABLE(table), &counter, GLOBAL_SETTINGS_CNT, &row, _desc)

static void create_settings_window(void)
{
	wnd_settings.pi = xmalloc_zero(GLOBAL_SETTINGS_CNT * sizeof(struct prop_item));
	int counter = 0;
	int row = 0;

	wnd_settings.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title( GTK_WINDOW(wnd_settings.window), "Global settings" );
	g_signal_connect( G_OBJECT(wnd_settings.window), "delete_event", G_CALLBACK(gtk_widget_destroy), NULL );
	g_signal_connect( G_OBJECT(wnd_settings.window), "destroy", G_CALLBACK(destroy_settings_window), NULL);
	GtkWidget * vbox = gtk_vbox_new( FALSE, 0 );
	gtk_container_add( GTK_CONTAINER(wnd_settings.window), vbox );
	
	GtkWidget * frame = gtk_frame_new("Editor settings");
	gtk_container_set_border_width( GTK_CONTAINER(frame), 5 );
	gtk_box_pack_start( GTK_BOX(vbox), frame, FALSE, TRUE, 0 );
	GtkWidget * table = gtk_table_new( 1, 3, FALSE );
	gtk_container_set_border_width( GTK_CONTAINER(table), 5 );
	gtk_container_add( GTK_CONTAINER(frame), table );

	SW_ADD_PROP( string_tolerance, "Snap tolerance (in pixels)" );
	SW_ADD_PROP( string_grid_x1, "Grid x1" );
	SW_ADD_PROP( string_grid_y1, "Grid y1" );
	SW_ADD_PROP( string_grid_x2, "Grid x2" );
	SW_ADD_PROP( string_grid_y2, "Grid y2" );
	SW_ADD_PROP( string_grid_color, "Grid color" );
	SW_ADD_PROP( string_rulerx_res, "Ruler x-resolution" );
	SW_ADD_PROP( string_rulery_res, "Ruler y-resolution" );
	SW_ADD_PROP( string_drag_image, "Panning: drag the image" );
	SW_ADD_PROP( string_enable_fifi, "Enable Fifi" );

	frame = gtk_frame_new("External programs");
	gtk_container_set_border_width( GTK_CONTAINER(frame), 5 );
	gtk_box_pack_start( GTK_BOX(vbox), frame, FALSE, TRUE, 0 );
	table = gtk_table_new( 1, 3, FALSE );
	gtk_container_set_border_width( GTK_CONTAINER(table), 5 );
	gtk_container_add( GTK_CONTAINER(frame), table );
	row = 0;
	
	SW_ADD_PROP( string_editor, "Your favourite text editor" );
	SW_ADD_PROP( string_editor_in_term, "Run the text editor in a terminal?" );
	SW_ADD_PROP( string_browser, "Your favourite WWW browser" );
	SW_ADD_PROP( string_browser_in_term, "Run the browser in a terminal?" );
	SW_ADD_PROP( string_terminal, "Your favourite terminal" );
	
	frame = gtk_frame_new("Advanced");
	gtk_container_set_border_width( GTK_CONTAINER(frame), 5 );
	gtk_box_pack_start( GTK_BOX(vbox), frame, FALSE, TRUE, 0 );
	table = gtk_table_new( 1, 3, FALSE );
	gtk_container_set_border_width( GTK_CONTAINER(table), 5 );
	gtk_container_add( GTK_CONTAINER(frame), table );
	row = 0;
	
	SW_ADD_PROP( string_max_history_volume, "Maximum volume of undo history (in KB)" );
	SW_ADD_PROP( string_max_history_count, "Maximum items of undo history" );
	SW_ADD_PROP( string_max_cache_memory, "Maximum memory for cache" );
	SW_ADD_PROP( string_max_cache_count, "Maximum items in cache" );
	SW_ADD_PROP( string_min_cache_count, "Minimum items in cache" );

	wnd_settings.oh.data = &wnd_settings;
	wnd_settings.oh.obj_prop_change = sync_settings_window;
	obj_hook_add( &wnd_settings.oh );

	gtk_widget_show_all( GTK_WIDGET(wnd_settings.window) );
}

/***** STYLE WINDOW *****/
#define DISP_PROP_CNT 7
struct wnd_style wnd_style;
static void destroy_style_window( GtkWidget * widget, gpointer data );
static void sync_style_window(struct o *o, struct go_hook * h );
static void create_style_window(void);

void on_show_style_window( void * subj UNUSED )
{
	if ( wnd_style.window )
	{
		gtk_widget_destroy( wnd_style.window );
		wnd_style.window = 0;
	}
	else create_style_window();
}

static void destroy_style_window( GtkWidget * widget UNUSED, gpointer data UNUSED )
{
	wnd_style.window = NULL;
	xfree(wnd_style.pi);
	wnd_style.pi = NULL;
	go_hook_remove( PROP_STORE_TLO(ps_recycler), &wnd_style.gh );
	refresh_gui();
}

static void sync_style_window(struct o *o, struct go_hook * h UNUSED)
{
	if ( o != PROP_STORE_O(ps_recycler) ) return;
	for ( uns counter = 0; counter < DISP_PROP_CNT; counter++ )
		if ( wnd_style.pi[counter].o )	 // FIXME: nemelo by byt potreba, vsechny stylove properties by mely byt nastaveny od startu
			prop_sync( wnd_style.pi + counter );
}

#define ST_ADD_PROP(_key) settings_add_prop(_key, wnd_style.pi + counter, PROP_STORE_O(ps_recycler), GTK_TABLE(table),\
	&counter, DISP_PROP_CNT, &row, _key->text)

static void create_style_window(void)
{
	wnd_style.pi = xmalloc_zero(DISP_PROP_CNT * sizeof(struct prop_item));
	int counter = 0;
	int row = 0;

	wnd_style.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title( GTK_WINDOW(wnd_style.window), "Style palette" );
	g_signal_connect( G_OBJECT(wnd_style.window), "delete_event", G_CALLBACK(gtk_widget_destroy), NULL );
	g_signal_connect( G_OBJECT(wnd_style.window), "destroy", G_CALLBACK(destroy_style_window), NULL);
	GtkWidget * vbox = gtk_vbox_new( FALSE, 0 );
	gtk_container_add( GTK_CONTAINER(wnd_style.window), vbox );
	
	GtkWidget * frame = gtk_frame_new("Current style properties");
	gtk_container_set_border_width( GTK_CONTAINER(frame), 5 );
	gtk_box_pack_start( GTK_BOX(vbox), frame, FALSE, TRUE, 0 );
	GtkWidget * table = gtk_table_new( 1, 3, FALSE );
	gtk_container_set_border_width( GTK_CONTAINER(table), 5 );
	gtk_container_add( GTK_CONTAINER(frame), table );

	ST_ADD_PROP( string_stroke_style );
	ST_ADD_PROP( string_stroke_cap );
	ST_ADD_PROP( string_stroke_join );
	ST_ADD_PROP( string_stroke_width );
	ST_ADD_PROP( string_stroke_color );
	ST_ADD_PROP( string_fill_color );
	ST_ADD_PROP( string_invisible );

	wnd_style.gh.data = &wnd_style;
	wnd_style.gh.go_prop_change = sync_style_window;
	go_hook_add( PROP_STORE_TLO(ps_recycler), &wnd_style.gh );

	gtk_widget_show_all( GTK_WIDGET(wnd_style.window) );
}

/***** ABOUT WINDOW *****/

void on_show_about_window( void * subj UNUSED )
{
	GtkWidget * dialog = gtk_dialog_new_with_buttons( "About VRR", NULL,
		GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
		GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, NULL );
	gtk_dialog_set_default_response( GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT );

	GtkBox * main_vbox = GTK_BOX(gtk_vbox_new( FALSE, 5 ));
	gtk_container_set_border_width( GTK_CONTAINER(main_vbox), 10 );
	gtk_container_add( GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), GTK_WIDGET(main_vbox) );

	GtkWidget * logo = gtk_image_new_from_file( LIB_PATH"logo.png" );
	gtk_box_pack_start( main_vbox, logo, TRUE, TRUE, 0 );

	GtkWidget * label1 = gtk_label_new( "VRR "PACKAGE_VERSION" - a Vector gRaphic editoR" );
	PangoAttrList * pal = pango_attr_list_new();
	pango_attr_list_change( pal, pango_attr_weight_new(PANGO_WEIGHT_BOLD) );
	gtk_label_set_attributes( GTK_LABEL(label1), pal );

	gtk_box_pack_start( main_vbox, label1, FALSE, FALSE, 10 );

	GtkWidget * label2 = gtk_label_new( "Copyright (c) 2003-2007 The VRR Team <vrr@ucw.cz>\n"
		"This program is distributed under GNU General Public License.\n\n"
		"This program has been developed by students of the Faculty of Mathematics and Physics, "
		"Charles University, Prague, Czech Republic.\n\n"
		"Authors:\nMartin Mare\305\241 (advisor), Ji\305\231\303\255 Fink, Pavel Charv\303\241t, Eva Ondr\303\241\304\215kov\303\241, "
		"Tom\303\241\305\241 Valla, Zuzana Vl\304\215kov\303\241, Ond\305\231ej `Santiago' Zaj\303\255\304\215ek" );
	gtk_misc_set_alignment( GTK_MISC(label2), 0.5, 0.5 );
	gtk_label_set_line_wrap( GTK_LABEL(label2), TRUE );
	gtk_box_pack_start( main_vbox, label2, TRUE, TRUE, 0 );

	gtk_widget_show_all(dialog);
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
}

/**** CREATION OF SCHEME PROXY ****/

void on_create_proxy( void * subj )
{
	ASSERT_GROUP(context.group);
	struct go * go = subj ? (struct go*)subj : group_get_first_selected_go( context.group );
	ASSERT_O(go);

	GtkWidget * dialog = gtk_dialog_new_with_buttons( "Create Scheme proxy", NULL,
		GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
		GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL );
	gtk_dialog_set_default_response( GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT );

	GtkBox * main_vbox = GTK_BOX(gtk_vbox_new( FALSE, 5 ));
	gtk_container_set_border_width( GTK_CONTAINER(main_vbox), 10 );
	gtk_container_add( GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), GTK_WIDGET(main_vbox) );
	string n = go_get_name(go);
	string t = string_go_type[go->type];
	char buf[40 + n->length + t->length];
	sprintf(buf, "Enter a string identifier for %s \"%s\":", t->text, n->text);
	gtk_box_pack_start( main_vbox, gtk_label_new(buf), FALSE, FALSE, 0 );
	GtkWidget * entry = gtk_entry_new();
	gtk_box_pack_start( main_vbox, entry, FALSE, FALSE, 0 );

	gtk_widget_show_all(dialog);
	if ( gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT )
	{
		const char * ident = gtk_entry_get_text(GTK_ENTRY(entry));
		scm_c_define( ident, o2scm(go2o(go)) );
	}
	gtk_widget_destroy(dialog);
}
