/** @file
 * GUI - Implementation of the main GUI callbacks of menu commands,
 * of the main functions etc.
 *
 * Copyright (c) 2004-2006    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG
#include "lib/lib.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <locale.h>
#ifdef USE_LIBPAPER
#include <paper.h>
#endif

#include <gtk/gtk.h>

#include "kernel/kernel.h"

#include "gui/main.h"
#include "gui/cmdmgr.h"
#include "gui/visualisation.h"
#include "scheme/glt_gui.h"
#include "gui/properties.h"
#include "gui/dialogs.h"
#include "vcl/vcl.h"

#include "build/path.h"

struct context context;

GtkWidget * main_window;
struct console *console_window;

GList *global_units_glist;
GList *global_multiplier_glist;
GSList * wuh_list;

static int closing_all;

GdkPixbuf * icon_arrow_down, * icon_arrow_up, * icon_arrow_left, * icon_arrow_right;
GdkPixbuf * icon_pencil_on, * icon_modified;
GdkPixbuf * icon_rotation_left, * icon_rotation_right, * icon_zoom_in, * icon_zoom_out;

static int noname_counter;
static char noname_buf[20];

//Hook calbacks for handling selection changes
static void mh_init(void);
static void mh_finish(void);
static void mh_go_selected( struct o * o, struct go_hook * hook );
static void mh_go_unselected( struct o * o, struct go_hook * hook );
static void mh_obj_sel_changed( struct o * o, struct obj_hook * hook );
static void mh_obj_added( struct o * o, struct obj_hook * hook );
static void mh_obj_after_delete( struct o * o, struct o * owner UNUSED, struct o * prev UNUSED, struct obj_hook * hook UNUSED );
static void mh_go_after_delete( struct o * o, struct o * owner UNUSED, struct o * prev UNUSED, struct go_hook * gh UNUSED );
static void mh_add_go_hook( struct obj * obj );

static void resources_init(void);
static void resources_finish(void);

static void gui_save_document( struct obj * doc, const char * filename );

/******* CALLBACKS *********/

void on_delete( void * subj )
{
	char buf[TRANS_ERR_SIZE];
	if ( subj && ((struct o*)subj)->kind == T_GO || context.group && group_get_first_selected_go(context.group))
	{
		ASSERT_TLO(context.tlo);
		ASSERT_GROUP(context.group);
		struct go * go = (struct go*)subj;
		if (go) ASSERT_O(go);

		TRANS_BEGIN_MAIN(context.tlo, "Delete", buf)
			if (go) go_select(go);
			group_remove_rehang_selected_go(context.group);
		TRANS_FAILED
			statusbar_err(NULL, "Cannot delete: %s", buf);
		TRANS_END
	} else {
		TRANS_BEGIN_MAIN(tlo_universum, "Delete", buf)
			if (subj)
			{ ASSERT_O(subj); obj_select(subj); }
			obj_unlink_selected();
		TRANS_FAILED
			statusbar_err(NULL, "Cannot delete: %s", buf);
		TRANS_END
	}
}

void on_go_enclose_sel( void * subj UNUSED )
{
	ASSERT_TLO(context.tlo);
	ASSERT_GROUP(context.group);
	char s[TRANS_ERR_SIZE];

	TRANS_BEGIN_MAIN( context.tlo, "Add dependent to selection", s)
		//tlo_add_selection_down();  FIXME OLD
	TRANS_FAILED
		statusbar_err(NULL, "Cannot add dependent to selection: %s.", s);
	TRANS_END
}

void on_obj_undo( void * subj UNUSED )
{
	char s[TRANS_ERR_SIZE];
	if (!trans_undo(tlo_universum, s)) statusbar_err(NULL, "%s", s);
}

void on_obj_redo( void * subj UNUSED )
{
	char s[TRANS_ERR_SIZE];
	if (!trans_redo(tlo_universum, s)) statusbar_err(NULL, "%s", s);
}

void on_tlo_undo( void * subj )
{
	char s[TRANS_ERR_SIZE];
	if ( !subj ) subj = obj_get_first_selected();
	ASSERT_TLO( (struct obj_tlo*)subj );
	if (!trans_undo( (struct obj_tlo*)subj, s )) statusbar_err(NULL, "%s", s);
}

void on_tlo_redo( void * subj )
{
	char s[TRANS_ERR_SIZE];
	if ( !subj ) subj = obj_get_first_selected();
	ASSERT_TLO( (struct obj_tlo*)subj );
	if (!trans_redo( (struct obj_tlo*)subj, s )) statusbar_err(NULL, "%s", s);
}

void on_copy( void * subj )
{
	char s[TRANS_ERR_SIZE], ss[TRANS_ERR_SIZE];
	ASSERT_TLO(context.tlo);
	if ( context.tlo == tlo_clipboard ) return;

	TRANS_BEGIN_MAIN( context.tlo, "Copy", ss)
		if (subj) go_select(void2go(subj));
		if (!clipboard_copy( context.group, "Copy", s, 0 )) statusbar_err(NULL, "%s", s);
	TRANS_FAILED
		statusbar_err(NULL, "Cannot copy: %s", ss);
	TRANS_END
}

void on_cut( void * subj )
{
	char s[TRANS_ERR_SIZE];
	ASSERT_TLO(context.tlo);
	if ( context.tlo == tlo_clipboard ) return;

	TRANS_BEGIN_MAIN( context.tlo, "Cut", s)
		if (subj) go_select(void2go(subj));
		clipboard_cut( context.group, 0 );
	TRANS_FAILED
		statusbar_err(NULL, "Cannot cut: %s", s);
	TRANS_END
}

void on_paste( void * subj UNUSED )
{
	char s[TRANS_ERR_SIZE];
	ASSERT_TLO(context.tlo);
	if ( context.tlo != tlo_clipboard )
		TRANS_BEGIN_MAIN( context.tlo, "Paste", s)
			clipboard_paste( context.group, NULL, 0 );
		TRANS_FAILED
			statusbar_err(NULL, "Cannot paste: %s", s);
		TRANS_END
}

#define GET_PARAM_GRP(_grp) \
	struct go_group * volatile _grp; \
	if ( subj ) _grp = (struct go_group*)subj; \
	else { \
		ASSERT(context.group); \
		_grp = (struct go_group*) group_get_first_selected_go(context.group); \
	} \
	ASSERT_GROUP(_grp);

void on_group_select_all( void * subj UNUSED )
{
	char s[TRANS_ERR_SIZE];
	struct go_group * group = context.group;
	ASSERT_GROUP(group);

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Select all", s );
		group_select_all_go(group);
	TRANS_FAILED
		statusbar_err(NULL, "Cannot select: %s", s);
	TRANS_END
}

void on_group_select_none( void * subj UNUSED )
{
	char s[TRANS_ERR_SIZE];
	struct go_group * group = context.group;
	ASSERT_GROUP(group);

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Unselect all", s );
		group_unselect_all_go(group);
	TRANS_FAILED
		statusbar_err(NULL, "Cannot unselect: %s", s);
	TRANS_END
}

void on_group_selected( void * subj UNUSED )
{
	ASSERT(context.group);
	char s[TRANS_ERR_SIZE];
	struct go_group * group = context.group, * new_group;

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Group selected", s )
		new_group = go_group_new();
		go_link_prev( &new_group->go, group, group_get_first_selected_go(group) );
		group_relink_selected_go( group, new_group, NULL );
		go_select(&new_group->go);
	TRANS_FAILED
		statusbar_err(NULL, "Cannot group: %s", s);
	TRANS_END
}
void on_group_ungroup( void * subj )
{
	char s[TRANS_ERR_SIZE];
	GET_PARAM_GRP(group);
	struct go_group * up_group;

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Ungroup", s )
		if ( !(up_group = go_get_group(&group->go)) )
			trans_fail("This is the top group!\n");
		group_relink_all_go( group, up_group, &group->go );
		go_unlink(&group->go);
	TRANS_FAILED
		statusbar_err(NULL, "Cannot ungroup: %s", s);
	TRANS_END
}

void on_group_pathize_selected( void * subj UNUSED )
{
	ASSERT(context.group);
	char s[TRANS_ERR_SIZE];
	struct go_group * group = context.group;
	struct go * new_go;

	TRANS_BEGIN_MAIN( go_get_tlo(&group->go), "Pathize selection", s )
		new_go = &group_pathize_selected_go(group)->group.go;
		group_unselect_all_go(group);
		go_select(new_go);
		gui_prop_recycle( go2o(new_go) );
		change_context( context.w, context.group, go2o(new_go) );
	TRANS_FAILED
		statusbar_err(NULL, "Cannot pathize: %s", s);
	TRANS_END
}

#define GET_ARG_GO \
	struct go * volatile go; \
	ASSERT(context.group); \
	if (subj) go = void2go(subj); else go = group_get_first_selected_go(context.group); \
	ASSERT_GO(go)

void on_move_up( void * subj )
{
	GET_ARG_GO;
	struct go * volatile prev = group_get_prev_go(context.group, go);
	ASSERT(go);
	prev = group_get_prev_go(context.group, prev);
	char s[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( go_get_tlo(go), "Move up", s )
		go_relink( go, context.group, prev );
	TRANS_FAILED
		statusbar_err(NULL, "Cannot move up: %s", s);
	TRANS_END
}
void on_move_down( void * subj )
{
	GET_ARG_GO;
	char s[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( go_get_tlo(go), "Move down", s )
		go_relink( go, context.group, group_get_next_go(context.group, go) );
	TRANS_FAILED
		statusbar_err(NULL, "Cannot move down: %s", s);
	TRANS_END
}
void on_move_front( void * subj )
{
	GET_ARG_GO;
	if ( go == group_get_first_go(context.group) ) return;
	char s[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( go_get_tlo(go), "Move to front", s )
		go_relink( go, context.group, NULL );
	TRANS_FAILED
		statusbar_err(NULL, "Cannot move to front: %s", s);
	TRANS_END
}
void on_move_back( void * subj )
{
	GET_ARG_GO;
	if ( go == group_get_last_go(context.group) ) return;
	char s[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( go_get_tlo(go), "Move to back", s )
		go_relink( go, context.group, group_get_last_go(context.group) );
	TRANS_FAILED
		statusbar_err(NULL, "Cannot move to back: %s", s);
	TRANS_END
}

#define GET_ARG_OBJ \
	struct obj * volatile obj; \
	if (subj) obj = void2obj(subj); else obj = obj_get_first_selected(); \
	ASSERT_OBJ(obj)

void on_page_move_up( void * subj )
{
	GET_ARG_OBJ;	
	struct obj * prev = obj_get_prev_son(obj);
	if (!prev)
		return;

	char s[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( tlo_universum, "Move page up", s )
		obj_relink_after( obj, obj->father, obj_get_prev_son(prev) );
	TRANS_FAILED
		statusbar_err(NULL, "Cannot move page up: %s", s);
	TRANS_END
}
void on_page_move_down( void * subj )
{
	GET_ARG_OBJ;
	struct obj * next = obj_get_next_son(obj);
	if (!next)
		return;

	char s[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( tlo_universum, "Move page down", s )
		obj_relink_after( obj, obj->father, next );
	TRANS_FAILED
		statusbar_err(NULL, "Cannot move page down: %s", s);
	TRANS_END
}
void on_page_move_front( void * subj )
{
	GET_ARG_OBJ;
	char s[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( tlo_universum, "Move page to front", s )
		obj_relink_after( obj, obj->father, NULL );
	TRANS_FAILED
		statusbar_err(NULL, "Cannot move page to front: %s", s);
	TRANS_END
}
void on_page_move_back( void * subj )
{
	GET_ARG_OBJ;
	struct obj * last = obj_get_last_son(obj->father);
	if (last == obj)
		return;

	char s[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( tlo_universum, "Move page to back", s )
		obj_relink_after( obj, obj->father, obj_get_last_son(obj->father) );
	TRANS_FAILED
		statusbar_err(NULL, "Cannot move page to back: %s", s);
	TRANS_END

}

void on_obj_copy( void * subj UNUSED )
{
	char s[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( tlo_universum, "Copy", s )
		clipboard_obj_copy();
	TRANS_FAILED
		statusbar_err(NULL, "Cannot copy: %s", s);
	TRANS_END
}
void on_obj_cut( void * subj UNUSED )
{
	char s[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( tlo_universum, "Cut", s )
		clipboard_obj_cut();
	TRANS_FAILED
		statusbar_err(NULL, "Cannot cut: %s", s);
	TRANS_END
}
void on_obj_paste( void * subj )
{
	struct obj * volatile obj;
        if (subj) obj = void2obj(subj); else obj = obj_get_first_selected();
	if (!obj) obj = obj_universum;
	ASSERT_OBJ(obj);

	char s[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( tlo_universum, "Paste", s )
		clipboard_obj_paste( obj, NULL );
	TRANS_FAILED
		statusbar_err(NULL, "Cannot paste: %s", s);
	TRANS_END
}

void on_change_prop( void * subj )
{
	struct o * o = NULL;
	if ( subj ) o = void2o(subj);
	if (!o && context.group)
	{
		ASSERT_GROUP(context.group);
		o = (struct o*)group_get_first_selected_go(context.group);
	}
	if (!o)
		o = (struct o*)obj_get_first_selected();
	if (o)
		prop_window_create(o);
}
void on_edit_text( void * subj )
{
	ASSERT_GROUP(context.group);
	struct go * go = subj ? (struct go*)subj : group_get_first_selected_go( context.group );
	ASSERT_O(go);
	text_edit_create( go2o(go), string_text );
}

void on_snap_hangers( void * subj UNUSED )
{
	factory.snap_set ^= SNAP_HANGERS;
	for ( GSList * l = view_list; l; l = l->next ) transform_update((struct view*)l->data );
}
void on_snap_grid( void * subj UNUSED )
{
	factory.snap_set ^= SNAP_GRID;
	for ( GSList * l = view_list; l; l = l->next )
	{
		view_update_grid( (struct view*)l->data );
		transform_update( (struct view*)l->data );
	}
}
void on_snap_lines( void * subj UNUSED )
{
	factory.snap_set ^= SNAP_LINES;
	for ( GSList * l = view_list; l; l = l->next ) transform_update((struct view*)l->data );
}
void on_snap_inters( void * subj UNUSED )
{
	factory.snap_set ^= SNAP_INTERS;
	for ( GSList * l = view_list; l; l = l->next ) transform_update((struct view*)l->data );
}
void on_snap_depend( void * subj UNUSED ){ factory.snap_set ^= SNAP_DEPENDENCY; }

void on_document_rerun_tex( void * subj )
{
	struct obj * obj;
	if (subj)
	{
		ASSERT_DOC(subj);
		WALK_OBJ_RECURSIVE(subj, obj)
			if (obj->type == OT_TLO)
				on_page_rerun_tex(obj);
	} else {
		struct obj * ooo;
		WALK_OBJ_SELECTED(ooo)
			if (ooo->type == OT_DOCUMENT)
				WALK_OBJ_RECURSIVE(ooo, obj)
					if (obj->type == OT_TLO)
						on_page_rerun_tex(obj);
	}
}
void on_page_rerun_tex( void * subj )
{
	char buf[TRANS_ERR_SIZE];
	if (subj)
	{
		ASSERT_TLO(subj);
		TRANS_BEGIN_MAIN(subj, "Rerun TeX", buf)
			group_rerun_tex_process(tlo_get_tlg(subj));
		TRANS_FAILED
			statusbar_err( NULL, "Cannot rerun TeX: %s\n", buf );
		TRANS_END
	} else {
		struct obj * volatile obj;
		WALK_OBJ_SELECTED(obj)
			if (obj->type == OT_TLO)
			{
			TRANS_BEGIN_MAIN(obj2tlo(obj), "Rerun TeX", buf)
				group_rerun_tex_process(tlo_get_tlg(obj2tlo(obj)));
			TRANS_FAILED
				statusbar_err( NULL, "Cannot rerun TeX: %s\n", buf );
			TRANS_END
			}
	}
}

void on_dump( void * subj UNUSED )
{
	obj_dump_subtree(obj_universum);
	if ( context.group )
		group_dump_subtree(context.group);
	if ( context.tlo )
		tlo_dump_tsort(context.tlo);
}

void on_file_new( void * subj UNUSED )
{
	char buf[TRANS_ERR_SIZE];

	TRANS_BEGIN_MAIN(tlo_universum, "New file", buf)
		struct obj_doc * doc = doc_new_named(get_new_noname());
		struct obj_tlo * tlo = tlo_new_named(get_new_page_noname(doc));
		obj_link_tail( &tlo->obj, &doc->obj );
		obj_link_tail( &doc->obj, obj_universum );
		doc_set_all_saved(doc);
		create_view(tlo_get_tlg(tlo));
	TRANS_FAILED
		statusbar_err(NULL, "Cannot create a new file: %s", buf);
	TRANS_END
}

static int prompt_save_file( struct obj_doc * doc )
{
	int result = 0;
	GtkWidget * q = gtk_message_dialog_new( NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE,
		"The document \"%s\" is unsaved. Do you wish to save it before closing it?", obj_lookup_name(&doc->obj) );
	gtk_dialog_add_buttons( GTK_DIALOG(q), "Save", GTK_RESPONSE_YES, "Cancel", GTK_RESPONSE_CANCEL,
		"Discard changes" , GTK_RESPONSE_NO, NULL );
	switch ( gtk_dialog_run(GTK_DIALOG(q)) )
	{
	case GTK_RESPONSE_NO: break;
	case GTK_RESPONSE_YES: result = 2; break;
	default: result = 1;
	}
	gtk_widget_destroy(GTK_WIDGET(q));
	if (result == 2) { result = 0; on_file_save(doc); }
	return result;
}

static int prompt_discard_changes(void)
{
	int result = 0;
	GtkWidget * q = gtk_message_dialog_new( NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE,
		"There are unsaved documents.\nDo you wish to quit anyway?" );
	gtk_dialog_add_buttons( GTK_DIALOG(q), "Cancel", GTK_RESPONSE_CANCEL, "Discard changes and quit", GTK_RESPONSE_YES,
		NULL );
	if ( gtk_dialog_run(GTK_DIALOG(q)) == GTK_RESPONSE_YES )
		result = 1;
	gtk_widget_destroy(GTK_WIDGET(q));
	return result;
}

void on_file_close( void * subj )
{
	char buf[TRANS_ERR_SIZE];
	if ( subj )
	{
		ASSERT_DOC( subj );
		if ( !doc_is_all_saved(subj) && prompt_save_file(subj) ) return;
		TRANS_BEGIN_MAIN( tlo_universum, "Close file", buf )
			obj_unlink( (struct obj*)subj );
		TRANS_FAILED
			statusbar_err(NULL, "Cannot close the file: %s", buf);
		TRANS_END
	} else {
		struct obj * obj;
		WALK_OBJ_SELECTED(obj)
			if ( obj->type == OT_DOCUMENT &&
				!doc_is_all_saved(obj2doc(obj)) && prompt_save_file(obj2doc(obj)) ) return;

		TRANS_BEGIN_MAIN(tlo_universum, "Close files", buf)
		  	obj_unlink_selected();
		TRANS_FAILED
			statusbar_err(NULL, "Cannot close files: %s", buf);
		TRANS_END
	}
}

void on_file_open( void * subj UNUSED )
{
	OPEN_DLG_START( "Open file ...", NULL, NULL )
	{
		char err[TRANS_ERR_SIZE];
		struct obj_doc * doc;
		struct obj * obj;
		TRANS_BEGIN_MAIN(tlo_universum, "Open file", err)
			doc = (struct obj_doc*)doc_load(filenames[i]);
			ASSERT_DOC(doc);
			statusbar_msg(NULL, "File %s loaded.", filenames[i]);
			prop_set_text( obj2o(&doc->obj), string_filename, filenames[i] );
			doc_set_all_saved(doc);
			if ((obj = obj_get_first_son(&doc->obj)) && obj->type == OT_TLO)
				create_view(tlo_get_tlg(obj2tlo(obj)));
		TRANS_FAILED
			statusbar_err(NULL, "Error loading file: %s", err);
			SHOW_ERROR_MESSAGE( GTK_WINDOW(main_window), "Error loading file %s: %s", filenames[i], err );
		TRANS_END
  	}
	OPEN_DLG_END;
}

static void gui_save_document( struct obj * doc, const char * filename )
{
	char err[TRANS_ERR_SIZE];
	ASSERT_DOC( doc );
	TRANS_BEGIN_MAIN(tlo_universum, "Save File", err)
		doc_save ( obj2doc(doc), filename);
		statusbar_msg( NULL, "File %s saved.", filename);
		prop_set_text( obj2o(doc), string_filename, filename );
		doc_set_all_saved( obj2doc(doc) );
	TRANS_FAILED
		statusbar_err( NULL, "Cannot save the file: %s", err);
		SHOW_ERROR_MESSAGE( GTK_WINDOW(main_window), "Error saving file %s: %s", filename, err );
	TRANS_END
}

void on_file_save( void * subj )
{
	char err[TRANS_ERR_SIZE];
	struct obj * volatile obj;
	if ( subj ) obj = subj; else obj = obj_get_first_selected();

	TRANS_BEGIN_MAIN(tlo_universum, "Save File", err)
		const char * filename = prop_lookup_text( obj2o(obj), string_filename );
		if ( !filename || filename == string_file_unsaved->text ) on_file_save_as(obj );
		else gui_save_document( obj, filename );
	TRANS_FAILED
		statusbar_err( NULL, "Cannot save the file: %s", err);
	TRANS_END
}

void on_file_save_as( void * subj )
{
	struct obj * obj;
	if ( subj ) obj = subj; else obj = obj_get_first_selected();
	ASSERT_DOC(obj);

	const char * on = prop_lookup_text( obj2o(obj), string_name );
	if (!on) BUG();
	char nn[strlen(on) + strlen (last_save_folder) + 10];
	sprintf( nn, "%s%s.vrr", last_save_folder, on );

	SAVE_DLG_START("Save file as ... ", NULL, nn )
	{ gui_save_document( obj, filename ); }
	SAVE_DLG_END;
}

void on_file_save_all( void * subj UNUSED )
{
	struct obj * obj;
	WALK_OBJ_RECURSIVE( obj_universum, obj )
		if ( obj->type == OT_DOCUMENT )
			on_file_save(obj);
}

void on_file_exit( void * subj UNUSED ) { main_quit(); }

void on_file_new_page( void * subj )
{
	char buf[TRANS_ERR_SIZE];
	struct obj * volatile obj;
	if ( subj ) obj = subj; else obj = obj_get_first_selected();

	ASSERT_DOC( obj );
	TRANS_BEGIN_MAIN(tlo_universum, "Create new page", buf)
		struct obj_tlo * tlo = tlo_new_named( get_new_page_noname(obj2doc(obj)) );
		obj_link_tail( &tlo->obj, obj );
		create_view(tlo_get_tlg(tlo));
	TRANS_FAILED
		statusbar_err( NULL, "Cannot create a new page: %s", buf);
	TRANS_END
}

void on_file_close_page( void * subj )
{
	char buf[TRANS_ERR_SIZE];
	if ( subj )
	{
		ASSERT_TLO( subj );
		TRANS_BEGIN_MAIN(tlo_universum, "Close page", buf)
			obj_unlink( subj );
		TRANS_FAILED
			statusbar_err( NULL, "Cannot close the page: %s", buf);
		TRANS_END
	} else {
		TRANS_BEGIN_MAIN(tlo_universum, "Close pages", buf)
			obj_unlink_selected();
		TRANS_FAILED
			statusbar_err( NULL, "Cannot close pages: %s", buf);
		TRANS_END
	}
}

void on_page_new_view( void * subj )
{
	struct o * o;
	if ( subj )
	{
		o = void2o(subj);
		ASSERT_TLO(o);
		create_view(tlo_get_tlg(obj2tlo(o2obj(o))));
		return;
	}
	struct obj * obj;
	WALK_OBJ_SELECTED(obj)
		if (obj->type == OT_TLO)
			create_view(tlo_get_tlg(obj2tlo(obj)));
}

void on_group_new_view( void * subj )
{
	GET_PARAM_GRP(group);
	create_view(group);
}

/**** basic transforms ****/

static void on_transform_basic( const struct geom_transform *t, void * subj, const char *desc )
{
	ASSERT_TLO(context.tlo);
	ASSERT_GROUP(context.group);
	ASSERT(desc);
	struct go * volatile go = NULL;
	if (subj)
		go = void2go(subj);

	char errs[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( context.tlo, desc, errs )
		if (go)
			go_select(go);

		struct geom_rectangle r, bbox;
		geom_rectangle_set_empty(&bbox);
		WALK_GROUP_SELECTED_GO (context.group, go)
		{
			if (go_get_exact_bbox_after_trans(go, &geom_transform_identity, &r))
				trans_fail("Numeric error in transformation");
			DBG("Got bbox %.3f, %.3f, %.3f, %.3f\n", r.ll.x, r.ll.y, r.ur.x, r.ur.y );
			if (!GEOM_ETEST(geom_rectangle_merge(&r, &bbox, &bbox)))
				trans_fail("Numeric error in transformation");
		}
		struct geom_transform2 tt, m;
		tt.primary = *t;
		if (!GEOM_ETEST(geom_transform2_set_move( &m, fmidfinr(bbox.ur.x, bbox.ll.x), fmidfinr(bbox.ur.y, bbox.ll.y) )) ||
		    !GEOM_ETEST(geom_transform_merge(&tt.primary, &m.primary, &tt.primary)) ||
		    !GEOM_ETEST(geom_transform_merge(&m.inverted, &tt.primary, &tt.primary)) ||
		    !GEOM_ETEST(geom_transform_invert(&tt.primary, &tt.inverted)))
			trans_fail("Numeric error in transformation");

		tsort_selected_transform(context.group, &tt);
	TRANS_FAILED
		statusbar_err( NULL, "Cannot transform: %s", errs );
	TRANS_END;
}

void on_rotate90( void * subj ) { on_transform_basic( &geom_transform_rot90, subj, "Rotate by 90 degrees" ); }
void on_rotate180( void * subj ) { on_transform_basic( &geom_transform_rot180, subj, "Rotate by 180 degrees" ); }
void on_rotate270( void * subj ) { on_transform_basic( &geom_transform_rot270, subj, "Rotate by 270 degrees" ); }
void on_flip_vertical( void * subj ) { on_transform_basic( &geom_transform_vflip, subj, "Flip vertically" ); }
void on_flip_horizontal( void * subj ) { on_transform_basic( &geom_transform_hflip, subj, "Flip horizontally" ); }

/**** windows *****/

void on_show_main_window( void * subj UNUSED )
{
	gtk_widget_show( main_window );
}

void on_show_property_window (void *subj UNUSED)
{
	if (!context_property_window)
		prop_window_create(NULL);
	else
		gtk_widget_destroy(context_property_window->window);
}

void on_show_console_window( void * subj UNUSED )
{
	if (!console_window)
	{
		console_create();
	}
	else if ( console_window->visible )
	{
		gtk_widget_hide( console_window->window );
		console_window->visible = 0;
	}
	else
	{
		gtk_widget_show( console_window->window );
		console_window->visible = 1;
	}
}

void on_show_universe_browser( void * subj UNUSED )
{
	if (!(wnd_univ_browser.window)){
		create_universe_browser();
		return;
	}
	if ( wnd_univ_browser.visible ){
		gtk_widget_hide( wnd_univ_browser.window );
		wnd_univ_browser.visible = 0;
	} else {
		gtk_widget_show( wnd_univ_browser.window );
		wnd_univ_browser.visible = 1;
	}
}

void on_show_undo_history( void * subj UNUSED )
{
	create_wnd_undo_history(tlo_universum);
}

void on_view_clipboard( void * subj UNUSED ){ create_view( tlo_get_tlg(tlo_clipboard) ); }


void on_undo_history( void * subj )
{
	if ( !subj ) subj = obj_get_first_selected();
	ASSERT_TLO( subj );
	create_wnd_undo_history( subj );
}

static void open_in_browser(const char * url)
{
	const char * browser = prop_lookup_text( obj2o(obj_universum), string_browser );
	if (!browser) { statusbar_err(NULL, "Please set the WWW browser in the Settings first!"); return; }
	uns run_in_term = prop_lookup_uns( obj2o(obj_universum), string_browser_in_term, 2 );
	if (run_in_term == 2) { statusbar_err(NULL, "Please set if to run the browser in a terminal first!"); return; }
	const char * term = NULL;
	if (run_in_term)
	{
		term = prop_lookup_text( obj2o(obj_universum), string_terminal );
		if (!term) { statusbar_err(NULL, "Please set the terminal in the Settings first!"); return; }
	}

	switch ( fork() )
	{
	case -1: statusbar_err(NULL, "Cannot fork!"); break;
	case 0:
	{
		if (run_in_term)
			execlp( term, term, "-e", browser, url, NULL );
		else
			execlp( browser, browser, url, NULL );

		if (run_in_term)
			err( "Error running the browser '%s' in terminal '%s' on '%s': %s\n", browser, term, url, strerror(errno));
		else
			err( "Error running the browser '%s' on '%s': %s\n", browser, url, strerror(errno));
		exit(1);
	}
	default: ;
	}
}

void on_open_www( void * subj UNUSED) { open_in_browser(VRR_HOMEPAGE); }
void on_help( void * subj UNUSED) { open_in_browser("file://"DOC_PATH"manual/manual.html"); }

void on_kerntest( void * subj UNUSED ) { }


void main_quit(void)
{
	struct obj * obj;
	WALK_OBJ_SON(obj_universum, obj)
		if ( obj->type == OT_DOCUMENT && !doc_is_all_saved(obj2doc(obj)) )
		{
			if (!prompt_discard_changes()) return;
			break;
		}

	closing_all = 1;
	DBG("Close the UB window if open\n");
	if ( wnd_univ_browser.window )
		gtk_widget_destroy( wnd_univ_browser.window );
	DBG("Destroying the UB menu\n");
	if ( wnd_univ_browser.menu )
		gtk_object_destroy( GTK_OBJECT(wnd_univ_browser.menu) );
	DBG("Close all open views\n");
	while ( view_list )
		close_view( (struct view*)view_list->data );
	DBG("Destroying the view menu\n");
	if ( view_menu )
		gtk_object_destroy( GTK_OBJECT(view_menu) );
	DBG("Close all undo history windows if open\n");
	while ( wuh_list )
		gtk_widget_destroy( ((struct wnd_undo_history*)wuh_list->data)->window );
	DBG("Close all text edit windows if open\n");
	while ( text_edit_list )
		gtk_widget_destroy( ((struct text_edit*)text_edit_list->data)->window );
	DBG("Close settings window if open\n");
	if ( wnd_settings.window ) gtk_widget_destroy( wnd_settings.window );
	DBG("Close style window if open\n");
	if ( wnd_style.window ) gtk_widget_destroy( wnd_style.window );

	gui_fonts_finish();

	DBG("Quit\n");
	gtk_main_quit();
}

char * get_new_noname(void)
{
	sprintf( noname_buf, "Untitled %d", ++noname_counter );
	return noname_buf;
}
char * get_new_page_noname( struct obj_doc * doc )
{
	uns cnt = 1;
	for ( struct obj * ooo = obj_get_first_son( &doc->obj ); ooo; ooo = obj_get_next_son(ooo) ) cnt++;
	sprintf( noname_buf, "Page %u", cnt );
	return noname_buf;
}

void refresh_gui(void)
{
	if ( closing_all ) return;
	main_window_refresh_menu();
	main_window_refresh_toolbar();
	ub_refresh_menu();
	view_refresh_menu();
	for ( GSList * l = view_list; l; l = l->next )
	{
		struct view * vw = (struct view*)l->data;
		view_refresh_toolbar(vw);
		if ( vw->tlo == context.tlo ) gtk_widget_show( vw->image_on );
		else gtk_widget_hide( vw->image_on );
		if ( tlo_is_saved(vw->tlo) ) gtk_widget_hide( vw->image_modified );
		else gtk_widget_show( vw->image_modified );
		transform_update(vw);
	}
	for ( GSList * l = wuh_list; l; l = l->next )
		wuh_refresh_toolbar( (struct wnd_undo_history*)l->data );
}

GdkPixbuf * icon_load( const char * filename )
{
	GError * err = NULL;
	GdkPixbuf * pb;
	if (!(pb = gdk_pixbuf_new_from_file(filename, &err)))
	{
		statusbar_err(NULL, "Error loading icon: %s", err->message);
		g_error_free(err);
	}
	return pb;
}

GtkWidget * image_get_from_icon( GdkPixbuf * icon )
{
	if ( icon ) return gtk_image_new_from_pixbuf(icon);
	else return gtk_image_new_from_stock("gtk-missing-image", GTK_ICON_SIZE_MENU);
}

static void resources_init(void)
{
	icon_arrow_up = icon_load(LIB_PATH"up.png");
	icon_arrow_down = icon_load(LIB_PATH"down.png");
	icon_arrow_left = icon_load(LIB_PATH"left.png");
	icon_arrow_right = icon_load(LIB_PATH"right.png");
	icon_pencil_on = icon_load(LIB_PATH"pencil-on.png");
	icon_modified = icon_load(LIB_PATH"modified.png");
	icon_rotation_left = icon_load(LIB_PATH"rotation_left.png");
	icon_rotation_right = icon_load(LIB_PATH"rotation_right.png");
	icon_zoom_out = icon_load(LIB_PATH"zoom_out.png");
	icon_zoom_in = icon_load(LIB_PATH"zoom_in.png");
}
static void resources_finish(void) { }

static void inner_main(int argc UNUSED, char **argv UNUSED)
{
	guilelink_gui_init ();

	vcl_init();

	resources_init();
	mh_init();
	command_manager_init();
#ifdef USE_LIBPAPER
	paperinit ();
#endif

	view_list = wuh_list = NULL;
	prop_window_list = NULL;
	text_edit_list = NULL;
	wnd_univ_browser.menu = view_menu = NULL;
	wnd_plugins.visible = 0;
	wnd_settings.window = NULL;

	main_window_create();
	gui_prop_init();
	factory_init();
	trans_clear_undo(tlo_universum); //clear the initial transactions
	set_fifi_enabled(prop_lookup_uns( obj2o(obj_universum), string_enable_fifi, ENABLE_FIFI ));

	struct obj * obj, * o2;
	WALK_OBJ_RECURSIVE(obj_universum, obj)
		if (obj->type == OT_DOCUMENT && (o2 = obj_get_first_son(obj)) && o2->type == OT_TLO)
			create_view(tlo_get_tlg(obj2tlo(o2)));
	refresh_gui();

	gtk_main();

	factory_finish();
	gui_prop_finish();
	mh_finish();
	command_manager_finish();
	resources_finish();
	kernel_finish();
#ifdef USE_LIBPAPER
	paperdone();
#endif
}

int main( int argc, char * * argv )
{
	gtk_disable_setlocale();
	setlocale (LC_CTYPE, "");
	gtk_init( &argc, &argv ); //read GTK options first

	kernel_init(argc, argv, 1, inner_main);

	ASSERT(0);
}

/** Outputs a formatted message to the statusbar of the giver window.
 * If the window is NULL, tries to use the context window and then the main window.
 * If none of them is non-NULL, writes to stderr. */
void statusbar_msg(struct window * window, const char * fmt, ... )
{
	ASSERT(fmt);
	va_list ap;
	char buf[MAX_ERR_MSG_LENGTH];
	va_start( ap, fmt );
	int cnt = vsnprintf( buf, MAX_ERR_MSG_LENGTH, fmt, ap );
	va_end( ap );

	if(unlikely( cnt < 0 || cnt >= MAX_ERR_MSG_LENGTH ))
	{ DBG("Statusbar message too long.\n"); return; }
	if (!cnt) return;

	if ( window || (window = (struct window*)context.w) || (window = (struct window*)&wnd_main, wnd_main.window) )
	{
		ASSERT(GTK_IS_STATUSBAR(window->statusbar));
		gtk_statusbar_pop(GTK_STATUSBAR(window->statusbar), window->context_id);
		gtk_statusbar_push(GTK_STATUSBAR(window->statusbar), window->context_id, buf);
	}
	else err("%s\n", buf);
}

void statusbar_clear(struct window *window)
{
	if ( window || (window = (struct window*)context.w) || (window = (struct window*)&wnd_main, wnd_main.window) )
	{
		gtk_statusbar_pop( GTK_STATUSBAR(window->statusbar), window->context_id);
		gtk_statusbar_push( GTK_STATUSBAR(window->statusbar), window->context_id, " ");
	}
}

// cuts white chars form beginning and end of the string
char *str_trim(char *str)
{
	char *ptr=str + strlen(str);

	while((ptr > str) && isspace(*(ptr-1)))
		ptr--;
	if(ptr > str)
		*ptr = 0;
	ptr = str;
	while(*ptr && isspace(*ptr))
		ptr++;
	memmove(str, ptr, strlen(ptr) + 1);

	return str;
}

/* Main GUI hooks */
void mh_add_go_hook( struct obj * obj )
{
	struct go_hook * gh = xmalloc_zero(sizeof(struct go_hook));
	gh->data = (void *)&mh_init;
	gh->go_unlink = mh_go_after_delete;
	gh->go_selected = mh_go_selected;
	gh->go_unselected = mh_go_unselected;
	DBG("Main hooks: Adding a GO hook to object %p\n", obj );
	go_hook_add(obj2tlo(obj), gh);
}

void mh_init(void)
{
	struct obj_hook * oh = xmalloc_zero(sizeof(struct obj_hook));
	oh->data = (void *)&mh_init;
	oh->obj_link = mh_obj_added;
	oh->obj_unlink = mh_obj_after_delete;
	oh->obj_selected = oh->obj_unselected = mh_obj_sel_changed;
	DBG("Main hooks: adding the obj hook\n");
	obj_hook_add(oh);

	struct obj * ooo;
	WALK_OBJ_RECURSIVE( obj_universum, ooo ) if ( ooo->type == OT_TLO ) mh_add_go_hook(ooo);
}
void mh_finish(void)
{
	DBG("Main hooks: removing the obj hook\n");
	obj_hook_remove_by_data((void *)&mh_init);

	struct obj* ooo;
	struct go_hook * h;
	WALK_OBJ_RECURSIVE( obj_universum, ooo )
		if ( ooo->type == OT_TLO )
		{
			DBG("Main hooks: Removing a GO hook from %p\n", ooo );
			if ( h = go_hook_remove_by_data( obj2tlo(ooo), (void *)&mh_init )) xfree(h); else BUG();
		}
}
void mh_go_selected( struct o * o, struct go_hook * hook UNUSED )
{
	struct go * go = o2go(o);
	if ( go_get_group(go) != context.group ) return;

	if ( ++context.sel_cnt[go->type] == 1 ) context.sel_mask |= GOT_TO_VTM(go->type);
	context.sel_cnt_all++;
}
void mh_go_unselected( struct o * o, struct go_hook * hook UNUSED )
{
	struct go * go = o2go(o);
	if ( go_get_group(go) != context.group ) return;

	ASSERT( context.sel_cnt[go->type] > 0  &&  context.sel_cnt_all > 0 );
	if ( --context.sel_cnt[go->type] == 0 ) context.sel_mask &= ~(GOT_TO_VTM(go->type));
	context.sel_cnt_all--;
}
void mh_obj_sel_changed( struct o * o UNUSED, struct obj_hook * hook UNUSED ) { context.msel_valid = 0; }

void mh_obj_added( struct o * o, struct obj_hook * hook UNUSED )
{
	DBG("Main hooks: Object %p added\n", o );
	struct obj * oo, * obj = o2obj(o);
	WALK_OBJ_RECURSIVE( obj, oo )
		if ( oo->type == OT_TLO )
		{
			mh_add_go_hook(oo);
		}
}
void mh_obj_after_delete( struct o * o, struct o * owner UNUSED, struct o * prev UNUSED, struct obj_hook * hook UNUSED )
{
	struct obj * oo, * obj = o2obj(o);
	void * h;
	DBG("Main hooks: Object %p after delete\n", o );
	DBG("Main hooks: Removing an OBJ hook from %p\n", o );
	WALK_OBJ_RECURSIVE( obj, oo )
		if ( oo->type == OT_TLO )
		{
			DBG("Main hooks: Removing a GO hook from %p\n", oo);
			if ( h = go_hook_remove_by_data( obj2tlo(oo), (void *)&mh_init )) xfree(h); else BUG();
		}

	if ( context.tlo && o_is_ancestor( obj2o(obj), obj2o(&context.tlo->obj) ))
		change_context( context.w, NULL, NULL );
}
void mh_go_after_delete( struct o * o, struct o * owner UNUSED, struct o * prev UNUSED, struct go_hook * gh UNUSED )
{
	if ( o == context.o ) change_context( context.w, context.group, NULL );
	else if ( context.group && o == go2o(&context.group->go) ) change_context( context.w, NULL, NULL );
}

void win_o_init( struct window * window, uns type )
{
	ASSERT(window);
	DBG("Init window %p, type %u\n", window, type );
	window->type = type;
	window->proxy = SCM_BOOL_F;
	window->statusbar = gtk_statusbar_new();
	window->context_id = gtk_statusbar_get_context_id( GTK_STATUSBAR(window->statusbar), "VRR!");
}

void win_o_finish( struct window * window )
{
	ASSERT(window);
	DBG("Finish window %p, type %u\n", window, window->type );
	window_proxy_clean(window);
}
