#ifndef __VRR_MAIN_H__
#define __VRR_MAIN_H__

/** @file
 * GUI - The main gui structures, windows, and functions
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

/**
 * \defgroup GUI
 * The Graphic User Interface.
 */

#include "lib/lib.h"

#include <gtk/gtk.h>
#include <libguile.h>
#include "kernel/kernel.h"
#include "gui/visualisation.h"

/** Maximal length of error mssage (includes also additional arguments).  */
#define MAX_ERR_MSG_LENGTH		TRANS_ERR_SIZE + 200
/** URL of the VRR homepage */
#define VRR_HOMEPAGE			"http://vrr.ucw.cz/"

/** The snap tolerance in pixels. */
#define TOLERANCE_PIX			10
/** Text editor, which can be used by vrr to modify text files */
#define EXTERNAL_TEXT_EDITOR		"vim"
/** Do we use a text editor which must be run in a terminal? */
#define TEXT_EDITOR_RUN_IN_TERM		1
/** Which command should be use to open web browser for help?  */
#define WWW_BROWSER			"links"
/** Do we use a browser which must be run in a terminal? */
#define WWW_BROWSER_RUN_IN_TERM		1
/** The terminal to run browser and/or text editor, if needed */
#define FAVOURITE_TERMINAL		"rxvt"
/** The x-coordinate of the first grid vector */
#define GRID_X1				10.0
/** The y-coordinate of the first grid vector */
#define GRID_Y1				0.0
/** The x-coordinate of the second grid vector */
#define GRID_X2				0.0
/** The y-coordinate of the second grid vector */
#define GRID_Y2				10.0
/** The default grid color */
#define GRID_COLOR			(rgba_uns2uns(128, 128, 128, 255))
/** The resolution of the horizontal ruler */
#define RULERX_RES			10.0
/** The resolution of the vertical ruler */
#define RULERY_RES			10.0
/** Whether to drag the image, not the view */
#define PANNING_DRAG_IMAGE		0
/** Whether to enable Fifi */
#define ENABLE_FIFI			0

/** A macro that defines a GValue and initializes it to zero. */
#define SAFE_G_VALUE(_x) GValue _x; bzero(&_x, sizeof(_x))

/** Flags for snap settings. */
enum
{
	SNAP_HANGERS		= 1,
	SNAP_GRID		= 2,
	SNAP_LINES		= 4,
	SNAP_INTERS		= 8,
	SNAP_DEPENDENCY		= 16
};

extern GtkWidget * view_menu;		//!< The popup menu of all views.
extern GSList * view_list;		//!< The list of all existing views.
extern GtkWidget * main_window; 	//!< The one and the only main window instance.
extern struct console *console_window;	//!< Console (the one and the only one :))

/** The window type identifier used in the gerenal window structure struct window. */
enum window_type
{
	WT_VIEW,
	WT_UNDO_HISTORY,
	WT_UNIVERSE_BROWSER,
	WT_PROPERTIES,
	WT_PLUGINS,
	WT_TEXT_EDITOR,
	WT_CONSOLE,
	WT_MAIN_WINDOW,
	WT_UNITS,
};

/** The common data for most GUI windows. They are in the beginning of the structure
 * to enable typecasts. */
#define WIN_O_MAGIC u8 type; SCM proxy; GtkWidget * statusbar; gint context_id

/** The general window structure used for most GUI windows. */
struct window { WIN_O_MAGIC; };
/** Initialize the window with the given type, create the statusbar, initialize scheme proxy. */
void win_o_init( struct window * window, uns type );
/** Finalize the window. */
void win_o_finish( struct window * window );

/** The context structure. Contains the currently significant objects in the GUI. */
struct context {
	struct window * w;	//!< The current window
	struct obj_doc * doc;	//!< The context document; typically the father of the context tlo
	struct obj_tlo * tlo;	//!< The current TLO for nonmeta transactions, ancestor of the context group
	struct go_group * group; //!< the group inside the context.tlo for which go selection is calculated
	struct o * o;		//!< The current object (meta or nonmeta)
	int sel_cnt[GOT_MAX];	//!< The array of counts of selected gos in the context.group; per each go type
	int sel_cnt_all;	//!< The total count of selected gos in the context.group
	int msel_cnt_all;	//!< The count of all selected meta objects
	int sel_valid;		//!< A flag if the current selection is valid
	int msel_valid;		//!< A flag if the current meta selection is valid
	uns sel_mask;		//!< The bitmask of selected gos in the context.group
	uns msel_mask;		//!< The bitmask of selected objects
};
/** The main GUI context */
extern struct context context;
/** Changes the current context and does some change notifications by calling some functions */
void change_context( struct window * window, struct go_group * group, struct o * o );

/* ***** MAIN WINDOW *****/
/** Main window structure */
struct wnd_main 
{
	WIN_O_MAGIC;
	GtkWidget * window;
	GtkWidget * menubar, * menubox;
	GtkWidget * toolbar, * toolbox;
};
/** The Main window instance */
extern struct wnd_main wnd_main;
/** Create the main window (the only instance in wnd_main)*/
void main_window_create(void);
/** Show the main window (the only instance in wnd_main)*/
void main_window_show(void);

void main_window_refresh_menu(void);
void main_window_refresh_toolbar(void);

/* ***** UNIVERSE BROWSER *****/

struct VrrKernelStore;
/** The Universe browser window structure. */
struct wnd_univ_browser{
	WIN_O_MAGIC;
	GtkWidget * window;			//! The GTK window.
	int visible;				//! A flag if the window is created and visible.
	struct VrrKernelStore * tree_store;	//! The kernel-gui interface structure.
	GtkWidget * tree_view;			//! The TreeView widget to display the kernel structure.
	GtkWidget * menu;			//! The popup menu.
};
/** The one and the only instance of the Universe browser window, */
extern struct wnd_univ_browser wnd_univ_browser;

/** Create the Universe browser window. The window data structure is the global wnd_univ_browser. */
void create_universe_browser(void);
/** Refresh the menu (and rebuild, if necessary. */
void ub_refresh_menu(void);
/** Show/hide the context icons where needed. */
void ub_context_changed(struct context * old_context);

/* ***** OPTIONS WINDOW ******/
/** The Global settings window strucure */
struct wnd_settings{
	GtkWidget *window;
	struct prop_item * pi;
	struct obj_hook oh;
};
/** The one and the only Global settings window instance */
extern struct wnd_settings wnd_settings;

/* ***** STYLE WINDOW ******/
/** The Style window structure */
struct wnd_style
{
	GtkWidget * window;
	struct prop_item * pi;
	struct go_hook gh;
};
/** The one and the only Style window instance */
extern struct wnd_style wnd_style;

/* ***** PLUGIN WINDOW *****/
/** The Plugin window structure */
struct wnd_plugins
{
	WIN_O_MAGIC;
	GtkWidget * window;
	int visible;
	GtkListStore * plugin_list;
	GtkTreeStore * function_list;
	GtkWidget * plugin_view;
	GtkWidget * function_view;
	GtkWidget * unload_button;
};
/** The one and the only Plugin window instance */
extern struct wnd_plugins wnd_plugins;

/* ***** THE VIEW ******/
/** The internal moving modes of the View */
enum moving_mode
{
	MOVE_NONE,
	MOVE_FACTORY,
	MOVE_TRANSFORM,
	MOVE_TRANSFORM_2,
	MOVE_PANNING,
	MOVE_PANNING_NAV,
	MOVE_PANNING_IMG,
	MOVE_SELECT,
	MOVE_CNT
};

struct view_navigator;
struct visualisation;
/** The View window structure */
struct view{
	WIN_O_MAGIC;
	struct go_group * group;	//!< the top level group
	struct obj_tlo *tlo;		//!< the parent tlo of the top level group
	struct obj_hook obj_hook;	//!< The object hook
	struct go_hook go_hook;		//!< The go hook for the displayed tlo
	struct visualisation vis;	//!< The View's visualisation
	int tf_drag_on;			//!< The drag mode of Santiago's tool
	struct geom_point transform_origin;	//!< The transformation origin of the Select/Transform tool
	struct geom_point axis;		//!< The second skew axis point of the Select/Transofom tool
	uns transform_state;		//!< The state of the Select/Transofom tool
	real tolerance;			//!< The snap tolerance recomputed to coords (for the current zoom)
	struct geom_transform2 grid;	//!< The grid transformation matrix.
	struct vcl_affinity * fifi;	//!< The secondary cursor marking the snap point

	GtkWidget *window;		//!< The GTK window
	GtkWidget *table;		//!< the main widget table
 	GtkWidget *vruler;		//!< vertical ruler
	GtkWidget *hruler;		//!< horizontal rulers
	GtkWidget *toolbar;		//!< toolbar
	GtkWidget *toolbar2;		//!< toolbar 2
	GtkAdjustment *vadj;		//!< adjustment (vertical scrollbar)
	GtkAdjustment *hadj;		//!< adjustment (horizontal scrollbar)
	GtkWidget *vscrollbar;		//!< pointer to the vertical scrollbar
	GtkWidget *hscrollbar;		//!< pointer to the horizontal scrollbar
	GtkWidget *image_on;		//!< The context indicator image
	GtkWidget *image_modified;	//!< The modified indicator image
	struct view_navigator * navigator;	//!< the navigator (of orientation in view) window struct
	
	GtkWidget * zoombar;		//!< A label to display zoom, rotation etc.
	GtkWidget * factory_statusbar;	//!< A statusbar for factory messages.
	GtkWidget * factory_box;	//!< A container for factory property editor widget.
	uns factory_box_full;		//!< A flag if the factory box contains some widgets.
	GtkWidget * coords_label;	//!< A label to diplay the current coords.
	struct prop_item * pi;		//!< A prop_item structure for the factory property editor widget.
};
/** The View navigator structure */
struct view_navigator
{
	GtkWidget * window;
	GtkSpinButton * zoom_entry, * rot_entry;
	real zoom, rotation;		//!< The internal saved values
	struct visualisation vis;	//!< The navigator's visualisation of the viewed image
	struct vcl_path * box;		//!< The green rectangle box
	struct view * parent_view;	//!< The parent view
};
/** Create a view for the given group */
struct view * create_view(struct go_group * group);
/** Close the view and deallocate all resources (some of them need to be deallocated before the window is destroyed) */
void close_view(struct view * vw);
void view_refresh_toolbar(struct view * vw);
void view_refresh_menu(void);
void view_update_grid(struct view * vw);

void transform_update( struct view *vw );
void view_get_size (struct view * vw, uns *width, uns *height);

//MOVING FUNCTIONS
/** Start moving in view. Set the given moving mode and the cursor of cursor_type to the view window.
 * @ai pix the pixel coordinates of the startpoint
 */
void view_moving_start(struct geom_point pix, struct view * vw, enum moving_mode mode, GdkCursorType cursor_type);
/** End the moving, unset the cursor and set moving state to MOVE_NONE */
void view_moving_end(void);
/** Process the button release event of the moving drag */
void view_moving_release( GdkEventButton *event, struct view * vw );
/** Save the last move coordinates and set the events_pending flag on. */
void view_moving_motion( GdkEventMotion *event, struct view *vw );
/** Process the click event -- save the coordinates */
void view_clicked( GdkEventButton * event, struct view * vw );
/** Start the rectangular selection */
void view_selection_start( struct geom_point coords, struct geom_point coords_pix, struct view * vw );
/** Start the transformation drag move */
int view_transform_start( GdkEventButton *event, struct geom_point coords, struct geom_point coords_pix, struct view * vw );
/** Start the Santiago's transformation drag move */
void view_s_transform_start( GdkEventButton *event, struct geom_point coords, struct geom_point coords_pix, struct view * view );
/** Start the panning move */
void view_panning_start( struct geom_point coords_pix, struct view * vw );
/** Start the panning move in the View navigator */
void view_panning_nav_start( struct geom_point coords, struct geom_point coords_pix, struct view * vw );

/** Set Fifi on or off
 * @ai on Whether to show Fifi
 */
void set_fifi_on(uns on);
/** Enable or disable Fifi
 * @ai enabled Whether to enable Fifi
 */
void set_fifi_enabled(uns enabled);
/** Set or unset the shift mode of the tranformation box */
void transform_set_shift( int on, struct view * vw );
/** Update and refresh the transformation box and gadgets */
void transform_update( struct view *vw );
/** Recalculate the dimension mode of the transformation bbox (empty, 
 * wide, high) and redisplay it. */
void transform_recalc_dimen( struct view * vw );
/** Recalculate the transformation box and gadget positions */ 
void transform_recalc_all( struct view * vw );
/** The view moving mode of the one view where there was a move started recently */
extern enum moving_mode moving_mode;

/* ****** UNDO HISTORY WINDOW *****/

struct VrrUndoStore;
/** The Undo history window structure */
struct wnd_undo_history{
	WIN_O_MAGIC;
	GtkWidget * window;
	GtkWidget * vbox;
	GtkWidget * undo_view;
	struct VrrUndoStore * undo_model;
	GtkWidget * toolbar;
	struct obj_tlo * tlo;
	struct obj_hook obj_hook;
	struct trans_hook trans_hook;
};
extern GSList * wuh_list; //!< The list of opened undo history windows.

/** Create the undo history window.
 * \returns A pointer to the allocated window structure.
 */
struct wnd_undo_history * create_wnd_undo_history(struct obj_tlo *tlo);
void wuh_refresh_toolbar( struct wnd_undo_history * wuh );

/* ****** TEXT EDITOR *******/
/** The Text editor window structure */
struct text_edit
{
	WIN_O_MAGIC;
	uns mode;
	struct o * o;
	struct obj_tlo * parent_tlo;	//!< The parent tlo of the edited o (or tlo_universum)
	struct prop_item * pi;
	string text_key;	//!< The key identified of the text property edited
	uns prop_count;		//!< The count of displayed properties
	GtkWidget * window;
	GtkTextBuffer * text_buf;
	GtkWidget * modified_icon;
	struct obj_hook obj_hook;
	struct go_hook go_hook;
};
/** The list of all existing text editor windows */
extern GSList * text_edit_list;
/** The list of all fonts found by the Fontconfig */
extern GtkListStore *font_list;
/** Destroy the allocated font list */
void gui_fonts_finish(void);
/** Fill the font list using the Fontconfig */
int fill_font_list(void);
/** Convert the TreeIter of the font list to font_id */
uns fonts_iter_to_font_id( GtkTreeIter * ti );
/** Compare the TreeIter of the font list with the font filename */
int fonts_compare_iter_with( GtkTreeIter * ti, char * filename );
/** Return the default or least recently used font_id */
int fonts_get_default(void);
/** Create the text editor window
 * @ai the object whose properties to edit
 * @ai the text property key
 */
void text_edit_create( struct o * o, string text_key );

/* ****** SCHEME CONSOLE *******/

enum console_state
{
        STATE_NORMAL,
	STATE_FSEARCH_FIRST,
	STATE_RSEARCH_FIRST,
	STATE_FSEARCH,
	STATE_RSEARCH
};

struct console_string {
        struct snode list;
        char *data;
};

struct console_search_stack
{
        enum console_state state;
	s16 search_buffer_used;
        s16 search_buffer_chars;
        GtkTextMark *mark;
        struct console_search_stack *next;
};

struct console {
	WIN_O_MAGIC;
	GtkWidget *window;
	GtkTextView *text;
	GtkTextBuffer *buf;
	GtkTextTag *input_tag, *output_tag,  *edit_tag;
	GtkTextMark *open_mark, *close_mark;

	int visible;
        int blockread;
        enum console_state state;

        char *search_buffer;
        s16 search_buffer_size;
        s16 search_buffer_used;
        s16 search_buffer_chars;
        struct console_search_stack *search_stack;

	SCM port;
};

void console_create (void);
void console_init (void);

/* **** OPERATION FACTORY ******/

/** The kind of the input object of a factory state */
enum of_input_kind
{
	OFIK_NONE,
	OFIK_DPR,
	OFIK_PROP,
	OFIK_SEL,
	OFIK_TRANSFORM,
	OFIK_TF,
};

enum of_go_role { OFR_DISPLAY, OFR_CNT };
/** The GO Factory state flags */
enum ofs_flags {
	OFF_FULL = 1, //!< If the state has its required input ready
	OFF_START_MOVE = 2, //!< If the state does start the move
	OFF_END_MOVE = 4, //!< If the state does end the move
	OFF_PATHABLE = 8, //!< If the state creates an object which can be contained in a path
	OFF_HIDE_HANGERS = 16 //!< If the state does not work with hangers, hide them
};
/** The structure for one GO factory state */
struct of_state {
	uns flags;			//!< A combination of the enum ofs_flags values
	enum of_input_kind input_kind;	//!< Specifies what input is wanted from the user.
	uns input_type;			//!< The input type specific for the input kind. For input_kind == OFIT_DPR is of dist_pass_item_type,
					//!< for input_kind == OFIT_PROP is of prop_type. For other kinds is ignored.
	uns input_subtype;		//!< If the input_kind is OFIT_PROP, this specifies the property subtype of the input value.
	union {
		struct dist_pass_result dpr;	//!< The snap result -- hanger, coords, go, ...
		struct prop prop;		//!< The parameter obtained from the user in a way different from snap, any value that can be stored as property
	} value;			//!< The value obtained from the user: a point, go, number, etc.
	char * description;		//!< A sententious description of what is wanted from the user.
	void (* action_func)(struct of_state *);	//!< A function to be called when leaving this state.
	void (* cleanup_func)(struct of_state *);	//!< Cleanup actions made in this step.
	struct of_state * prev;         //!< The previous state in case of a step back.
	struct of_state * next;         //!< The next state in case of success.
	struct undo_gui * prev_undo_item;	//!< Pointer to the last undo item before any action in this state.
	struct go * tmp_go[OFR_CNT];	//!< Temporarily created gos
};
/** The start state which must be the previous state of all the starting ones */
extern struct of_state ofs_start;
/** The ending state which must be the next state for all last states */
extern struct of_state ofs_end;

/** The GO factory flags */
enum of_flags {
	OFF_OPERATION_ACTIVE = 1,	//!< If there is an active operation (not in the initial state, but with some input already obtained
	OFF_PATH_ON = 2,		//!< If the path mode is on
	OFF_PATH_ACTIVE = 4,		//!< If there is a path created
	OFF_PATH_DESTROYED = 8		//!< A path is just being destroyed by the factory
};
/** The one and the only GO Factory instance */
extern struct factory {
	uns flags;
	struct obj_tlo * tlo;		//!< The tlo for the current factory operation
	struct go_group * group;	//!< The group where to link newly created objects
	struct of_state * state;	//!< The current state
	struct of_state * start_state;	//!< The starting state of the current operation
	string desc;			//!< The description of the current operation
	real snap_tolerance;		//!< The current snap tolerance in coords (for the current view)
	uns snap_set;			//!< Bitmask of snap settings
	uns snap_tol_pix; 		//!< The maximum permitted snap distance in pixels
	struct hanger * hanger_cursor;	//!< The mouse cursor hanger, used when moving some objects continually during creation.
	struct go_path * path;		//!< The path which is beaing created (if any).
} factory;

//PUBLIC Operation Factory states
extern struct of_state ofs_select;
extern struct of_state ofs_transform_x;
extern struct of_state ofs_point;
extern struct of_state ofs_decorator_pt;
extern struct of_state ofs_arrow;
extern struct of_state ofs_point_inters;
extern struct of_state ofs_ngon_1;
extern struct of_state ofs_segment_1;
extern struct of_state ofs_bez_cub_1;
extern struct of_state ofs_bez_quad_1;
extern struct of_state ofs_arc_3p_1;
extern struct of_state ofs_arc_center_radius_1;
extern struct of_state ofs_circ_3p_1;
extern struct of_state ofs_circ_center_pt_1;
extern struct of_state ofs_circ_center_radius_1;
extern struct of_state ofs_ellipse_3p_smallest_1;
extern struct of_state ofs_ellipse_3p_ecc_1;
extern struct of_state ofs_ellarc_center_radii_1;
extern struct of_state ofs_ellipse_center_radii_1;
extern struct of_state ofs_ellipse_foci_pt_1;
extern struct of_state ofs_ellipse_xy1ecc_1;
extern struct of_state ofs_transform_1;
extern struct of_state ofs_rehang_1, ofs_rehang_2;
extern struct of_state ofs_text_1;
extern struct of_state ofs_tex_text_1;

//OPERATION FACTORY FUNCTIONS
void factory_init(void);
void factory_finish(void);
/** Start a new factory operation 
 * @ai state The state to start with
 */
void factory_op_start( struct go_group * group, struct of_state * state, char * desc );
/** Break the current operation -- return to the state before it */
void factory_op_break( void );
/** Move the factory cursor to the new point position */
void factory_moving_step(struct geom_point po);
/** Move to the next state. This function should be called when the input value for
 * the current state is ready. */
void factory_op_step( void );
/** Merge all partial transactions for the current operation */
int factory_merge_trans(void);
/** Move to the previous state and undo the effect of the last step */
void factory_op_step_back(void);
/** Output a factory message into the factory statusbar */
void factory_msg(const char * text);

struct go * factory_get_created_go(void);

/** Snap the point using the current snap settings.
 * @ai pos The position to be snapped
 * @ao dpr The result of the dist pass -- the nearest suitable object found
 * @ai grid The view grid
 * @ai tlo The parent tlo inside which to snap
 * @ai maxdist The maximum snap distance
 */
void snap_point( struct geom_point * pos, struct dist_pass_result * dpr, struct geom_transform2 * grid, struct obj_tlo * tlo, real maxdist );
/** Snap the point to an object using the current snap settings.
 * @ai pos The position to be snapped
 * @ao dpr The result of the dist pass -- the nearest suitable object found
 * @ai tlo The parent tlo inside which to snap
 * @ai maxdist The maximum snap distance
 */
void snap_to_go( struct geom_point * pos, struct dist_pass_result * dpr, struct obj_tlo * tlo, real maxdist );
/** Snap the point to an anchor using the current snap settings.
 * @ai pos The position to be snapped
 * @ao dpr The result of the dist pass -- the nearest suitable object found
 * @ai tlo The parent tlo inside which to snap
 * @ai maxdist The maximum snap distance
 */
void snap_to_anchor( struct geom_point * pos, struct dist_pass_result * dpr, struct obj_tlo * tlo, real maxdist );

/* ***** MISC ******/
/** Get a new name for an untitled document.
 * \returns The pointer to a static buffer (should not be freed).
 */
char * get_new_noname(void);
/** Get the name for a new untitled page in the document.
 * \returns The pointer to a static buffer (should not be freed).
 */
char * get_new_page_noname(struct obj_doc * doc);
/** Update all the GUI widgets whose values depend on context or similar settings.
 * Sync all menus (enable/disable) the menu items, indicator icons, ...
 */
void refresh_gui(void);

/** Closes all opened documents and does all neccessary cleaning
 * before quitting. After that, terminates the program.
 */
void main_quit(void);

//RESOURCES
extern GdkPixbuf * icon_arrow_down, * icon_arrow_up, * icon_arrow_left, * icon_arrow_right;
extern GdkPixbuf * icon_pencil_on, * icon_modified;
extern GdkPixbuf * icon_rotation_left, * icon_rotation_right, * icon_zoom_in, * icon_zoom_out;

//FUNCTIONS
GdkPixbuf * icon_load( const char * filename );
GtkWidget * image_get_from_icon( GdkPixbuf * icon );

void statusbar_clear(struct window * window);
void statusbar_msg(struct window * window, const char * fmt, ... ) FORMAT(2,3);
#define statusbar_err(_win,...) statusbar_msg(_win, ":-( "__VA_ARGS__)

/** \brief (Un)selects some objects according to the rect, go and group.
 * If the rect is non-NULL, performs the rectangular selection within the rectangle rect. Otherwise, (un)selects the go.
 * The (un)selection is done according to the keys pressed (passed in the event). Can also unselect all other gos within
 * the given group, if non-NULL. */
void modify_selection( struct go * go, struct obj_tlo * tlo, GdkEventButton * event, struct geom_rectangle * rect, struct go_group * group );
/** (Un)selects the obj according to the keys pressed (passed in the event. */
void modify_meta_selection( struct obj * obj, GdkEventButton * event );

/** Allocates and fills memory with val.  */
float *new_float(float val);

/** Cuts white chars from beginning and end of the str.  */
char *str_trim(char *str);


/* ****** CALLBACKS ******/

void on_file_new( void * subj );
void on_file_open( void * subj );
void on_svg_file_import( void * subj );
void on_import_ipe5( void * subj );
void on_file_save( void * subj );
void on_file_save_all( void * subj );
void on_file_save_as( void * subj );
void on_file_close( void * subj );
void on_file_new_page( void * subj );
void on_file_close_page( void * subj );
void on_group_new_view( void * subj );
void on_page_new_view( void * subj );
void on_document_rerun_tex( void * subj );
void on_page_rerun_tex( void * subj );

void on_view_fit_selected( void * subj );
void on_view_fit_all( void * subj );
void on_view_zoom_in( void * subj );
void on_view_zoom_out( void * subj );
void on_view_zoom_reset( void * subj );

void on_export_pdf( void * subj );
void on_export_ps( void * subj );
void on_export_eps( void * subj );
void on_export_svg( void * subj );

void on_set_vrr_options( void * subj );
void on_help( void * subj);
void on_open_www( void * subj);
void on_show_about_window( void * subj );
void on_file_exit( void * subj );
void on_show_property_window( void * subj );
void on_show_console_window( void * subj );
void on_show_universe_browser( void * subj );
void on_show_main_window( void * subj );
void on_show_undo_history( void * subj );
void on_view_clipboard( void * subj );
void on_show_plugin_window( void * subj );
void on_show_unit_window( void * subj );
void on_show_settings_window( void * subj );
void on_show_style_window( void * subj );

void on_copy( void * subj );
void on_cut( void * subj );
void on_paste( void * subj );

void on_obj_copy( void * subj );
void on_obj_cut( void * subj );
void on_obj_paste( void * subj );

void on_group_selected( void * subj );
void on_group_ungroup( void * subj );
void on_group_pathize_selected( void * subj );

void on_group_select_all( void * subj );
void on_group_select_none( void * subj );

void on_move_up( void * subj );
void on_move_down( void * subj );
void on_move_front( void * subj );
void on_move_back( void * subj );

void on_page_move_up( void * subj );
void on_page_move_down( void * subj );
void on_page_move_front( void * subj );
void on_page_move_back( void * subj );

void on_create_proxy( void * subj );

/** UNDO and REDO */
void on_obj_undo( void * subj );
void on_obj_redo( void * subj );
void on_tlo_undo( void * subj );
void on_tlo_redo( void * subj );
void on_hist_jump( void * subj );

void on_delete( void * subj );
void on_go_enclose_sel( void * subj );

void on_rotate90( void * subj );
void on_rotate180( void * subj );
void on_rotate270( void * subj );
void on_flip_vertical( void * subj );
void on_flip_horizontal( void * subj );

/** Change properties */
void on_change_prop( void * subj );
void on_edit_text( void * subj );

void on_undo_history( void * subj );
void on_dump( void * subj );
void on_kerntest( void * subj );

/* GO creating callbacks. Implemented in creatego.c */
void on_random_point( void * subj );
void on_random_segment( void * subj );
void on_random_bezier_quad( void * subj );
void on_random_bezier_cubic( void * subj );

void on_path_switch( void * subj );
void on_path_end_close( void * subj );
void on_path_end( void * subj );
void on_step_back( void * subj );
void on_break_op( void * subj );

/* Snap modes */
void on_snap_hangers( void * subj );
void on_snap_grid( void * subj );
void on_snap_lines( void * subj );
void on_snap_inters( void * subj );
void on_snap_depend( void * subj );

/* MISC */

/** For an obj, returns his father obj. For a go in a top level
 * group, returns the tlo. For other gos, returns their father group. */
struct o * o_get_gui_ancestor(struct o * o);

#endif
