/** @file
 * GUI - Implementation of the Plugin window
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG
#include "lib/lib.h"

#include "kernel/kernel.h"

#include <gtk/gtk.h>
#include "gui/main.h"
#include "gui/cmdmgr.h"
#include "scheme/glt_gui.h"
#include "gui/dialogs.h"

enum { PLUGIN_NAME_COL, PLUGIN_PTR_COL, PLUGIN_COL_CNT };
enum { PLUG_FUNC_NAME_COL, PLUG_FUNC_FAMILY_COL, PLUG_FUNC_COL_CNT };

struct wnd_plugins wnd_plugins;
static void create_plugin_window(void);
static void append_plugin_to_lists(struct plugin_rec * plugin);
static void refresh_lists(void);

static void on_unload_plugin( GtkWidget * w, gpointer data );
static void on_load_plugin( GtkWidget * w, gpointer data );
static void plugin_cursor_changed( GtkTreeView * t, gpointer data );

static struct plugin_rec * selected_plugin;

void on_show_plugin_window( void * subj UNUSED )
{
	if ( wnd_plugins.visible )
	{
		gtk_widget_destroy( wnd_plugins.window );
		wnd_plugins.visible = 0;
	}
	else
		create_plugin_window();
}

static void destroy_plugin_window( GtkWidget *widget UNUSED, gpointer data UNUSED )
{
	wnd_plugins.visible = 0;
	plugin_change_callback = NULL;
	selected_plugin = NULL;
	win_o_finish( (struct window*)&wnd_plugins );
	refresh_gui();
}

static void create_plugin_window(void)
{
	win_o_init( (struct window*)&wnd_plugins, WT_PLUGINS );

	wnd_plugins.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title( GTK_WINDOW(wnd_plugins.window), "Plugin manager" );
	gtk_window_add_accel_group( GTK_WINDOW(wnd_plugins.window), accel_group );
	gtk_widget_set_size_request( wnd_plugins.window, -1, 200 );

	GtkWidget * table = gtk_table_new( 2, 3, FALSE );
	gtk_container_add( GTK_CONTAINER(wnd_plugins.window), table );

	GtkWidget * sw = gtk_scrolled_window_new( NULL, NULL );
	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW(sw), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC );
	gtk_table_attach( GTK_TABLE(table), sw, 0, 1, 0, 1, GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0 );

	wnd_plugins.plugin_list = GTK_LIST_STORE(gtk_list_store_new( PLUGIN_COL_CNT, G_TYPE_STRING, G_TYPE_POINTER ));
	wnd_plugins.plugin_view = gtk_tree_view_new_with_model( GTK_TREE_MODEL(wnd_plugins.plugin_list) );

	GtkTreeViewColumn * column = gtk_tree_view_column_new_with_attributes( "Loaded plugins:",
		gtk_cell_renderer_text_new(), "text", PLUGIN_NAME_COL, NULL );
	gtk_tree_view_append_column( GTK_TREE_VIEW(wnd_plugins.plugin_view), column );
	gtk_tree_selection_set_mode( gtk_tree_view_get_selection(GTK_TREE_VIEW(wnd_plugins.plugin_view)), GTK_SELECTION_SINGLE );

	gtk_container_add( GTK_CONTAINER(sw), wnd_plugins.plugin_view );
	g_signal_connect( G_OBJECT(wnd_plugins.plugin_view), "cursor-changed", G_CALLBACK(plugin_cursor_changed), NULL );

	sw = gtk_scrolled_window_new( NULL, NULL );
	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW(sw), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC );
	gtk_table_attach( GTK_TABLE(table), sw, 1, 2, 0, 2, GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0 );

	wnd_plugins.function_list = GTK_TREE_STORE(gtk_tree_store_new( PLUG_FUNC_COL_CNT, G_TYPE_STRING, G_TYPE_STRING ));
	wnd_plugins.function_view = gtk_tree_view_new_with_model( GTK_TREE_MODEL(wnd_plugins.function_list ));

	column = gtk_tree_view_column_new_with_attributes( "Functions available in the plugins:",
		gtk_cell_renderer_text_new(), "text", PLUG_FUNC_NAME_COL, "family", PLUG_FUNC_FAMILY_COL, NULL );
	gtk_tree_view_append_column( GTK_TREE_VIEW(wnd_plugins.function_view), column );
	gtk_tree_selection_set_mode( gtk_tree_view_get_selection(GTK_TREE_VIEW(wnd_plugins.function_view)), GTK_SELECTION_NONE );

	gtk_container_add( GTK_CONTAINER(sw), wnd_plugins.function_view );

	GtkWidget * hbutbox = gtk_hbutton_box_new();
	gtk_button_box_set_layout( GTK_BUTTON_BOX(hbutbox), GTK_BUTTONBOX_SPREAD );
	gtk_table_attach( GTK_TABLE(table), hbutbox, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 5, 5 );

	GtkWidget * load_button = gtk_button_new_with_label("Load ...");
	gtk_box_pack_start( GTK_BOX(hbutbox), load_button, FALSE, FALSE, 0 );
	wnd_plugins.unload_button = gtk_button_new_with_label("Unload");
	gtk_box_pack_start( GTK_BOX(hbutbox), wnd_plugins.unload_button, FALSE, FALSE, 0 );
	gtk_widget_set_sensitive( wnd_plugins.unload_button, FALSE );

	gtk_table_attach( GTK_TABLE(table), wnd_plugins.statusbar, 0, 2, 2, 3, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0 );

	g_signal_connect( G_OBJECT(load_button), "clicked", G_CALLBACK(on_load_plugin), NULL );
	g_signal_connect( G_OBJECT(wnd_plugins.unload_button), "clicked", G_CALLBACK(on_unload_plugin), NULL );

	g_signal_connect( G_OBJECT(wnd_plugins.window), "delete_event", G_CALLBACK(gtk_widget_destroy), NULL );
	g_signal_connect( G_OBJECT(wnd_plugins.window), "destroy", G_CALLBACK(destroy_plugin_window), NULL);

	selected_plugin = NULL;
	plugin_change_callback = refresh_lists;
	refresh_lists();

	gtk_widget_show_all( GTK_WIDGET(wnd_plugins.window) );
	wnd_plugins.visible = 1;
}

static void append_plugin_to_lists(struct plugin_rec *plugin)
{
	GtkTreeIter ti, tif;

	gtk_list_store_append( wnd_plugins.plugin_list, &ti );
	gtk_list_store_set( wnd_plugins.plugin_list, &ti, PLUGIN_NAME_COL, plugin->name, PLUGIN_PTR_COL, plugin, -1 );

	gtk_tree_store_append( wnd_plugins.function_list, &ti, NULL );
	gtk_tree_store_set( wnd_plugins.function_list, &ti, PLUG_FUNC_NAME_COL, plugin->name, -1 );

	struct plugin_function * fc;
	char buf[1000];
	CLIST_WALK(fc, plugin->func_list) {
		gtk_tree_store_append( wnd_plugins.function_list, &tif, &ti );
		gtk_tree_store_set( wnd_plugins.function_list, &tif,
			PLUG_FUNC_NAME_COL, plugin_function_prototype(fc, buf, 1000),
			PLUG_FUNC_FAMILY_COL, "Monospace", -1 );
	}
}

static void refresh_lists(void)
{
	gtk_tree_store_clear( wnd_plugins.function_list );
	gtk_list_store_clear( wnd_plugins.plugin_list );

	struct plugin_rec *plug;
	CLIST_WALK(plug, plugin_list)
		append_plugin_to_lists(plug);
}

static void on_load_plugin( GtkWidget * w UNUSED, gpointer data UNUSED )
{
	plugin_change_callback = NULL;
	OPEN_DLG_START("Load plugin ...", NULL, NULL)
	{
		char err_msg[TRANS_ERR_SIZE];
		TRANS_BEGIN_ANONYMOUS (err_msg)
			plugin_load(filenames[i]);
		TRANS_FAILED
			SHOW_ERROR_MESSAGE( GTK_WINDOW(wnd_plugins.window), "Cannot load plugin: %s", err_msg );
		TRANS_END
	}
	OPEN_DLG_END;
	plugin_change_callback = refresh_lists;
	refresh_lists();
}

static void on_unload_plugin( GtkWidget * w UNUSED, gpointer data UNUSED )
{
	ASSERT(selected_plugin);
	char err_msg[TRANS_ERR_SIZE];
	TRANS_BEGIN_ANONYMOUS (err_msg)
		plugin_unload(selected_plugin);
	TRANS_FAILED
		SHOW_ERROR_MESSAGE( GTK_WINDOW(wnd_plugins.window), "Cannot unload plugin: %s", err_msg );
	TRANS_END
}

static void plugin_cursor_changed( GtkTreeView * t UNUSED, gpointer data UNUSED )
{
	GtkTreePath * tp;
	GtkTreeIter selected_row;
	gtk_tree_view_get_cursor( GTK_TREE_VIEW(wnd_plugins.plugin_view), &tp, NULL );

	if (!tp) return;
	if ( gtk_tree_model_get_iter( GTK_TREE_MODEL(wnd_plugins.plugin_list), &selected_row, tp ) )
	{
		SAFE_G_VALUE(ptr);
		gtk_tree_model_get_value( GTK_TREE_MODEL(wnd_plugins.plugin_list), &selected_row, PLUGIN_PTR_COL, &ptr );
		selected_plugin = (struct plugin_rec*)g_value_get_pointer(&ptr);
		ASSERT(selected_plugin);
		gtk_widget_set_sensitive( wnd_plugins.unload_button, selected_plugin->flags & PLUGIN_UNLOADABLE );
		g_value_unset( &ptr );
	}
	gtk_tree_path_free(tp);
}
