/** @file
 * GUI
 * The implementation of the Property editor and the Context property editor.
 * Routines for creating individual property editor widgets that can be
 * inserted into another window.
 * Definitions of property editor widgets for property subtypes.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG
#undef BASIC_DEBUG
#include "lib/lib.h"
#include <stdio.h>
#include <string.h>

#include <gtk/gtk.h>

#include "kernel/kernel.h"
#include "font/font.h"
#include "gui/properties.h"
#include "gui/selectcolor.h"
#include "scheme/glt_gui.h"
#include "gui/dialogs.h"

#ifdef BASIC_DEBUG
#define BASIC_DBG msg
#else
#define BASIC_DBG DBG
#endif

#define CPW context_property_window

enum { PW_NONEMPTY = 1, PW_CONTEXT = 2, PW_SELECTION = 4, PW_MERGED = 8 };

enum { PWT_BUG, PWT_SPIN_UNS, PWT_SPIN_REAL, PWT_COMBO, PWT_CHECKBOX, PWT_ENTRY, PWT_FUNC };

#define COLOR_INNER_TO_GTK(x) ((x) << 8)
#define COLOR_GTK_TO_INNER(x) ((x) >> 8)
#define PROP_EPSILON 0.0001

struct pst_data
{
	uns widget_type;
	char * description;
	union
	{
		struct { gdouble lower, upper, step, page; } spin;
		struct { uns max; string * strings; } combo;
		struct { uns dummy; } nothing;
		struct {
			void (*create_edit_func)( struct prop_item * pi );
			void (*update_edit_func)( struct prop_item * pi, struct prop * prop );
		} func;
	} data;
};

#define FN(_subtype,_what) _subtype##_edit_##_what
#define CREATE_F(_subtype)		static void FN(_subtype,create)( struct prop_item * pi )
#define UPDATE_F(_subtype,_u,_uu)	static void FN(_subtype,update)( struct prop_item * pi _u, struct prop * prop _uu )
#define CALLB_F(_subtype	)	static void FN(_subtype,callback)( GtkWidget * widget UNUSED, struct prop_item * pi )
#define GEN_FH(x) CREATE_F(x); UPDATE_F(x,,)
#define GEN_FN(x) FN(x,create), FN(x,update)

GEN_FH(color);
GEN_FH(font);
CALLB_F(color);
CALLB_F(font);
static struct pst_data pst_uns[PTU_MAX] =
{
	[PTU_UNSPECIFIED] = { PWT_SPIN_UNS, "unspecified", { .spin = { 0, REAL_MAX, 1.0, 10.0 }}},
	[PTU_COLOR] = { PWT_FUNC, "color", { .func = { GEN_FN(color) }}},
	[PTU_FONT] = { PWT_FUNC, "font", { .func = { GEN_FN(font) }}},
	[PTU_CONIC_TYPE_0P] = { PWT_COMBO, NULL, { .combo = { GEOM_CONIC_ARC_0P, string_conic_type }}},
	[PTU_CONIC_TYPE_1P] = { PWT_COMBO, NULL, { .combo = { GEOM_CONIC_ARC_1P, string_conic_type }}},
	[PTU_CONIC_TYPE_2P] = { PWT_COMBO, NULL, { .combo = { GEOM_CONIC_ARC_2P, string_conic_type }}},
	[PTU_CONIC_TYPE_3P] = { PWT_COMBO, NULL, { .combo = { GEOM_CONIC_ARC_3P, string_conic_type }}},
	[PTU_BOOLEAN] = { PWT_CHECKBOX, "boolean", { .nothing = {0}}},
	[PTU_CAP_STYLE] = { PWT_COMBO, NULL, { .combo = { PSC_MAX - 1, string_stroke_cap_type }}},
	[PTU_JOIN_STYLE] = { PWT_COMBO, NULL, { .combo = { PSJ_MAX - 1, string_stroke_join_type }}},
	[PTU_STROKE_STYLE] = { PWT_COMBO, NULL, { .combo = { PSS_MAX - 1, string_stroke_style_type }}},
	[PTU_ALIGNMENT_X] = { PWT_COMBO, NULL, { .combo = { PTAX_MAX - 1, string_alignment_x_type }}},
	[PTU_ALIGNMENT_Y] = { PWT_COMBO, NULL, { .combo = { PTAY_MAX - 1, string_alignment_y_type }}},
	[PTU_ARROW_FRONT] = { PWT_COMBO, NULL, { .combo = { PTAF_MAX - 1, string_arrow_front_type }}},
	[PTU_ARROW_BACK] = { PWT_COMBO, NULL, { .combo = { PTAB_MAX - 1, string_arrow_back_type }}},
	[PTU_ARROW_ALIGN] = { PWT_COMBO, NULL, { .combo = { PTAA_MAX - 1, string_arrow_align_type }}},
};

GEN_FH(ptr);
static struct pst_data pst_ptr[PTP_MAX] =
{
	[PTP_UNSPECIFIED] = { PWT_FUNC, NULL, { .func = { GEN_FN(ptr) }}},
	[PTP_TRANSFORM] = { PWT_FUNC, NULL, { .func = { GEN_FN(ptr) }}},
	[PTP_TEX_PROCESS] = { PWT_FUNC, NULL, { .func = { GEN_FN(ptr) }}},
};

GEN_FH(filename);
CALLB_F(filename);
GEN_FH(largetext);
CALLB_F(largetext);
static struct pst_data pst_string[PTS_MAX] =
{
	[PTS_UNSPECIFIED] = { PWT_ENTRY, "unspecified", { .nothing = {0}}},
	[PTS_FILE_NAME] = { PWT_FUNC, "file name", { .func = { GEN_FN(filename) }}},
	[PTS_LARGE_TEXT] = { PWT_FUNC, "large text", { .func = { GEN_FN(largetext) }}},
};

static struct pst_data pst_real[PTR_MAX] =
{
	[PTR_UNSPECIFIED] = { PWT_SPIN_REAL, "unspecified", { .spin = { -REAL_MAX, REAL_MAX, 1.0, 10.0 }}},
	[PTR_POSITIVE] = { PWT_SPIN_REAL, "positive", { .spin = { 0.0001, REAL_MAX, 1.0, 10.0 }}},
	[PTR_COORDINATE] = { PWT_SPIN_REAL, "coordinate", { .spin = { -REAL_MAX, REAL_MAX, 1.0, 10.0 }}},
	[PTR_LENGTH] = { PWT_SPIN_REAL, "length", { .spin = { 0.0, REAL_MAX, 1.0, 10.0 }}},
	[PTR_ANGLE_PI] = { PWT_SPIN_REAL, "angle 0 ... Pi", { .spin = { 0.0, R_PI, R_PI/12, R_PI/4 }}},
	[PTR_ANGLE_2PI] = { PWT_SPIN_REAL, "angle 0 ... 2Pi", { .spin = { 0.0, R_2PI, R_PI/12, R_PI/4 }}},
	[PTR_ANGLE_4PI] = { PWT_SPIN_REAL, "angle -2Pi ... 2Pi", { .spin = { -R_2PI, R_2PI, R_PI/12, R_PI/4 }}},
	[PTR_REFERENCE] = { PWT_SPIN_REAL, "reference", { .spin = { 0.0, 1.0, 0.1, 0.2 }}}
};

static struct pst_data * pst_info[PT_MAX]  = { pst_uns, pst_ptr, pst_string, pst_real };

static GtkListStore * ptu_list, * pts_list, * ptr_list;

GSList * prop_window_list;

/** The one context window, if exists */
struct prop_window * context_property_window = NULL;

static void prop_window_destroy( GtkWidget * w UNUSED, struct prop_window * pw );
static void pw_empty( struct prop_window * pw );
static void pw_fill( struct prop_window * pw, struct o * o );
static void pw_refill( struct prop_window * pw );
static inline void pw_maybe_refill( struct o * o, struct prop_window * pw );
static void pw_unlink( struct prop_window * pw, struct o * o );
static void pw_merge( struct prop_window * pw, struct go * go );
static void pw_unmerge( struct prop_window * pw, struct go * go );
static void prop_value_changed( GtkWidget * w, struct prop_item * pi );
static void prop_unit_changed( GtkWidget * w, struct prop_item * pi );
static void prop_sync_prop( struct prop_item * pi, struct prop * prop );

static void fill_prop_type_list( GtkListStore ** list, struct pst_data * data, int cnt );
static void add_new_property( GtkWidget * w UNUSED, struct prop_window * pw );
static void delete_property( GtkWidget * w UNUSED, struct prop_window * pw );
static void manage_units( GtkWidget * w UNUSED, struct prop_window * pw );

static void prop_go_unlink(struct o *o, struct o * owner UNUSED, struct o * prev UNUSED, struct go_hook * h);
static void prop_go_create_prop(struct o *o, struct prop* p, struct go_hook * h);
static void prop_go_change_prop(struct o *o, struct go_hook * h);
static void prop_go_selected(struct o *o, struct go_hook * h);
static void prop_go_unselected(struct o *o, struct go_hook * h);
static void prop_go_delete_prop(struct o *o, struct prop* p, struct go_hook * h);
static void prop_obj_unlink(struct o *o, struct o * owner UNUSED, struct o * prev UNUSED, struct obj_hook * h);
static void prop_obj_create_prop(struct o *o, struct prop* p, struct obj_hook * h);
static void prop_obj_change_prop(struct o *o, struct obj_hook * h);
static void prop_obj_delete_prop(struct o *o, struct prop* p, struct obj_hook * h);
static void prop_unit_event(uns u, struct unit_hook *h);
static gboolean entry_focus_out( GtkWidget * w, GdkEventFocus * event UNUSED, struct prop_item * pi );
static gboolean prop_label_pressed( GtkWidget * w, GdkEventButton *event, struct prop_item * pi );
static void gui_prop_remember( struct o * o, struct prop * prop );

#define PROP_MULT ((prop_subtype2quantity[prop->type][prop->subtype] == PQ_NONE) ? 1.0 : prop_get_unit_multiplier(pi->o,prop))

#define REF(_name, _string) string_##_name = string_lookup_ref(_string)

static string string_cm, string_m, string_pi, string_inch, string_pt;
string GUI_STRINGS;

static int changing_something;
PROP_STORE_DEFINE(ps_global);
PROP_STORE_DEFINE(ps_recycler);

void prop_store_set( PROP_STORE_DEFINE(ps), const char * name, uns type, prop_value pv )
{
	char buf[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( PROP_STORE_TLO(ps), "Store property", buf )
		prop_create( PROP_STORE_O(ps), string_lookup(name), type, 0, PROP_UNIT_MAX, pv );
	TRANS_FAILED
		DBG("Could not store value %s\n", name);
	TRANS_END
}

static void set_default_setting( string key, uns type, uns subtype, prop_value pv )
{
	struct o * o = obj2o(obj_universum);
	struct prop * prop = prop_find(o, key );
	if ( prop && (prop->type != type) )
	{
		prop_delete(o, prop);
		prop = NULL;
	}
	if ( !prop )
		prop_create(o, key, type, subtype, PROP_UNIT_MAX, pv );
}

void gui_prop_init(void)
{
	REF(tolerance, "snap-tolerance");
	REF(editor, "text-editor");
	REF(editor_in_term, "run-text-editor-in-terminal");
	REF(browser, "www-browser");
	REF(browser_in_term, "run-www-browser-in-terminal");
	REF(terminal, "terminal");
	REF(grid_x1, "grid-x1");
	REF(grid_y1, "grid-y1");
	REF(grid_x2, "grid-x2");
	REF(grid_y2, "grid-y2");
	REF(grid_color, "grid-color");
	REF(rulerx_res, "ruler-x-res");
	REF(rulery_res, "ruler-y-res");
	REF(drag_image, "drag-image");
	REF(enable_fifi, "enable-fifi");

	char buf[TRANS_ERR_SIZE];
	char buf2[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( tlo_universum, "Initialize properties", buf )
		//pokusne pridani jednotek
		//TODO: ukladat nekam do nastaveni
		unit_set( REF(cm, "cm"), 10.0, PQ_LENGTH );
		unit_set( REF(m, "m"), 1000.0, PQ_LENGTH );
		unit_set( REF(inch, "inch"), 25.4, PQ_LENGTH );
		unit_set( REF(pt, "pt"), 25.4/72, PQ_LENGTH );
		unit_set( REF(pi, "Pi"), R_PI, PQ_ANGLE );

		PROP_STORE_NEW(ps_recycler);
		PROP_STORE_NEW(ps_global);
		TRANS_BEGIN_MAIN( PROP_STORE_TLO(ps_recycler), "Initialize recycler", buf2 )
			// FIXME: osklivy hack do doby, nez budou hotove styly
			struct o * o = go2o(&go_path_new()->group.go);
			struct prop * prop;
			WALK_PROP(o, prop)
				gui_prop_remember( o, prop );
		TRANS_FAILED
			DBG("Error initializing recycler: %s\n", buf2);
		TRANS_END

		//set defaults for unloaded settings
		set_default_setting(string_tolerance, PT_UNS, PTU_UNSPECIFIED, (prop_value){ .u = TOLERANCE_PIX });
		set_default_setting(string_editor, PT_STRING, PTS_UNSPECIFIED, (prop_value){ .s = string_lookup(EXTERNAL_TEXT_EDITOR) });
		set_default_setting(string_editor_in_term, PT_UNS, PTU_BOOLEAN, (prop_value){ .u = TEXT_EDITOR_RUN_IN_TERM });
		set_default_setting(string_browser, PT_STRING, PTS_UNSPECIFIED, (prop_value){ .s = string_lookup(WWW_BROWSER) });
		set_default_setting(string_browser_in_term, PT_UNS, PTU_BOOLEAN, (prop_value){ .u = WWW_BROWSER_RUN_IN_TERM });
		set_default_setting(string_terminal, PT_STRING, PTS_UNSPECIFIED, (prop_value){ .s = string_lookup(FAVOURITE_TERMINAL) });
		set_default_setting(string_grid_x1, PT_REAL, PTR_COORDINATE, (prop_value){ .r = GRID_X1 });
		set_default_setting(string_grid_y1, PT_REAL, PTR_COORDINATE, (prop_value){ .r = GRID_Y1 });
		set_default_setting(string_grid_x2, PT_REAL, PTR_COORDINATE, (prop_value){ .r = GRID_X2 });
		set_default_setting(string_grid_y2, PT_REAL, PTR_COORDINATE, (prop_value){ .r = GRID_Y2 });
		set_default_setting(string_grid_color, PT_UNS, PTU_COLOR, (prop_value){ .u = GRID_COLOR });
		set_default_setting(string_rulerx_res, PT_REAL, PTR_COORDINATE, (prop_value){ .r = RULERX_RES });
		set_default_setting(string_rulery_res, PT_REAL, PTR_COORDINATE, (prop_value){ .r = RULERY_RES });
		set_default_setting(string_drag_image, PT_UNS, PTU_BOOLEAN, (prop_value){ .u = PANNING_DRAG_IMAGE });
		set_default_setting(string_enable_fifi, PT_UNS, PTU_BOOLEAN, (prop_value){ .u = ENABLE_FIFI });

	TRANS_FAILED
	TRANS_END

	create_unit_quant_lists();
	fill_prop_type_list( &ptu_list, pst_uns, PTU_MAX );
	fill_prop_type_list( &pts_list, pst_string, PTS_MAX );
	fill_prop_type_list( &ptr_list, pst_real, PTR_MAX );
}

static void fill_prop_type_list( GtkListStore ** list, struct pst_data * data, int cnt )
{
	*list = GTK_LIST_STORE(gtk_list_store_new( 2, G_TYPE_STRING, G_TYPE_INT ));
	for (int index = 0; index < cnt; index++)
	{
		GtkTreeIter ti;
		if (data[index].description)
		{
			gtk_list_store_append( *list, &ti );
			gtk_list_store_set( *list, &ti, 0, data[index].description, 1, index, -1 );
		}
	}
}

void gui_prop_finish(void)
{
	while ( prop_window_list )
		gtk_widget_destroy( ((struct prop_window*)prop_window_list->data)->window );
	if ( context_property_window )
		gtk_widget_destroy(context_property_window->window);

	//TODO: presunout pokusne jednotky na lepsi misto
	string_unref(string_cm);
	string_unref(string_m);
	string_unref(string_pi);
	string_unref(string_inch);
	string_unref(string_pt);

	string array[] = { GUI_STRINGS };
	for(int i = sizeof(array) / sizeof(string) - 1; i >= 0; i--)
		string_unref(array[i]);

	PROP_STORE_DESTROY(ps_recycler);
	PROP_STORE_DESTROY(ps_global);

	destroy_unit_quant_lists();
	g_object_unref( G_OBJECT(ptu_list) );
	g_object_unref( G_OBJECT(pts_list) );
	g_object_unref( G_OBJECT(ptr_list) );
}

static inline int prop_is_recyclable( struct o * o, struct prop * prop )
{
	return (prop->flags & PTF_RECYCLABLE) &&
		(prop->key != string_fill_color || o->kind != T_GO ||
		(o->type != GOT_TEXT && o->type != GOT_TEX_TEXT));
}

void gui_prop_recycle( struct o * o )
{
	DBG("Recycle properties for o %p\n", o );
	struct prop * volatile prop;
	struct prop * rec_prop;
	struct o * or = PROP_STORE_O(ps_recycler);
	uns unit;
	char buf[TRANS_ERR_SIZE];
	WALK_PROP(o, prop)
	// check if the ps_recycler has the same property (including the same type and subtype)
	// if so, change the o's property accordingly
		if ( prop_is_recyclable(o, prop) && (rec_prop = prop_find(or, prop->key)) &&
			rec_prop->type == prop->type && rec_prop->subtype == prop->subtype )
		{
			// the change can fail for various reasons (e.g. if the value is invalid for this o)
			DBG("Recycle property %s\n", prop->key->text );
			TRANS_BEGIN(buf)
				prop_set( o, prop->key, prop_get_value(or, rec_prop) );
				unit = prop_get_unit( or, rec_prop );
				if (unit != PROP_UNIT_MAX) prop_set_unit( o, prop, unit );
			TRANS_FAILED
				DBG("Could not recycle property %s: %s\n", prop->key->text, buf);
			TRANS_END
		}
}

void gui_prop_recycler_set( string key, uns type, uns subtype, uns unit, prop_value val )
{
	struct prop * rec_prop;
	struct o * or = PROP_STORE_O(ps_recycler);
	char buf[TRANS_ERR_SIZE];
	// we want to rewrite the old property and delete it if incompatible
	TRANS_BEGIN_MAIN( PROP_STORE_TLO(ps_recycler), "Remember the property for further use", buf )
		if ( (rec_prop = prop_find(or, key)) && (rec_prop->type != type || rec_prop->subtype != subtype) )
		{
			DBG("Property %s remembered with different type/subtype, delete\n", key->text);
			prop_delete( or, rec_prop );
		}
		DBG("Remember property %s\n", key->text );
		prop_create( or, key, type, subtype, unit, val);
	TRANS_FAILED
		DBG("Could not remember property %s: %s\n", key->text, buf);
	TRANS_END
}

static void gui_prop_remember( struct o * o, struct prop * prop )
{
	if (!prop_is_recyclable(o,prop)) return;
	gui_prop_recycler_set( prop->key, prop->type, prop->subtype, prop_get_unit(o, prop), prop_get_value(o, prop));
}

void prop_window_create( struct o * o )
{
	if ( !o && context_property_window )
	{
		DBG("Context property window already exists, will not be created.\n");
		gtk_window_present( GTK_WINDOW(context_property_window->window) );
		return;
	}

	struct prop_window * pw = xmalloc_zero(sizeof(struct prop_window));
	win_o_init( (struct window*)pw, WT_PROPERTIES );
	pw->flags = 0;
	pw->items = NULL;
	if ( !o )
	{ pw->flags |= PW_CONTEXT; context_property_window = pw; }

	pw->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	GtkWidget * vbox = gtk_vbox_new( FALSE, 5 );
	gtk_container_add( GTK_CONTAINER(pw->window), vbox );
	g_signal_connect( G_OBJECT(pw->window), "destroy", G_CALLBACK(prop_window_destroy), pw );

	pw->table = GTK_TABLE(gtk_table_new(1, 3, FALSE));
	gtk_container_set_border_width( GTK_CONTAINER(pw->table), 5 );
	gtk_box_pack_start( GTK_BOX(vbox), GTK_WIDGET(pw->table), TRUE, TRUE, 5 );

	GtkWidget * hbutbox = gtk_hbutton_box_new();
	gtk_button_box_set_layout( GTK_BUTTON_BOX(hbutbox), GTK_BUTTONBOX_SPREAD );
	gtk_box_pack_start( GTK_BOX(vbox), hbutbox, FALSE, TRUE, 0 );

	GtkWidget * button;
	gtk_box_pack_start( GTK_BOX(hbutbox), pw->button_add = gtk_button_new_with_label( "Add ..." ), FALSE, FALSE, 0 );
	g_signal_connect( G_OBJECT(pw->button_add), "clicked", G_CALLBACK(add_new_property), pw );
	gtk_box_pack_start( GTK_BOX(hbutbox), pw->button_delete = gtk_button_new_with_label( "Delete" ), FALSE, FALSE, 0 );
	g_signal_connect( G_OBJECT(pw->button_delete), "clicked", G_CALLBACK(delete_property), pw );
	gtk_box_pack_start( GTK_BOX(hbutbox), button = gtk_button_new_with_label( "Units ..." ), FALSE, FALSE, 0 );
	g_signal_connect( G_OBJECT(button), "clicked", G_CALLBACK(manage_units), pw );

	gtk_box_pack_start( GTK_BOX(vbox), pw->statusbar, FALSE, FALSE, 0 );

	prop_window_list = g_slist_prepend( prop_window_list, (gpointer)pw );
	pw_fill( pw, o ? o : NULL );

	if ( pw->flags & PW_CONTEXT )
	{
		if ( context.tlo )
		{
			BASIC_DBG("PW: add a hook for the context.tlo\n");
			go_hook_add(pw->tlo = context.tlo, &pw->go_hook);
		} else
			pw->tlo = NULL;
	}

	gtk_widget_show_all(pw->window);

	pw->go_hook.data = pw;
	pw->go_hook.go_unlink = prop_go_unlink;
	pw->go_hook.go_prop_create = prop_go_create_prop;
	pw->go_hook.go_prop_delete = prop_go_delete_prop;
	pw->go_hook.go_prop_change = prop_go_change_prop;
	pw->go_hook.go_selected = prop_go_selected;
	pw->go_hook.go_unselected = prop_go_unselected;

	pw->obj_hook.data = pw;
	pw->obj_hook.obj_unlink = prop_obj_unlink;
	pw->obj_hook.obj_prop_create = prop_obj_create_prop;
	pw->obj_hook.obj_prop_delete = prop_obj_delete_prop;
	pw->obj_hook.obj_prop_change = prop_obj_change_prop;

	pw->unit_hook.data = pw;
	pw->unit_hook.add = prop_unit_event;
	pw->unit_hook.deleted = prop_unit_event;
	pw->unit_hook.changed = prop_unit_event;
	pw->unit_hook.def = prop_unit_event;
	unit_hook_add(&pw->unit_hook);
}

static void prop_window_destroy( GtkWidget * w UNUSED, struct prop_window * pw )
{
	pw_empty(pw);
	if ( pw->flags & PW_CONTEXT ) 
	{
		context_property_window = NULL;
		if (pw->tlo) go_hook_remove(pw->tlo, &pw->go_hook);
		refresh_gui();
	}
	prop_window_list = g_slist_remove( prop_window_list, (gpointer)pw );
	unit_hook_remove(&pw->unit_hook);
	win_o_finish((struct window*)pw);
	xfree(pw);
}
static void prop_widget_destroy( GtkWidget * w, gpointer data UNUSED ){ gtk_widget_destroy(w); }
static void pw_empty( struct prop_window * pw )
{
	ASSERT(pw);
	BASIC_DBG("Removing property obj hook\n");
	obj_hook_remove(&pw->obj_hook);

	if (!(pw->flags & PW_NONEMPTY )) return;

	//destroy all prop_items of the widgets
	gtk_container_foreach( GTK_CONTAINER(pw->table), prop_widget_destroy, NULL );
	if ( !(pw->flags & PW_CONTEXT) )
	{
		if ( pw->tlo )
		{
			BASIC_DBG("Removing property go hook from tlo %p\n", pw->tlo);
			go_hook_remove(pw->tlo, &pw->go_hook);
			pw->tlo = NULL;
		}
	}
	gtk_table_resize( GTK_TABLE(pw->table), 1, 3 );
	pw->flags &= ~PW_NONEMPTY;
	for ( uns i = 0; i < pw->prop_count; i++ )
		if ( pw->items[i].ptr )
		{
			BASIC_DBG("Freeing ptr for item %u\n", i);
			xfree( pw->items[i].ptr );
		}
	xfree(pw->items);
	pw->items = NULL;
	pw->o = NULL;
}

static void pw_set_window_title( struct prop_window * pw )
{
	if ( pw->flags & PW_NONEMPTY )
	{
		if ( pw->flags & PW_MERGED )
			gtk_window_set_title( GTK_WINDOW(pw->window), "Properties of selection" );
		else
		{
			string s = o_get_name(pw->o);
			char * name = (s == string_noname) ? "" : s->text;
			const char * tn = ((pw->o->kind == T_OBJ) ? string_obj_type : string_go_type)[pw->o->type]->text;
			char buf[strlen(name) + strlen(tn) + 25];
			sprintf( buf, "Properties of %s %s", tn, name );
			gtk_window_set_title( GTK_WINDOW(pw->window), buf );
		}
	} else gtk_window_set_title( GTK_WINDOW(pw->window), "Properties" );
}

/** Go through the filled prop_items in the prop_window and seek them
 * in the o. Remove all prop_items whose properties have not been found.
 * \return 1 iff there was a removed property. */
static int pw_merge_prop_items( struct prop_window * pw, struct o * o )
{
	struct prop_item * pi;
	struct prop * prop;
	uns next_pos = 0, cur_pos = 0;
	for ( ; cur_pos < pw->prop_count; cur_pos++ )
	{
		pi = pw->items + cur_pos;
		if ( (prop = prop_find(o, pi->key)) && prop->type == pi->type && prop->subtype == pi->subtype )
		{
			if (next_pos < cur_pos)
				pw->items[next_pos] = pw->items[cur_pos];
			next_pos++;
		}
	}
	pw->prop_count = next_pos;
	return next_pos < cur_pos;
}

/** Fills in the prop_item array with the intersection of properties for all displayed
 * gos. Supposes that pw->o is the one displayed o (if the PW_MERGED flag is unset)
 * or the first selected go (if the PW_MERGED flag is set). */
static void pw_fill_prop_items( struct prop_window * pw )
{
	ASSERT(pw && pw->o);
	pw->prop_count = prop_get_count(pw->o);
	//allocate the number of properties of the first object, which will not exceed
	//the total number of properties in the intersection
	pw->items = xmalloc_zero(sizeof(struct prop_item) * pw->prop_count);

	// fill the properties of the first o first
	struct prop * prop;
	struct prop_item * pi;
	uns counter = 0;
	WALK_PROP(pw->o, prop)
	{
		ASSERT(prop && prop->key);
		pi = pw->items + counter++;
		pi->o = pw->o;
		pi->key = prop->key;
		pi->type = prop->type;
		pi->subtype = prop->subtype;
		pi->pw = (struct window*)pw;
	}

	if ( !(pw->flags & PW_MERGED) )
		return;

	struct go * go;
	ASSERT(context.group && (go = group_get_first_selected_go(context.group)));
	for ( go = group_get_next_selected_go(context.group, group_get_first_selected_go(context.group));
		go; go = group_get_next_selected_go(context.group, go) )
		pw_merge_prop_items(pw, go2o(go));
}

/** Fills the window with properties.
 * If the o is non-NULL, it fills the window with its properties.
 * If it is NULL, it uses the selected objects for the current context.group,
 * or the context.o, if there is no selection.
 * Otherwise, displays the "no object to display" message. */
static void pw_fill( struct prop_window * pw, struct o * o )
{
	ASSERT(pw);
	pw->o = o;
	// determine which objects to display
	pw->flags &= ~(PW_NONEMPTY | PW_SELECTION | PW_MERGED );
	if ( !pw->o )
	{
		ASSERT( pw->flags & PW_CONTEXT );
		if ( context.group && (pw->o = (struct o*)group_get_first_selected_go(context.group)) )
		{
			pw->flags |= PW_SELECTION | PW_NONEMPTY;
			if ( group_get_next_selected_go( context.group, o2go(pw->o) ))
				pw->flags |= PW_MERGED;
		}
		else 
			if ( context.o )
			{
				pw->o = context.o;
				pw->flags |= PW_NONEMPTY;
			}
	}
	else pw->flags |= PW_NONEMPTY;

	pw_set_window_title(pw);

	BASIC_DBG("Adding a property obj hook\n");
	obj_hook_add(&pw->obj_hook);
	if ( !(pw->flags & PW_CONTEXT) )
	{
		ASSERT(pw->o);
		if ( o->kind == T_GO )
		{
			go_hook_add(pw->tlo = go_get_tlo(o2go(o)), &pw->go_hook);
			BASIC_DBG("Adding a property go hook to tlo %p\n", pw->tlo);
		} else
			pw->tlo = NULL;
	}

	if ( !pw->o )
	{
		pw->prop_count = 0;
		GtkWidget * label = gtk_label_new("no object to display");
		gtk_widget_set_size_request( label, -1, 100 );
		gtk_table_attach( GTK_TABLE(pw->table), label, 0, 3, 0, 1,
			GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0 );
		gtk_widget_set_sensitive( pw->button_add, FALSE );
		gtk_widget_set_sensitive( pw->button_delete, FALSE );
		gtk_widget_show_all( GTK_WIDGET(pw->table) );
	} else {
		gtk_widget_set_sensitive( pw->button_add, TRUE );
		gtk_widget_set_sensitive( pw->button_delete, TRUE );
		pw_fill_prop_items(pw);
		pw_refill(pw);
	}
}

/** Removes all property widgets from the window prop_window and creates them again.
 * Does not recalculate/refill the inner struktures (prop_item array etc.) */
static void pw_refill( struct prop_window * pw )
{
	ASSERT(pw);
	gtk_container_foreach( GTK_CONTAINER(pw->table), prop_widget_destroy, NULL );
	gtk_table_resize( GTK_TABLE(pw->table), pw->prop_count, 3 );

	// fill in the table with displayed properties
	struct prop_item * pi;
	for (uns counter = 0; counter < pw->prop_count; counter++ )
	{
		pi = pw->items + counter;
		prop_value_edit_create( pi );
		prop_unit_edit_create( pi );
		prop_sync( pi );

		GtkWidget * label = gtk_label_new(pi->key->text);
		GtkWidget * eb = gtk_event_box_new();
		gtk_container_add( GTK_CONTAINER(eb), label );
		g_object_set( G_OBJECT(label), "xalign", 0.0, NULL );
		gtk_widget_set_events( eb, GDK_BUTTON_PRESS_MASK );
		g_signal_connect( G_OBJECT(eb), "button-press-event", G_CALLBACK(prop_label_pressed), pi );
		gtk_table_attach( pw->table, eb, 0, 1, counter, counter+1, GTK_FILL, GTK_FILL, 5, 0 );
		if (pi->unit_edit)
		{
			gtk_table_attach( pw->table, pi->value_edit, 1, 2, counter, counter+1, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0 );
			gtk_table_attach( pw->table, pi->unit_edit, 2, 3, counter, counter+1, GTK_FILL, GTK_FILL, 0, 0 );
		} else
		gtk_table_attach( pw->table, pi->value_edit, 1, 3, counter, counter+1, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0 );
	}
	gtk_widget_show_all( GTK_WIDGET(pw->table) );
}

void prop_item_init( struct prop_item * pi, struct o * o, string key )
{
	struct prop * prop = prop_find( o, key );
	ASSERT(prop);
	pi->o = o;
	pi->key = key;
	pi->type = prop->type;
	pi->subtype = prop->subtype;
	pi->pw = NULL;
	prop_value_edit_create( pi );
	prop_unit_edit_create( pi );
	prop_sync( pi );
}

void prop_value_edit_create( struct prop_item * pi )
{
	ASSERT( pi );
	struct pst_data data = pst_info[pi->type][pi->subtype];

	switch( data.widget_type )
	{
	case PWT_BUG: BUG(); break;
	case PWT_SPIN_UNS:
		pi->value_edit = gtk_spin_button_new_with_range( data.data.spin.lower, data.data.spin.upper, data.data.spin.step );
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(pi->value_edit), 0.0 );
		gtk_spin_button_set_digits(GTK_SPIN_BUTTON(pi->value_edit), 0 );
		//g_object_set( G_OBJECT(pi->value_edit), "xalign", 1.0, NULL );
		g_signal_connect( G_OBJECT(pi->value_edit), "value-changed", G_CALLBACK(prop_value_changed), pi);
		break;
	case PWT_SPIN_REAL:
		pi->value_edit = gtk_spin_button_new_with_range( data.data.spin.lower, data.data.spin.upper, data.data.spin.step );
		gtk_spin_button_set_digits(GTK_SPIN_BUTTON(pi->value_edit), 4 );
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(pi->value_edit), 0.0 );
		//g_object_set( G_OBJECT(pi->value_edit), "xalign", 1.0, NULL );
		g_signal_connect( G_OBJECT(pi->value_edit), "value-changed", G_CALLBACK(prop_value_changed), pi);
		break;
	case PWT_COMBO:
		pi->value_edit = gtk_combo_box_new_text();
		for ( uns i = 0; i <= data.data.combo.max; i++ )
			gtk_combo_box_append_text( GTK_COMBO_BOX(pi->value_edit), data.data.combo.strings[i]->text );
		g_signal_connect( G_OBJECT(pi->value_edit), "changed", G_CALLBACK(prop_value_changed), pi );
		break;
	case PWT_CHECKBOX:
		pi->value_edit = gtk_check_button_new();
		g_signal_connect( G_OBJECT(pi->value_edit), "toggled", G_CALLBACK(prop_value_changed), pi);
		break;
	case PWT_ENTRY:
		pi->value_edit = gtk_entry_new();
		g_signal_connect(G_OBJECT(pi->value_edit), "focus-out-event", G_CALLBACK(entry_focus_out), pi);
		g_signal_connect(G_OBJECT(pi->value_edit), "activate", G_CALLBACK(prop_value_changed), pi);
		break;
	case PWT_FUNC:
		data.data.func.create_edit_func( pi );
		break;
	}
	ASSERT(pi->value_edit);
}

static gdouble tame_number( real x, real mult, int * was_bad )
{
	DBG("Taming number %f with mult %f ", x, mult );
	x *= mult;
	if (isinf(x)) { x = signbit(x) ? -REAL_MAX : REAL_MAX; if (was_bad) *was_bad = 1; }
	else { if (was_bad) *was_bad = 0; }
	DBG(" = %f\n", (gdouble)x );
	return (gdouble)x; //FIXME: staci to takhle?
}

static real untame_number( gdouble y, real mult, int * was_bad )
{
	DBG("Untaming number %f with mult %f ", y, mult );
	real x = (real)y * mult;
	if (isinf(x)) { x = signbit(x) ? -REAL_MAX : REAL_MAX; if (was_bad) *was_bad = 1; }
	else { if (was_bad) *was_bad = 0; }
	DBG(" = %f\n", x );
	return x;
}

#define TRANS_PROP_CHANGE(_val) \
{ \
	char buf[TRANS_ERR_SIZE]; \
	char transname[strlen(pi->key->text) + 25]; \
	sprintf( transname, "Change %s", pi->key->text ); \
	if (!(pi->flags & PIF_FACTORY)) factory_op_break(); \
	TRANS_BEGIN_MAIN( pi->o->kind == T_OBJ ? tlo_universum : go_get_tlo(o2go(pi->o)), transname, buf) \
		BASIC_DBG("\n*** Changing the value of property %s\n", pi->key->text ); \
		if ( pi->pw && (((struct prop_window*)pi->pw)->flags & PW_MERGED) ) \
		{ \
			ASSERT(context.group); \
			BASIC_DBG("\n*** Changing the property of many selected gos\n"); \
			struct go * one_go; \
			WALK_GROUP_SELECTED_GO(context.group, one_go) \
				prop_set( go2o(one_go), pi->key, _val ); \
		} \
		else prop_set( pi->o, pi->key, _val ); \
		gui_prop_remember( pi->o, prop ); \
	TRANS_FAILED \
		statusbar_err( pi->pw, "Could not change %s: %s", pi->key->text, buf); \
		prop_sync_prop(pi, prop); /*return to the original value*/ \
	TRANS_END \
}

#define FIND_PROP \
	ASSERT( pi && pi->o && pi->key && pi->value_edit ); \
	struct prop *prop; \
	if (!(prop = prop_find( pi->o, pi->key ))) \
	{ statusbar_err( pi->pw, "Property which should be changed does not exist."); return; }

static void prop_value_changed( GtkWidget * w UNUSED, struct prop_item * pi )
{
	if (changing_something) return; // prevent unwanted changes when the original value was clamped due to range change
	FIND_PROP;
	struct pst_data data = pst_info[prop->type][prop->subtype];
	volatile prop_value new_value;
	volatile string s = NULL;

	switch ( data.widget_type )
	{
	case PWT_BUG: BUG(); break;
	case PWT_SPIN_UNS:
		new_value.u = (uns)gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(pi->value_edit));
		break;
	case PWT_SPIN_REAL:
	{
		int bad;
		new_value.r = untame_number(gtk_spin_button_get_value(GTK_SPIN_BUTTON(pi->value_edit)), PROP_MULT, &bad );
		if (bad) statusbar_err( pi->pw, "Too big value in property %s, changed to 0.", pi->key->text);
		break;
	}
	case PWT_COMBO:
	{
		int active = gtk_combo_box_get_active(GTK_COMBO_BOX(pi->value_edit));
		if (active == -1) new_value.u = prop_get_value(pi->o, prop).u;
		else new_value.u = (uns)active;
		break;
	}
	case PWT_CHECKBOX:
		new_value.u = (uns)gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pi->value_edit));
		break;
	case PWT_ENTRY:
		s = new_value.s = string_lookup_ref(gtk_entry_get_text(GTK_ENTRY(pi->value_edit)));
		break;
	default: BUG();
	}

	real diff;
	if ( (prop->type == PT_REAL &&
			(diff = fabs(new_value.r - prop_get_value(pi->o,prop).r)) > untame_number( PROP_EPSILON, PROP_MULT, NULL)) ||
		(prop->type != PT_REAL && !prop_value_compare( prop->type, prop_get_value(pi->o, prop), new_value )))
	{
		if (prop->type == PT_REAL) { DBG("Difference is %f\n", diff); }
		TRANS_PROP_CHANGE(new_value);
	}

	if (s) string_unref(s);
}

void prop_sync( struct prop_item * pi )
{ ASSERT( pi && pi->o && pi->key ); prop_sync_prop( pi, prop_find(pi->o, pi->key) ); }

static void prop_sync_prop( struct prop_item * pi, struct prop * prop )
{
	ASSERT( pi && prop && pi->o && pi->key && pi->key == prop->key && pi->value_edit );
	//I suppose that the count of properties and the order of the list has not
	//changed without letting me know
	ASSERT( pi->type == prop->type && pi->subtype == prop->subtype );

	struct pst_data data = pst_info[prop->type][prop->subtype];
	prop_value new_value = prop_get_value( pi->o, prop );

	switch ( data.widget_type )
	{
	case PWT_BUG: BUG(); break;
	case PWT_SPIN_UNS:
		if (new_value.u != (uns)gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(pi->value_edit)))
			gtk_spin_button_set_value(GTK_SPIN_BUTTON(pi->value_edit), (gdouble)new_value.u );
		break;
	case PWT_SPIN_REAL:
	{
		int bad;
		real mult = 1.0 / PROP_MULT;
		DBG("Property %s has unit %d, 1/mult = %f \n", pi->key->text, prop_get_unit(pi->o, prop), mult );
		gdouble x = tame_number(new_value.r, mult, &bad);
		if (bad) statusbar_err( pi->pw, "Too big value in property %s, changed to 0.", pi->key->text);
		gdouble diff = fabs(x - gtk_spin_button_get_value(GTK_SPIN_BUTTON(pi->value_edit)));
		DBG("Difference is %f\n", diff );
		if ( diff > PROP_EPSILON)
		{
			gtk_spin_button_set_value(GTK_SPIN_BUTTON(pi->value_edit), x );
			DBG("Setting the value of %s to %f\n", pi->key->text, x );
		}
		//set range
		//x = tame_number(data.data.spin.step, mult, NULL);
		x = data.data.spin.step;
		gdouble inc;
		gtk_spin_button_get_increments(GTK_SPIN_BUTTON(pi->value_edit), &inc, NULL);
		if ( fabs(x - inc) > PROP_EPSILON )
		{
			changing_something = 1;
			gtk_spin_button_set_range(GTK_SPIN_BUTTON(pi->value_edit),
				tame_number(data.data.spin.lower, mult, NULL), tame_number(data.data.spin.upper, mult, NULL) );
			gtk_spin_button_set_increments(GTK_SPIN_BUTTON(pi->value_edit),
				data.data.spin.step, data.data.spin.page );
			changing_something = 0;
		}
		break;
	}
	case PWT_COMBO:
		if ((int)new_value.u != gtk_combo_box_get_active(GTK_COMBO_BOX(pi->value_edit)))
			gtk_combo_box_set_active(GTK_COMBO_BOX(pi->value_edit), new_value.u);
		break;
	case PWT_CHECKBOX:
		if (new_value.u != (uns)gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pi->value_edit)))
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pi->value_edit), new_value.u > 0 );
		break;
	case PWT_ENTRY:
		if (strcmp(new_value.s->text, gtk_entry_get_text(GTK_ENTRY(pi->value_edit))))
			gtk_entry_set_text( GTK_ENTRY(pi->value_edit), new_value.s->text );
		break;
	case PWT_FUNC:
		data.data.func.update_edit_func(pi, prop);
		break;
	}
	gtk_widget_set_sensitive( pi->value_edit, !(prop->flags & PTF_READ_ONLY));
	if ( pi->unit_edit )
	{
		// find out the index of the unit in our combobox
		GtkTreeIter ti;
		uns q = prop_subtype2quantity[pi->type][pi->subtype];
		if (unit_to_tree_iter( GTK_TREE_MODEL(*(unit_quant_lists[q])), prop_get_unit(pi->o, prop), &ti))
			BUG();
		gtk_combo_box_set_active_iter( GTK_COMBO_BOX(pi->unit_edit), &ti );
	}
}

void prop_unit_edit_create( struct prop_item * pi )
{
	ASSERT(pi);
	uns quantity = prop_subtype2quantity[pi->type][pi->subtype];
	if (quantity == PQ_NONE)
	{ pi->unit_edit = NULL; return; }

	pi->unit_edit = gtk_combo_box_new_with_model(GTK_TREE_MODEL(*(unit_quant_lists[quantity])));
	GtkCellRenderer * cr = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start( GTK_CELL_LAYOUT(pi->unit_edit), cr, FALSE );
	gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(pi->unit_edit), cr, "text", 0 );

	//gtk_combo_box_set_active( GTK_COMBO_BOX(pi->unit_edit), prop_get_unit( pi->o, prop ) );
	g_signal_connect( G_OBJECT(pi->unit_edit), "changed", G_CALLBACK(prop_unit_changed), pi );
}

void prop_unit_changed( GtkWidget * w UNUSED, struct prop_item * pi )
{
	ASSERT( pi && pi->o && pi->key && pi->unit_edit );
	struct prop *prop;
	if (!(prop = prop_find( pi->o, pi->key )))
	{ statusbar_err( pi->pw, "Property whose unit should be changed does not exist."); return; }

	volatile uns unit = 0;
	uns q = prop_subtype2quantity[prop->type][prop->subtype];
	GtkTreeIter ti;
	if (!gtk_combo_box_get_active_iter(GTK_COMBO_BOX(pi->unit_edit), &ti ))
		BUG();
	if (tree_iter_to_unit( GTK_TREE_MODEL(*(unit_quant_lists[q])), &ti, (uns*)&unit ))
		BUG();

	if ( unit != prop_get_unit(pi->o, prop) )
	{
		if (!(pi->flags & PIF_FACTORY))
			factory_op_break();
		char buf[TRANS_ERR_SIZE];
		TRANS_BEGIN_MAIN( pi->o->kind == T_OBJ ? tlo_universum : go_get_tlo(o2go(pi->o)), "Change unit", buf)
			BASIC_DBG("\n*** Set unit of property %s to %d\n", pi->key->text, unit );
			prop_set_unit( pi->o, prop, unit );
			gui_prop_remember( pi->o, prop );
		TRANS_FAILED
			statusbar_err( pi->pw, "Cannot change property unit: %s", buf);
			prop_sync_prop(pi, prop); //return to the original value
		TRANS_END
		prop_sync_prop(pi, prop); //FIXME: nemelo by byt potreba, mel by se volat hook
	}
}

struct new_property
{
	GtkComboBox * type_combo, * subtype_combo;
	GtkWidget * value_box;
	struct prop_item pi;
};
static void add_new_property_prop_change( struct o *o, struct go_hook* h )
{
	if (changing_something)
		return;
	struct new_property * np = (struct new_property*)h->data;
	if (np->pi.o == o)
		prop_sync( &np->pi );
}
static void add_new_property_type_changed( GtkComboBox * type_combo, struct new_property * np )
{
	ASSERT(np);
	GtkListStore * ls[] = { ptu_list, pts_list, ptr_list };
	int i = gtk_combo_box_get_active(type_combo);
	if (i < 0 || i > 2)
		i = 0;
	gtk_combo_box_set_model( np->subtype_combo, GTK_TREE_MODEL(ls[i]) );
	gtk_combo_box_set_active( np->subtype_combo, 0 );
}
static void add_new_property_subtype_changed( GtkComboBox * subtype_combo, struct new_property * np )
{
	ASSERT(np);
	GtkListStore * ls = GTK_LIST_STORE(gtk_combo_box_get_model(subtype_combo));
	np->pi.type = (ls == ptu_list) ? PT_UNS : ((ls == ptr_list) ? PT_REAL : PT_STRING );
	//get subtype
	GtkTreeIter ti;
	if (!gtk_combo_box_get_active_iter(subtype_combo, &ti))
		return;
	SAFE_G_VALUE(st);
	gtk_tree_model_get_value( GTK_TREE_MODEL(ls), &ti, 1, &st );
	np->pi.subtype = (uns)(g_value_get_int(&st));
	g_value_unset(&st);

	if (np->pi.key)
		string_unref(np->pi.key);
	char buf[50];
	sprintf( buf, "%u-%u-new", np->pi.type, np->pi.subtype );
	np->pi.key = string_lookup_ref(buf);

	// get the old value, if there was any; otherwise, use the default value
	prop_value pv = (np->pi.type == PT_UNS) ? (prop_value){ .u = 0 } :
		((np->pi.type == PT_REAL) ? (prop_value){ .r = 0.0 } : (prop_value){ .s = string_lookup("") } );
	if (np->pi.type == PT_UNS && np->pi.subtype == PTU_FONT )
		pv.u = (uns)fonts_get_default();
	pv = prop_lookup( PROP_STORE_O(ps_recycler), np->pi.key, np->pi.type, pv );
	changing_something = 1;
	gui_prop_recycler_set( np->pi.key, np->pi.type, np->pi.subtype, PROP_UNIT_MAX, pv );
	changing_something = 0;

	gtk_container_foreach( GTK_CONTAINER(np->value_box), prop_widget_destroy, NULL );
	prop_value_edit_create(&np->pi);
	prop_unit_edit_create(&np->pi);
	// editing large text when adding the property looks exteremely pathological...
	if (np->pi.type != PT_STRING || np->pi.subtype != PTS_LARGE_TEXT)
		gtk_box_pack_start(GTK_BOX(np->value_box), np->pi.value_edit, TRUE, TRUE, 0 );
	if (np->pi.unit_edit)
		gtk_box_pack_start(GTK_BOX(np->value_box), np->pi.unit_edit, FALSE, TRUE, 0 );
	prop_sync( &np->pi );
	gtk_widget_show_all(np->value_box);
}
static void add_new_property( GtkWidget * w UNUSED, struct prop_window * pw )
{
	ASSERT(pw->o);
	GtkWidget * dialog = gtk_dialog_new_with_buttons( "Add new property", GTK_WINDOW(pw->window),
		GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
		GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
	gtk_dialog_set_default_response( GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT );
	
	GtkBox * main_vbox = GTK_BOX(gtk_vbox_new( FALSE, 5 ));
	gtk_container_add( GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), GTK_WIDGET(main_vbox) );

	struct new_property np;
	np.pi.o = PROP_STORE_O(ps_recycler);
	np.pi.key = NULL;
	np.pi.pw = NULL;
	GtkTable * table = GTK_TABLE(gtk_table_new( 4, 2, FALSE ));
	gtk_container_set_border_width( GTK_CONTAINER(table), 5 );
	gtk_box_pack_start( main_vbox, GTK_WIDGET(table), TRUE, TRUE, 5 );
	gtk_table_attach( table, gtk_label_new("Key:"), 0, 1, 0, 1, GTK_FILL, GTK_FILL, 5, 0 );
	GtkWidget * name_entry = gtk_entry_new();
	gtk_table_attach( table, name_entry, 1, 2, 0, 1, GTK_FILL, GTK_FILL, 0, 0 );
	gtk_table_attach( table, gtk_label_new("Type:"), 0, 1, 1, 2, GTK_FILL, GTK_FILL, 5, 0 );
	GtkWidget * type_combo = gtk_combo_box_new_text();
	np.type_combo = GTK_COMBO_BOX(type_combo);
	gtk_combo_box_append_text( np.type_combo, "integer" );
	gtk_combo_box_append_text( np.type_combo, "string" );
	gtk_combo_box_append_text( np.type_combo, "real" );
	gtk_combo_box_set_active( np.type_combo, 0 );
	gtk_table_attach( table, type_combo, 1, 2, 1, 2, GTK_FILL, GTK_FILL, 0, 0 );

	gtk_table_attach( table, gtk_label_new("Subtype:"), 0, 1, 2, 3, GTK_FILL, GTK_FILL, 5, 0 );
	GtkWidget * subtype_combo = gtk_combo_box_new_with_model(GTK_TREE_MODEL(ptu_list));
	np.subtype_combo = GTK_COMBO_BOX(subtype_combo);
	GtkCellRenderer * cr = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start( GTK_CELL_LAYOUT(subtype_combo), cr, FALSE );
	gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(subtype_combo), cr, "text", 0 );
	gtk_combo_box_set_active( GTK_COMBO_BOX(subtype_combo), 0 );
	gtk_table_attach( table, subtype_combo, 1, 2, 2, 3, GTK_FILL, GTK_FILL, 0, 0 );

	gtk_table_attach( table, gtk_label_new("Value:"), 0, 1, 3, 4, GTK_FILL, GTK_FILL, 5, 0 );
	gtk_table_attach( table, np.value_box = gtk_hbox_new( FALSE, 0 ), 1, 2, 3, 4, GTK_FILL, GTK_FILL, 0, 0 );

	g_signal_connect( G_OBJECT(type_combo), "changed", G_CALLBACK(add_new_property_type_changed), &np );
	g_signal_connect( G_OBJECT(subtype_combo), "changed", G_CALLBACK(add_new_property_subtype_changed), &np );
	add_new_property_subtype_changed( GTK_COMBO_BOX(subtype_combo), &np );

	struct go_hook gh;
	bzero(&gh, sizeof(struct go_hook));
	gh.data = &np;
	gh.go_prop_change = add_new_property_prop_change;
	go_hook_add( PROP_STORE_TLO(ps_recycler), &gh );

	gtk_widget_show_all(dialog);
	if ( gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT )
	{
		string key = string_lookup(gtk_entry_get_text(GTK_ENTRY(name_entry)));

		char errs[TRANS_ERR_SIZE];
		changing_something = 1;
		statusbar_msg( (struct window*)pw, " " );
		TRANS_BEGIN_MAIN( pw->tlo ? pw->tlo : tlo_universum, "Add property", errs )
			BASIC_DBG("*** Add property %s\n", key->text);
			struct prop * prop = prop_find( np.pi.o, np.pi.key );
			if (!prop)
				trans_fail("Internal error.");
			struct go * go;
			if ( pw->flags & PW_MERGED )
			{
				ASSERT(context.group);
				WALK_GROUP_SELECTED_GO(context.group, go)
					prop_create(go2o(go), key, prop->type, prop->subtype,
						prop_get_unit(np.pi.o, prop), prop_get_value(np.pi.o, prop));
			}
			else
				prop_create(pw->o, key, prop->type, prop->subtype,
					prop_get_unit(np.pi.o, prop), prop_get_value(np.pi.o, prop));
		TRANS_FAILED
			statusbar_err( (struct window*)pw, "Cannot add: %s", errs );
		TRANS_END
		changing_something = 0;
		pw_maybe_refill(pw->o, pw);
	}
	gtk_widget_destroy(dialog);
	if (np.pi.key)
		string_unref(np.pi.key);
	go_hook_remove( PROP_STORE_TLO(ps_recycler), &gh );
}
static void delete_property( GtkWidget * w UNUSED, struct prop_window * pw )
{
	ASSERT(pw && pw->o);
	ASSERT(!(pw->flags & PW_MERGED) || context.group);
	char errs[TRANS_ERR_SIZE];
	// I suppose that all merged gos have the same tlo (even the same group)
	changing_something = 1;
	statusbar_msg( (struct window*)pw, " " );
	TRANS_BEGIN_MAIN( pw->tlo ? pw->tlo : tlo_universum, "Delete properties", errs )
		for (uns u = 0; u < pw->prop_count; u++ )
		{
			struct prop_item * pi = pw->items + u;
			if ( pi->flags & PIF_SELECTED )
			{
				BASIC_DBG("*** Delete property %s\n", pi->key->text);
				struct go * go;
				if ( pw->flags & PW_MERGED )
					WALK_GROUP_SELECTED_GO(context.group, go)
						prop_find_delete(go2o(go), pi->key);
				else
					prop_find_delete(pw->o, pi->key);
			}
		}
	TRANS_FAILED
		statusbar_err( (struct window*)pw, "%s", errs );
	TRANS_END
	changing_something = 0;
	pw_maybe_refill(pw->o, pw);
}
static void manage_units( GtkWidget * w UNUSED, struct prop_window * pw UNUSED )
{
	if ( wnd_units.visible )
		gtk_window_present( GTK_WINDOW(wnd_units.window) );
	else
		on_show_unit_window(NULL);
}

/* CONTEXT AND SELECTION CHANGES HANDLING */

#define O_IS_IMPORTANT_FOR_SEL(_o) (o_is_selected(_o) && _o->kind == T_GO && go_get_group(o2go(_o)) == context.group)

#define O_IS_IMPORTANT(_o) ((pw->flags & PW_SELECTION) && O_IS_IMPORTANT_FOR_SEL(_o) || !(pw->flags & PW_SELECTION) && _o == pw->o )

void prop_window_context_changed(struct context * old_context)
{
	if ( !CPW ) return;
	if ( (CPW->flags & PW_SELECTION) && context.tlo != old_context->tlo || context.group != old_context->group ||
		!(CPW->flags & PW_SELECTION) && context.o != CPW->o)
	{
		BASIC_DBG("PW: Context changed\n");
		if ( context.tlo != old_context->tlo )
		{
			BASIC_DBG("PW: the context tlo changed, rehook\n");
			if ( CPW->tlo ) go_hook_remove(CPW->tlo, &CPW->go_hook);
			if ( CPW->tlo = context.tlo ) go_hook_add(CPW->tlo, &CPW->go_hook);
		}
		
		pw_empty( CPW );
		pw_fill( CPW, NULL );
	}
}
static void pw_unlink(struct prop_window * pw, struct o * o)
{
	if ( pw->flags & PW_SELECTION )
	{
		if ( O_IS_IMPORTANT_FOR_SEL(o) )
			pw_unmerge(pw, o2go(o));
	} else
		if (pw->o && o_is_ancestor(o, pw->o))
		{
			BASIC_DBG("PW: unlink\n");
			if ( pw->flags & PW_CONTEXT ) { pw_empty(pw); pw_fill(pw, NULL); }
			else gtk_widget_destroy(pw->window);
		}
}
static inline void pw_maybe_refill( struct o * o, struct prop_window * pw )
{ if ( !changing_something && O_IS_IMPORTANT(o) ) { BASIC_DBG("PW: refill with %p\n", o); pw_empty(pw); pw_fill(pw, o); } }

static void pw_sync(struct o *o, struct prop_window * pw)
{
	if ( o == pw->o )
	{
		BASIC_DBG("Syncing the property window with o %p\n", o );
		for (uns u = 0; u < pw->prop_count; u++ )
			prop_sync(pw->items + u);
		pw_set_window_title(pw);
	}
}
static void pw_merge( struct prop_window * pw, struct go * go )
{
	if ( (pw->flags & PW_CONTEXT) && context.group && go_get_group(go) == context.group )
	{
		BASIC_DBG("Adding the go %p to property window\n", go);
		if ( !(pw->flags & PW_SELECTION) )
		{
			BASIC_DBG("Go is the only selected one\n");
			if ( pw->o != go2o(go) )
			{
				pw_empty(pw);
				pw_fill(pw, NULL);
			} else 
				pw->flags |= PW_SELECTION;
		} else {
			BASIC_DBG("There are some selected gos already\n");
			if (pw_merge_prop_items(pw, go2o(go))) pw_refill(pw);

			pw->flags |= PW_MERGED;
			pw_set_window_title(pw); //change the title if there was one selected go only
		}
	}
}
static void pw_unmerge( struct prop_window * pw, struct go * go )
{
	if ( (pw->flags & PW_SELECTION) && context.group && go_get_group(go) == context.group )
	{
		struct go * first;
		BASIC_DBG("Removing the go %p from property window\n", go);
		if ( !(first = group_get_first_selected_go(context.group)) )
		{
			BASIC_DBG("There is no selected go left\n");
			if ( go2o(go) != context.o )
			{ pw_empty(pw); pw_fill(pw, NULL); }
			else
				pw->flags &= ~PW_SELECTION;
		} else {
			BASIC_DBG("There are some selected gos yet\n");
			pw->o = go2o(first); //keep the pointer point to the first selected go
		}
	}
}
static void prop_go_unlink(struct o *o, struct o * owner UNUSED, struct o * prev UNUSED, struct go_hook * h)
{ pw_unlink((struct prop_window*)h->data,o); }
static void prop_obj_unlink(struct o *o, struct o * owner UNUSED, struct o * prev UNUSED, struct obj_hook * h)
{ pw_unlink((struct prop_window*)h->data,o); }
static void prop_go_create_prop(struct o *o, struct prop* p UNUSED, struct go_hook * h)
{ pw_maybe_refill(o,(struct prop_window*)h->data); }
static void prop_go_delete_prop(struct o *o, struct prop* p UNUSED, struct go_hook * h)
{ pw_maybe_refill(o,(struct prop_window*)h->data); }
static void prop_obj_create_prop(struct o *o, struct prop* p UNUSED, struct obj_hook * h)
{ pw_maybe_refill(o,(struct prop_window*)h->data); }
static void prop_obj_delete_prop(struct o *o, struct prop* p UNUSED, struct obj_hook * h)
{ pw_maybe_refill(o,(struct prop_window*)h->data); }
static void prop_go_change_prop(struct o *o, struct go_hook * h) { pw_sync(o,(struct prop_window*)h->data);}
static void prop_obj_change_prop(struct o *o, struct obj_hook * h) { pw_sync(o,(struct prop_window*)h->data);}
static void prop_go_selected(struct o *o, struct go_hook * h) { pw_merge((struct prop_window*)h->data, o2go(o));}
static void prop_go_unselected(struct o *o, struct go_hook * h) { pw_unmerge((struct prop_window*)h->data, o2go(o));}
static void prop_unit_event(uns u UNUSED, struct unit_hook *h)
{
	struct prop_window * pw = (struct prop_window*)h->data;
	pw_sync(pw->o, pw);
}

static gboolean entry_focus_out( GtkWidget * w, GdkEventFocus * event UNUSED, struct prop_item * pi )
{ prop_value_changed(w, pi); return FALSE; }

static void update_buffer( GtkWidget * area, struct prop_item * pi, struct prop * prop )
{
	ASSERT(pi);
	if (!prop)
		prop = prop_find( pi->o, pi->key );
	ASSERT(prop);

	int width = area->allocation.width;
	int height = area->allocation.height;
	if ( !pi->ptr )
	{
		BASIC_DBG("Allocating buffer for property item %s\n", pi->key->text);
		pi->ptr = xmalloc( 3 * width * height );
	}
	else
		pi->ptr = xrealloc( pi->ptr, 3 * width * height );

	// fill the color buffer
	guchar * buf = pi->ptr;
	uns r, g, b, alpha;
	uns2rgba_uns( prop_get_value(pi->o, prop).u, &r, &g, &b, &alpha);

	int i, j, under, k = 0;
	for ( i = 0; i < height; i++ ){
		for ( j = 0; j < width; j++, k+=3 ){
			under = ( ( j / SQ_SIZE + i / SQ_SIZE ) & 1 ) * ( 255 - alpha );
			buf[k] = r * alpha / 255 +  under;
			buf[k+1] = g * alpha / 255 + under;
			buf[k+2] = b * alpha / 255 + under;
		}
	}
}

static gboolean redraw_area( GtkWidget * area, GdkEventExpose *event UNUSED, struct prop_item * pi )
{
	ASSERT(pi);
	if ( !pi->ptr || sizeof(pi->ptr) != area->allocation.width * area->allocation.height * 3 )
		update_buffer(area, pi, NULL);
	gdk_draw_rgb_image( area->window, area->style->fg_gc[GTK_STATE_NORMAL],
		0, 0, area->allocation.width, area->allocation.height, GDK_RGB_DITHER_MAX,
		pi->ptr, area->allocation.width * 3 );
	return TRUE;
}

CREATE_F(color)
{
	GtkWidget * w = pi->value_edit = gtk_button_new();
	GtkWidget * child = gtk_drawing_area_new();
	g_signal_connect( G_OBJECT(child), "expose-event", G_CALLBACK(redraw_area), pi );
	gtk_widget_set_size_request( child, 30, 10 );
	gtk_container_add( GTK_CONTAINER(pi->value_edit), child );
	g_signal_connect( G_OBJECT(w), "clicked", G_CALLBACK(color_edit_callback), pi );
}

UPDATE_F(color,,)
{
	GtkWidget * area = gtk_bin_get_child(GTK_BIN(pi->value_edit));
	update_buffer(area, pi, prop);
	gtk_widget_queue_draw(area);
}
CALLB_F(color)
{
	FIND_PROP;
	GtkWidget *color_selector = wnd_select_color_new();
	GtkWidget *dialog = gtk_dialog_new_with_buttons("Color selection", NULL,
			GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
			GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
	gtk_dialog_set_default_response( GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT );
	gtk_container_add (GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), color_selector);
	gtk_widget_show_all(dialog);
	gtk_window_resize(GTK_WINDOW(dialog), 400, 300);

	uns r, g, b, a;
	uns2rgba_uns( prop_get_value(pi->o, prop).u, &r, &g, &b, &a );
	wnd_select_color_set_rgb((WndSelectColor *) color_selector, (int)r, (int)g, (int)b );
	wnd_select_color_set_alpha((WndSelectColor *) color_selector, (int)a );

	if ( gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT )
	{
		int na, nr, ng, nb;
		wnd_select_color_get_rgb((WndSelectColor*) color_selector, &nr, &ng, &nb );
		wnd_select_color_get_alpha((WndSelectColor *) color_selector, &na );
		wnd_select_color_push_recent( nr, ng, nb, na );
		prop_value new_val;
		new_val.u = rgba_uns2uns( nr, ng, nb, na );
		TRANS_PROP_CHANGE(new_val);
	}
	gtk_widget_destroy(dialog);
}

CREATE_F(filename)
{
	pi->value_edit = gtk_button_new_with_label("");
	gtk_button_set_relief(GTK_BUTTON(pi->value_edit), GTK_RELIEF_NONE);
	g_signal_connect(G_OBJECT(pi->value_edit), "clicked", G_CALLBACK(filename_edit_callback), pi);
}
UPDATE_F(filename,,)
{
	char * s = prop_get_value(pi->o, prop).s->text;
	if (!(*s) || s == string_file_unsaved->text )
		s = "Edit ...";
	gtk_button_set_label( GTK_BUTTON(pi->value_edit), s );
}

CALLB_F(filename)
{
	FIND_PROP;
	SAVE_DLG_START( "Choose the filename ...", NULL, prop_get_value(pi->o, prop).s->text )
	{
		prop_value new_val;
		new_val.s = string_lookup_ref(filename);
		TRANS_PROP_CHANGE(new_val);
		string_unref(new_val.s);
	}
	SAVE_DLG_END;
}

CREATE_F(largetext)
{
	pi->value_edit = gtk_button_new_with_label("Edit ...");
	gtk_button_set_relief (GTK_BUTTON(pi->value_edit), GTK_RELIEF_NONE);
	g_signal_connect (G_OBJECT (pi->value_edit), "clicked", G_CALLBACK(largetext_edit_callback), pi);
}
UPDATE_F(largetext,UNUSED,UNUSED) {}
CALLB_F(largetext) { text_edit_create( pi->o, pi->key ); }

CREATE_F(ptr)
{
	pi->value_edit = gtk_label_new("");
	gtk_label_set_justify( GTK_LABEL(pi->value_edit), GTK_JUSTIFY_LEFT );
}
UPDATE_F(ptr,,UNUSED)
{
#ifdef GUI_DEBUG
	char pstr[100];
	sprintf(pstr, "%p", prop_get_value(pi->o, prop).v);
	gtk_label_set_text( GTK_LABEL(pi->value_edit), pstr );
#else
	gtk_label_set_text( GTK_LABEL(pi->value_edit), "" );
#endif
}

CREATE_F(font)
{
	ASSERT(pi);
	if (!font_list && !fill_font_list())
		statusbar_err(pi->pw, "Could not load fonts.");

	GtkWidget * w = pi->value_edit = gtk_combo_box_new_with_model(GTK_TREE_MODEL(font_list));
	GtkCellRenderer * cr = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start( GTK_CELL_LAYOUT(w), cr, FALSE );
	gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(w), cr, "text", 1 );
	g_signal_connect( G_OBJECT(w), "changed", G_CALLBACK (font_edit_callback), pi);
}
UPDATE_F(font,,)
{
	int fnt_id = prop_get_value(pi->o, prop).u;
	if (!font_id_valid(fnt_id))
		return;
	char *filename = (char *) font_filename (fnt_id);
	ASSERT(filename);
	GtkTreeIter ti;
	if (gtk_combo_box_get_active_iter( GTK_COMBO_BOX(pi->value_edit), &ti ))
	{
		if (!fonts_compare_iter_with(&ti, filename))
			// the font is the same, no change needed
			return;
	}
	// walk the list and find the item
	int result, success = 0;
	for ( result = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(font_list), &ti); result;
		result = gtk_tree_model_iter_next(GTK_TREE_MODEL(font_list), &ti) )
	{
		if (!fonts_compare_iter_with(&ti, filename))
		{
			success = 1;
			gtk_combo_box_set_active_iter( GTK_COMBO_BOX(pi->value_edit), &ti);
			break;
		}
	}
	if (!success)
		statusbar_err( pi->pw, "Unknown font %s!", filename );
}
CALLB_F(font)
{
	if (changing_something)
		return; // prevent unwanted changes when the original value was clamped due to range change
	FIND_PROP;

	uns new_font_id;
	GtkTreeIter ti;
	if (!gtk_combo_box_get_active_iter( GTK_COMBO_BOX(pi->value_edit), &ti ))
		return;

	new_font_id = fonts_iter_to_font_id(&ti);
	if ( new_font_id != prop_get_value(pi->o, prop).u && font_id_valid(new_font_id))
		TRANS_PROP_CHANGE((prop_value){ .u = new_font_id });
}

static gboolean prop_label_pressed( GtkWidget * label_box, GdkEventButton *event UNUSED, struct prop_item * pi )
{
	ASSERT(pi);

	if ( !(pi->flags & PIF_SELECTED) )
	{
		pi->flags |= PIF_SELECTED;
		GdkColor color;
		gdk_color_parse( "lightblue", &color );
		gtk_widget_modify_bg( label_box, GTK_STATE_NORMAL, &color );
	} else {
		pi->flags &= ~PIF_SELECTED;
		gtk_widget_modify_bg( label_box, GTK_STATE_NORMAL, NULL );
	}
	return TRUE;
}
