#ifndef _GUI_RULER_H_
#define _GUI_RULER_H_

/*
 * GUI - Rulers
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include "lib/lib.h"

#define MINIMUM_INCR		2

#define TYPE_RULER            (ruler_get_type ())
#define RULER(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), TYPE_RULER, Ruler))
#define RULER_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), TYPE_RULER, RulerClass))
#define IS_RULER(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), TYPE_RULER))
#define IS_RULER_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), TYPE_RULER))
#define RULER_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), TYPE_RULER, RulerClass))

typedef struct _Ruler        Ruler;
typedef struct _RulerClass   RulerClass;

typedef double draw_line_fn (Ruler *ruler, GdkGC *gc, double cur, int ruler_width);
typedef void print_label_fn (Ruler *ruler, PangoLayout *layout, double pos, int digit_height, char *text, int digit_offset);

extern double scale[3];

/**
 * All distances below are in 1/72nd's of an inch. (According to
 * Adobe that's a point, but points are really 1/72.27 in.)
 */
struct _Ruler
{
  GtkWidget widget;	/**< The ruler widget */

  GdkPixmap *backing_store;	/**< pixmap for ruler painting */
  GdkGC *non_gr_exp_gc;	/* graphical */
  gint xsrc;	/**< horizontal position */
  gint ysrc;	/**< vertical position */
  gint slider_size; /**< size of the slider */

  gdouble position; /**< The position of the mark on the ruler (in pixels) */
  gdouble offset; /**< pixel coords of 0 */
  gdouble resolution; /**< pixels per selected unit */
};



/**
 Ruler class definition.
 */
struct _RulerClass
{
  GtkWidgetClass parent_class;	/**< parent class */

  void (* draw_ticks) (Ruler *ruler);	/**< method which draws the ruler - lines and numbers */
  void (* draw_pos)   (Ruler *ruler);	/**< method which draws actual position of cursor */

  void (* pressed) (Ruler *ruler);
};

/**
 Get type of ruler.
 \returns unique_type for this class
*/
GtkType   ruler_get_type   (void) G_GNUC_CONST;

void ruler_set_position (Ruler *ruler, gdouble position);
void ruler_set_offset (Ruler *ruler, gdouble offset);
void ruler_set_resolution (Ruler *ruler, gdouble resolution);

void compute_and_draw_ticks (Ruler *ruler, PangoLayout *layout,
		int digit_width, int width, int height,
		draw_line_fn draw_line, print_label_fn print_label,
		int digit_offset);

static inline gdouble ruler_pixels_to_coords (Ruler *ruler, gdouble p)
{ return ((p - ruler -> offset) / ruler -> resolution); }
static inline gdouble ruler_coords_to_pixels (Ruler *ruler, gdouble c)
{ return (c * ruler -> resolution +  ruler -> offset); }

/**
 Draws lines and numbers of ruler.
 \param ruler drawed ruler
*/
void    ruler_draw_ticks (Ruler      *ruler);
/**
 Draws actual position of cursor.
 \param ruler modified ruler
*/
void    ruler_draw_pos   (Ruler      *ruler);

/**
 Redrawing of whole ruler.
 \param widget ruler to be exposed
 \param event cause of ruler exposition
 \returns Is this event already handled?
*/
gint ruler_expose (GtkWidget      *widget,
		  GdkEventExpose *event);

////////////////////////////////////////////////////////


#define TYPE_HORIZ_RULER	    (horiz_ruler_get_type ())
#define HORIZ_RULER(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), TYPE_HORIZ_RULER, HorizRuler))
#define HORIZ_RULER_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), TYPE_HORIZ_RULER, HorizRulerClass))
#define IS_HORIZ_RULER(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), TYPE_HORIZ_RULER))
#define IS_HORIZ_RULER_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), TYPE_HORIZ_RULER))
#define HORIZ_RULER_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), TYPE_HORIZ_RULER, HorizRulerClass))


typedef struct _HorizRuler       HorizRuler;
typedef struct _HorizRulerClass  HorizRulerClass;

/**
 Horizontal ruler.
*/
struct _HorizRuler
{
  Ruler ruler;
};

/**
 Horizontal ruler class.
*/
struct _HorizRulerClass
{
  RulerClass parent_class;
};

/**
 Gets type of horizontal ruler.
 \return unique type for horizontal ruler.
*/
GtkType    horiz_ruler_get_type (void) G_GNUC_CONST;

/**
 Craetes new horizonal ruler and sets default values of its vars.
 \returns pointer to newly created ruler
*/
GtkWidget* horiz_ruler_new      (void);


/**
 Handles cursor motion.
 \param widget ruler to be redrawn
 \param event the cause of ruler's change
 \returns Is this event already handled?
*/
int horiz_ruler_motion_notify (GtkWidget *widget,
			  GdkEventMotion *event);


////////////////////////////////////////////////////////

#define TYPE_VERT_RULER		   (vert_ruler_get_type())
#define VERT_RULER(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), TYPE_VERT_RULER, VertRuler))
#define VERT_RULER_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), TYPE_VERT_RULER, VertRulerClass))
#define IS_VERT_RULER(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), TYPE_VERT_RULER))
#define IS_VERT_RULER_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), TYPE_VERT_RULER))
#define VERT_RULER_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), TYPE_VERT_RULER, VertRulerClass))

/**
 Vertical ruler.
*/
typedef struct _VertRuler       VertRuler;

/**
 Vertical ruler class.
*/
typedef struct _VertRulerClass  VertRulerClass;

/**
 Vertical ruler widget.
 */
struct _VertRuler
{
  Ruler ruler;	/**< Ruler */
};

/**
 The parent of RulerClass.
 */
struct _VertRulerClass
{
  RulerClass parent_class;
};

/**
 Gets type of vertical ruler.
 \returns unique type for vertical ruler
*/
GtkType vert_ruler_get_type (void) G_GNUC_CONST;

/**
 Creates new vertical ruler.
 \return pointer to created ruler.
*/
GtkWidget *vert_ruler_new      (void);

/**
 Handles cursor motion.
 \param widget vertical ruler to be redrawn
 \param event the cause of ruler's change
 \returns Is this event already handled?
*/
int vert_ruler_motion_notify (GtkWidget      *widget,
			  GdkEventMotion *event);


#endif
