/** @file
 * GUI - The implemetation of the Select color dialog.
 *
 * Copyright (c) 2004-2007    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG
#include "lib/lib.h"

#include <stdio.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include "gui/selectcolor.h"
#include "kernel/kernel.h"

#define BORDER_WIDTH 5
#define TABLE_SPACINGS 3

#define ARROW_HEIGHT 5
#define ARROW_WIDTH 5

static char * COLOR_NAMES[] = { "Red:", "Green:", "Blue:",
	"Hue:", "Saturation:", "Value:",
	"Cyan:", "Magenta:", "Yellow:", "Black:",
	"Alpha:", "Result:" };

static enum colors RGB_TABLE[] = { COLOR_RED, COLOR_GREEN, COLOR_BLUE };
static enum colors HSV_TABLE[] = { COLOR_HUE, COLOR_SATURATION, COLOR_VALUE };
static enum colors CMYK_TABLE[] = { COLOR_CYAN, COLOR_MAGENTA, COLOR_YELLOW, COLOR_BLACK };
static enum colors RESULT_TABLE[] = { COLOR_ALPHA, COLOR_RESULT };

static uns recent_colors[RECENT_COL_TAB_W * RECENT_COL_TAB_H];
static uns recent_tab;

#define DROPPER_WIDTH 17
#define DROPPER_HEIGHT 17
#define DROPPER_X_HOT 2
#define DROPPER_Y_HOT 16

#define REDRAW(x) update_area( wnd, x );\
	redraw_area( wnd->drawing_area[x], NULL, wnd )

#define REDRAW_ONLY(x) redraw_area( wnd->drawing_area[x], NULL, wnd )

static const guchar dropper_bits[] = {
	0xff, 0x8f, 0x01, 0xff, 0x77, 0x01, 0xff, 0xfb, 0x00, 0xff, 0xf8, 0x00,
	0x7f, 0xff, 0x00, 0xff, 0x7e, 0x01, 0xff, 0x9d, 0x01, 0xff, 0xd8, 0x01,
	0x7f, 0xd4, 0x01, 0x3f, 0xee, 0x01, 0x1f, 0xff, 0x01, 0x8f, 0xff, 0x01,
	0xc7, 0xff, 0x01, 0xe3, 0xff, 0x01, 0xf3, 0xff, 0x01, 0xfd, 0xff, 0x01,
	0xff, 0xff, 0x01, };

static const guchar dropper_mask[] = {
	0x00, 0x70, 0x00, 0x00, 0xf8, 0x00, 0x00, 0xfc, 0x01, 0x00, 0xff, 0x01,
	0x80, 0xff, 0x01, 0x00, 0xff, 0x00, 0x00, 0x7f, 0x00, 0x80, 0x3f, 0x00,
	0xc0, 0x3f, 0x00, 0xe0, 0x13, 0x00, 0xf0, 0x01, 0x00, 0xf8, 0x00, 0x00,
	0x7c, 0x00, 0x00, 0x3e, 0x00, 0x00, 0x1e, 0x00, 0x00, 0x0d, 0x00, 0x00,
	0x02, 0x00, 0x00, };

/************************* Private function declarations ********************************/
static GtkWidget * create_color_model_table( WndSelectColor * wnd, enum colors * rows, int count );
static GtkWidget * create_recent_colors_table( WndSelectColor * wnd );

static void set_cmyk_from_rgb( WndSelectColor * wnd );
static void set_hsv_from_rgb( WndSelectColor * wnd );
static void set_rgb_from_hsv( WndSelectColor * wnd );
static void set_rgb_from_cmyk( WndSelectColor * wnd );
static void set_hexet_from_rgb( WndSelectColor * wnd );
static void set_rgb_from_hexet( WndSelectColor * wnd );
static void fill_color_buf_rgb( guchar * buf, int colcnt, int rowcnt, int which_color, int red, int green, int blue );
static void fill_color_buf_cmy( guchar * buf, int colcnt, int rowcnt, int which_color, int cyan, int magenta, int yellow, int black );
static void fill_color_buf_alpha( guchar * buf, int colcnt, int rowcnt, int red, int green, int blue );
static void fill_color_buf_black( guchar * buf, int colcnt, int rowcnt, int cyan, int magenta, int yellow );
static void fill_color_buf_saturation( guchar * buf, int colcnt, int rowcnt, int red, int green, int blue );
static void fill_color_buf_value( guchar * buf, int colcnt, int rowcnt, int red, int green, int blue );
static void fill_color_buf_hue( guchar * buf, int colcnt, int rowcnt );
static void fill_color_buf_result( guchar * buf, int colcnt, int rowcnt, int red, int green, int blue, int alpha );

/* Dropper */
static void start_dropper( WndSelectColor * wnd );
static void finish_dropper( WndSelectColor * wnd );
static gboolean mouse_release( GtkWidget * widget, GdkEventButton * but, WndSelectColor * wnd );
static gboolean key_press( GtkWidget * widget, GdkEventKey * but, WndSelectColor * wnd );
static GdkCursor * make_dropper_cursor( void );
static void grab_color_at_mouse( WndSelectColor * wnd, gint x_root, gint y_root );

/* Callbacks */
static void something_changed( GtkWidget * spin, WndSelectColor * wnd );
static void update_area( WndSelectColor * wnd, enum colors which );

static gboolean on_area_click( GtkWidget * da, GdkEventButton * but, WndSelectColor * wnd );
static gboolean on_recent_area_click( GtkWidget * da, GdkEventButton * but, WndSelectColor * wnd );
static void on_page_switched( GtkNotebook * notebook, GtkNotebookPage * p, int page_num, gpointer u);
static gboolean redraw_area( GtkWidget * da, GdkEventExpose *event, WndSelectColor * wnd );
static gboolean on_hexet_changed( GtkWidget * da, GdkEventKey * event, WndSelectColor * wnd );

/* GTK */
static void wnd_select_color_class_init( WndSelectColorClass  *klass );
static void wnd_select_color_init( WndSelectColor * wnd );
static void wnd_select_color_destroy( GtkObject * object );
static void wnd_select_color_size_request( GtkWidget * wnd, GtkRequisition * requisition );
static void wnd_select_color_size_allocate( GtkWidget * wnd, GtkAllocation * allocation );
static GObject* wnd_select_color_constructor( GType type, guint n_construct_properties, GObjectConstructParam *construct_params);
static GtkBinClass *parent_class = NULL;

/*************** Function definitions  **************************************/

GtkType wnd_select_color_get_type( void )
{
	static GtkType my_type = 0;

	DBG("GetType called\n");
	if (!my_type) {
		static const GtkTypeInfo my_info = {
			"WndSelectColor",
			sizeof(WndSelectColor),
			sizeof(WndSelectColorClass),
			(GtkClassInitFunc) wnd_select_color_class_init,
			(GtkObjectInitFunc) wnd_select_color_init,
			NULL, NULL, NULL };

		my_type = gtk_type_unique( gtk_bin_get_type(), &my_info );
	}
	DBG("GetType OK\n");
	return my_type;
}

static void wnd_select_color_class_init (WndSelectColorClass * klass)
{
	GObjectClass *g_object_class;
	GtkObjectClass *object_class;
	GtkWidgetClass *widget_class;
	GtkContainerClass *container_class;

	DBG("Clas init\n");
	g_object_class = G_OBJECT_CLASS (klass);
	object_class = (GtkObjectClass*) klass;
	widget_class = (GtkWidgetClass*) klass;
	container_class = (GtkContainerClass*) klass;

	parent_class = g_type_class_peek_parent (klass);

	widget_class->size_request = wnd_select_color_size_request;
	widget_class->size_allocate = wnd_select_color_size_allocate;
	g_object_class->constructor = wnd_select_color_constructor;
	object_class->destroy = wnd_select_color_destroy;
	DBG("Clas init OK\n");
}

static void wnd_select_color_init( WndSelectColor *wnd )
{
	GTK_WIDGET_SET_FLAGS (wnd, GTK_NO_WINDOW);
}

static GObject* wnd_select_color_constructor( GType type, guint n_construct_properties, GObjectConstructParam *construct_params)
{
	GObject *object;
	WndSelectColor *wnd;
	GtkWidget * widget, * vbox, * table;
	int i;
	DBG("Constructor\n");
	object = (* G_OBJECT_CLASS (parent_class)->constructor) (type, n_construct_properties, construct_params);

	DBG("Get the window\n");
	wnd = GTK_VRR_SELECT_COLOR( object );
	wnd->change_in_progress = 1;

	for ( i = 0; i < COLOR_RESULT; i++ ){

		DBG("Drawing area %d\n", i);
		wnd->drawing_area[i] = gtk_drawing_area_new();
		g_object_set_data( G_OBJECT( wnd->drawing_area[i]), "vrrindex", (gpointer)(addr_int_t)i );
		g_signal_connect( G_OBJECT( wnd->drawing_area[i] ),
				"expose-event", G_CALLBACK( redraw_area ), wnd );
		gtk_widget_add_events( wnd->drawing_area[i], GDK_BUTTON_PRESS_MASK |
				GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK );
		g_signal_connect( G_OBJECT( wnd->drawing_area[i] ),
				"button-release-event", G_CALLBACK( on_area_click ), wnd );
		/*g_signal_connect( G_OBJECT( wnd->drawing_area[i] ),
				"motion-notify-event", G_CALLBACK( on_area_click ), wnd );*/
		wnd->buf[i] = NULL;

		DBG("Adjustment %d\n", i);
		wnd->adj[i] = GTK_ADJUSTMENT( gtk_adjustment_new( 0.0, 0.0, 1.0, 0.01, 0.1, 0.1 ) );
	}
	DBG("Drawing area RESULT\n");
	wnd->drawing_area[COLOR_RESULT] = gtk_drawing_area_new();
	g_object_set_data( G_OBJECT( wnd->drawing_area[COLOR_RESULT]), "vrrindex", (gpointer)COLOR_RESULT );
	g_signal_connect( G_OBJECT( wnd->drawing_area[COLOR_RESULT] ),
			"expose-event", G_CALLBACK( redraw_area ), wnd );
	wnd->buf[COLOR_RESULT] = NULL;

	/* Drawing areas for recent colors */
	for ( i = COLOR_RECENT_START; i < COLOR_MAX; i++ ){
		DBG("Drawing area %d\n", i);
		wnd->drawing_area[i] = gtk_drawing_area_new();
		g_object_set_data( G_OBJECT( wnd->drawing_area[i]), "vrrindex", (gpointer)(addr_int_t)i );
		g_signal_connect( G_OBJECT( wnd->drawing_area[i] ),
				"expose-event", G_CALLBACK( redraw_area ), wnd );
		gtk_widget_add_events( wnd->drawing_area[i], GDK_BUTTON_PRESS_MASK );
		g_signal_connect( G_OBJECT( wnd->drawing_area[i] ),
				"button-press-event", G_CALLBACK( on_recent_area_click ), wnd );
		wnd->buf[i] = NULL;
	}

	/* Frame */
	DBG("Frame\n");
	widget = gtk_frame_new( "Color" );
	gtk_container_set_border_width( GTK_CONTAINER( widget ), BORDER_WIDTH );
	gtk_container_add( GTK_CONTAINER( wnd ), widget );

	vbox = gtk_vbox_new( FALSE, BORDER_WIDTH );
	gtk_container_set_border_width( GTK_CONTAINER( vbox ), BORDER_WIDTH );
	gtk_container_add( GTK_CONTAINER( widget ), vbox );

	/* Tab notebook */
	DBG("Notebook\n");
	wnd->notebook = GTK_NOTEBOOK( gtk_notebook_new() );
	gtk_box_pack_start( GTK_BOX( vbox ), GTK_WIDGET( wnd->notebook ),
			TRUE, TRUE, 0 );

	DBG("Page RGB\n");
	table = create_color_model_table( wnd, RGB_TABLE, 3 );
	widget = gtk_label_new( "Hexet:" );
	gtk_misc_set_alignment( GTK_MISC( widget ), 1.0, 0.5 );
	gtk_table_attach( GTK_TABLE(table), widget, 0, 1, 3, 4, GTK_FILL, GTK_FILL,
				TABLE_SPACINGS, TABLE_SPACINGS );
	wnd->hexet_entry = GTK_ENTRY( gtk_entry_new_with_max_length( 6 ) );
	gtk_entry_set_width_chars( wnd->hexet_entry, 6 );
	g_signal_connect( G_OBJECT( wnd->hexet_entry ), "key-press-event",
			G_CALLBACK( on_hexet_changed ), wnd );
	gtk_table_attach( GTK_TABLE(table), GTK_WIDGET( wnd->hexet_entry ), 1, 2, 3, 4, 0, GTK_FILL,
			TABLE_SPACINGS, TABLE_SPACINGS );

	gtk_notebook_append_page( wnd->notebook, table, gtk_label_new( "RGB" ) );

	DBG("Page HSV\n");
	gtk_notebook_append_page( wnd->notebook,
		create_color_model_table( wnd, HSV_TABLE, 3 ),
		gtk_label_new( "HSV" ));

	DBG("Page CMYK\n");
	gtk_notebook_append_page( wnd->notebook,
		create_color_model_table( wnd, CMYK_TABLE, 4 ),
		gtk_label_new( "CMYK" ));

	DBG("Page Recent\n");
	gtk_notebook_append_page( wnd->notebook,
		create_recent_colors_table( wnd ), gtk_label_new( "Recent" ));

/*	DBG("Page GTK\n");
	widget = gtk_color_selection_new();
	gtk_color_selection_set_has_opacity_control( GTK_COLOR_SELECTION(widget), TRUE );
	gtk_color_selection_set_has_palette( GTK_COLOR_SELECTION(widget), TRUE );
	gtk_notebook_append_page( wnd->notebook, widget,
			gtk_label_new("GTK") );*/

	/* dropper */
	wnd->dropper = NULL;
	wnd->dropper_button = gtk_button_new_with_label( "Dropper" );

	/* RESULT TABLE */
	gtk_box_pack_start( GTK_BOX( vbox ), create_color_model_table( wnd, RESULT_TABLE, 2 ),
			FALSE, FALSE, 0 );

	/* signals */
	DBG("Connect the signals\n");
	g_signal_connect_swapped( G_OBJECT( wnd->dropper_button ),
			"clicked", G_CALLBACK( start_dropper ), wnd );

	wnd->change_in_progress = 0;

	DBG("Set the color\n");
	wnd_select_color_set_rgb( wnd, 100, 20, 200 );
	wnd_select_color_set_alpha( wnd, 255 );

	gtk_widget_show_all( GTK_WIDGET(wnd->notebook) ); // workaround a GTK "historical" bug
	gtk_notebook_set_current_page( wnd->notebook, recent_tab );
	g_signal_connect( G_OBJECT( wnd->notebook ), "switch-page", G_CALLBACK( on_page_switched ), wnd );
	DBG("OK\n");

	return object;
}

GtkWidget* wnd_select_color_new(void)
{
	return GTK_WIDGET( gtk_type_new( wnd_select_color_get_type()));
}

static void wnd_select_color_size_request( GtkWidget * wnd, GtkRequisition * requisition )
{
	gtk_widget_size_request( GTK_BIN(wnd)->child, requisition );
}
static void wnd_select_color_size_allocate( GtkWidget * wnd, GtkAllocation * allocation )
{
	wnd->allocation = *allocation;
	gtk_widget_size_allocate( GTK_BIN(wnd)->child, allocation );
}

/*******************************************************************************/
/* Functions for setting and getting colors */

void wnd_select_color_get_rgb( WndSelectColor * wnd, int * red, int * green, int * blue ){
	g_return_if_fail( GTK_IS_VRR_SELECT_COLOR( wnd ) );
	if ( red ) *red = INNER_TO_OUTER( gtk_adjustment_get_value( wnd->adj[COLOR_RED] ) );
	if ( green ) *green = INNER_TO_OUTER( gtk_adjustment_get_value( wnd->adj[COLOR_GREEN] ) );
	if ( blue ) *blue = INNER_TO_OUTER( gtk_adjustment_get_value( wnd->adj[COLOR_BLUE] ) );
}

void wnd_select_color_get_alpha( WndSelectColor * wnd, int * alpha ){
	g_return_if_fail( GTK_IS_VRR_SELECT_COLOR( wnd ) );
	*alpha = INNER_TO_OUTER( gtk_adjustment_get_value( wnd->adj[COLOR_ALPHA] ));
}

void wnd_select_color_set_alpha( WndSelectColor * wnd, int alpha ){
	g_return_if_fail( GTK_IS_VRR_SELECT_COLOR( wnd ) );
	ASSERT( alpha >= COLOR_MIN_VALUE && alpha <= COLOR_MAX_VALUE );

	wnd->change_in_progress = 1;
	gtk_adjustment_set_value( wnd->adj[COLOR_ALPHA], OUTER_TO_INNER( alpha ));
	REDRAW( COLOR_RESULT );
	wnd->change_in_progress = 0;
}

void wnd_select_color_set_rgb( WndSelectColor * wnd, int red, int green, int blue ){

	int i;

	g_return_if_fail( GTK_IS_VRR_SELECT_COLOR( wnd ) );
	ASSERT( red >= COLOR_MIN_VALUE && red <= COLOR_MAX_VALUE );
	ASSERT( green >= COLOR_MIN_VALUE && green <= COLOR_MAX_VALUE );
	ASSERT( blue >= COLOR_MIN_VALUE && blue <= COLOR_MAX_VALUE );

	wnd->change_in_progress = 1;

	gtk_adjustment_set_value( wnd->adj[COLOR_RED], OUTER_TO_INNER( red ));
	gtk_adjustment_set_value( wnd->adj[COLOR_GREEN], OUTER_TO_INNER( green ));
	gtk_adjustment_set_value( wnd->adj[COLOR_BLUE], OUTER_TO_INNER( blue ));

	set_hsv_from_rgb( wnd );
	set_cmyk_from_rgb( wnd );
	set_hexet_from_rgb( wnd );

	for ( i = 0; i < COLOR_MAX; i++ ){
		REDRAW( i );
	}

	wnd->change_in_progress = 0;
}

void wnd_select_color_push_recent( uns red, uns green, uns blue, uns alpha )
{
	uns color = rgba_uns2uns( red, green, blue, alpha );
	uns tmp, bak = color;
	int i, max = RECENT_COL_TAB_W * RECENT_COL_TAB_H;
	for ( i = 0; i < max; i++ )
	{
		tmp = recent_colors[i];
		recent_colors[i] = color;
		color = tmp;
		if ( color == bak ) break;
	}
}

static GtkWidget * create_color_model_table( WndSelectColor * wnd, enum colors * rows, int count ){

	GtkTable * table = GTK_TABLE( gtk_table_new( count, 3, FALSE ) );
	GtkWidget * label;
	GtkWidget * widget, * frame;
	int i;

	ASSERT( rows );

	gtk_table_set_row_spacings( table, TABLE_SPACINGS );
	gtk_table_set_col_spacings( table, TABLE_SPACINGS );

	for ( i = 0; i < count; i++ ){
		/* label */
		label = gtk_label_new( COLOR_NAMES[rows[i]] );
		gtk_misc_set_alignment( GTK_MISC( label ), 1.0, 0.5 );
		gtk_table_attach( table, label, 0, 1, i, i+1, GTK_FILL, GTK_FILL,
				TABLE_SPACINGS, TABLE_SPACINGS );

		frame = gtk_frame_new( NULL );
		gtk_widget_set_size_request( frame, -1, 30 );
		gtk_frame_set_shadow_type( GTK_FRAME(frame), GTK_SHADOW_IN );
		gtk_container_add( GTK_CONTAINER( frame ), GTK_WIDGET(wnd->drawing_area[rows[i]]) );
		gtk_table_attach( table, frame, 1, 2, i, i+1, GTK_FILL | GTK_EXPAND, GTK_FILL, TABLE_SPACINGS, TABLE_SPACINGS );

		if ( rows[i] != COLOR_RESULT ){
			widget = gtk_spin_button_new( wnd->adj[rows[i]], 0.01, 2 );
			g_object_set_data( G_OBJECT( widget ), "vrrindex", (gpointer)(rows[i]) );
			g_signal_connect( G_OBJECT(widget), "value-changed",
				G_CALLBACK( something_changed ), wnd );
		} else {
			widget = wnd->dropper_button;
		}
		gtk_table_attach( table, widget, 2, 3, i, i+1, GTK_FILL, GTK_FILL, TABLE_SPACINGS, TABLE_SPACINGS );

	}
	return GTK_WIDGET(table);
}

static GtkWidget * create_recent_colors_table( WndSelectColor * wnd )
{
	GtkTable * table = GTK_TABLE( gtk_table_new( RECENT_COL_TAB_W, RECENT_COL_TAB_H, FALSE ) );
	GtkWidget * frame;
	int i, j;

	//gtk_table_set_row_spacings( table, TABLE_SPACINGS );
	//gtk_table_set_col_spacings( table, TABLE_SPACINGS );

	for ( i = 0; i < RECENT_COL_TAB_H; i++ ){
		for ( j = 0; j < RECENT_COL_TAB_W; j++ ){
			frame = gtk_frame_new( NULL );
			gtk_widget_set_size_request( frame, 25, 25 );
			gtk_frame_set_shadow_type( GTK_FRAME(frame), GTK_SHADOW_IN );
			gtk_container_add( GTK_CONTAINER( frame ),
				GTK_WIDGET(wnd->drawing_area[COLOR_RECENT_START + i*RECENT_COL_TAB_W + j]) );
			gtk_table_attach( table, frame, j, j+1, i, i+1, GTK_FILL | GTK_EXPAND, GTK_FILL,
				TABLE_SPACINGS, TABLE_SPACINGS );
		}
	}
	return GTK_WIDGET(table);
}

static void wnd_select_color_destroy( GtkObject * object ){
	int i;
	WndSelectColor * wnd = GTK_VRR_SELECT_COLOR( object );
	DBG("Destroying the buffers.\n");
	for ( i=0; i < COLOR_MAX; i++ )
		if ( wnd->buf[i] ){
			xfree( wnd->buf[i] );
			wnd->buf[i] = NULL;
		}

	DBG("OK, buffers freed\n");
	if ( wnd->dropper )
		gtk_widget_destroy( wnd->dropper );

	DBG("Now destroy the object\n");
	(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
	DBG("Object destroyed, everything OK, the weather is nice :-)\n");
}

/* Callbacks */

static void something_changed( GtkWidget * spin, WndSelectColor * wnd )
{
	int j;
	addr_int_t which = (addr_int_t) g_object_get_data( G_OBJECT( spin ), "vrrindex" );
	if ( wnd->change_in_progress ) return;

	DBG("User changed value of %d\n", which );
	wnd->change_in_progress = 1;
	switch( which ){
		case COLOR_RED:
		case COLOR_GREEN:
		case COLOR_BLUE:
			set_hsv_from_rgb( wnd );
			set_cmyk_from_rgb( wnd );
			set_hexet_from_rgb( wnd );
			for ( j = 0; j < COLOR_MAX; j++ ) update_area( wnd, j);
			REDRAW_ONLY(COLOR_BLUE); REDRAW_ONLY(COLOR_GREEN);
			REDRAW_ONLY(COLOR_RED); REDRAW_ONLY(COLOR_ALPHA);
			REDRAW_ONLY(COLOR_RESULT);
			break;
		case COLOR_HUE:
		case COLOR_SATURATION:
		case COLOR_VALUE:
			set_rgb_from_hsv( wnd );
			set_cmyk_from_rgb( wnd );
			set_hexet_from_rgb( wnd );
			for ( j = 0; j < COLOR_MAX; j++ ) update_area( wnd, j);
			REDRAW_ONLY(COLOR_HUE); REDRAW_ONLY(COLOR_SATURATION);
			REDRAW_ONLY(COLOR_VALUE); REDRAW_ONLY(COLOR_ALPHA);
			REDRAW_ONLY(COLOR_RESULT);
			break;
		case COLOR_CYAN:
		case COLOR_MAGENTA:
		case COLOR_YELLOW:
		case COLOR_BLACK:
			set_rgb_from_cmyk( wnd );
			set_hsv_from_rgb( wnd );
			set_hexet_from_rgb( wnd );
			for ( j = 0; j < COLOR_MAX; j++ ) update_area( wnd, j);
			REDRAW_ONLY(COLOR_CYAN); REDRAW_ONLY(COLOR_MAGENTA);
			REDRAW_ONLY(COLOR_YELLOW); REDRAW_ONLY(COLOR_BLACK);
			REDRAW_ONLY(COLOR_ALPHA);
			REDRAW_ONLY(COLOR_RESULT);
			break;
		default: REDRAW_ONLY(which); REDRAW(COLOR_RESULT);
	}
	wnd->change_in_progress = 0;
}

static gboolean on_area_click( GtkWidget * da, GdkEventButton * but UNUSED, WndSelectColor * wnd ){
	addr_int_t which = (addr_int_t) g_object_get_data( G_OBJECT(da), "vrrindex" );
	double value = (double)(but->x) / (da->allocation.width);
	DBG( "CLICK on %d with value %.3f!", which, value );
	gtk_adjustment_set_value( wnd->adj[which], CLAMP( value, 0.0, 1.0 ) );
	return TRUE;
}

static gboolean on_recent_area_click( GtkWidget * da, GdkEventButton * but UNUSED, WndSelectColor * wnd )
{
	addr_int_t which = (addr_int_t) g_object_get_data( G_OBJECT(da), "vrrindex" );
	DBG("Click on recent area #%d\n", which );
	ASSERT( which >= COLOR_RECENT_START );
	wnd_select_color_set_alpha( wnd, uns2a_uns(recent_colors[ which - COLOR_RECENT_START ]));
	wnd_select_color_set_rgb( wnd, uns2r_uns(recent_colors[ which - COLOR_RECENT_START ]),
		uns2g_uns(recent_colors[ which - COLOR_RECENT_START ]),
		uns2b_uns(recent_colors[ which - COLOR_RECENT_START ]) );

	return TRUE;
}

static void on_page_switched( GtkNotebook * notebook UNUSED, GtkNotebookPage * p UNUSED, int page_num, gpointer u UNUSED )
{ recent_tab = page_num; }

static gboolean on_hexet_changed( GtkWidget * da UNUSED, GdkEventKey * event, WndSelectColor * wnd )
{
	int j;
	if ( event->keyval == GDK_Return || event->keyval == GDK_KP_Enter ){
		DBG("Hexet enter\n");
		wnd->change_in_progress = 1;
		/* recalculate */
		set_rgb_from_hexet( wnd );
		set_hsv_from_rgb( wnd );
		set_cmyk_from_rgb( wnd );
		/* refill the color areas */
		for ( j = 0; j < COLOR_MAX; j++ ) update_area( wnd, j );
		/* redraw */
		REDRAW_ONLY(COLOR_BLUE); REDRAW_ONLY(COLOR_GREEN);
		REDRAW_ONLY(COLOR_RED); REDRAW_ONLY(COLOR_ALPHA);
		REDRAW_ONLY(COLOR_RESULT);

		wnd->change_in_progress = 0;
		return TRUE;
	} else
		return FALSE;
}

#define GET_VALUE(x) gtk_adjustment_get_value( wnd->adj[x] ) * 255

static void update_area( WndSelectColor * wnd, enum colors which )
{
	ASSERT( which < COLOR_MAX );
	int width = GTK_WIDGET(wnd->drawing_area[which])->allocation.width;
	int height = GTK_WIDGET(wnd->drawing_area[which])->allocation.height;

	if ( !wnd->buf[which] ){
		DBG("Buffer for %d NULL, allocationg new.\n", which );
		wnd->buf[which] = xmalloc( 3 * width * height );
	} else {
		if ( wnd->buf_width[which] != width ||
			wnd->buf_height[which] != height ){
			DBG("Buffer of %d has different size, reallocating\n", which);
			wnd->buf[which] = xrealloc( wnd->buf[which],
				       	3 * width * height );
		}
	}
	wnd->buf_width[which] = width;
	wnd->buf_height[which] = height;

	switch ( which ){
	case COLOR_RED:
	case COLOR_GREEN:
	case COLOR_BLUE:
		fill_color_buf_rgb( wnd->buf[which], width, height, which,
			GET_VALUE(COLOR_RED), GET_VALUE(COLOR_GREEN), GET_VALUE(COLOR_BLUE) );
		break;
	case COLOR_ALPHA:
		fill_color_buf_alpha( wnd->buf[which], width, height,
			GET_VALUE(COLOR_RED), GET_VALUE(COLOR_GREEN), GET_VALUE(COLOR_BLUE) );
		break;
	case COLOR_HUE:
		fill_color_buf_hue( wnd->buf[which], width, height );
		break;
	case COLOR_SATURATION:
		fill_color_buf_saturation( wnd->buf[which], width, height,
			GET_VALUE(COLOR_RED), GET_VALUE(COLOR_GREEN), GET_VALUE(COLOR_BLUE) );
		break;
	case COLOR_VALUE:
		fill_color_buf_value( wnd->buf[which], width, height,
			GET_VALUE(COLOR_RED), GET_VALUE(COLOR_GREEN), GET_VALUE(COLOR_BLUE) );
		break;
	case COLOR_CYAN:
	case COLOR_MAGENTA:
	case COLOR_YELLOW:
		fill_color_buf_cmy( wnd->buf[which], width, height, which,
			GET_VALUE(COLOR_CYAN), GET_VALUE(COLOR_MAGENTA), GET_VALUE(COLOR_YELLOW),
			GET_VALUE(COLOR_BLACK) );
		break;
	case COLOR_BLACK:
		fill_color_buf_black( wnd->buf[which], width, height,
			GET_VALUE(COLOR_CYAN), GET_VALUE(COLOR_MAGENTA), GET_VALUE(COLOR_YELLOW) );
		break;
	case COLOR_RESULT:
		fill_color_buf_result( wnd->buf[which], width, height,
			GET_VALUE(COLOR_RED), GET_VALUE(COLOR_GREEN), GET_VALUE(COLOR_BLUE), GET_VALUE(COLOR_ALPHA) );
		break;
	default: // A buffer for recently used colors
		ASSERT(which < COLOR_MAX);
		fill_color_buf_result( wnd->buf[which], width, height,
			(int)uns2r_uns(recent_colors[which - COLOR_RECENT_START]),
			(int)uns2g_uns(recent_colors[which - COLOR_RECENT_START]),
			(int)uns2b_uns(recent_colors[which - COLOR_RECENT_START]),
			(int)uns2a_uns(recent_colors[which - COLOR_RECENT_START]));
	}
}

static gboolean redraw_area( GtkWidget * da, GdkEventExpose *event UNUSED, WndSelectColor * wnd )
{
	addr_int_t which = (addr_int_t) g_object_get_data( G_OBJECT(da), "vrrindex" );
	GdkPoint pts[3];
	if ( ! GDK_IS_DRAWABLE( da->window ) ) return FALSE;
	DBG("Redrawing area %d\n", which );

	if ( !wnd->buf[which] ||
		wnd->buf_width[which] != da->allocation.width || wnd->buf_width[which] != da->allocation.width )
			update_area( wnd, which );

	gdk_draw_rgb_image( da->window, da->style->fg_gc[GTK_STATE_NORMAL],
		0, 0, wnd->buf_width[which], wnd->buf_height[which], GDK_RGB_DITHER_MAX,
		wnd->buf[which], wnd->buf_width[which] * 3 );

	if ( which >= COLOR_RESULT ) return TRUE;

	/* the arrow */
	pts[0].x = gtk_adjustment_get_value( wnd->adj[which] ) * da->allocation.width;
       	pts[1].y = pts[2].y = da->allocation.height;
	pts[0].y = pts[1].y - ARROW_HEIGHT;
	pts[1].x = pts[0].x - ARROW_WIDTH;
	pts[2].x = pts[0].x + ARROW_WIDTH;
	gdk_draw_polygon( da->window, da->style->black_gc, TRUE, pts, 3 );
	gdk_draw_polygon( da->window, da->style->white_gc, FALSE, pts, 3 );
	pts[1].y = pts[2].y = -1;
	pts[0].y = ARROW_HEIGHT - 1;
	gdk_draw_polygon( da->window, da->style->black_gc, TRUE, pts, 3 );
	gdk_draw_polygon( da->window, da->style->white_gc, FALSE, pts, 3 );
	return TRUE;
}



/* private */
static void set_hexet_from_rgb( WndSelectColor * wnd )
{
	char hexet[7];
	double red = gtk_adjustment_get_value( wnd->adj[COLOR_RED] );
	double green = gtk_adjustment_get_value( wnd->adj[COLOR_GREEN] );
	double blue = gtk_adjustment_get_value( wnd->adj[COLOR_BLUE] );

	sprintf( hexet, "%.2X%.2X%.2X", (int)(red * 255.0 + 0.5),
			(int)(green * 255.0 + 0.5), (int)(blue * 255.0 + 0.5) );
	gtk_entry_set_text( wnd->hexet_entry, hexet );
}
static void set_rgb_from_hexet( WndSelectColor * wnd )
{
	unsigned int all;
	if ( sscanf( gtk_entry_get_text( wnd->hexet_entry ), "%X", &all ) == 1 ){
		gtk_adjustment_set_value( wnd->adj[COLOR_RED], (all >> 16) / 255.0 );
		gtk_adjustment_set_value( wnd->adj[COLOR_GREEN], ((all >> 8) % 256) / 255.0 );
		gtk_adjustment_set_value( wnd->adj[COLOR_BLUE], ( all % 256 ) / 255.0 );
	} else
		DBG("Wrong value of hexet!\n");
}
static void set_hsv_from_rgb( WndSelectColor * wnd ){

	double red = gtk_adjustment_get_value( wnd->adj[COLOR_RED] );
	double green = gtk_adjustment_get_value( wnd->adj[COLOR_GREEN] );
	double blue = gtk_adjustment_get_value( wnd->adj[COLOR_BLUE] );

	double min = MIN( red, MIN( blue, green ));
	double value = MAX( red, MAX( blue, green ));
	double delta = value - min;
	double hue, saturation;

	DBG( "Converting from RED:%.3f, GREEN:%.3f, BLUE:%.3f", red, green, blue );

	if ( ALMOST_ZERO(delta) ){
		hue = 0; saturation = 0;
	} else {
		saturation = delta / value; /* this should not do any
					       harm since value >= delta */
		if ( ALMOST_ZERO( value - red ) )
			hue = (green - blue) / 6 / delta;
		else if ( ALMOST_ZERO( value - green ) )
			hue = 0.3333 + ( blue - red ) / 6 / delta;
		else
			hue = 0.6667 + ( red - green ) / 6 / delta;

		if ( hue > 1 ) hue -= 1;
		if ( hue < 0 ) hue += 1;
	}

	DBG( "Converted to H:%.3f, S:%.3f, V:%.3f", hue, saturation, value );
	gtk_adjustment_set_value( wnd->adj[COLOR_HUE], hue );
	gtk_adjustment_set_value( wnd->adj[COLOR_SATURATION], saturation );
	gtk_adjustment_set_value( wnd->adj[COLOR_VALUE], value );
}

static void set_cmyk_from_rgb( WndSelectColor * wnd ){

	double red = gtk_adjustment_get_value( wnd->adj[COLOR_RED] );
	double green = gtk_adjustment_get_value( wnd->adj[COLOR_GREEN] );
	double blue = gtk_adjustment_get_value( wnd->adj[COLOR_BLUE] );

	double black;

	black = MAX( red, MAX(  green, blue ));
	if ( ALMOST_ZERO( black ) ){
		gtk_adjustment_set_value( wnd->adj[COLOR_CYAN], 0 );
		gtk_adjustment_set_value( wnd->adj[COLOR_MAGENTA], 0 );
		gtk_adjustment_set_value( wnd->adj[COLOR_YELLOW], 0 );
	} else {
		gtk_adjustment_set_value( wnd->adj[COLOR_CYAN], 1 - red / black );
		gtk_adjustment_set_value( wnd->adj[COLOR_MAGENTA], 1 - green / black );
		gtk_adjustment_set_value( wnd->adj[COLOR_YELLOW], 1 - blue / black );
	}
	gtk_adjustment_set_value( wnd->adj[COLOR_BLACK], 1 - black );

}

static void set_rgb_from_hsv( WndSelectColor * wnd ){

	double saturation = gtk_adjustment_get_value( wnd->adj[COLOR_SATURATION] );
	double hue = gtk_adjustment_get_value( wnd->adj[COLOR_HUE] );
	double value = gtk_adjustment_get_value( wnd->adj[COLOR_VALUE] );
	double red, green, blue;

	DBG( "Converting from H:%.3f, S:%.3f, V:%.3f", hue, saturation, value );

	if ( ALMOST_ZERO( saturation ) )
		red = blue = green = value;
	else {
		double var_h = 6 * hue;
		int var_i = (int)( var_h );
		double var_1 = value * ( 1 - saturation );
		double var_2 = value * ( 1 - saturation * ( var_h - var_i ));
		double var_3 = value * ( 1 - saturation * ( 1 - var_h + var_i ));

		switch( var_i ){
		case 0: red = value; green = var_3; blue = var_1; break;
		case 1: red = var_2; green = value; blue = var_1; break;
		case 2: red = var_1; green = value; blue = var_3; break;
		case 3: red = var_1; green = var_2; blue = value; break;
		case 4: red = var_3; green = var_1; blue = value; break;
		default: red = value; green = var_1; blue = var_2; break;
		}
	}

	DBG( "Converted to RED:%.3f, GREEN:%.3f, BLUE:%.3f", red, green, blue );

	gtk_adjustment_set_value( wnd->adj[COLOR_RED], red );
	gtk_adjustment_set_value( wnd->adj[COLOR_GREEN], green );
	gtk_adjustment_set_value( wnd->adj[COLOR_BLUE], blue );
}

static void set_rgb_from_cmyk( WndSelectColor * wnd )
{
	double cyan = gtk_adjustment_get_value( wnd->adj[COLOR_CYAN] );
	double magenta = gtk_adjustment_get_value( wnd->adj[COLOR_MAGENTA] );
	double yellow = gtk_adjustment_get_value( wnd->adj[COLOR_YELLOW] );
	double black = gtk_adjustment_get_value( wnd->adj[COLOR_BLACK] );
	double red, green, blue;

	DBG("Converting from CYAN:%.3f, MAGENTA:%.3f, YELLOW:%.3f, BLACK:%.3f\n",
			cyan, magenta, yellow, black );

	black = 1 - black;
	red = black - black * cyan;
	green = black - magenta * black;
	blue = black - yellow * black;

	DBG( "Converted to RED:%.3f, GREEN:%.3f, BLUE:%.3f", red, green, blue );
	gtk_adjustment_set_value( wnd->adj[COLOR_RED], red );
	gtk_adjustment_set_value( wnd->adj[COLOR_GREEN], green );
	gtk_adjustment_set_value( wnd->adj[COLOR_BLUE], blue );

}

/* Functions for filling color buffers with rainbows :) */
static void fill_color_buf_rgb( guchar * buf, int colcnt, int rowcnt, int which_color, int red, int green, int blue ){
	int width = 3 * colcnt;
	int i, j;
	int total = width * rowcnt;
	int half = total / 2;
	ASSERT( which_color <= COLOR_BLUE );

	for ( i = 0, j = 0; i < total; i+=3, j+=1 ){
		if ( i < half ){
			buf[i] = red;
			buf[i + 1] = green;
			buf[i + 2] = blue;
		} else {
			buf[i] = buf[i+1] = buf[i+2] = 0;
		}
		buf[i + which_color] = (j % colcnt * 256) / colcnt;
	}
}

static void fill_color_buf_alpha( guchar * buf, int colcnt, int rowcnt, int red, int green, int blue )
{
	int i, j, k;
	int width = 3 * colcnt, total = width * rowcnt;
	int under;
	double alpha;
	for ( i = j = k = 0; i < total; i+=3, j+=1 ){
		if ( j >= colcnt ){
			j = 0; k += 1;
		}

		alpha = ((double)j) / colcnt;
		under = ( ( j / SQ_SIZE + k / SQ_SIZE ) & 1 ) * ( 1 - alpha ) * 255;
		buf[i] = red * alpha +  under;
		buf[i+1] = green * alpha + under;
		buf[i+2] = blue * alpha + under;
	}
}

#define FILL_PART(s,r,g,b) for ( j=0; j<s; j++, k+=3 ){\
	buf[k] = r; buf[k+1] = g; buf[k+2] = b;\
	}

static void fill_color_buf_hue( guchar * buf, int colcnt, int rowcnt )
{
	int i, j, k = 0;

	int size1 = colcnt / 6;
	int size2 = colcnt / 3 - size1;
	int size3 = colcnt / 2 - size2 - size1;
	int size4 = (2 * colcnt) / 3; /* - size3 - size2 - size1;*/
	int size5 = (5 * colcnt) / 6 - size4;
	int size6 = colcnt - size5 - size4;
	size4 -= size3 + size2 + size1;

	ASSERT( buf );
	for ( i = 0; i < rowcnt; i++ ){
		FILL_PART( size1, 255, j * 255 / size1, 0 );
		FILL_PART( size2, 255 - j * 255 / size2, 255, 0 );
		FILL_PART( size3, 0, 255, 255 * j / size3 );
		FILL_PART( size4, 0, 255 - j * 255 / size4, 255 );
		FILL_PART( size5, j * 255 / size5, 0, 255 );
		FILL_PART( size6, 255, 0, 255 - j * 255 / size6 );
	}
}

static void fill_color_buf_saturation( guchar * buf, int colcnt, int rowcnt, int red, int green, int blue )
{
	int max = MAX( red, MAX( green, blue ) );
	int min = MIN( red, MIN( green, blue ) );

	int rdelta = max - red;
	int gdelta = max - green;
	int bdelta = max - blue;
	int delta = (max - min) * colcnt;
	int i, j, k = 0;
	for ( i = 0; i < rowcnt; i++ ){
		for ( j = 0; j < colcnt; j++, k+=3 ){
			if ( delta != 0 ){
				buf[k] =  max - rdelta * j * max / delta;
				buf[k+1] = max - gdelta * j * max / delta;
				buf[k+2] = max - bdelta * j * max / delta;
			} else {
				buf[k] = buf[k+1] = buf[k+2] = max;
			}
		}
	}
}

static void fill_color_buf_value( guchar * buf, int colcnt, int rowcnt, int red, int green, int blue )
{
	int max = MAX( red, MAX( green, blue ) ) * colcnt;
	int i, j, k = 0;
	for ( i = 0; i < rowcnt; i++ ){
		for ( j = 0; j < colcnt; j++, k+=3 ){
			if ( max != 0 ){
				buf[k] = red * j * 255 /  max;
				buf[k+1] = green * j * 255 / max;
				buf[k+2] = blue * j * 255 / max;
			} else {
				buf[k] = buf[k+1] = buf[k+2] = 0;
			}
		}
	}
}

static void fill_color_buf_cmy( guchar * buf, int colcnt, int rowcnt, int which_color, int cyan, int magenta, int yellow, int black )
{
	int i, j, k = 0;
	int half = rowcnt / 2;
	int bl = 255 - black;
	for ( i = 0; i < rowcnt; i++ ){
		for ( j = 0; j < colcnt; j++, k+=3 ){
			if ( i < half ){
				buf[k] =  bl - cyan * bl / 255;
				buf[k+1] = bl - magenta * bl / 255;
				buf[k+2] = bl - yellow * bl / 255;
			} else {
				buf[k] = buf[k+1] = buf[k+2] = 255;
			}
			buf[k + which_color - COLOR_CYAN] = bl - j * bl / colcnt;
		}
	}
}

static void fill_color_buf_black( guchar * buf, int colcnt, int rowcnt, int cyan, int magenta, int yellow )
{
	int i, j, k = 0;
	int bl;
	for ( i = 0; i < rowcnt; i++ ){
		for ( j = 0; j < colcnt; j++, k+=3 ){
			bl = 255 - 255 * j / colcnt;
			buf[k] =  bl - cyan * bl / 255;
			buf[k+1] = bl - magenta * bl / 255;
			buf[k+2] = bl - yellow * bl / 255;
		}
	}
}

static void fill_color_buf_result( guchar * buf, int colcnt, int rowcnt, int red, int green, int blue, int alpha )
{
	int i, j, k = 0;
	int under;

	DBG("Redrawing final color, alpha is %d\n", alpha );
	for ( i = 0; i < rowcnt; i++ ){
		for ( j = 0; j < colcnt; j++, k+=3 ){
			under = ( ( j / SQ_SIZE + i / SQ_SIZE ) & 1 ) * ( 255 - alpha );
			buf[k] = red * alpha / 255 +  under;
			buf[k+1] = green * alpha / 255 + under;
			buf[k+2] = blue * alpha / 255 + under;
		}
	}
	DBG("Fill OK\n");
}

/* Dropper */
static void start_dropper( WndSelectColor * wnd )
{
	GdkCursor *picker_cursor;
	GdkGrabStatus grab_status;
	GtkWidget * dropper = wnd->dropper;

	if ( !dropper ){
		DBG("Create new dropper\n");
		dropper = gtk_invisible_new();
		gtk_widget_add_events ( dropper, GDK_BUTTON_RELEASE_MASK | GDK_BUTTON_PRESS_MASK |
				GDK_POINTER_MOTION_MASK);
		gtk_widget_show ( dropper );
		wnd->dropper = dropper;
	}
	DBG("Grab keyboard\n");
	if ( gdk_keyboard_grab( dropper->window, FALSE, gtk_get_current_event_time ()) != GDK_GRAB_SUCCESS )
		return;
	DBG("Make cursor\n");
	picker_cursor = make_dropper_cursor();
	DBG("Grab pointer\n");
	grab_status = gdk_pointer_grab( dropper->window, FALSE,
		GDK_BUTTON_RELEASE_MASK | GDK_BUTTON_PRESS_MASK | GDK_POINTER_MOTION_MASK,
		NULL, picker_cursor, gtk_get_current_event_time() );
	gdk_cursor_unref (picker_cursor);

	if ( grab_status != GDK_GRAB_SUCCESS ) {
		gdk_keyboard_ungrab( GDK_CURRENT_TIME );
	        return;
	}
	gtk_grab_add( dropper );

	g_signal_connect( dropper, "button_press_event", G_CALLBACK (mouse_release), wnd );
	g_signal_connect( dropper, "key_press_event", G_CALLBACK (key_press), wnd );

	DBG("Dropper started.\n");
}

static void finish_dropper( WndSelectColor * wnd )
{
	guint32 time = gtk_get_current_event_time ();
	DBG("Finish dropper\n");
	gdk_keyboard_ungrab( time );
	gdk_pointer_ungrab( time );
	gtk_grab_remove( wnd->dropper );
	g_signal_handlers_disconnect_by_func ( wnd->dropper, (void *)mouse_release, wnd );
	g_signal_handlers_disconnect_by_func ( wnd->dropper, (void *)key_press, wnd );
	gtk_widget_destroy( wnd->dropper );
	wnd->dropper = NULL;
	DBG("Dropper finished\n");
}

static gboolean mouse_release( GtkWidget * widget UNUSED, GdkEventButton * but, WndSelectColor * wnd )
{
	if ( but->button != 1) return FALSE;

        grab_color_at_mouse( wnd, but->x_root, but->y_root );
	finish_dropper( wnd );

	return TRUE;
}

static gboolean key_press( GtkWidget * widget UNUSED, GdkEventKey * but, WndSelectColor * wnd )
{
	if ( but->keyval == GDK_Escape ){
		finish_dropper( wnd );
		return TRUE;
	}
	return FALSE;
}

static GdkCursor * make_dropper_cursor(void )
{
	GdkCursor *cursor;
	GdkColor bg = { 0, 0xffff, 0xffff, 0xffff };
	GdkColor fg = { 0, 0x0000, 0x0000, 0x0000 };
	GdkPixmap *pixmap = gdk_bitmap_create_from_data( NULL,
			(gchar *) dropper_bits, DROPPER_WIDTH, DROPPER_HEIGHT );
	GdkPixmap *mask =  gdk_bitmap_create_from_data( NULL,
			(gchar *) dropper_mask, DROPPER_WIDTH, DROPPER_HEIGHT );
	cursor = gdk_cursor_new_from_pixmap (pixmap, mask, &fg, &bg,
	                DROPPER_X_HOT, DROPPER_Y_HOT);
	g_object_unref (pixmap);
	g_object_unref (mask);

	return cursor;
}
static void grab_color_at_mouse( WndSelectColor * wnd, gint x_root, gint y_root )
{
	GdkImage *image;
	guint32 pixel;
	GdkColor color;
	GdkColormap *colormap = gdk_colormap_get_system();

	image = gdk_image_get( GDK_ROOT_PARENT(), x_root, y_root, 1, 1);
	pixel = gdk_image_get_pixel (image, 0, 0);
	g_object_unref (image);

	gdk_colormap_query_color (colormap, pixel, &color);

	DBG("Setting the color.\n");
	wnd_select_color_set_rgb( wnd, GDK_TO_OUTER( color.red ), GDK_TO_OUTER( color.green ), GDK_TO_OUTER( color.blue ) );
}

