#undef LOCAL_DEBUG
#include "lib/lib.h"

#include "kernel/kernel.h"

#include <gtk/gtk.h>
#include "gui/main.h"
#include "gui/cmdmgr.h"
#include "scheme/glt_gui.h"
#include "gui/properties.h"

/** \file gui/units.c
 * Implementation of the Unit manager window and maintenance of unit lists per quantities.
 */

struct wnd_units wnd_units;
static void create_unit_window(void);

static void on_new_unit( GtkWidget * w, gpointer data );
static void on_edit_unit( GtkWidget * w, gpointer data );
static void on_set_default_unit( GtkWidget * w, gpointer data );
static void on_delete_unit( GtkWidget * w, gpointer data );
static void unit_cursor_changed( GtkTreeView * t, gpointer data );

/** The unit lists which are used in unit combo boxes as well as in the unit manager. */
static GtkListStore * uq_none_list, * uq_len_list, * uq_angle_list, * uq_ref_list;
/** The array of all unit lists */
GtkListStore ** unit_quant_lists[PQ_MAX] = { &uq_none_list, &uq_len_list, &uq_angle_list, &uq_ref_list };

static uns current_quantity;
static int selected_unit = -1;
uns example_subtypes[PQ_MAX] = { PTR_UNSPECIFIED, PTR_COORDINATE, PTR_ANGLE_4PI, PTR_REFERENCE };

static void uw_unit_added(uns slot, struct unit_hook *h);
static void uw_unit_deleted(uns slot, struct unit_hook *h);
static void uw_unit_changed(uns slot, struct unit_hook *h);
static void uw_unit_default_changed(uns quantity, struct unit_hook *h);

static struct unit_hook uw_unit_hook = { .data = &uw_unit_hook, .add = uw_unit_added, .deleted = uw_unit_deleted,
	.changed = uw_unit_changed, .def = uw_unit_default_changed };
	
static char * mult_desc[] = { NULL, "Multiplier (against millimeters)", "Multiplier (against radians)", "Multiplier" };

void create_unit_quant_lists(void)
{
	int i;
	for ( i = 0; i < PQ_MAX; i++ )
		*(unit_quant_lists[i]) = GTK_LIST_STORE(gtk_list_store_new( 4, G_TYPE_STRING, G_TYPE_FLOAT,
			G_TYPE_INT, G_TYPE_INT ));

	//fill the unit lists
	for ( i = 0; i < PROP_UNIT_MAX; i++ )
	{
		if (!unit_array[i].name)
			continue;

		GtkTreeIter ti;
		GtkListStore * li = *(unit_quant_lists[unit_array[i].quantity]);
		gtk_list_store_append( li, &ti );
		gtk_list_store_set( li, &ti,
			UNIT_NAME_COL, unit_array[i].name->text,
			UNIT_MULT_COL, (float)unit_array[i].multiplier,
			UNIT_INDEX_COL, i,
			UNIT_WEIGHT_COL, ((uns)i == unit_default[unit_array[i].quantity]) ? 1000 : 400,
			-1 );
	}
	unit_hook_add( &uw_unit_hook );
}
void destroy_unit_quant_lists(void)
{
	for ( int i = 0; i < PQ_MAX; i++ )
                g_object_unref( G_OBJECT(*(unit_quant_lists[i])) );
	unit_hook_remove( &uw_unit_hook );
}

void on_show_unit_window( void * subj UNUSED )
{
	if ( wnd_units.visible )
	{
		gtk_widget_destroy( wnd_units.window );
		wnd_units.visible = 0;
	}
	else
		create_unit_window();
}

static void destroy_unit_window( GtkWidget *widget UNUSED, gpointer data UNUSED )
{
	wnd_units.visible = 0;
	refresh_gui();
}

static void quantity_combo_changed( GtkWidget * w UNUSED, gpointer data UNUSED )
{
	int i = gtk_combo_box_get_active( wnd_units.quantity_combo );
	if ( i < 0 || i > 2 )
		i = 0;
	current_quantity = i + 1;
	gtk_tree_view_set_model( GTK_TREE_VIEW(wnd_units.unit_view), GTK_TREE_MODEL(*(unit_quant_lists[current_quantity])));
	unit_cursor_changed( GTK_TREE_VIEW(wnd_units.unit_view), NULL );
	g_object_set( G_OBJECT(wnd_units.unit_view_mult_col), "title", mult_desc[current_quantity], NULL );
}

static void create_unit_window(void)
{
	win_o_init( (struct window*)&wnd_units, WT_UNITS );

	wnd_units.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title( GTK_WINDOW(wnd_units.window), "Unit manager" );
	gtk_window_add_accel_group( GTK_WINDOW(wnd_units.window), accel_group );
	gtk_widget_set_size_request( wnd_units.window, -1, 300 );

	GtkWidget * table = gtk_table_new( 4, 2, FALSE );
	gtk_container_add( GTK_CONTAINER(wnd_units.window), table );

	gtk_table_attach( GTK_TABLE(table), gtk_label_new("Displayed units for quantity:"), 0, 1, 0, 1,
		GTK_EXPAND | GTK_FILL, GTK_FILL, 5, 5 );

	wnd_units.quantity_combo = GTK_COMBO_BOX(gtk_combo_box_new_text());
	gtk_combo_box_append_text( wnd_units.quantity_combo, "length" );
	gtk_combo_box_append_text( wnd_units.quantity_combo, "angle" );
	gtk_combo_box_append_text( wnd_units.quantity_combo, "reference" );
	gtk_combo_box_set_active( wnd_units.quantity_combo, 0 );
	gtk_table_attach( GTK_TABLE(table), GTK_WIDGET(wnd_units.quantity_combo), 1, 2, 0, 1,
		GTK_EXPAND | GTK_FILL, GTK_FILL, 5, 5 );
	g_signal_connect( G_OBJECT(wnd_units.quantity_combo), "changed", G_CALLBACK(quantity_combo_changed), NULL );

	GtkWidget * sw = gtk_scrolled_window_new( NULL, NULL );
	gtk_container_set_border_width( GTK_CONTAINER(sw), 5 );
	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW(sw), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC );
	gtk_table_attach( GTK_TABLE(table), sw, 0, 2, 1, 2, GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0 );

	current_quantity = 1;
	wnd_units.unit_view = gtk_tree_view_new_with_model( GTK_TREE_MODEL(uq_len_list) );

	GtkTreeViewColumn * column = gtk_tree_view_column_new_with_attributes( "Name",
		gtk_cell_renderer_text_new(), "text", UNIT_NAME_COL, "weight", UNIT_WEIGHT_COL, NULL );
	g_object_set( G_OBJECT(column), "max-width", 100, NULL );
	gtk_tree_view_append_column( GTK_TREE_VIEW(wnd_units.unit_view), column );
#ifdef GUI_DEBUG
	column = gtk_tree_view_column_new_with_attributes( "Index",
		gtk_cell_renderer_text_new(), "text", UNIT_INDEX_COL, NULL );
	gtk_tree_view_append_column( GTK_TREE_VIEW(wnd_units.unit_view), column );
#endif
	column = gtk_tree_view_column_new_with_attributes( mult_desc[PQ_LENGTH],
		gtk_cell_renderer_text_new(), "text", UNIT_MULT_COL, NULL );
	gtk_tree_view_append_column( GTK_TREE_VIEW(wnd_units.unit_view), column );
	gtk_tree_selection_set_mode( gtk_tree_view_get_selection(GTK_TREE_VIEW(wnd_units.unit_view)), GTK_SELECTION_SINGLE );
	wnd_units.unit_view_mult_col = column;

	gtk_container_add( GTK_CONTAINER(sw), wnd_units.unit_view );
	g_signal_connect( G_OBJECT(wnd_units.unit_view), "cursor-changed", G_CALLBACK(unit_cursor_changed), NULL );

	GtkWidget * hbutbox = gtk_hbutton_box_new();
	gtk_button_box_set_layout( GTK_BUTTON_BOX(hbutbox), GTK_BUTTONBOX_SPREAD );
	gtk_table_attach( GTK_TABLE(table), hbutbox, 0, 2, 2, 3, GTK_EXPAND | GTK_FILL, GTK_FILL, 5, 5 );

	GtkWidget * new_button = gtk_button_new_with_label("New ...");
	gtk_box_pack_start( GTK_BOX(hbutbox), new_button, FALSE, FALSE, 0 );
	wnd_units.edit_button = gtk_button_new_with_label("Edit ...");
	gtk_box_pack_start( GTK_BOX(hbutbox), wnd_units.edit_button, FALSE, FALSE, 0 );
	wnd_units.delete_button = gtk_button_new_with_label("Delete");
	gtk_box_pack_start( GTK_BOX(hbutbox), wnd_units.delete_button, FALSE, FALSE, 0 );
	wnd_units.default_button = gtk_button_new_with_label("Set as default");
	gtk_box_pack_start( GTK_BOX(hbutbox), wnd_units.default_button, FALSE, FALSE, 0 );

	g_signal_connect( G_OBJECT(new_button), "clicked", G_CALLBACK(on_new_unit), NULL );
	g_signal_connect( G_OBJECT(wnd_units.edit_button), "clicked", G_CALLBACK(on_edit_unit), NULL );
	g_signal_connect( G_OBJECT(wnd_units.delete_button), "clicked", G_CALLBACK(on_delete_unit), NULL );
	g_signal_connect( G_OBJECT(wnd_units.default_button), "clicked", G_CALLBACK(on_set_default_unit), NULL );

	g_signal_connect( G_OBJECT(wnd_units.window), "delete_event", G_CALLBACK(gtk_widget_destroy), NULL );
	g_signal_connect( G_OBJECT(wnd_units.window), "destroy", G_CALLBACK(destroy_unit_window), NULL);

	gtk_table_attach( GTK_TABLE(table), wnd_units.statusbar, 0, 2, 3, 4, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0 );

	gtk_widget_show_all( GTK_WIDGET(wnd_units.window) );
	wnd_units.visible = 1;

	unit_cursor_changed( GTK_TREE_VIEW(wnd_units.unit_view), NULL );
}

static void new_or_edit_unit_dialog( uns * unit )
{
	GtkWidget * dialog = gtk_dialog_new_with_buttons( unit ? "Edit unit" : "Add new unit",
		GTK_WINDOW(wnd_units.window),
		GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR,
		GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
	gtk_dialog_set_default_response( GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT );

	GtkBox * main_vbox = GTK_BOX(gtk_vbox_new( FALSE, 5 ));
	gtk_container_add( GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), GTK_WIDGET(main_vbox) );

	GtkTable * table = GTK_TABLE(gtk_table_new( 3, 2, FALSE ));
	gtk_container_set_border_width( GTK_CONTAINER(table), 5 );
	gtk_box_pack_start( main_vbox, GTK_WIDGET(table), TRUE, TRUE, 5 );

	gtk_table_attach( table, gtk_label_new("Name:"), 0, 1, 0, 1, GTK_FILL, GTK_FILL, 5, 0 );
	GtkWidget * name_entry = gtk_entry_new();
	if (unit)
		gtk_entry_set_text( GTK_ENTRY(name_entry), unit_array[*unit].name->text );

	gtk_table_attach( table, name_entry, 1, 2, 0, 1, GTK_FILL, GTK_FILL, 0, 0 );
	gtk_table_attach( table, gtk_label_new("Quantity:"), 0, 1, 1, 2, GTK_FILL, GTK_FILL, 5, 10 );
	gtk_table_attach( table, gtk_label_new(gtk_combo_box_get_active_text(wnd_units.quantity_combo)),
		1, 2, 1, 2, GTK_FILL, GTK_FILL, 0, 10 );

	GtkBox * mult_box = GTK_BOX(gtk_hbox_new( FALSE, 0 ));
	gtk_table_attach( table, gtk_label_new("="), 0, 1, 2, 3, GTK_FILL, GTK_FILL, 5, 0 );
	gtk_table_attach( table, GTK_WIDGET(mult_box), 1, 2, 2, 3, GTK_FILL, GTK_FILL, 0, 0 );

	struct prop_item pi;
	pi.o = PROP_STORE_O(ps_recycler);
	pi.key = string_lookup_ref("edit-unit-multiplier");
	pi.type = PT_REAL;
	pi.subtype = example_subtypes[current_quantity];
	prop_value pv;
	if (unit)
		pv.r = unit_array[*unit].multiplier;
	else
		pv = prop_lookup( PROP_STORE_O(ps_recycler), pi.key, pi.type, (prop_value){ .r = 1.0 } );
	gui_prop_recycler_set( pi.key, pi.type, pi.subtype, PROP_UNIT_MAX, pv );

	prop_value_edit_create( &pi );
	prop_unit_edit_create( &pi );
	prop_sync( &pi );
	gtk_box_pack_start( mult_box, pi.value_edit, TRUE, TRUE, 0 );
	if ( pi.unit_edit )
		gtk_box_pack_start( mult_box, pi.unit_edit, TRUE, TRUE, 0 );

	gtk_widget_show_all(dialog);
	if ( gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT )
	{
		string name = string_lookup(gtk_entry_get_text(GTK_ENTRY(name_entry)));

		char errs[TRANS_ERR_SIZE];
		statusbar_msg( (struct window*)&wnd_units, " " );
		TRANS_BEGIN_MAIN( tlo_universum, unit ? "Edit unit" : "Add unit", errs )
			struct prop * prop = prop_find( pi.o, pi.key );
			if (!prop)
				trans_fail("Internal error.");
			if (unit && unit_array[*unit].name != name)
				unit_change_name(*unit, name);
			unit_set( name, prop_lookup_real( PROP_STORE_O(ps_recycler), pi.key, 1.0 ), current_quantity );
		TRANS_FAILED
			statusbar_err( (struct window*)&wnd_units, "Cannot add/change: %s", errs );
		TRANS_END
	}
	gtk_widget_destroy(dialog);
	string_unref(pi.key);
}

static void on_new_unit( GtkWidget * w UNUSED, gpointer data UNUSED )
{ new_or_edit_unit_dialog(NULL); }

static void on_edit_unit( GtkWidget * w UNUSED, gpointer data UNUSED )
{
	if (selected_unit >= 0)
	{
		uns u = (uns)selected_unit;
		new_or_edit_unit_dialog(&u);
	}
}

static void on_set_default_unit( GtkWidget * w UNUSED, gpointer data UNUSED )
{
	if (selected_unit < 0)
		return;
	
	char errs[TRANS_ERR_SIZE];
	statusbar_msg( (struct window*)&wnd_units, " " );
	TRANS_BEGIN_MAIN( tlo_universum, "Set unit as default", errs )
		unit_default_set( unit_array[selected_unit].quantity, selected_unit );
	TRANS_FAILED
		statusbar_err( (struct window*)&wnd_units, "Cannot set default: %s", errs );
	TRANS_END
}

static void on_delete_unit( GtkWidget * w UNUSED, gpointer data UNUSED )
{
	if (selected_unit < 0)
		return;
	
	char errs[TRANS_ERR_SIZE];
	statusbar_msg( (struct window*)&wnd_units, " " );
	TRANS_BEGIN_MAIN( tlo_universum, "Delete unit", errs )
		unit_delete( selected_unit );
	TRANS_FAILED
		statusbar_err( (struct window*)&wnd_units, "Cannot delete: %s", errs );
	TRANS_END
}

static void unit_cursor_changed( GtkTreeView * t UNUSED, gpointer data UNUSED )
{
	GtkTreePath * tp;
	GtkTreeIter selected_row;
	gtk_tree_view_get_cursor( GTK_TREE_VIEW(wnd_units.unit_view), &tp, NULL );

	if (!tp)
		selected_unit = -1;
	else
	{
		if ( gtk_tree_model_get_iter( GTK_TREE_MODEL(*(unit_quant_lists[current_quantity])), &selected_row, tp ) )
		{
			SAFE_G_VALUE(index);
			gtk_tree_model_get_value( GTK_TREE_MODEL(*(unit_quant_lists[current_quantity])),
				&selected_row, UNIT_INDEX_COL, &index );
			selected_unit = g_value_get_int(&index);
			g_value_unset( &index );
		}
		gtk_tree_path_free(tp);
	}

	DBG("Selected unit = %d\n", selected_unit );
	if (selected_unit >= 0)
	{
		int is_default = (uns)selected_unit == unit_default[unit_array[selected_unit].quantity];
		gtk_widget_set_sensitive( wnd_units.default_button, !is_default );
		gtk_widget_set_sensitive( wnd_units.delete_button, !is_default );
		gtk_widget_set_sensitive( wnd_units.edit_button, TRUE );
	} else {
		gtk_widget_set_sensitive( wnd_units.delete_button, FALSE );
		gtk_widget_set_sensitive( wnd_units.default_button, FALSE );
		gtk_widget_set_sensitive( wnd_units.edit_button, FALSE );
	}
}

static void uw_unit_added(uns slot, struct unit_hook *h )
{ uw_unit_changed(slot, h); }

#define WALK_TREE_MODEL_START(_tm, _tmm, _ti, _unit) \
	GtkTreeModel * _tmm = _tm; \
	GtkTreeIter _ti; \
	int result, _unit; \
	SAFE_G_VALUE(index); \
	for ( result = gtk_tree_model_get_iter_first(_tm, &_ti); result; result = gtk_tree_model_iter_next(_tm, &_ti) ) \
	{ \
		gtk_tree_model_get_value(_tm, &_ti, UNIT_INDEX_COL, &index ); \
		_unit = g_value_get_int(&index); \
		g_value_unset( &index ); \
		ASSERT(_unit >= 0);

#define WALK_TREE_MODEL_END }

static void uw_unit_deleted(uns slot, struct unit_hook *h UNUSED )
{
	GtkTreeIter ti;
	GtkListStore * ls = *(unit_quant_lists[unit_array[slot].quantity]);
	if (unit_to_tree_iter( GTK_TREE_MODEL(ls), slot, &ti))
		return;
	gtk_list_store_remove( ls, &ti );
	unit_cursor_changed( GTK_TREE_VIEW(wnd_units.unit_view), NULL );
}
static void uw_unit_changed(uns slot, struct unit_hook *h UNUSED )
{
	GtkTreeIter ti;
	GtkListStore * ls = *(unit_quant_lists[unit_array[slot].quantity]);
	if (unit_to_tree_iter( GTK_TREE_MODEL(ls), slot, &ti))
	{
		// unit not found -- it was probably added or moved from another quantity
		// search other lists and remove the deleted instance, if any
		for (uns u = 0; u < PQ_MAX; u++)
		{
			GtkListStore * ls2 = *unit_quant_lists[u];
			if (!unit_to_tree_iter( GTK_TREE_MODEL(ls2), slot, &ti))
			{
				gtk_list_store_remove( ls2, &ti );
				break;
			}
		}
		gtk_list_store_append( ls, &ti );
	}
	gtk_list_store_set( ls, &ti,
		UNIT_NAME_COL, unit_array[slot].name->text,
		UNIT_MULT_COL, (float)unit_array[slot].multiplier,
		UNIT_INDEX_COL, slot,
		UNIT_WEIGHT_COL, ((uns)slot == unit_default[unit_array[slot].quantity]) ? 1000 : 400,
		-1 );
	unit_cursor_changed( GTK_TREE_VIEW(wnd_units.unit_view), NULL );
}
static void uw_unit_default_changed(uns quantity, struct unit_hook *h UNUSED )
{
	WALK_TREE_MODEL_START( GTK_TREE_MODEL(*(unit_quant_lists[quantity])), tm, ti, unit )
		gtk_list_store_set( GTK_LIST_STORE(tm), &ti,
			UNIT_WEIGHT_COL, ((uns)unit == unit_default[quantity]) ? 1000 : 400, -1 );
	WALK_TREE_MODEL_END
	unit_cursor_changed( GTK_TREE_VIEW(wnd_units.unit_view), NULL );
}

int tree_iter_to_unit( GtkTreeModel * tm, GtkTreeIter * ti, uns * unit )
{
	SAFE_G_VALUE(index);
	gtk_tree_model_get_value(tm, ti, UNIT_INDEX_COL, &index );
	int iu = g_value_get_int(&index);
	*unit = (uns)iu;
	g_value_unset( &index );
	return (iu < 0);
}

int unit_to_tree_iter( GtkTreeModel * tm, uns unit, GtkTreeIter * ti )
{
	int success = 0;
	WALK_TREE_MODEL_START( tm, tmm UNUSED, ti2, u2 )
		if ( unit == (uns)u2 )
		{
			success = 1;
			*ti = ti2;
			break;
		}
	WALK_TREE_MODEL_END
	return !success;
}
