/*
 * GUI - Vertical ruler
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "lib/lib.h"
#include "gui/ruler.h"
#include <string.h>

#define RULER_WIDTH           14

#define ROUND(x) ((int) ((x) + 0.5))

static void vert_ruler_class_init    (VertRulerClass *klass);
static void vert_ruler_init          (VertRuler      *vert_ruler);
int vert_ruler_motion_notify (GtkWidget      *widget,
				      GdkEventMotion *event);
static void vert_ruler_draw_ticks    (Ruler       *ruler);
static void vert_ruler_draw_pos      (Ruler       *ruler);
static gint vert_ruler_expose (GtkWidget      *widget,
		                  GdkEventExpose *event UNUSED);


GtkType vert_ruler_get_type (void)
{
  static GtkType vert_ruler_type = 0;

  if (!vert_ruler_type)
    {
      static const GtkTypeInfo vert_ruler_info = {
	"VertRuler",
	sizeof(VertRuler),
	sizeof(VertRulerClass),
	(GtkClassInitFunc) vert_ruler_class_init,
	(GtkObjectInitFunc) vert_ruler_init,
	0, 0, 0 };

      vert_ruler_type = gtk_type_unique (ruler_get_type(),
		      &vert_ruler_info);
    }
    DBG("Vert_ruler gettype finished\n");

  return vert_ruler_type;
}

static void vert_ruler_class_init (VertRulerClass *klass)
{
  GtkWidgetClass *widget_class;
  RulerClass *ruler_class;

  widget_class = (GtkWidgetClass *) klass;
  ruler_class = (RulerClass *) klass;

  widget_class->motion_notify_event = vert_ruler_motion_notify;
  widget_class->expose_event = vert_ruler_expose;

  ruler_class->draw_ticks = vert_ruler_draw_ticks;
  ruler_class->draw_pos = vert_ruler_draw_pos;
}

static void vert_ruler_init (VertRuler *vert_ruler)
{
  GtkWidget *widget;

  widget = GTK_WIDGET (vert_ruler);
  widget->requisition.width = widget->style->xthickness * 2 + RULER_WIDTH;
  widget->requisition.height = widget->style->ythickness * 2 + 1;
}

GtkWidget* vert_ruler_new (void)
{
  return GTK_WIDGET (gtk_type_new(vert_ruler_get_type()));
}


int vert_ruler_motion_notify (GtkWidget      *widget,
			  GdkEventMotion *event)
{
  Ruler *ruler;
  gint y;

  ruler = RULER (widget);

  if (event->is_hint)
    gdk_window_get_pointer (widget->window, NULL, &y, NULL);
  else
    y = event->y;

  ruler->position = y;
  g_object_notify (G_OBJECT (ruler), "position");

  /*  Make sure the ruler has been allocated already  */
  if (ruler->backing_store)
    ruler_draw_pos (ruler);

  return FALSE;
}

static void print_label(Ruler *ruler, PangoLayout *layout, double pos, int digit_height, char *text, int digit_offset)
{
	int i, tmp;
	PangoRectangle logical_rect;
	GtkWidget *widget = GTK_WIDGET (ruler);
	int xthickness;

    xthickness = widget->style->xthickness;
	tmp = strlen (text);
	for (i = 0; i < tmp; i++)
 	{
	    pango_layout_set_text (layout, text + i, 1);
    	  pango_layout_get_extents (layout, NULL, &logical_rect);
	      gtk_paint_layout (widget->style,
                    ruler->backing_store,
                    GTK_WIDGET_STATE (widget),
                    FALSE,
                    NULL,
                    widget,
                    "vert_ruler",
					xthickness,
					pos + digit_height * i + 2 + PANGO_PIXELS (logical_rect.y - digit_offset),
                    layout);
    }
}

static double draw_line (Ruler *ruler, GdkGC *gc, double cur, int width)
{
	double pos;
	int xthickness;
	GtkWidget *widget = GTK_WIDGET (ruler);
	int ruler_width;

    xthickness = widget->style->xthickness;
	ruler_width = widget->allocation.width - 2 * xthickness;
	pos = ROUND (ruler_coords_to_pixels (ruler, cur));
    gdk_draw_line (ruler->backing_store, gc,
              ruler_width, pos, ruler_width - width, pos);

	return pos;
}

static void vert_ruler_draw_ticks (Ruler *ruler)
{
  GtkWidget *widget;
  GdkGC *gc, *bg_gc;
  gint width, height;
  gint ruler_width;
  gint ythickness, xthickness;
  gint digit_height;
  gint digit_offset;
  PangoLayout *layout;
  PangoRectangle logical_rect, ink_rect;

  if (!GTK_WIDGET_DRAWABLE (ruler))
    return;

  widget = GTK_WIDGET (ruler);

  gc = widget->style->fg_gc[GTK_STATE_NORMAL];
  bg_gc = widget->style->bg_gc[GTK_STATE_NORMAL];

  ythickness = widget->style->ythickness;
  xthickness = widget->style->xthickness;

  layout = gtk_widget_create_pango_layout (widget, "012456789");
  pango_layout_get_extents (layout, &ink_rect, &logical_rect);

  digit_height = PANGO_PIXELS (ink_rect.height) + 2;
  digit_offset = ink_rect.y;

  height = widget->allocation.height;
  width = widget->allocation.width - ythickness * 2;
  ruler_width = widget->allocation.width;

  gtk_paint_box (widget->style, ruler->backing_store,
		 GTK_STATE_NORMAL, GTK_SHADOW_OUT,
		 NULL, widget, "vert_ruler",
		 0, 0,
		 widget->allocation.width, widget->allocation.height);

  gdk_draw_line (ruler->backing_store, gc,
		 ruler_width - 2 * xthickness,
		 ythickness,
		 ruler_width - 2 * xthickness,
		 widget->allocation.height - ythickness);

  compute_and_draw_ticks (ruler, layout, digit_height, height, width, draw_line, print_label, digit_offset);

  g_object_unref (layout);
}

static gint vert_ruler_expose (GtkWidget      *widget,
		                  GdkEventExpose *event)
{
	ruler_expose(widget, event);
	return 0;
}

static void vert_ruler_draw_pos (Ruler *ruler)
{
  GtkWidget *widget;
  GdkGC *gc;
  int i;
  gint x, y;
  gint width, height;
  gint bs_width, bs_height;
  gint xthickness;
  gint ythickness;
  gdouble increment;

  if (GTK_WIDGET_DRAWABLE (ruler))
    {
      widget = GTK_WIDGET (ruler);

      gc = widget->style->fg_gc[GTK_STATE_NORMAL];
      xthickness = widget->style->xthickness;
      ythickness = widget->style->ythickness;
      width = widget->allocation.width - xthickness * 2;
      height = widget->allocation.height;

      bs_height = width / 2;
      bs_height |= 1;  /* make sure it's odd */
      bs_width = bs_height / 2 + 1;

      if ((bs_width > 0) && (bs_height > 0))
	{
	  /*  If a backing store exists, restore the ruler  */
	  if (ruler->backing_store && ruler->non_gr_exp_gc)
	    gdk_draw_drawable (ruler->widget.window,
			       ruler->non_gr_exp_gc,
			       ruler->backing_store,
			       ruler->xsrc, ruler->ysrc,
			       ruler->xsrc, ruler->ysrc,
			       bs_width, bs_height);

	  increment = ruler->resolution;

	  x = (width + bs_width) / 2 + xthickness - 1;
	  y = ruler -> position + (ythickness - bs_height) / 2 - 1;

	  for (i = 0; i < bs_width; i++)
	    gdk_draw_line (widget->window, gc,
			   x + i, y + i,
			   x + i, y + bs_height - i - 1);

	  ruler->xsrc = x;
	  ruler->ysrc = y;
	}
    }
}

