/** @file
 * GUI - Implemetation of the View window and the View navigator window
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#undef LOCAL_DEBUG
#include "lib/lib.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <ctype.h>

#include "kernel/kernel.h"

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "gui/main.h"
#include "gui/cmdmgr.h"
#include "gui/ruler.h"
#include "gui/visualisation.h"
#include "gui/properties.h"
#include "scheme/glt_gui.h"

#include "vcl/vcl.h"

#define PAGE_INCREMENT_X 5
#define PAGE_INCREMENT_Y 5

GtkWidget * view_menu;
GSList * view_list;

/** Refresh the zoom and rotation label */
static void set_zoom_label( struct view * vw, real zoom, real rotation );
/** Set the current view into context and make its snap tolerance value
 * the current one. */
static void set_view_as_active( struct view * vw );
/** Make sure that there is a not yet active operation of the GO factory
 * for the view's tlo and group */
static void ensure_op(void);
/** A callback to handle all mouse button press events. Sets the view to the context. */
static gboolean button_press_callback(GtkWidget *widget UNUSED, GdkEventButton *event, struct view *vw);
/** A callback to handle all mouse button release events. Sets the view to the context. */
static gboolean button_release_callback(GtkWidget *widget UNUSED, GdkEventButton *event, struct view *vw);
/** A callback to handle all key press events. Sets the view to the context. */
static gboolean key_press_callback(GtkWidget *widget UNUSED, GdkEventKey *event, struct view * vw);

/** A hook handler for selection and other changes of objects. */
static void view_sel_changed_2(struct go *go, const struct changed_data_generic *data UNUSED, struct go_hook* h);
static void view_obj_before_delete( struct o* o, struct o * owner UNUSED, struct o * prev UNUSED, struct obj_hook *hook );
static void view_go_before_delete( struct o* o, struct o* owner UNUSED, struct o* prev UNUSED, struct go_hook *hook );
static void view_obj_prop_change( struct o * o, struct obj_hook *hook );

static void create_view_navigator(struct view * vw);
static void close_view_navigator(struct view * vw);
static void view_nav_update( struct view_navigator * vwn, real zoom, real rotation );
static void view_nav_redraw( struct view_navigator * vwn );
static void view_nav_resize( GtkWidget *w UNUSED, GdkRectangle *a UNUSED, struct view_navigator *vwn );
static void view_nav_value_changed( GtkWidget * w UNUSED, struct view_navigator * vwn );
static gboolean view_nav_motion(GtkWidget *widget UNUSED, GdkEventMotion *event, struct view_navigator *vwn);
static gboolean view_nav_button_press(GtkWidget *widget UNUSED, GdkEventButton *event, struct view_navigator *vwn);
static int change_in_progress;

/* WINDOW AND CALLBACKS */

static void view_recompute_tolerance( struct view * vw )
{
	struct geom_point pi0, pit, po0, pot;
	pi0.x = 0; pi0.y = 0;
	pit.x = factory.snap_tol_pix; pit.y = 0;
	visualisation_pixels_to_coords( &vw->vis, &pi0, &po0 );
	visualisation_pixels_to_coords( &vw->vis, &pit, &pot );
	vw->tolerance = hypot( po0.x - pot.x, po0.y - pot.y );
	if ( (struct window*)vw == context.w )
		factory.snap_tolerance = vw->tolerance;
	DBG("Recomputed tolerance, %d pixels = %.3f coords.\n", factory.snap_tol_pix, vw->tolerance );
}

static void update_rulers (struct view *vw)
{
	struct geom_vector px, py;
	struct geom_transform t = geom_transform_identity;

	vcl_inverted_apply_matrix (vw -> vis . area, &t);
	px . dx = t.coef[0][0]; px . dy = t.coef[1][0];
	py . dx = t.coef[0][1]; py . dy = t.coef[1][1];

	ruler_set_resolution (RULER (vw -> hruler),  1 / hypot (px . dx, px . dy) *
		prop_lookup_real( obj2o(obj_universum), string_rulerx_res, RULERX_RES) );
	ruler_set_resolution (RULER (vw -> vruler), -1 / hypot (py . dx, py . dy) *
		prop_lookup_real( obj2o(obj_universum), string_rulery_res, RULERY_RES) );
}

static void visualisation_changed (struct view *vw)
{
	GtkWidget *visw = vw->vis.canvas->gtk->widget;
	gint width = visw->allocation.width;
	gint height = visw->allocation.height;
	struct geom_rectangle bbox1, bbox2, bbox3;

	struct geom_transform t = geom_transform_identity;
        vcl_primary_apply_matrix(vw->vis.area, &t);

        struct geom_point po = { t.coef[0][2], t.coef[1][2] };

	// FIXME: possible numeric errors
	visualisation_get_visible_bbox (&(vw -> vis), &bbox1);
	geom_rectangle_transform (&bbox1, &t, &bbox2);
	bbox1 . ll . x = bbox1 . ur . x = width / 2.0;
	bbox1 . ll . y = bbox1 . ur . y = height / 2.0;
	geom_rectangle_merge (&bbox1, &bbox2, &bbox3);

	GtkAdjustment * adj = vw -> hadj;
	adj -> lower = bbox3 . ll . x - width / 2.0 - po . x;
	adj -> upper = bbox3 . ur . x + width / 2.0 - po . x;
	adj -> value = - po . x;
	gtk_widget_queue_draw (vw -> hscrollbar);
	ruler_set_offset (RULER (vw -> hruler), po . x);

	adj = vw -> vadj;
	adj -> lower = bbox3 . ll . y - height / 2.0 - po . y;
	adj -> upper = bbox3 . ur . y + height / 2.0 - po . y;
	adj -> value = - po . y;
	gtk_widget_queue_draw (vw -> vscrollbar);
	ruler_set_offset (RULER (vw -> vruler), po . y);

	//calculate the zoom and rotation
	struct geom_point ux = { 1.0, 0.0 }, zero = { 0.0, 0.0 };
	visualisation_coords_to_vcoords( &vw->vis, &ux, &ux );
	visualisation_coords_to_vcoords( &vw->vis, &zero, &zero );
	real xzoom = hypot( ux.x - zero.x , ux.y - zero.y ) * 100.0;
	real xrot = (real)atan2( (double)(ux.y - zero.y), (double)(ux.x - zero.x) ) / R_PI * 180.0;
	set_zoom_label( vw, xzoom, xrot );

	if ( vw->navigator )
	{ view_nav_update( vw->navigator, xzoom, xrot ); view_nav_redraw( vw->navigator ); }

	view_recompute_tolerance(vw);
	transform_recalc_dimen(vw);
	update_rulers (vw);
}

static void view_size_allocate (GtkWidget *widget UNUSED, GdkRectangle *allocation UNUSED, struct view *vw)
{
	GtkWidget *visw = vw->vis.canvas->gtk->widget;
	gint height = visw -> allocation . height;
	gint width = visw -> allocation . width;

	GtkAdjustment * adj = vw -> hadj;
	real delta = width - adj -> page_size;

	adj -> page_size = width;
	adj -> lower -= delta / 2;
	adj -> value -= delta / 2;
	adj -> upper += delta / 2;

	if (!IS_ZERO(delta))
	{
		gtk_widget_queue_draw (vw -> hscrollbar);
		ruler_set_offset (RULER (vw -> hruler), - adj -> value);
	}

	adj = vw -> vadj;
	delta = height - adj -> page_size;

	adj -> page_size = height;
	adj -> lower -= delta / 2;
	adj -> value -= delta / 2;
	adj -> upper += delta / 2;

	if (!IS_ZERO(delta))
	{
		gtk_widget_queue_draw (vw -> vscrollbar);
		ruler_set_offset (RULER (vw -> vruler), - adj -> value);
	}
	if ( vw->navigator ) view_nav_redraw( vw->navigator );
}

/* SCALE/MOVE/ZOOM */

void on_view_fit_selected( void * subj )
{
	ASSERT(subj && ((struct window*)subj)->type == WT_VIEW);
	struct view * vw = (struct view*)subj;
	struct geom_rectangle t;
	visualisation_get_selected_bbox( &vw->vis, &t );
	DBG("Selected bbox obtained: [ %.3f, %.3f ] [ %.3f %.3f ]\n", t.ll.x, t.ll.y, t.ur.x, t.ur.y );
	visualisation_vcoords_to_coords( &vw->vis, &t.ll, &t.ll );
	visualisation_vcoords_to_coords( &vw->vis, &t.ur, &t.ur );
	DBG("Selected coords obtained: [ %.3f, %.3f ] [ %.3f %.3f ]\n", t.ll.x, t.ll.y, t.ur.x, t.ur.y );
	if (visualisation_fit_box( &vw->vis, t ))
		statusbar_msg( (struct window*)vw, "The selected bbox is empty.");
}

void on_view_fit_all( void * subj )
{
	ASSERT(subj && ((struct window*)subj)->type == WT_VIEW);
	struct view * vw = (struct view*)subj;
	struct geom_rectangle t;
	visualisation_get_visible_bbox( &vw->vis, &t );
	DBG("Visible bbox obtained: [ %.3f, %.3f ] [ %.3f %.3f ]\n", t.ll.x, t.ll.y, t.ur.x, t.ur.y );
	if (visualisation_fit_box( &vw->vis, t ))
		statusbar_msg( (struct window*)vw, "The image is empty.");
}

static void view_zoom_in( GtkWidget * w UNUSED, struct view * vw )
{ visualisation_scale( &vw->vis, 1.25, 1.25 ); }

void on_view_zoom_in( void * subj )
{
	ASSERT(subj && ((struct window*)subj)->type == WT_VIEW);
	view_zoom_in( NULL, (struct view*)subj );
}
static void view_zoom_out( GtkWidget * w UNUSED, struct view * vw )
{ visualisation_scale( &vw->vis, 0.8, 0.8 ); }

void on_view_zoom_out( void * subj )
{
	ASSERT(subj && ((struct window*)subj)->type == WT_VIEW);
	view_zoom_out( NULL, (struct view*)subj );
}
static void view_zoom_reset( GtkWidget * w UNUSED, struct view * vw )
{ visualisation_set_orientation( &vw->vis, 0.0, 0.0, 1.0, 1.0, 0.0 ); }

void on_view_zoom_reset( void * subj )
{
	ASSERT(subj && ((struct window*)subj)->type == WT_VIEW);
	view_zoom_reset( NULL, (struct view*)subj );
}
static void view_move_up( GtkWidget * w UNUSED, struct view * vw )
{ visualisation_move( &vw->vis, 0, -PAGE_INCREMENT_Y ); }

static void view_move_down( GtkWidget * w UNUSED, struct view * vw )
{ visualisation_move( &vw->vis, 0, PAGE_INCREMENT_Y ); }

static void view_move_left( GtkWidget * w UNUSED, struct view * vw )
{ visualisation_move( &vw->vis, PAGE_INCREMENT_X, 0 ); }

static void view_move_right( GtkWidget * w UNUSED, struct view * vw )
{ visualisation_move ( &vw->vis, -PAGE_INCREMENT_X, 0 ); }

static void view_rotate_plus( GtkWidget * w UNUSED, struct view * vw )
{ visualisation_rotate( &vw->vis, 0.2 ); }

static void view_rotate_minus( GtkWidget * w UNUSED, struct view * vw )
{ visualisation_rotate( &vw->vis, -0.2 ); }

static void scrollbar_value_changed (GtkAdjustment *x UNUSED, struct view *vw)
{ visualisation_absolut_move_px (&(vw->vis), -vw->hadj->value, -vw->vadj->value); }

/* MOUSE CALLBACKS */

static gboolean scroll_callback(GtkWidget *widget UNUSED, GdkEventScroll *event, struct view *vw)
{
	switch ( event->direction )
	{
		case GDK_SCROLL_UP:	view_zoom_in( NULL, vw ); break;
		case GDK_SCROLL_DOWN:	view_zoom_out( NULL, vw ); break;
		default: break;
	}
	return TRUE;
}

static gboolean motion_notify_callback(GtkWidget *widget UNUSED, GdkEventMotion *event, struct view *vw)
{
	vert_ruler_motion_notify(vw->vruler, event);
	horiz_ruler_motion_notify(vw->hruler, event);

	//TODO: coords label
	
	view_moving_motion(event, vw);
	return TRUE;
}

static gboolean key_press_callback(GtkWidget *widget UNUSED, GdkEventKey *event, struct view * vw)
{
	set_view_as_active(vw);

	if ( vw->tlo != factory.tlo || factory.start_state != &ofs_select ) return FALSE;

	if ( event->keyval == GDK_Shift_L || event->keyval == GDK_Shift_R )
		transform_set_shift( event->type == GDK_KEY_PRESS, vw );

	return FALSE;
}

static gboolean ruler_button_press_callback (GtkWidget *widget UNUSED, GdkEventButton *event, struct view * vw)
{
	set_view_as_active(vw);

	if ( event->button == 3 ) //open menu
	{
		view_refresh_menu();
		gtk_menu_popup( GTK_MENU(view_menu), NULL, NULL, NULL, NULL, event->button, gtk_get_current_event_time() );
	}
	return TRUE;
}

static gboolean button_release_callback(GtkWidget *widget UNUSED, GdkEventButton *event, struct view *vw)
{ view_moving_release(event, vw); return TRUE; }

static gboolean button_press_callback(GtkWidget *widget UNUSED, GdkEventButton *event, struct view *vw)
{
	struct geom_point pi = {event -> x, event -> y};
	struct geom_point po;
	view_clicked( event, vw );

	visualisation_pixels_to_coords (&(vw -> vis), &pi, &po);
	DBG("%f %f\n", po . x, po . y);
	set_view_as_active(vw);

	if ( event->button == 1 ) switch ( factory.state->input_kind )
	{
	case OFIK_NONE: BUG();

	case OFIK_DPR:
		ensure_op();
		switch ( factory.state->input_type )
		{
		case DIST_PASS_HANGER:
			snap_point( &po, &factory.state->value.dpr, &vw->grid, context.tlo, vw->tolerance );
			factory_op_step();
			break;
	 	case DIST_PASS_ANCHOR:
		{
			if ( moving_mode != MOVE_NONE ) return TRUE; // prevent from misbehaviour in case of a double-click
			snap_to_anchor( &po, &factory.state->value.dpr, context.tlo, vw->tolerance );
			if ( factory.state->value.dpr.type == DIST_PASS_ANCHOR )
				factory_op_step();
			else
				//select
				view_selection_start( po, pi, vw ); 
			break;
		}
		case DIST_PASS_POINT:
			factory.state->value.dpr.type = DIST_PASS_POINT;
			factory.state->value.dpr.data.point.point = po;
			factory_op_step();
			break;
		case DIST_PASS_GO:
		{
			snap_to_go( &po, &factory.state->value.dpr, context.tlo, vw->tolerance );
			if ( factory.state->value.dpr.type == DIST_PASS_GO )
			{
				go_ref(factory.state->value.dpr.data.go.go);
				factory_op_step();
			}
			break;
		}
		}
		break;

	case OFIK_PROP:
		statusbar_err(NULL, "Do not click, enter the number below.");
		break;

	case OFIK_SEL:
		if ( moving_mode != MOVE_NONE ) return TRUE; // prevent from misbehaviour in case of a double-click
		ensure_op();
		if (!view_transform_start( event, po, pi, vw ))
			view_selection_start( po, pi, vw );
		break;

	case OFIK_TF:
		if ( moving_mode != MOVE_NONE ) return TRUE; // prevent from misbehaviour in case of a double-click
		ensure_op();
		view_s_transform_start( event, po, pi, vw );		
		break;

	case OFIK_TRANSFORM: break;
	default: BUG();
	}
	if ( moving_mode != MOVE_NONE ) return TRUE; // for sure

	if ( event->button == 2 )
	{
		if ( event->state & GDK_CONTROL_MASK )
		{
			snap_point( &po, NULL, &vw->grid, context.tlo, vw->tolerance );
			visualisation_center( &vw->vis, po.x, po.y );
		} else
			view_panning_start( pi, vw );
	}

	if ( event->button == 3 ) //open menu
	{
		view_refresh_menu();
		gtk_menu_popup( GTK_MENU(view_menu), NULL, NULL, NULL, NULL, event->button, gtk_get_current_event_time() );
	}
	return TRUE;
}

static void set_view_as_active( struct view * vw )
{
	if ( context.w != (struct window*)vw )
		change_context( (struct window*)vw, vw->group,
			(context.o && context.o->kind == T_GO && go_get_group(o2go(context.o)) == vw->group) ?
			context.o : NULL );

	factory.snap_tolerance = vw->tolerance;
}
static void ensure_op(void)
{
	if ( factory.group != context.group )
	{
		if ( factory.flags & OFF_OPERATION_ACTIVE ) factory_op_break();
		factory_op_start( context.group, factory.start_state, factory.desc->text );
	}
}

static void button_navigate_pressed(GtkWidget *button UNUSED, struct view *vw)
{
	if ( vw->navigator ) gtk_window_present( GTK_WINDOW(vw->navigator->window) );
	else create_view_navigator(vw);
}

static void set_zoom_label( struct view * view, real zoom, real rotation )
{
	char buf[MAX_ERR_MSG_LENGTH]; //FIXME: staci to?
	if (!GTK_IS_LABEL(view->zoombar)) return;
	sprintf( buf, " %.1f%%, %.1f* ", zoom, rotation );
	gtk_label_set_text( GTK_LABEL(view->zoombar), buf );
}

static void set_window_title(GtkWindow *window, struct obj_tlo * tlo, struct go_group * group)
{
	ASSERT(tlo && group);
	uns pos = 0, len = 0;
	struct o * o = (tlo_get_tlg(tlo) == group) ? void2o(tlo) : go2o(&group->go);
	struct o * o3 = o, * o2 = o_get_gui_ancestor(o);
	for ( ; o2; o3 = o, o = o2, o2 = o_get_gui_ancestor(o2) ) len += strlen(o_lookup_name(o)) + 3;
	//get the filename of the highest obj under the universum
	const char * filename = prop_lookup_text( o3, string_filename );
	if (!filename) filename = "";
	//assemble the title
	char buf[len + 5 + strlen(filename)];
	o = (tlo_get_tlg(tlo) == group) ? void2o(tlo) : go2o(&group->go);
	o2 = o_get_gui_ancestor(o);
	for ( ; o2; o = o2, o2 = o_get_gui_ancestor(o2) )
	{
		const char * name = o_lookup_name(o);
		sprintf( buf + pos, pos ? " - %s" : "%s", name );
		pos += strlen(name) + (pos ? 3 : 0);
	}
	sprintf( buf + pos, " (%s)", filename );
	gtk_window_set_title (window, buf);
}
static gboolean view_delete_event( GtkWidget *widget UNUSED, GdkEvent * event UNUSED, struct view *vw )
{
	change_context(NULL, NULL, NULL);
	factory_op_break();

	if (vw->fifi)
		visualisation_gui_fifi_destroy( &vw->vis, &vw->fifi );
	view_moving_release(NULL, vw); //end the move, if any
        visualisation_finish (&(vw -> vis)); // finish visualisation after that
	// (moving might have allocated some visualisation objects)

	if ( vw->navigator )
		close_view_navigator(vw);

	return FALSE;
}
void close_view( struct view * vw )
{
	view_delete_event( NULL, NULL, vw );
	gtk_widget_destroy( vw->window );
}
static void view_destroy( GtkWidget *widget UNUSED, struct view *vw )
{
	go_hook_remove( vw->tlo, &vw->go_hook );
	obj_hook_remove( &vw->obj_hook );
	obj_unref( &vw->tlo->obj );
	expand_unref( vw->tlo );
	go_unref( &vw->group->go );
	//remove the view from the list of all views
	view_list = g_slist_remove( view_list, (gpointer)vw );

	win_o_finish((struct window*)vw);
	xfree(vw->pi);
	xfree(vw);
	//refresh_gui();
}

// create view window
struct view *create_view(struct go_group * group)
{
	GtkWidget *vscrollbarbox, *hscrollbarbox, *hbox;
	GtkRange *range;
	struct view *vw = xmalloc_zero(sizeof(struct view));
	vw->pi = xmalloc_zero(sizeof(struct prop_item));

	win_o_init( (struct window*)vw, WT_VIEW );

	vw->tf_drag_on = 0;
	vw->group = group;
	vw->tlo = go_get_tlo(&group->go);
	vw->navigator = NULL;
	obj_ref(&vw->tlo->obj);
	expand_ref(vw->tlo);
	go_ref(&vw->group->go);

	// set hooks after visualisation hooks (hooks are added to start of the list)
	vw->obj_hook.data = vw;
	vw->obj_hook.obj_unlink = view_obj_before_delete;
	vw->obj_hook.obj_prop_change = view_obj_prop_change;
	obj_hook_add( &vw->obj_hook );

	vw->go_hook.data = vw;
	vw->go_hook.go_changed = view_sel_changed_2;
	vw->go_hook.go_unlink = view_go_before_delete;
	go_hook_add( vw->tlo, &vw->go_hook );

	vw->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	set_window_title(GTK_WINDOW(vw->window), vw->tlo, vw->group);
	gtk_widget_set_size_request (vw->window, 600, 620);
	gtk_window_add_accel_group( GTK_WINDOW( vw->window ), accel_group );
	g_signal_connect (G_OBJECT (vw->window), "destroy", G_CALLBACK (view_destroy), vw);
	g_signal_connect (G_OBJECT (vw->window), "delete_event", G_CALLBACK (view_delete_event), vw);
	g_signal_connect_after (G_OBJECT (vw->window), "size_allocate", G_CALLBACK(view_size_allocate), vw);

	GtkWidget * table;
	gtk_container_add(GTK_CONTAINER(vw->window), vw->table = table = gtk_table_new(5, 5, FALSE) );

	vw->toolbar = NULL;
	vw->toolbar2 = NULL;
	view_refresh_toolbar(vw);

	// vertical ruler
	vw->vruler = vert_ruler_new();
	gtk_table_attach(GTK_TABLE(table), vw->vruler, 2, 3, 1, 2, GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
	g_signal_connect(G_OBJECT (vw->vruler), "button-press-event", G_CALLBACK (ruler_button_press_callback), vw);

	// box for vertical scrollbar and its buttons
	vscrollbarbox = gtk_vbox_new(FALSE,0);

	// button up
	GtkWidget * button = gtk_button_new();
	gtk_container_add(GTK_CONTAINER(button), image_get_from_icon(icon_arrow_up));
	gtk_box_pack_start(GTK_BOX(vscrollbarbox), button, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(view_move_up), vw);

	// vertical scrollbar
	vw->vadj = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 0, 1, 1, 1));
	vw->vscrollbar = gtk_vscrollbar_new(vw->vadj);
	gtk_box_pack_start(GTK_BOX(vscrollbarbox), vw->vscrollbar, TRUE, TRUE, 0);
	gtk_table_attach(GTK_TABLE(table), vscrollbarbox, 4, 5, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
	range = GTK_RANGE(vw->vscrollbar);
	range->has_stepper_a = range->has_stepper_d = FALSE;

	// button down
	button = gtk_button_new();
	gtk_container_add(GTK_CONTAINER(button), image_get_from_icon(icon_arrow_down));
	gtk_box_pack_start(GTK_BOX(vscrollbarbox), button, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(view_move_down), vw);

	// horizontal ruler
	vw->hruler = horiz_ruler_new();
	gtk_table_attach (GTK_TABLE(table), vw->hruler, 3, 4, 0, 1, GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
	g_signal_connect(G_OBJECT (vw->hruler), "button-press-event", G_CALLBACK (ruler_button_press_callback), vw);

	// box for horizontal scrollbar and ints buttons
	hscrollbarbox = gtk_hbox_new(FALSE, 0);

	// button left
	button = gtk_button_new();
	gtk_container_add(GTK_CONTAINER(button), image_get_from_icon(icon_arrow_left));
	gtk_box_pack_start(GTK_BOX(hscrollbarbox), button, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(view_move_left), vw);

	// horizontal scrollbar
	// TODO pagesize
	vw->hadj = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 0, 1, 1, 1));
	vw->hscrollbar = gtk_hscrollbar_new(vw->hadj);
	gtk_box_pack_start(GTK_BOX(hscrollbarbox), vw->hscrollbar, TRUE, TRUE, 0);
	range = GTK_RANGE(vw->hscrollbar);
	range->has_stepper_a = range->has_stepper_d = FALSE;

	// button right
	button = gtk_button_new();
	gtk_container_add(GTK_CONTAINER(button), image_get_from_icon(icon_arrow_right));
	gtk_box_pack_start(GTK_BOX(hscrollbarbox), button, FALSE, FALSE, 0);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(view_move_right), vw);

	gtk_table_attach (GTK_TABLE(table), hscrollbarbox, 3, 4, 2, 3, GTK_FILL, GTK_FILL, 0, 0);

	g_signal_connect (G_OBJECT(vw->vscrollbar), "value-changed", G_CALLBACK(scrollbar_value_changed), vw);
	g_signal_connect (G_OBJECT(vw->hscrollbar), "value-changed", G_CALLBACK(scrollbar_value_changed), vw);

	// button (top, right)
	button = gtk_button_new();
	gtk_container_add(GTK_CONTAINER(button), image_get_from_icon(icon_rotation_right));
	gtk_table_attach(GTK_TABLE(table), button, 4, 5, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(view_rotate_minus), vw);

	// button (top, left)
	button = gtk_button_new();
	gtk_container_add(GTK_CONTAINER(button), image_get_from_icon(icon_rotation_left));
	gtk_table_attach(GTK_TABLE(table), button, 2, 3, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(view_rotate_plus), vw);

	// button (bottom, right)
	button = gtk_button_new();
	gtk_container_add(GTK_CONTAINER(button), image_get_from_icon(icon_zoom_out));
	gtk_table_attach(GTK_TABLE(table), button, 4, 5, 2, 3, GTK_FILL, GTK_FILL, 0, 0);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(button_navigate_pressed), vw);

	/*// button (bottom, left)
	button = gtk_button_new();
	gtk_container_add(GTK_CONTAINER(button), image_get_from_icon(icon_zoom_in));
	gtk_table_attach(GTK_TABLE(table), button, 2, 3, 2, 3, GTK_FILL, GTK_FILL, 0, 0);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(view_zoom_out), vw);*/

	//the factory statusbar
	gtk_table_attach( GTK_TABLE(table), hbox = gtk_hbox_new( FALSE, 0 ), 0, 5, 3, 4, GTK_FILL, GTK_FILL, 0, 0 );
	vw->factory_statusbar = gtk_statusbar_new();
	gtk_statusbar_set_has_resize_grip( GTK_STATUSBAR(vw->factory_statusbar), FALSE );
	//gtk_box_pack_start( GTK_BOX(hbox), vw->coords_label = gtk_label_new("tady budou souradnice"), FALSE, TRUE, 0 );
	
	// status bars
	gtk_table_attach( GTK_TABLE(table), hbox = gtk_hbox_new( FALSE, 0 ), 0, 5, 4, 5, GTK_FILL, GTK_FILL, 0, 0 );
	gtk_box_pack_start( GTK_BOX(hbox), vw->statusbar, TRUE, TRUE, 0 );
	gtk_box_pack_start( GTK_BOX(hbox), vw->factory_statusbar, TRUE, TRUE, 0 );
	gtk_box_pack_start( GTK_BOX(hbox), vw->factory_box = gtk_hbox_new( FALSE, 0 ), FALSE, TRUE, 0 );
	gtk_box_pack_start( GTK_BOX(hbox), vw->image_on = gtk_image_new_from_pixbuf( icon_pencil_on ), FALSE, TRUE, 0 );
	gtk_box_pack_start( GTK_BOX(hbox), vw->image_modified = gtk_image_new_from_pixbuf( icon_modified ), FALSE, TRUE, 0 );
	gtk_box_pack_start( GTK_BOX(hbox), vw->zoombar = gtk_label_new(""), FALSE, TRUE, 0 );
	gtk_statusbar_set_has_resize_grip( GTK_STATUSBAR(vw->statusbar), FALSE );

	//add the view into the list of all views
	view_list = g_slist_prepend( view_list, (gpointer)vw );

	gtk_table_attach(GTK_TABLE(table), visualisation_init (&(vw -> vis), vw->group, (simple_cb)visualisation_changed, vw),
			 3, 4, 1, 2, GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
	gtk_widget_add_events(vw ->vis.canvas->gtk-> widget, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK);
	g_signal_connect(G_OBJECT (vw -> vis . canvas -> gtk -> widget), "motion-notify-event", G_CALLBACK (motion_notify_callback), vw);
	g_signal_connect(G_OBJECT (vw -> vis . canvas -> gtk -> widget), "button-press-event", G_CALLBACK (button_press_callback), vw);
	g_signal_connect(G_OBJECT (vw -> vis . canvas -> gtk -> widget), "button-release-event", G_CALLBACK (button_release_callback), vw);
	g_signal_connect(G_OBJECT (vw->window), "key-press-event", G_CALLBACK (key_press_callback), vw);
	g_signal_connect(G_OBJECT (vw->window), "key-release-event", G_CALLBACK (key_press_callback), vw);
	g_signal_connect(G_OBJECT (vw -> vis . canvas -> gtk -> widget), "scroll-event", G_CALLBACK (scroll_callback), vw);

	gtk_widget_show_all (vw->window);
	gtk_widget_set_size_request (vw->window, -1, -1);
	gtk_window_set_resizable (GTK_WINDOW (vw->window), TRUE);

	visualisation_changed(vw); //initialize the rulers and scrollbars
	on_view_fit_all( vw );
	change_context( (struct window*)vw, vw->group, NULL );
	ensure_op();

	int err = geom_transform2_set_basis( &vw->grid, 0.0, 0.0, GRID_X1, GRID_X2, GRID_Y1, GRID_Y2 );
	ASSERT(!err);
	view_update_grid(vw);
	transform_update(vw);
	return vw;
}

void view_update_grid( struct view * vw )
{
	struct o * o = obj2o(obj_universum);
	if ( factory.snap_set & SNAP_GRID )
	{
		real x1 = prop_lookup_real( o, string_grid_x1, GRID_X1 );
		real y1 = prop_lookup_real( o, string_grid_y1, GRID_Y1 );
		real x2 = prop_lookup_real( o, string_grid_x2, GRID_X2 );
		real y2 = prop_lookup_real( o, string_grid_y2, GRID_Y2 );
	
		if ( x1 != vw->grid.primary.coef[0][0] || y1 != vw->grid.primary.coef[1][0] ||
			x2 != vw->grid.primary.coef[0][1] || y2 != vw->grid.primary.coef[1][1] )
		{
			struct geom_transform2 orig_grid = vw->grid;
			if (geom_transform2_set_basis( &vw->grid, 0.0, 0.0, x1, y1, x2, y2 ))
			{
				DBG("Erroneous grid %.3f, %.3f, %.3f, %.3f\n", x1, y1, x2, y2 );
				if (geom_transform2_set_basis( &vw->grid, 0.0, 0.0, x1, -y1, y1, x1 ))
				{
					DBG("Erroneous grid %.3f, %.3f, %.3f, %.3f\n", x1, y1, -y1, x1 );
					vw->grid = orig_grid;
				}
			}
		}
		visualisation_set_grid( &vw->vis, &vw->grid, prop_lookup_uns(o, string_grid_color, GRID_COLOR) );
	}
	else visualisation_unset_grid( &vw->vis );
}

void view_refresh_toolbar(struct view * vw)
{
	if ( !(vw->toolbar) || !sync_toolbar( vw->toolbar, CL_VIEW_TBAR ) ){
		struct cmddata cd = { (struct window*)vw, vw->group };
		DBG("View toolbar needs rebuilding!\n");
		if ( vw->toolbar ) gtk_container_remove( GTK_CONTAINER(vw->table), vw->toolbar );
		gtk_widget_show( vw->toolbar = create_toolbar( CL_VIEW_TBAR, 1, cd ) );// create toolbar
		gtk_table_attach(GTK_TABLE(vw->table), vw->toolbar, 0, 1, 0, 3, GTK_FILL, GTK_FILL, 0, 0);
		DBG("Rebuild OK\n");
	}
	if ( !(vw->toolbar2) || !sync_toolbar( vw->toolbar2, CL_VIEW_TBAR2 ) ){
		struct cmddata cd = { (struct window*)vw, vw->group };
		DBG("View toolbar2 needs rebuilding!\n");
		if ( vw->toolbar2 ) gtk_container_remove( GTK_CONTAINER(vw->table), vw->toolbar2 );
		gtk_widget_show( vw->toolbar2 = create_toolbar( CL_VIEW_TBAR2, 1, cd ) );// create toolbar
		gtk_table_attach(GTK_TABLE(vw->table), vw->toolbar2, 1, 2, 0, 3, GTK_FILL, GTK_FILL, 0, 0);
		DBG("Rebuild OK\n");
	}
}

void view_refresh_menu(void)
{
	if ( !(view_menu) || !sync_menu( view_menu, CL_VIEW_MENU ) ){
		struct cmddata cd = { NULL, NULL };
		DBG("View menu needs rebuilding!\n");
		if ( view_menu ) gtk_object_sink( GTK_OBJECT(view_menu) );
		view_menu = create_menu( CL_VIEW_MENU, 1, cd );
		DBG("Rebuild OK\n");
	}
}

/* HOOKS */

void view_sel_changed_2(struct go *go UNUSED, const struct changed_data_generic *data, struct go_hook* h)
{
	switch ( data->kind )
	{
	case CK_CHANGED:
		if (moving_mode != MOVE_TRANSFORM && moving_mode != MOVE_TRANSFORM_2)
			transform_recalc_all( (struct view*)h->data );
		break;
	default:
		transform_recalc_dimen( (struct view*)h->data );
	}
}
void view_obj_before_delete( struct o* o, struct o * owner UNUSED, struct o * prev UNUSED, struct obj_hook *hook )
{
	struct view * view = (struct view*)hook->data;
	if ( o_is_ancestor( o, go2o(&view->group->go) )) close_view( view );
}
static void view_go_before_delete( struct o* o, struct o* owner UNUSED, struct o* prev UNUSED, struct go_hook *hook )
{
	struct view * view = (struct view*)hook->data;
	if ( o_is_ancestor( o, go2o(&view->group->go) )) close_view( view );
}

static void view_obj_prop_change( struct o * o, struct obj_hook *hook )
{
	struct view * vw = (struct view*)hook->data;
	if ( o_is_ancestor( o, obj2o(&vw->tlo->obj)) )
		set_window_title( GTK_WINDOW(vw->window), vw->tlo, vw->group );

	if ( o == obj2o(obj_universum) )
	{
		// snap tolerance
		uns tolerance = prop_lookup_uns( o, string_tolerance, TOLERANCE_PIX );
		if ( tolerance != factory.snap_tol_pix )
		{
			factory.snap_tol_pix = tolerance;
			for ( GSList * l = view_list; l; l = l->next )
				view_recompute_tolerance( (struct view*)l->data );
		}
		// grid, rulers, fifi
		for ( GSList * l = view_list; l; l = l->next )
		{
			view_update_grid( (struct view*)l->data );
			update_rulers( (struct view*)l->data );
		}
		set_fifi_enabled(prop_lookup_uns( o, string_enable_fifi, ENABLE_FIFI ));
	}
}

/* VIEW NAVIGATOR WINDOW */

static gboolean view_navigator_delete_event( GtkWidget *widget UNUSED, GdkEvent * event UNUSED, struct view_navigator *vwn )
{
	view_moving_release(NULL, vwn->parent_view); //end the move, if any
	return FALSE;
}
static void close_view_navigator(struct view * vw)
{
	ASSERT(vw && vw->navigator);
	view_navigator_delete_event( NULL, NULL, vw->navigator );
	gtk_widget_destroy( vw->navigator->window );
}
static void destroy_view_navigator( GtkWidget * w UNUSED, struct view_navigator * vwn)
{
	visualisation_finish(&vwn->vis);
	vwn->parent_view->navigator = NULL;
	xfree(vwn);
}
static void create_view_navigator(struct view * vw)
{
	GtkWidget * table, * hbutbox, * button;
	struct view_navigator * vwn = xmalloc_zero(sizeof(struct view_navigator));
	vwn->parent_view = vw;
	vw->navigator = vwn;
	vwn->box = NULL;

	vwn->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	set_window_title( GTK_WINDOW(vwn->window), vw->tlo, vw->group );
	gtk_window_add_accel_group( GTK_WINDOW(vwn->window), accel_group );
	g_signal_connect( G_OBJECT(vwn->window), "destroy", G_CALLBACK(destroy_view_navigator), vwn);
	g_signal_connect( G_OBJECT(vwn->window), "delete_event", G_CALLBACK(view_navigator_delete_event), vwn);

	gtk_container_add( GTK_CONTAINER(vwn->window), table = gtk_table_new( 3, 4, FALSE ) );

	gtk_table_attach( GTK_TABLE(table), visualisation_init( &vwn->vis, vw->group, NULL, NULL ),
			0, 3, 0, 1, GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 5, 5 );
	gtk_widget_set_size_request( GTK_WIDGET(vwn->vis.canvas->gtk->widget), 100, 100 );
	g_signal_connect_after( G_OBJECT(vwn->vis.canvas->gtk->widget), "size_allocate", G_CALLBACK(view_nav_resize), vwn);
	gtk_widget_add_events( vwn->vis.canvas->gtk->widget, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK );
	g_signal_connect( G_OBJECT( vwn->vis.canvas->gtk->widget), "motion-notify-event", G_CALLBACK(view_nav_motion), vwn);
	g_signal_connect( G_OBJECT( vwn->vis.canvas->gtk->widget), "button-press-event", G_CALLBACK(view_nav_button_press), vwn);
	g_signal_connect( G_OBJECT( vwn->vis.canvas->gtk->widget), "scroll-event", G_CALLBACK (scroll_callback), vw);

	gtk_table_attach( GTK_TABLE(table), gtk_label_new( "Zoom:"), 0, 1, 1, 2, 0, 0, 5, 5 );
	vwn->zoom_entry = GTK_SPIN_BUTTON(button = gtk_spin_button_new_with_range( 5, 1000, 50 ));
	g_object_set( G_OBJECT(vwn->zoom_entry), "climb-rate", 1.0, "digits", 1, "numeric", TRUE, "xalign", 1.0, NULL );
	gtk_table_attach( GTK_TABLE(table), button, 1, 2, 1, 2, GTK_FILL, 0, 5, 5 );
	g_signal_connect( G_OBJECT(vwn->zoom_entry), "value-changed", G_CALLBACK(view_nav_value_changed), vwn);
	gtk_table_attach( GTK_TABLE(table), gtk_label_new( "%"), 2, 3, 1, 2, 0, 0, 5, 5 );

	gtk_table_attach( GTK_TABLE(table), gtk_label_new( "Rotation:"), 0, 1, 2, 3, 0, 0, 5, 5 );
	vwn->rot_entry = GTK_SPIN_BUTTON(button = gtk_spin_button_new_with_range( -190, 190, 10 ));
	g_object_set( G_OBJECT(vwn->rot_entry), "digits", 1, "numeric", TRUE, "xalign", 1.0, NULL );
	gtk_table_attach( GTK_TABLE(table), button, 1, 2, 2, 3, GTK_FILL, 0, 5, 5 );
	g_signal_connect( G_OBJECT(vwn->rot_entry), "value-changed", G_CALLBACK(view_nav_value_changed), vwn);
	gtk_table_attach( GTK_TABLE(table), gtk_label_new( "deg"), 2, 3, 2, 3, 0, 0, 5, 5 );

	gtk_table_attach( GTK_TABLE(table), hbutbox = gtk_hbutton_box_new(), 0, 3, 3, 4, 0, 0, 5, 5 );
	gtk_box_pack_start( GTK_BOX(hbutbox), button = gtk_button_new_with_label("Reset"), FALSE, TRUE, 0 );
	g_signal_connect( G_OBJECT(button), "clicked", G_CALLBACK(view_zoom_reset), vw );

	visualisation_set_show_anchors_mode( &vwn->vis, VIS_SHOW_NONE, NULL );
	visualisation_set_show_hangers_mode( &vwn->vis, VIS_SHOW_NONE, NULL );
	
	change_in_progress = 1;
	gtk_widget_show_all( vwn->window );
	change_in_progress = 0;

	GdkCursor * cursor = gdk_cursor_new( GDK_FLEUR );
	gdk_window_set_cursor( vwn->vis.canvas->gtk->widget->window, cursor );
	gdk_cursor_unref( cursor );

	//get parent zoom and rotation
	struct geom_point ux, zero;
	ux.x = 1;
	ux.y = zero.x = zero.y = 0;
	visualisation_coords_to_vcoords( &vw->vis, &ux, &ux );
	visualisation_coords_to_vcoords( &vw->vis, &zero, &zero );
	view_nav_update( vwn, hypot( ux.x - zero.x , ux.y - zero.y ) * 100.0, //zoom
		(real)atan2( (double)(ux.y - zero.y), (double)(ux.x - zero.x) ) / R_PI * 180.0 ); //rotation
}

void view_nav_redraw( struct view_navigator * vwn )
{
	struct geom_rectangle t;
	struct geom_point a, b, c;
	GtkWidget * visw = vwn->parent_view->vis.canvas->gtk->widget;
	int width = visw->allocation.width;
	int height = visw->allocation.height;

	visualisation_get_visible_bbox( &vwn->vis, &t );
	if ( t.ur.x > t.ll.x || t.ur.y > t.ll.y ) visualisation_fit_box( &vwn->vis, t );

	a.x = a.y = 0;
	b.x = width; b.y = 0;
	c.x = width; c.y = height;
	visualisation_pixels_to_coords( &vwn->parent_view->vis, &a, &a );
	visualisation_pixels_to_coords( &vwn->parent_view->vis, &b, &b );
	visualisation_pixels_to_coords( &vwn->parent_view->vis, &c, &c );

	visualisation_gui_rectangle_update( &vwn->vis, &vwn->box, a, b, c );
}
void view_nav_value_changed( GtkWidget * w UNUSED, struct view_navigator * vwn UNUSED )
{
	if ( change_in_progress ) return;

	real zoom = gtk_spin_button_get_value_as_float(vwn->zoom_entry);
	real rot = gtk_spin_button_get_value_as_float(vwn->rot_entry);
	//this is necessary so that it does not diverge due to numeric errors
	if ( fabs( zoom - vwn->zoom ) < 0.01 && fabs( rot - vwn->rotation ) < 0.01 ) return;

	struct geom_point center = { 0.0, 0.0 };
	visualisation_vcoords_to_coords( &vwn->parent_view->vis, &center, &center );
	zoom /= 100.0; rot = rot * R_PI / 180.0;
	DBG("Setting zoom = %.3f, rotation = %.3f\n", zoom, rot );
	visualisation_set_orientation( &vwn->parent_view->vis, center.x, center.y, zoom, zoom, rot );
}
void view_nav_update( struct view_navigator * vwn, real zoom, real rotation )
{
	change_in_progress = 1;
	gtk_spin_button_set_value( vwn->zoom_entry, vwn->zoom = zoom );
	gtk_spin_button_set_value( vwn->rot_entry, vwn->rotation = rotation );
	change_in_progress = 0;
}
void view_nav_resize( GtkWidget *w UNUSED, GdkRectangle *a UNUSED, struct view_navigator *vwn )
{ view_nav_redraw( vwn ); }

static gboolean view_nav_motion(GtkWidget *widget UNUSED, GdkEventMotion *event, struct view_navigator *vwn)
{
	if ( moving_mode == MOVE_PANNING_NAV )
		view_moving_motion( event, vwn->parent_view );
	return TRUE;
}
static gboolean view_nav_button_press(GtkWidget *widget UNUSED, GdkEventButton *event, struct view_navigator *vwn)
{
        struct geom_point pi = { event->x, event->y };
	struct geom_point po;
	visualisation_pixels_to_coords( &vwn->vis, &pi, &po );

	if ( event->state & GDK_CONTROL_MASK )
		visualisation_center( &vwn->parent_view->vis, po.x, po.y );
	else
		view_panning_nav_start( po, pi, vwn->parent_view );

	return TRUE;
}

void view_get_size (struct view * vw, uns *width, uns *height)
{
  gint w, h;
  GdkWindow *win = gtk_widget_get_parent_window (vw -> vis . canvas -> gtk -> widget);
  gdk_drawable_get_size (win, &w, &h);
  if (width != NULL) *width = w;
  if (height != NULL) *height = h;
}
