#undef LOCAL_DEBUG

/*
 * DVI file parser
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "lib/lib.h"
#include <stdio.h>
#include <strings.h>
#include <string.h>
#include <alloca.h>
#include <kpathsea/kpathsea.h>
#include "geomlib/types.h"
#include "lib/garr.h"
#include "import/import.h"
#include "font/font.h"

#define PT2MM (25.4 / 72.27)

/*
 * TODO:
 * - Maybe handle the multipage documents in a different manner (currently
 *   overwriting multiple pages).
 *   Design interface import_dvi(file, page, output) ???
 */


char dvi_err_msgbuf[300];

/*
 * We compute dimensions in TeX scaled points, mag=1000
 */
struct dvi_position {
	s32 h, v;		// horizontal and vertical position
	s32 w, x;		// horizontal spacing
	s32 y, z;		// vertical spacing
};

#define DVI_FONT_NAME_LEN 100
struct dvi_font {
	u32 font_number;
	int font_id;
	s32 width[256];
	s32 height[256];
	s32 depth[256];
	real font_size;
	char name[DVI_FONT_NAME_LEN+1];
};

static FILE *dvi_file;
static struct garr *dvi_pos_stack;
static uns dvi_pos_stack_top;
static struct dvi_position dvi_pos;
static struct dvi_font *dvi_font;
static real dvi_factor;			// multiplication factor to get actual dimensions in mm
static u32 dvi_mag;
static int dvi_inside_postamble;
static struct garr *dvi_glyphs;
static uns dvi_glyph_count;
static int dvi_error;


#define HASH_NODE struct dvi_font
#define HASH_PREFIX(x) fnt_##x
#define HASH_KEY_ATOMIC font_number
#define HASH_ATOMIC_TYPE u32
#define HASH_WANT_CLEANUP
#define HASH_WANT_FIND
#define HASH_WANT_NEW
#include "lib/sht.h"

#define GARR_PREFIX(x) pos_stack_##x
#define GARR_TYPE struct dvi_position
#include "lib/garr.h"

#define GARR_PREFIX(x) glyph_##x
#define GARR_TYPE struct tex_glyph
#include "lib/garr.h"


/*
 * Here we perform the BigEndian --> LittleEndian conversion.
 */

static inline u32 uload(byte *buf, int n)
{
	u32 r = 0;
	int i;

	for (i=0; i<n; i++)
		r = buf[i] | (r << 8);
	return r;
}

static inline s32 sload(byte *buf, int n)
{
	u32 u = uload(buf, n);
	if (u & (1 << (8*n-1)))
		u |= 0xffffffff << (8*n-1);
	return u;
}

static int skip(FILE *f, int n)
{
	if (fseek(f, n, SEEK_CUR)<0)
		return 0;
	else return 1;
}

/* convert the TeX fix_word format into scaled points */
static inline s32 fw2sp(s32 fw, u32 s)
{
	s32 r;

	/* hm... isn't it better to work in fix point? */
	r = (real)s * (real)fw / (real)(1 << 20);
	return r;
}

#define UGET(a,n) do { byte buf[n]; \
	if (fread(buf, 1, (n), f)<(n)) { dvi_error = DVI_ERR_TFM_CORRUPTED; \
		snprintf(dvi_err_msgbuf, sizeof(dvi_err_msgbuf)-1, "Font metric '%s' corrupted.", nm); \
		return; } \
	else (a) = uload(buf, (n)); }while(0)

#define SGET(a,n) do { byte buf[n]; \
	if (fread(buf, 1, (n), f)<(n)) { dvi_error = DVI_ERR_TFM_CORRUPTED; \
		snprintf(dvi_err_msgbuf, sizeof(dvi_err_msgbuf)-1, "Font metric '%s' corrupted.", nm); \
		return; } \
	else (a) = sload(buf, (n)); }while(0)

static void tfm_load(char *name, u32 font_number, u32 chk_sum, u32 s, u32 d)
{
	FILE *f;
	char *nm;
	u32 lf,lh,bc,ec,nw,nh,nd,ni,nl,nk,ne,np;
	uns i;

	nm = kpse_find_tfm(name);
	if (!nm || !(f=fopen(nm, "r"))) {
		dvi_error = DVI_ERR_TFM_NOTFOUND;
		snprintf(dvi_err_msgbuf, sizeof(dvi_err_msgbuf)-1, "Font metric '%s' not found.", nm ? nm : "");
		return;
	}

	/* scan the header */
	UGET(lf, 2); UGET(lh, 2); UGET(bc, 2); UGET(ec, 2);
	UGET(nw, 2); UGET(nh, 2); UGET(nd, 2); UGET(ni, 2);
	UGET(nl, 2); UGET(nk, 2); UGET(ne, 2); UGET(np, 2);

	DBG("TFM header: lf=%d lh=%d bc=%d ec=%d nw=%d nh=%d nd=%d ni=%d nl=%d nk=%d ne=%d np=%d\n",
		lf,lh,bc,ec,nw,nh,nd,ni,nl,nk,ne,np);

	if (!(bc<=ec+1 && ec<=255 && lf == 6+lh+(ec-bc+1)+nw+nh+nd+ni+nl+nk+ne+np)) {
		dvi_error = DVI_ERR_TFM_INVALID_HDR;
		snprintf(dvi_err_msgbuf, sizeof(dvi_err_msgbuf)-1, "Font metric '%s' has invalid header.", nm);
		return;
	}
	if (bc > 255) {
		bc = 1;
		ec = 0;
	}

	u32 checksum;
	UGET(checksum, 4);
	DBG("TFM checksum: 0x%x\n", checksum);
	if (chk_sum != checksum) {
		dvi_error = DVI_ERR_TFM_BAD_CHKSUM;
		snprintf(dvi_err_msgbuf, sizeof(dvi_err_msgbuf)-1, "Font metric '%s' has bad control checksum.", nm);
		return;
	}

	s32 size;
	real dsgn_size;
	SGET(size, 4);
	dsgn_size = fw2sp(size, 1);
	DBG("TFM design size: 0x%x = %f\n", size, dsgn_size);

	if(!skip(f, 4*lh-8)) {
		dvi_error = DVI_ERR_TFM_CORRUPTED;
		snprintf(dvi_err_msgbuf, sizeof(dvi_err_msgbuf)-1, "Font metric '%s' corrupted.", nm);
		return;
	}

	u32 *char_info;
	s32 *width, *height, *depth;

	char_info = alloca((ec-bc+1)*4);
	width = alloca(nw*4);
	height = alloca(nh*4);
	depth = alloca(nd*4);

	for (i=0; i<=ec-bc; i++)
		UGET(char_info[i], 4);
	for (i=0; i<nw; i++)
		SGET(width[i], 4);
	for (i=0; i<nh; i++)
		SGET(height[i], 4);
	for (i=0; i<nd; i++)
		SGET(depth[i], 4);

	width[0] = height[0] = depth[0] = 0;

	/* we don't need the rest of the TFM file */
	fclose(f);

	struct dvi_font *fnt = fnt_new(font_number);

	for (i=bc; i<=ec; i++) {
		fnt->width[i] = fw2sp(width[char_info[i-bc] >> 24], s);
		fnt->height[i] = fw2sp(height[(char_info[i-bc] >> 20) & 0xf], s);
		fnt->depth[i] = fw2sp(depth[(char_info[i-bc] >> 16) & 0xf], s);
		DBG("char %u: w=%d h=%d d=%d\n", i, fnt->width[i], fnt->height[i], fnt->depth[i]);
	}
	fnt->font_size = dsgn_size * dvi_mag / 1000.0 * s / (real) d;
	DBG("Fontsize=%f\n", fnt->font_size);
	strncpy(fnt->name, name, DVI_FONT_NAME_LEN);
	fnt->name[DVI_FONT_NAME_LEN] = 0;

	DBG("Looking for type1 font %s ...\n", fnt->name);
	string path = kpse_find_file(fnt->name, kpse_type1_format, true);
	if (!path) {
		dvi_error = DVI_ERR_NO_FONT;
		snprintf(dvi_err_msgbuf, sizeof(dvi_err_msgbuf)-1, "Type1 font '%s' not found.", fnt->name);
		return;
	}
	DBG("Found type1 font path %s\n", path);
	fnt->font_id = font_load_file(path);
	free(path);	// libkpathsea uses malloc() :-/
	free(nm);
}

#undef UGET
#undef SGET
#define UGET(a,n) do { byte buf[n]; \
	if (fread(buf, 1, (n), dvi_file)<(n)) { dvi_error = DVI_ERR_CORRUPTED; \
		strncpy(dvi_err_msgbuf, "DVI file corrupted.", sizeof(dvi_err_msgbuf)-1); \
		goto step_end; } \
	else (a) = uload(buf, (n)); }while(0)

#define SGET(a,n) do { byte buf[n]; \
	if (fread(buf, 1, (n), dvi_file)<(n)) { dvi_error = DVI_ERR_CORRUPTED; \
		strncpy(dvi_err_msgbuf, "DVI file corrupted.", sizeof(dvi_err_msgbuf)-1); \
		goto step_end; } \
	else (a) = sload(buf, (n)); }while(0)

static void step(void)
{
	byte opcode;

	dvi_error = 0;

	if (fread(&opcode, 1, 1, dvi_file)<1) {
		dvi_error = DVI_EOF;
		strncpy(dvi_err_msgbuf, "DVI EOF.", sizeof(dvi_err_msgbuf)-1);
		return;
	}

	DBG("[ opcode=%d h=%d v=%d w=%d x=%d y=%d z=%d\n", opcode,
		dvi_pos.h, dvi_pos.v, dvi_pos.w, dvi_pos.x, dvi_pos.y, dvi_pos.z);

	if (opcode <= 127) {
		DBG("set_char_%u\n", opcode);
		if (!dvi_font) {
			dvi_error = DVI_ERR_NO_FONT;
			strncpy(dvi_err_msgbuf, "Internal error in DVI file.", sizeof(dvi_err_msgbuf)-1);
			goto step_end;
		}

		struct tex_glyph *g = glyph_get(dvi_glyphs, dvi_glyph_count++);
		s32 w = dvi_font->width[opcode];

		g->is_rule = 0;
		g->font_id = dvi_font->font_id;
		g->font_size = PT2MM * dvi_font->font_size;	// original font size
		g->code = opcode;
		g->xpos = dvi_pos.h * dvi_factor;
		g->ypos = -dvi_pos.v * dvi_factor;

		struct font_ctl ctl;
		font_ctl_init(&ctl);
		ctl.font_id = g->font_id;
		ctl.font_size = g->font_size;
		ctl.charmap = FONT_CHARMAP_TEX;
		if (font_get_glyph_bbox(&ctl, g->code, &g->bbox)<0) {
			dvi_error = DVI_ERR_VRRFONT;
			snprintf(dvi_err_msgbuf, sizeof(dvi_err_msgbuf)-1,
				"Error computing bbox of font %d, charcode %d.", g->font_id, g->code);
		}
		font_ctl_cleanup(&ctl);
		DBGLN("Computed bbox [%f, %f] - [%f, %f]\n", g->bbox.ll.x, g->bbox.ll.y,
			g->bbox.ur.x, g->bbox.ur.y);

		dvi_pos.h += w;
	} else
	if (opcode >= 171 && opcode <= 234) {
		struct dvi_font *fnt;
		uns f;

		DBG("fnt_num_%d\n", opcode - 171);

		if (!dvi_inside_postamble) {	// skip the second definition in postamble
			f = opcode - 171;
			fnt = (struct dvi_font *)fnt_find(f);
			if (!fnt) {
				dvi_error = DVI_ERR_UNDEF_FONT;
				strncpy(dvi_err_msgbuf, "Internal error in DVI file.", sizeof(dvi_err_msgbuf)-1);
			} else
				dvi_font = fnt;
		}
	} else
	switch (opcode) {
	case 128:
	case 129:
	case 130:
	case 131:
		{
			u32 a;
			uns i = opcode - 128 +1;
			UGET(a, i);
			DBG("set%d: %u\n", i, a);
			if (!dvi_font) {
				dvi_error = DVI_ERR_NO_FONT;
				strncpy(dvi_err_msgbuf, "Internal error in DVI file.", sizeof(dvi_err_msgbuf)-1);
				break;
			}
			if (i == 1) {
				struct tex_glyph *g = glyph_get(dvi_glyphs, dvi_glyph_count++);
				s32 w = dvi_font->width[a];

				g->is_rule = 0;
				g->font_id = dvi_font->font_id;
				g->font_size = PT2MM * dvi_font->font_size;	// original font size
				g->code = a;
				g->xpos = dvi_pos.h * dvi_factor;
				g->ypos = -dvi_pos.v * dvi_factor;

				struct font_ctl ctl;
				font_ctl_init(&ctl);
				ctl.font_id = g->font_id;
				ctl.font_size = g->font_size;
				ctl.charmap = FONT_CHARMAP_TEX;
				if (font_get_glyph_bbox(&ctl, g->code, &g->bbox)<0) {
					dvi_error = DVI_ERR_VRRFONT;
					snprintf(dvi_err_msgbuf, sizeof(dvi_err_msgbuf)-1,
						"Error computing bbox of font %d, charcode %d.", g->font_id, g->code);
				}
				font_ctl_cleanup(&ctl);
				DBGLN("Computed bbox [%f, %f] - [%f, %f]\n", g->bbox.ll.x, g->bbox.ll.y,
					g->bbox.ur.x, g->bbox.ur.y);

				dvi_pos.h += w;
			}
			else {
				/*
				 * Fonts using more than 256 characters are unused by TeX
				 * (and not supported by our parser ;-) )
				 */
			}
		}
		break;

	case 132:
		{
			s32 a, b;
			SGET(a, 4);
			SGET(b, 4);
			DBG("set_rule: a=%d b=%d\b", a, b);
			dvi_pos.h += b;
			if (a>0 && b>0) {
				struct tex_glyph *g = glyph_get(dvi_glyphs, dvi_glyph_count++);

				g->is_rule = 1;
				g->xpos = dvi_pos.h * dvi_factor;
				g->ypos = -dvi_pos.v * dvi_factor;
				g->bbox.ll.x = 0;
				g->bbox.ll.y = 0;
				g->bbox.ur.x = b * dvi_factor;
				g->bbox.ur.y = a * dvi_factor;
				if (g->bbox.ll.x > g->bbox.ur.x) {
					real t = g->bbox.ll.x;
					g->bbox.ll.x = g->bbox.ur.x;
					g->bbox.ur.x = t;
				}
				if (g->bbox.ll.y > g->bbox.ur.y) {
					real t = g->bbox.ll.y;
					g->bbox.ll.y = g->bbox.ur.y;
					g->bbox.ur.y = t;
				}
			}
		}
		break;

	case 133:
	case 134:
	case 135:
	case 136:
		{
			u32 c;
			uns i = opcode - 133 +1;
			UGET(c, i);
			DBG("put%i: %u\n", i, c);
			if (i == 1) {
				struct tex_glyph *g = glyph_get(dvi_glyphs, dvi_glyph_count++);

				g->is_rule = 0;
				g->font_id = dvi_font->font_id;
				g->font_size = PT2MM * dvi_font->font_size;	// original font size
				g->code = c;
				g->xpos = dvi_pos.h * dvi_factor;
				g->ypos = -dvi_pos.v * dvi_factor;

				struct font_ctl ctl;
				font_ctl_init(&ctl);
				ctl.font_id = g->font_id;
				ctl.font_size = g->font_size;
				ctl.charmap = FONT_CHARMAP_TEX;
				if (font_get_glyph_bbox(&ctl, g->code, &g->bbox)<0) {
					dvi_error = DVI_ERR_VRRFONT;
					snprintf(dvi_err_msgbuf, sizeof(dvi_err_msgbuf)-1,
						"Error computing bbox of font %d, charcode %d.", g->font_id, g->code);
				}
				font_ctl_cleanup(&ctl);

				DBGLN("Computed bbox [%f, %f] - [%f, %f]\n", g->bbox.ll.x, g->bbox.ll.y,
					g->bbox.ur.x, g->bbox.ur.y);
			} else {
				/*
				 * Fonts using more than 256 characters are unused by TeX
				 * (and not supported by our parser ;-) )
				 */
			}
		}
		break;

	case 137:
		{
			s32 a, b;
			SGET(a, 4);
			SGET(b, 4);
			DBG("put_rule: a=%d b=%d\n", a, b);
			if (a>0 && b>0) {
				struct tex_glyph *g = glyph_get(dvi_glyphs, dvi_glyph_count++);

				g->is_rule = 1;
				g->xpos = dvi_pos.h * dvi_factor;
				g->ypos = -dvi_pos.v * dvi_factor;
				g->bbox.ll.x = 0;
				g->bbox.ll.y = 0;
				g->bbox.ur.x = b * dvi_factor;
				g->bbox.ur.y = a * dvi_factor;
				if (g->bbox.ll.x > g->bbox.ur.x) {
					real t = g->bbox.ll.x;
					g->bbox.ll.x = g->bbox.ur.x;
					g->bbox.ur.x = t;
				}
				if (g->bbox.ll.y > g->bbox.ur.y) {
					real t = g->bbox.ll.y;
					g->bbox.ll.y = g->bbox.ur.y;
					g->bbox.ur.y = t;
				}
				/*
				g->bbox.ll.x = 0;
				g->bbox.ll.y = -a * dvi_factor;
				g->bbox.ur.x = b * dvi_factor;
				g->bbox.ur.y = 0;
				*/
			}
		}
		break;

	case 138:
		DBG("nop\n");
		break;

	case 139:
		{
			int i;
			s32 c[10], p;
			for (i=0; i<10; i++)
				SGET(c[i], 4);
			SGET(p, 4);
			DBG("bop: %d %d %d %d %d %d %d %d %d %d p=%d\n",
				c[0],c[1],c[2],c[3],c[4],c[5],c[6],c[7],c[8],c[9], p);

			bzero(&dvi_pos, sizeof(dvi_pos));
			dvi_pos_stack_top = 0;
			dvi_font = NULL;	// Well, according to standard, this should be undefined ;-)
		}
		break;

	case 140:
		DBG("eop\n");
		if (dvi_pos_stack_top) {
			dvi_error = DVI_ERR_STACK_NONEMPTY;
			strncpy(dvi_err_msgbuf, "Internal error in DVI file.", sizeof(dvi_err_msgbuf)-1);
		}
		// FIXME: do the page output
		break;

	case 141:
		DBG("push\n");
		*pos_stack_get(dvi_pos_stack, dvi_pos_stack_top++) = dvi_pos;
		break;

	case 142:
		DBG("pop\n");
		if (!dvi_pos_stack_top) {
			dvi_error = DVI_ERR_STACK_UNDERFLOW;
			strncpy(dvi_err_msgbuf, "Internal error in DVI file.", sizeof(dvi_err_msgbuf)-1);
			break;
		}
		dvi_pos = *pos_stack_get(dvi_pos_stack, --dvi_pos_stack_top);
		break;

	case 143:
	case 144:
	case 145:
	case 146:
		{
			s32 b;
			uns i = opcode - 143 +1;
			SGET(b, i);
			DBG("right%d: %d\n", i, b);
			dvi_pos.h += b;
		}
		break;

	case 147:
		DBG("w0\n");
		dvi_pos.h += dvi_pos.w;
		break;

	case 148:
	case 149:
	case 150:
	case 151:
		{
			s32 b;
			uns i = opcode - 148 +1;
			SGET(b, i);
			DBG("w%d: %d\n", i, b);
			dvi_pos.w = b;
			dvi_pos.h += b;
		}
		break;

	case 152:
		DBG("x0\n");
		dvi_pos.h += dvi_pos.x;
		break;

	case 153:
	case 154:
	case 155:
	case 156:
		{
			s32 b;
			uns i = opcode - 153 +1;
			SGET(b, i);
			DBG("x%d: %d\n", i, b);
			dvi_pos.x = b;
			dvi_pos.h += b;
		}
		break;

	case 157:
	case 158:
	case 159:
	case 160:
		{
			s32 a;
			uns i = opcode -157 +1;
			SGET(a, i);
			DBG("down%u: %d\n", i, a);
			dvi_pos.v += a;
		}
		break;

	case 161:
		DBG("y0\n");
		dvi_pos.v += dvi_pos.y;
		break;

	case 162:
	case 163:
	case 164:
	case 165:
		{
			s32 a;
			uns i = opcode -162 +1;
			SGET(a, i);
			DBG("y%u: %d\n", i, a);
			dvi_pos.y = a;
			dvi_pos.v += a;
		}
		break;

	case 166:
		DBG("z0\n");
		dvi_pos.v += dvi_pos.z;
		break;

	case 167:
	case 168:
	case 169:
	case 170:
		{
			s32 a;
			uns i = opcode -167 +1;
			SGET(a, i);
			DBG("z%u: %d\n", i, a);
			dvi_pos.z = a;
			dvi_pos.v += a;
		}
		break;

	case 235:
	case 236:
	case 237:
	case 238:
		{
			u32 k;
			uns i = opcode - 235 +1;
			UGET(k, i);
			DBG("fnt%d: %u\n", i, k);
			if (i == 1 && !dvi_inside_postamble) {		// skip the second definition in postamble
				struct dvi_font *fnt;

				fnt = (struct dvi_font *)fnt_find(k);
				if (!fnt) {
					dvi_error = DVI_ERR_UNDEF_FONT;
					strncpy(dvi_err_msgbuf, "Internal error in DVI file.", sizeof(dvi_err_msgbuf)-1);
				} else
					dvi_font = fnt;
			} else {
				/* TeX doesn't use these commands */
			}
		}
		break;

	case 239:
	case 240:
	case 241:
	case 242:
		{
			u32 k;
			uns i = opcode - 239 +1;
			UGET(k, i);
			DBG("xxx%d: %u\n", i, k);
			if (!skip(dvi_file, k)) {
				dvi_error = DVI_ERR_CORRUPTED;
				strncpy(dvi_err_msgbuf, "DVI file corrupted.", sizeof(dvi_err_msgbuf)-1);
			}
			/*
			 * Well, TeX uses this command for inserting its \special,
			 * but since we don't expect such inputs, we can safely ignore it.
			 * Moreover, TeX manual documents xxx? as undefined.
			 */
		}
		break;

	case 243:
	case 244:
	case 245:
	case 246:
		{
			u32 k, c, s, d, a, l;
			uns i = opcode - 243 +1;
			char *n;

			UGET(k, i);
			UGET(c, 4);
			UGET(s, 4);
			UGET(d, 4);
			UGET(a, 1);
			UGET(l, 1);
			n = alloca(a+l+1);
			if (fread(n, 1, a+l, dvi_file)<a+l) {
				dvi_error = DVI_ERR_CORRUPTED;
				strncpy(dvi_err_msgbuf, "DVI file corrupted.", sizeof(dvi_err_msgbuf)-1);
				goto step_end;
			}
			n[a+l] = 0;

			DBG("fnt_def%d: k=%u c=%u s=%u d=%u a=%u l=%u n=%s\n", i, k, c, s, d, a, l, n);

			if (!fnt_find(k)) {		// don't allow redefinitions
				tfm_load(n, k, c, s, d);
			}
		}
		break;

	case 247:
		{
			u32 i, num, den, mag, k;

			UGET(i, 1);
			UGET(num, 4);
			UGET(den, 4);
			UGET(mag, 4);
			UGET(k, 1);
			if (!skip(dvi_file, k)) {	// skip the comment
				dvi_error = DVI_ERR_CORRUPTED;
				strncpy(dvi_err_msgbuf, "DVI file corrupted.", sizeof(dvi_err_msgbuf)-1);
				break;
			}

			DBG("pre: i=%u num=%d den=%u mag=%u k=%u\n", i, num, den, mag, k);
			dvi_factor = (real) mag * num / (10000000.0*den);
			dvi_mag = mag;
		}
		break;

	case 248:
		{
			u32 p, num, den, mag, l, u, s, t;
			UGET(p, 4);
			UGET(num, 4);
			UGET(den, 4);
			UGET(mag, 4);
			UGET(l, 4);
			UGET(u, 4);
			UGET(s, 2);
			UGET(t, 2);

			DBG("post: p=%u num=%u den=%u mag=%u l=%u u=%u s=%u t=%u\n",
				p, num, den, mag, l, u, s, t);
			dvi_inside_postamble = 1;
		}
		break;

	case 249:
		{
			u32 q, i;
			UGET(q, 4);
			UGET(i, 1);

			DBG("post_post: q=%x i=%d\n", q, i);
			dvi_inside_postamble = 0;
			fseek(dvi_file, 0, SEEK_END);		// skip the remaining 223's
		}

		break;

	default:
		DBG("undefined opcode=%u\n", opcode);
	}
step_end:
	DBG("h=%d v=%d w=%d x=%d y=%d z=%d ]\n", dvi_pos.h, dvi_pos.v, dvi_pos.w, dvi_pos.x, dvi_pos.y, dvi_pos.z);
}

int import_dvi(FILE *f, struct garr *glyphs_out)
{
	ASSERT(f);
	ASSERT(glyphs_out);

	dvi_file = f;
	dvi_pos_stack = pos_stack_new(256);
	dvi_glyphs = glyphs_out;
	fnt_init();
	dvi_pos_stack_top = dvi_factor = dvi_mag = dvi_inside_postamble =
		dvi_glyph_count = dvi_error = 0;
	bzero(&dvi_pos, sizeof(dvi_pos));
	dvi_font = NULL;


	while (!dvi_error) {
		step();
	}
	if (dvi_error && dvi_error != DVI_EOF) {
		msg("DVI parsing finished with error code %d: %s\n", dvi_error, dvi_err_msgbuf);
		goto dvi_end;
	}

	uns i;
	real maxx, maxy, minx, miny;

	/* compute the global bbox */
	maxx = maxy = -1/0.0;
	minx = miny = 1/0.0;
	DBG("Listing encountered glyphs:\n");
	for (i=0; i<dvi_glyph_count; i++) {
		struct tex_glyph *g = glyph_get(dvi_glyphs, i);

		DBG("%d: ", i);
		maxx = MAX(maxx, g->xpos + MAX(g->bbox.ur.x, g->bbox.ll.x));
		maxy = MAX(maxy, g->ypos + MAX(g->bbox.ur.y, g->bbox.ll.y));
		minx = MIN(minx, g->xpos + MIN(g->bbox.ur.x, g->bbox.ll.x));
		miny = MIN(miny, g->ypos + MIN(g->bbox.ur.y, g->bbox.ll.y));
		if (g->is_rule)
			DBG("rule [%f, %f] [%f, %f]\n", g->bbox.ll.x, g->bbox.ll.y, g->bbox.ur.x, g->bbox.ur.y);
		else
			DBG("char_code=%d '%c' fntid=%d size=%f xpos=%f ypos=%f bbox [%f, %f] [%f, %f]\n",
				g->code, g->code, g->font_id, g->font_size, g->xpos, g->ypos,
				g->bbox.ll.x, g->bbox.ll.y, g->bbox.ur.x, g->bbox.ur.y);
	}
	DBG("Scanned %d glyphs, global bbox: x=%f..%f y=%f..%f\n", dvi_glyph_count, minx, maxx, miny, maxy);

	/* Now normalize all coordinates */
	for (i=0; i<dvi_glyph_count; i++) {
		struct tex_glyph *g = glyph_get(dvi_glyphs, i);

		g->xpos -= minx;
		g->ypos -= miny;
		DBG("glyph %d (%c) normalized coords: [%f, %f]\n", i, (g->is_rule ? '|' : g->code), g->xpos, g->ypos);
	}

	dvi_error = 0;

dvi_end:
	pos_stack_destroy(dvi_pos_stack);
	fnt_cleanup();

	if (dvi_error < 0)
		return dvi_error;
	return dvi_glyph_count;
}

