#ifndef _KERNEL_FUNCTIONS_
#define _KERNEL_FUNCTIONS_

/** @file
 *
 * 	Kernels' functions
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */


/*
 *  Map of this header file.
 *
 * x OBJ and GO handling
 *   - Verify macros and type-cast functions
 *   - Reference
 *   - Creating functions
 *   - Linking
 *   - Selectings
 *   - Transformation
 *
 * x Movement functions and walk macros
 *   - Movement in OBJ hierarchy
 *   - Movement in GROUP
 *   - Movement in O (OBJ and GO)
 *   - Movement in hanger/anchor dependence
 *   - Movement in selections
 *
 * x OBJ and GO hooks handling
 *
 * x Clipboard
 *
 * x Dumping functions
 *
 * x Save and load function
 *
 * x Miscellaneous
 *   - Kernel initialization and cleanup
 *
 */


SCHEME SO_obj_tlo go_get_tlo(SO_go go);
static inline struct go_info* go_get_info(struct go* g)
	{ return &go_info[g->type][g->subtype]; }
static inline int hanger_is_smart(struct hanger *h)
	{ return h->info->go_offset == 0; }
static inline struct hanger_smart* hanger2smart(struct hanger *h)
	{ ASSERT(h && hanger_is_smart(h)); return (struct hanger_smart*)h; }
static inline clist* go_get_lhanger(struct go *go)
	{ int i = go_get_info(go)->hs_offset; return i ? SKIP_OFFSET(clist, go, i) : NULL; }
static inline struct hanger_smart* go_lhanger_get_first_hanger_smart(clist *l)
	{ return l->head.next == &l->head ? NULL : SKIP_BACK(struct hanger_smart, nhanger, l->head.next); }
static inline struct hanger_smart* go_lhanger_get_next_hanger_smart(clist *l, struct hanger_smart *hs)
	{ return hs->nhanger.next == &l->head ? NULL : SKIP_BACK(struct hanger_smart, nhanger, hs->nhanger.next); }



/***********************************
 *                                 *
 *      OBJ handling functions     *
 *                                 *
 ***********************************/

/*
 * Verify macros and type-cast functions
 */

#ifdef DEBUG_KERNEL_SLOW
#define ASSERT_O(_o) ASSERT((_o) && (((struct o*)(_o))->kind == T_OBJ || ((struct o*)(_o))->kind == T_GO))
#define ASSERT_GO(_o) ASSERT((_o) && ((struct o*)(_o))->kind == T_GO)
#define ASSERT_GROUP(_group) ASSERT((_group) && ((struct o*)(_group))->kind == T_GO && ((struct o*)(_group))->type == GOT_GROUP)
static inline struct go* assert_go_in_tlo(struct go *go, struct obj_tlo *tlo)
	{ ASSERT_GO(go); ASSERT_TLO(tlo); ASSERT(go_get_tlo(go) == tlo); return go; }
#else
#define ASSERT_O(_o) do {} while(0)
#define ASSERT_GO(_o) do {} while(0)
#define ASSERT_GROUP(_o) do {} while(0)
static inline struct go* assert_go_in_tlo(struct go *go, struct obj_tlo *tlo UNUSED)
	{ return go; }
#endif

#define ASSERT_GO_IN_TLO(_go, _tlo) ASSERT(go_get_tlo(_go) == _tlo)
#define ASSERT_GO_IN_TLO_EXPAND(_go, _tlo) ASSERT(go_get_tlo(_go) == _tlo && _tlo->expand)

static inline struct go* o2go(struct o* o)
	{ ASSERT_GO(o); return (struct go*)o; }
static inline struct o* go2o(struct go* g)
	{ ASSERT_GO(g); return (struct o*)g; }

static inline struct go_group* go2group(struct go* go)
	{ ASSERT_GROUP(go); return (struct go_group*)go; }

static inline struct o* void2o(void *o)
	{ ASSERT_O(o); return o; }
static inline struct go* void2go(void *o)
	{ ASSERT_GO(o); return o; }

static inline struct geom_item* go2geom(struct go *go)
	{ ASSERT_GO(go); return SKIP_OFFSET(struct geom_item, go, OFFSETOF(struct go_group, geom)); }
static inline struct go* geom2go(struct geom_item *geom)
	{ struct go *go = SKIP_OFFSET(struct go, geom, -OFFSETOF(struct go_group, geom)); ASSERT_GO(go); return go; }



/*
 * Reference functions
 */

/* Macro DEBUG_REF_O_ARGS is used for debugging. You can ignore it. */
struct obj* obj_ref(struct obj *obj DEBUG_REF_O_ARGS);
void obj_unref(struct obj *obj DEBUG_REF_O_ARGS);

struct go* go_ref(struct go *go DEBUG_REF_O_ARGS);
void go_unref(struct go *go DEBUG_REF_O_ARGS);

struct o* o_ref(struct o *o DEBUG_REF_O_ARGS);
void o_unref(struct o *o DEBUG_REF_O_ARGS);

struct hanger* hanger_ref(struct hanger *hanger DEBUG_REF_O_ARGS);
void hanger_unref(struct hanger *hanger DEBUG_REF_O_ARGS);



/*
 * Creating functions
 */


struct hanger* hanger_new_mc(struct geom_point *gp);
SCHEME static inline SHANGER hanger_new_mc_xy(SREAL x, SREAL y)
	{ struct geom_point gp = (struct geom_point) { x, y }; return hanger_new_mc(&gp); }
SCHEME SHANGER hanger_new_mc_curve(void);


/*
 * Selecting functions
 */

SCHEME SO_go go_select(SO_go g); STRANS
SCHEME SO_go go_unselect(SO_go g); STRANS
SCHEME SVOID tlo_unselect_all_go(void); STRANS
SCHEME SVOID group_select_all_go(SO_go_group group); STRANS
SCHEME SVOID group_unselect_all_go(SO_go_group group); STRANS

void o_select(struct o* o);
void o_unselect(struct o* o);
void o_unselect_all(void);

SCHEME static inline SBOOL go_is_selected(SO_go o)
	{ return o->flags & GOF_SELECTED; }
SCHEME static inline SBOOL o_is_selected(SO_o o)
	{ return o->kind == T_OBJ ? obj_is_selected(o2obj(o)) : go_is_selected(o2go(o)); }



/**********************************************
 *                                            *
 *    Movement functions and walk macros      *
 *                                            *
 **********************************************/

/*
 * Movement in GROUP
 */

SCHEME static inline SO_go_group tlo_get_tlg(SO_obj_tlo tlo)
	{ return &tlo->tlg->group; }

/* Basic GROUP movement */
#ifdef STUPID_GROUP
SCHEME static inline SO_go group_get_first_go(SO_go_group group)
	{ cnode *n = clist_head(&group->lmember); return n ? SKIP_BACK(struct go, nmember, n) : NULL; }
SCHEME static inline SO_go group_get_next_go(SO_go_group group, SO_go go)
	{ cnode *n = clist_next(&group->lmember, &go->nmember); return n ? SKIP_BACK(struct go, nmember, n) : NULL; }
SCHEME static inline SO_go group_get_last_go(SO_go_group group)
	{ cnode *n = clist_tail(&group->lmember); return n ? SKIP_BACK(struct go, nmember, n) : NULL; }
SCHEME static inline SO_go group_get_prev_go(SO_go_group group, SO_go go)
	{ cnode *n = clist_prev(&group->lmember, &go->nmember); return n ? SKIP_BACK(struct go, nmember, n) : NULL; }
SCHEME static inline SO_go_group go_get_group(SO_go go)
	{ return go->owner; }
SCHEME static inline SO_go_tlg go_get_root(SO_go go)
	{ while(go->owner) go = &go->owner->go; return (struct go_tlg*)go; }
#else
static inline struct go* group_get_first_go(struct go_group *group)
	{ ccnode *n = cclist_head(&group->geom.group.list); return n ? geom2go(SKIP_BACK(struct geom_item, node, n)) : NULL; }
static inline struct go* group_get_next_go(struct go_group *group, struct go *go)
	{ ccnode *n = cclist_next(&group->geom.group.list, &go2geom(go)->node); return n ? geom2go(SKIP_BACK(struct geom_item, node, n)) : NULL; }
static inline struct go* group_get_last_go(struct go_group *group)
	{ ccnode *n = cclist_tail(&group->geom.group.list); return n ? geom2go(SKIP_BACK(struct geom_item, node, n)) : NULL; }
static inline struct go* group_get_prev_go(struct go_group *group, struct go *go)
	{ ccnode *n = cclist_prev(&group->geom.group.list, &go2geom(go)->node); return n ? geom2go(SKIP_BACK(struct geom_item, node, n)) : NULL; }
static inline struct go_group* go_get_group(struct go *go)
	{ struct geom_item *i = &go2geom(go)->owner->item; return i ? go2group(geom2go(i)) : NULL; }
static inline struct go_tlg* go_get_root(struct go* go)
	{ return (struct go_tlg*)go2group(geom2go(geom_tree_root(go2geom(go)))); }
#endif
#define WALK_GROUP_GO(_group, _walk) for(_walk = group_get_first_go(_group); _walk; _walk = group_get_next_go(_group, _walk))
#define WALK_GROUP_GO_BACKWARDS(_group, _walk) for(_walk = group_get_last_go(_group); _walk; _walk = group_get_prev_go(_group, _walk))

/* Recursive OBJ movement */
struct go* go_recursive_next(const struct go *root, struct go *walk);
#define WALK_GO_RECURSIVE(_root, _walk) for(_walk = _root; _walk; _walk = go_recursive_next(_root, _walk))
#define WALK_GO_RECURSIVE_WITHOUT_ROOT(_root, _walk) for(_walk = go_recursive_next(_root, _root); _walk; _walk = go_recursive_next(_root, _walk))

struct go* go_recursive_prev(const struct go *root, struct go *walk);
#define WALK_GO_RECURSIVE_BACKWARDS(_root, _walk) for(_walk = _root; _walk; _walk = go_recursive_prev(_root, _walk))
#define WALK_GO_RECURSIVE_WITHOUT_ROOT_BACKWARDS(_root, _walk) for(_walk = go_recursive_prev(_root, _root); _walk; _walk = go_recursive_prev(_root, _walk))

/* Miscellaneous GROUP movement */
static inline uns group_get_rank_go(struct go_group *group, struct go *go)
	{ uns n = -1; for(; go; go = group_get_prev_go(group, go)) n++; return n; }
static inline int group_is_go_descentant(struct go_group *group, struct go *go)
	{ do { if(&group->go == go) return 1; } while((go = &go_get_group(go)->go)); return 0; }
static inline uns go_is_linked(struct go *go)
	{ return go_get_root(go)->group.go.subtype == GOST_GROUP_TLG; }
SCHEME static inline SUNS group_get_count_go(SO_go_group group)
	{ struct go *go; uns u = 0; WALK_GROUP_GO(group, go) u++; return u; }

SCHEME static inline SBOOL go_is_ancestor(SO_go anc, SO_go des) 
	{ for(; des; des = &go_get_group(des)->go) if (anc == des) return 1; return 0; }

/*
 * Dependencies movement
 */

/* Hanger - GO movement */
#define WALK_GO_HANGER_BEGIN(_go, _h) do { \
	struct go_info *_gi = go_get_info(_go); \
	clist *_l = go_get_lhanger(_go);	\
	int _i = _gi->h_count; \
	for(_h = _i ? SKIP_OFFSET(struct hanger, _go, _gi->h_offset) : (_l ? &go_lhanger_get_first_hanger_smart(_l)->h : NULL); \
			_h; \
			_i--, _h = _i > 0 ? _h + 1 : (_i < 0 ? &go_lhanger_get_next_hanger_smart(_l, hanger2smart(_h))->h : \
				(_l ? &go_lhanger_get_first_hanger_smart(_l)->h : NULL))) \
					do {
#define WALK_GO_HANGER_END } while(0); } while(0);

SCHEME static inline SO_go hanger_get_go(SHANGER h)
	{ return void2go((hanger_is_smart(h) ? hanger2smart(h)->go : SKIP_OFFSET(struct go, h, -h->info->go_offset))); }

SCHEME static inline SHANGER go_get_hanger(SO_go go, SUNS n)
	{ struct go_info *gi = go_get_info(go); return (gi->h_count > n) ? SKIP_OFFSET(struct hanger, go, gi->h_offset) + n : NULL; }
SCHEME static inline SHANGER go_get_hanger_curve(SO_go go)
	{ int i = go_get_info(go)->h_curve; return i == -1 ? NULL : go_get_hanger(go, i); }
SCHEME static inline SHANGER go_get_hanger_start(SO_go go)
	{ int i = go_get_info(go)->h_start; return i == -1 ? NULL : go_get_hanger(go, i); }
SCHEME static inline SHANGER go_get_hanger_end(SO_go go)
	{ int i = go_get_info(go)->h_end; return i == -1 ? NULL : go_get_hanger(go, i); }
SCHEME static inline SHANGER go_get_hanger_center(SO_go go)
	{ int i = go_get_info(go)->h_center; return i == -1 ? NULL : go_get_hanger(go, i); }

SCHEME static inline SHANGER go_get_hanger_by_name(SO_go go, SSYMBOLS name)
	{ struct hanger *h; WALK_GO_HANGER_BEGIN(go, h) if(h->info->name == name) return h; WALK_GO_HANGER_END return NULL; }

/* Anchor - hanger movement */
SCHEME static inline SANCHOR hanger_get_first_anchor(SHANGER h)
	{ cnode *n = clist_head(&h->lanchor); return n ? SKIP_BACK(struct anchor, nanchor, n) : NULL; }
SCHEME static inline SANCHOR hanger_get_next_anchor(SANCHOR a)
	{ cnode *n = clist_next(&a->h->lanchor, &a->nanchor); return n ? SKIP_BACK(struct anchor, nanchor, n) : NULL; }
#define WALK_HANGER_ANCHOR(_hanger, _walk) for(_walk = hanger_get_first_anchor(_hanger); _walk; _walk = hanger_get_next_anchor(_walk))


/* Anchor - GO movement */
SCHEME static inline SO_go anchor_get_go(SANCHOR a)
	{ return void2go(SKIP_OFFSET(struct go, a, -a->info->offset)); }
SCHEME static inline SANCHOR go_get_anchor(SO_go go, SUNS anchor)
	{ struct go_info *gi = go_get_info(go); return anchor < gi->a_count ? SKIP_OFFSET(struct anchor, go, gi->a_offset) + anchor : NULL; }
static inline struct anchor* go_get_first_anchor(struct go *go)
	{ return SKIP_OFFSET(struct anchor, go, go_get_info(go)->a_offset); }
#define WALK_GO_ANCHOR(_go, _walk, _i) for(_i = -go_get_info(_go)->a_count, _walk = go_get_first_anchor(_go); _i; _i++, _walk++)

SCHEME static inline SANCHOR go_get_anchor_by_name(SO_go go, SSYMBOLS name)
	{ struct anchor *a; int i; WALK_GO_ANCHOR(go, a, i) if(a->info->name == name) return a; return NULL; }
SCHEME static inline SANCHOR go_get_anchor_start(SO_go go)
	{ int i = go_get_info(go)->a_start; return i == -1 ? NULL : go_get_anchor(go, i); }
SCHEME static inline SANCHOR go_get_anchor_end(SO_go go)
	{ int i = go_get_info(go)->a_end; return i == -1 ? NULL : go_get_anchor(go, i); }

/*
 * Movement in selections
 */

SCHEME static inline SO_go group_get_first_selected_go(SO_go_group group)
	{ cnode *n = clist_head(&group->lselected); return n ? SKIP_BACK(struct go, nselected, n) : NULL; }
SCHEME static inline SO_go group_get_next_selected_go(SO_go_group group, SO_go go)
	{ cnode *n = clist_next(&group->lselected, &go->nselected); return n ? SKIP_BACK(struct go, nselected, n) : NULL; }
#define WALK_GROUP_SELECTED_GO(_group, _walk) for(_walk = group_get_first_selected_go(_group); _walk; _walk = group_get_next_selected_go(_group, _walk))



/*
 * Movement in O (OBJ and GO)
 */

struct o* o_get_ancestor(struct o* o);

/* Type of _o must be a point to structure which is derived from o. Type of _walk must be struct o*. */
#define WALK_O_ANCESTOR(_o, _walk) for(_walk = void2o(_o); _walk; _walk = o_get_ancestor(_walk))

static inline int o_is_ancestor(struct o *ancestor, struct o *descendant)
	{ struct o *walk; WALK_O_ANCESTOR(descendant, walk) if(walk == ancestor) return 1; return 0; }

SCHEME static inline SUNS o_is_linked(SO_o o)
	{ return o->kind == T_OBJ ? obj_is_ancestor(obj_universum, o2obj(o)) : go_is_linked(o2go(o)); }

SCHEME static inline SO_obj_tlo o_get_trans_tlo(SO_o o)
	{ ASSERT_O(o); return o->kind == T_OBJ ? tlo_universum : go_get_tlo(o2go(o)); }

/*
 * Linking functions
 */

/* If after if NULL then go is inserted in the begin of group. */
SCHEME SO_go go_link_after(SO_go go, SO_go_group group, SO_go after) STRANS ;
SCHEME static inline SO_go go_link_head(SO_go go, SO_go_group group) STRANS
	{ return go_link_after(go, group, NULL); }
SCHEME static inline SO_go go_link_prev(SO_go go, SO_go_group group, SO_go prev) STRANS
	{ return go_link_after(go, group, prev ? group_get_prev_go(group, prev) : group_get_last_go(group)); }
SCHEME static inline SO_go go_link_tail(SO_go go, SO_go_group group) STRANS
	{ return go_link_after(go, group, group_get_last_go(group)); }

SCHEME SO_go go_unlink(SO_go go) STRANS ;
SCHEME SO_go go_relink(SO_go go, SO_go_group group, SO_go after) STRANS ;

SCHEME SVOID group_relink_selected_go(SO_go_group source, SO_go_group target, SO_go after) STRANS ;
SCHEME SVOID group_relink_all_go(SO_go_group source, SO_go_group target, SO_go after) STRANS ;

SCHEME SVOID tlo_unlink_all_go(void) STRANS ;
SCHEME SVOID group_unlink_selected_go(SO_go_group group) STRANS ;
SCHEME SVOID go_remove_rehang(SO_go go) STRANS ;
SCHEME SVOID group_remove_rehang_selected_go(SO_go_group group); STRANS
uns go_remove_carefully(struct go *go);

SCHEME SVOID anchor_rehang(SANCHOR a, SHANGER h); STRANS
int hanger_mc_move(struct hanger* h, struct geom_point* gp);

SCHEME SO_go_path group_pathize_selected_go(SO_go_group group) STRANS ;

/*
 * Miscellaneous
 */
	
SCHEME static inline SUNS go_is_invisible(SO_go go)
	{ return go->flags & GOF_INVISIBLE; }
SCHEME static inline SUNS go_is_styles_properties_used(SO_go go)
	{ go = &go_get_group(go)->go; return !go || go->type != GOT_GROUP || go->subtype != GOST_GROUP_PATH; }

/**********************************************
 *                                            *
 *     OBJ and GO hooks handling functions    *
 *                                            *
 **********************************************/

static inline struct go_hook* go_hook_add(struct obj_tlo *tlo, struct go_hook *hook)
	{ slist_add(&tlo->go_hooks, &hook->n); return hook; }

static inline struct go_hook* go_hook_remove(struct obj_tlo *tlo, struct go_hook *hook)
	{ slist_remove(&tlo->go_hooks, &hook->n); return hook; }

struct go_hook* go_hook_remove_by_data(struct obj_tlo *tlo, void * data);

static inline struct go_hook* go_hook_fill(
		struct go_hook *h,
		void* data,
		void (*g_link) (struct o *go, struct go_hook* h),
		void (*g_unlink) (struct o *go, struct o *owner, struct o *prev, struct go_hook* h),
		void (*g_relink) (struct o *go, struct go_group *old_group, struct go *old_prev, struct go_hook* h),
		void (*go_selected) (struct o *go, struct go_hook* h),
		void (*go_unselected) (struct o *go, struct go_hook* h),
		void (*go_prop_create) (struct o *go, struct prop* p, struct go_hook* h),
		void (*go_prop_change) (struct o *go, struct go_hook* h),
		void (*go_prop_delete) (struct o *go, struct prop* p, struct go_hook* h),
		void (*go_changed) (struct go *go, const struct changed_data_generic *data, struct go_hook* h))
	{ h->data = data; h->go_link = g_link; h->go_unlink = g_unlink; h->go_relink = g_relink; h->go_selected = go_selected; h->go_unselected = go_unselected; h->go_prop_create = go_prop_create; h->go_prop_change = go_prop_change; h->go_prop_delete = go_prop_delete; h->go_changed = go_changed; return h; }


/*
 * Miscellaneous
 */

SCHEME static inline SBOOL hanger_is_mc(SHANGER h)
	{ struct go *go = hanger_get_go(h); return go->type == GOT_GROUP && (go->subtype == GOST_GROUP_TLG || go->subtype == GOST_GROUP_TEMP); }
SCHEME static inline SBOOL hanger_is_position(SHANGER h)
	{ return h->info->quantity == QT_POINT; }
SCHEME static inline SBOOL hanger_is_curve(SHANGER h)
	{ return h->info->quantity == QT_CURVE; }
static inline quantity hanger_get_quantity(struct hanger *h)
	{ ASSERT(!hanger_is_curve(h)); return h->info->q_offset ? *SKIP_OFFSET(quantity, h, h->info->q_offset) : h->info->get_quantity(h); }
static inline quantity* hanger_get_quantity_ptr(struct hanger *h, quantity *q)
	{ *q = hanger_get_quantity(h); return q; }
static inline struct geom_point hanger_get_position(struct hanger *h)
	{ ASSERT(hanger_is_position(h)); return hanger_get_quantity(h).p; }
static inline struct geom_point* hanger_get_position_ptr(struct hanger *h, struct geom_point *gp)
	{ *gp = hanger_get_position(h); return gp; }

SCHEME SBOOL go_has_dependent(SO_go go);
SCHEME SUNS go_get_count_dependents(SO_go go);
SCHEME SO_go go_get_owner_and_member(SO_go go, SO_go_group group);


// RTREE & expansion

SCHEME SVOID expand_ref(SO_obj_tlo tlo);
SCHEME SVOID expand_unref(SO_obj_tlo tlo);

const struct geom_rectangle* go_get_rtree_bbox(struct go *go);
const struct geom_rectangle* group_get_selected_rtree_bbox(struct go_group *group);
const struct geom_rectangle* group_get_rtree_bbox(struct go_group *group);
const struct geom_rectangle* tlo_get_rtree_bbox(struct obj_tlo *tlo);

int go_get_exact_bbox_after_trans (struct go *go, const struct geom_transform *t, struct geom_rectangle *bbox);

#ifdef STUPID_GROUP
#define WALK_GROUP_SMART(_group, _down, _go, _owner, _pre, _in, _post) \
	do { \
		 _owner = _group; cnode *_n = &_owner->lmember.head; \
NEXT: \
		while((_n = clist_next(&_owner->lmember, _n))) { \
			_go = SKIP_BACK(struct go, nmember, _n); \
			if(_go->type == GOT_GROUP) \
		  		{ if(_down) { _owner = go2group(_go); goto DOWN; } } \
			else do { _in; } while(0); \
		} \
		if(_owner == _group) break; \
		do { _post; } while(0); \
		_n = &_owner->go.nmember; _owner = _owner->go.owner; goto NEXT;  \
DOWN: \
		do { _pre; } while(0); \
		_n = &_owner->lmember.head; goto NEXT; \
	} while(0)
#else
#define WALK_GROUP_SMART(_group, _down, _go, _owner, _pre, _in, _post) \
	do { \
		 _owner = _group; ccnode *_n = &_owner->geom.group.list.head; \
NEXT: \
		while((_n = cclist_next(&_owner->geom.group.list, _n))) { \
			_go = geom2go(SKIP_BACK(struct geom_item, node, _n)); \
			if(_go->type == GOT_GROUP) \
		  		{ if(_down) { _owner = go2group(_go); goto DOWN; } } \
			else do { _in; } while(0); \
		} \
		if(_owner == _group) break; \
		do { _post; } while(0); \
		_n = &_owner->geom.group.item.node; _owner = go2group(geom2go(&_owner->geom.group.item.owner->item)); goto NEXT;  \
DOWN: \
		do { _pre; } while(0); \
		_n = &_owner->geom.group.list.head; goto NEXT; \
	} while(0)
#endif


// FIXME: Lepsi implementaci pomoci rtree a stromu go-cek
#define WGSB(_r, _go, _bbox) ((_r = go_get_rtree_bbox(_go)) && geom_rectangle_have_intersection(_r, _bbox))

#define WALK_GROUP_SMART_BBOX(_group, _bbox, _go, _owner, _pre, _in, _post) \
	do { const struct geom_rectangle *_r; \
		WALK_GROUP_SMART(_group, WGSB(_r,_go,_bbox), _go, _owner, _pre, if(WGSB(_r,_go,_bbox)) do { _in; } while(0), _post); \
	} while(0)



/**********************
 *                    *
 *    Clipboard       *
 *                    *
 **********************/


SCHEME SVOID clipboard_copy_selected_go(SO_go_group source, SO_go_group target, SO_go after2, SBOOL reversed);
uns clipboard_copy(struct go_group *source, const char* name_trans, char* error_string, uns reversed);
void clipboard_paste(struct go_group *target, struct go *after, uns reversed);
void clipboard_cut(struct go_group *source, uns reversed);

struct obj_tlo* clipboard_fork_tlo(struct obj_tlo *source);
struct obj* clipboard_fork_obj(struct obj *source);
void clipboard_obj_copy(void);
void clipboard_obj_cut(void);
void clipboard_obj_paste(struct obj *target, struct obj *after);


/*****************************
 *                           *
 *    Dumping functions      *
 *                           *
 *****************************/

SCHEME SVOID o_dump(SO_o o);
SCHEME SVOID o_dump_prop(SO_o o);

SCHEME SVOID tlo_dump_bbox(SO_obj_tlo tlo);
SCHEME SVOID tlo_dump_tsort(SO_obj_tlo tlo);

SCHEME SVOID go_dump_flag(SO_go go);
SCHEME SVOID go_dump_hanger(SO_go go);
SCHEME SVOID go_dump_anchor(SO_go go);
SCHEME SVOID go_dump_bbox(SO_go go);
SCHEME SVOID go_dump_type(SO_go go);
SCHEME SVOID go_dump_all(SO_go go);

SCHEME SVOID group_dump_go(SO_go_group group);
SCHEME SVOID group_dump_subtree(SO_go_group root);
SCHEME SVOID group_dump_selected_go(SO_go_group group);

SCHEME SVOID hanger_dump(SHANGER h);
SCHEME SVOID hanger_dump_anchor(SHANGER h);
SCHEME SVOID anchor_dump(SANCHOR a);

SCHEME SVOID kernel_dump_memory(void);


#endif
