/*
 * Kernel - GO object
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

//#define LOCAL_DEBUG
#include "kernel/internal.h"
#include "lib/hashfunc.h"

#ifdef DEBUG_EXPAND
#define DBG_EXPAND(...) msg(__VA_ARGS__)
#else
#define DBG_EXPAND(...) do { } while(0)
#endif

#define ASSERT_TLO_CACHE(_tlo) ASSERT_TLO(_tlo); ASSERT(_tlo->expand_ref_count)
#define ASSERT_GO_CACHE_TRANS(_go) ASSERT_TLO_CACHE(trans_tlo); ASSERT(go_get_tlo(void2go(_go)) == trans_tlo)
#define ASSERT_GO_CACHE_WORK(_go) ASSERT_TLO(trans_tlo); ASSERT(&go_get_root(void2go(_go))->group == tlo_get_tlg(trans_tlo))
#define ASSERT_GO_CACHE_INFO(_go) ASSERT(go_get_tlo(void2go(_go))->expand_ref_count)

struct geom_point* point_transform(struct geom_point *p, const struct geom_transform *t)
{
	if(geom_point_transform(p, t, p) != 0)
		TRANS_FAIL(("Numeric error."), ("geom_point_transform failed"));
	return p;
}

UNUSED static void rectangle_dump(struct geom_rectangle *gr)
{
	msg("%f %f %f %f", gr->ll.x, gr->ll.y, gr->ur.x, gr->ur.y);
}

int go_approx(struct geom_approx *approx, struct go *go, uns flags)
{
	struct geom_expansion *e = go_expand(go);
	geom_approx_path(approx, &e->path.path, flags);
	geom_expansion_unref(e);
	return 1;
}

#define HASH_NODE struct rtree_bbox
#define HASH_PREFIX(x) rtree_bbox_ ## x
#define HASH_KEY_ATOMIC go
#define HASH_WANT_CLEANUP
#define HASH_WANT_FIND
#define HASH_WANT_INSERT
#define HASH_WANT_DELETE
#define HASH_WANT_REMOVE
#define HASH_GIVE_ALLOC
static inline void rtree_bbox_free(void *p) { ffree(p); }
#define HASH_GIVE_HASHFN
static inline uns rtree_bbox_hash(struct go *go) { return hash_ptr(go); }
#define HASH_CONSERVE_SPACE
#define HASH_ATOMIC_TYPE struct go*
#include "lib/sht.h"

const uns go_rtree_bbox_size = sizeof(struct rtree_bbox_bucket);

#define RECOMPUTE(_id) case GO_TYPE(_id): GLUEE(GO_NAME(_id), _recompute) ((GO_STRUCT(_id)*)go, &new_bbox); break;
#define SUBRECOMPUTE(_id, _sid) case GO_SUBTYPE(_id, _sid): GLUEE(GO_SUBNAME(_id, _sid), _recompute) ((GO_SUBSTRUCT(_id, _sid)*)spec, &new_bbox); break;

void go_recompute(struct go *go)
{
	ASSERT_GO_CACHE_WORK(go);
	DBG("GO recompute: %s at %p\n", go_lookup_name(go), go);

	struct rtree_bbox *rb = rtree_bbox_find(go);
	struct geom_rectangle old_bbox = rb ? rb->rtree.bbox : (struct geom_rectangle) { { 1, 1 }, { 0, 0 } },
								 new_bbox;	

	if ((! rb) && (go -> type == GOT_GROUP) && (go -> subtype == GOST_GROUP_PATH))
	{
		old_bbox = ((struct go_group *) go) -> bbox_all;
	}
								 
	
	GO_TYPE_SWITCH(
		GO_SUBTYPE_SWITCH(GROUP,
			SUBRECOMPUTE(GROUP, GROUP)
			SUBRECOMPUTE(GROUP, PATH)
			SUBRECOMPUTE(GROUP, TLG))
		RECOMPUTE(POINT)
		RECOMPUTE(SEGMENT)
		RECOMPUTE(BEZIER)
		GO_SUBTYPE_SWITCH(ELARC,
			SUBRECOMPUTE(ELARC, XYR)
			SUBRECOMPUTE(ELARC, FOCI)
			SUBRECOMPUTE(ELARC, 3SMALL)
			SUBRECOMPUTE(ELARC, 3ECC)
			SUBRECOMPUTE(ELARC, XY1ECC))
		RECOMPUTE(PARPOINT)
		RECOMPUTE(INTERSECTION)
		RECOMPUTE(TEXT)
		RECOMPUTE(TEX)
		RECOMPUTE(DECPOINT)
		RECOMPUTE(ARROW)
	)
		
	if(rb)
		geom_rtree_update(&trans_tlo->rtree, &rb->rtree, 0, &new_bbox);
	else if(trans_tlo->expand_ref_count && go->type != GOT_GROUP)
	{
		struct rtree_bbox_bucket *rbb = fmalloc(trans_tlo->frtree);
		rtree_bbox_insert(go, rbb);
		geom_rtree_insert(&trans_tlo->rtree, &rbb->n.rtree, &new_bbox);
	}
	hook_call_prop_change(go2o(go));
	hook_call_go_changed(go, &old_bbox, &new_bbox);
}

#define TRANSFORM(_id) case GO_TYPE(_id): GLUEE(GO_NAME(_id), _transform) ((GO_STRUCT(_id)*)go, t, &new_bbox); break;
#define SUBTRANSFORM(_id, _sid) case GO_SUBTYPE(_id, _sid): GLUEE(GO_SUBNAME(_id, _sid), _transform) ((GO_SUBSTRUCT(_id, _sid)*)spec, t, &new_bbox); break;

void go_transform(struct go *go, struct geom_transform2 *t2)
{
	ASSERT_GO_CACHE_WORK(go);
	DBG("GO transform: %s at %p\n", go_lookup_name(go), go);

	struct rtree_bbox *rb = rtree_bbox_find(go);
	struct geom_transform *t = &(t2->primary);
	struct geom_rectangle old_bbox = rb ? rb->rtree.bbox : (struct geom_rectangle) { { 1, 1 }, { 0, 0 } },
								 new_bbox;	

	GO_TYPE_SWITCH(
		GO_SUBTYPE_SWITCH(GROUP,
			SUBTRANSFORM(GROUP, GROUP)
			SUBTRANSFORM(GROUP, PATH)
			SUBTRANSFORM(GROUP, TLG))
		TRANSFORM(POINT)
		TRANSFORM(SEGMENT)
		TRANSFORM(BEZIER)
		GO_SUBTYPE_SWITCH(ELARC,
			SUBTRANSFORM(ELARC, XYR)
			SUBTRANSFORM(ELARC, FOCI)
			SUBTRANSFORM(ELARC, 3SMALL)
			SUBTRANSFORM(ELARC, 3ECC)
			SUBTRANSFORM(ELARC, XY1ECC))
		TRANSFORM(PARPOINT)
		TRANSFORM(INTERSECTION)
		TRANSFORM(TEXT)
		TRANSFORM(TEX)
		RECOMPUTE(DECPOINT)
		RECOMPUTE(ARROW)
		);

	if(rb)
		geom_rtree_update(&trans_tlo->rtree, &rb->rtree, 0, &new_bbox);
	else if(trans_tlo->expand_ref_count && go->type != GOT_GROUP)
	{
		struct rtree_bbox_bucket *rbb = fmalloc(trans_tlo->frtree);
		rtree_bbox_insert(go, rbb);
		geom_rtree_insert(&trans_tlo->rtree, &rbb->n.rtree, &new_bbox);
	}
	hook_call_prop_change(go2o(go));
	hook_call_go_transformed(go, &old_bbox, &new_bbox, t2);
}

static int
one_hanger_bbox (struct hanger *h, const struct geom_transform *t, struct geom_rectangle *bbox)
{
  int retval = 0;
  struct geom_point p = hanger_get_position (h);
  retval = geom_point_transform (&p, t, &p);
  bbox -> ll = bbox -> ur = p;
  return (retval);
}

int go_get_exact_bbox_after_trans (struct go *go, const struct geom_transform *t, struct geom_rectangle *bbox)
{
  int retval = 0;

  switch (go -> type)
    {
    case GOT_GROUP:
      {
	struct go *child;
	struct geom_rectangle child_bbox;

	geom_rectangle_set_empty (bbox);
	WALK_GO_RECURSIVE_WITHOUT_ROOT (go, child)
	  {
	    retval = retval || go_get_exact_bbox_after_trans (child, t, &child_bbox);
	    retval = retval || geom_rectangle_merge (&child_bbox, bbox, bbox);
	  }
      break;
      }
    case GOT_POINT:
      retval = one_hanger_bbox (((struct go_point *) go) -> h, t, bbox);
      break;
    case GOT_PARAMETRIC_POINT:
      retval = one_hanger_bbox (((struct go_parametric_point *) go) -> h, t, bbox);
      break;
    case GOT_INTERSECTION:
      retval = one_hanger_bbox (((struct go_intersection *) go) -> h, t, bbox);
      break;
    case GOT_SEGMENT:
      {
	struct geom_point p1 = hanger_get_position (((struct go_segment *) go) -> h + 1);
	struct geom_point p2 = hanger_get_position (((struct go_segment *) go) -> h + 2);
	retval = retval || geom_point_transform (&p1, t, &p1);
	retval = retval || geom_point_transform (&p2, t, &p2);
	retval = retval || geom_rectangle_init_from_points (bbox, &p1, &p2);
	break;
      }
    case GOT_BEZIER:
      {
	struct geom_bezier b;
	GEOM_ITEM_CREATE(&b, bezier);

	retval = geom_bezier_transform (&(((struct go_bezier *) go) -> geom), (void *)t, &b);
	retval = retval || geom_bezier_bbox (&b, bbox);

	geom_bezier_destroy(&b);
	break;
      }
    case GOT_TEXT: // FIXME - return should be exact bbox, not this
      go_text_get_text_bbox ((struct go_text *) go, bbox);
      retval = geom_rectangle_transform (bbox, t, bbox);
      break;
    case GOT_TEX_TEXT:  // FIXME - return should be exact bbox, not this
      go_tex_text_get_text_bbox ((struct go_tex_text *) go, bbox);
      retval = geom_rectangle_transform (bbox, t, bbox);
      break;
    default:
      {
	struct geom_path p;
	GEOM_ITEM_CREATE (&p, path);

	struct geom_expansion *e = go_expand (go);
	ASSERT (e != NULL);

	retval = geom_path_transform (&e->path.path, (void *)t, &p);
	retval = retval || geom_path_bbox (&p, bbox);
	go_expansion_unref(e);

	geom_path_destroy(&p);
	break;
      }
    }

  real sw = prop_lookup_real(go2o(go), string_stroke_width, 0);
  retval = retval || geom_rectangle_enlarge (bbox, sw, bbox);
  return (retval);
}

void go_altered(struct go* go)
{
	ASSERT_TLO(trans_tlo);
	ASSERT(go_get_tlo(go) == trans_tlo);
	if(&go_get_root(go)->group != tlo_get_tlg(trans_tlo))
		return;
	hook_call_prop_change(go2o(go));
	if(!trans_tlo->expand_ref_count)
		return;
	hook_call_go_altered(go, go_get_rtree_bbox(go));
}


static void group_compute_rtree_bbox(struct go_group *group)
{
	struct go *go;
	const struct geom_rectangle *r;
	geom_rectangle_set_empty(&group->bbox_all);
	geom_rectangle_set_empty(&group->bbox_sel);

	WALK_GROUP_GO(group, go)
	{
		r = go_get_rtree_bbox(go);
		geom_rectangle_merge(r, &group->bbox_all, &group->bbox_all);
		if(go_is_selected(go))
			geom_rectangle_merge(r, &group->bbox_sel, &group->bbox_sel);
	}
#ifdef LOCAL_DEBUG
	msg("group_compute_rtree_bbox %s at %p: all = ", go_lookup_name(&group->go), group);
	rectangle_dump(&group->bbox_all);
	msg(", sel = ");
	rectangle_dump(&group->bbox_sel);
	msg("\n");
#endif
}

const struct geom_rectangle* group_get_rtree_bbox(struct go_group *group)
{
	ASSERT_GROUP(group);
	ASSERT(group->go.type != GOST_GROUP_TEMP && group->go.type != GOST_GROUP_ZOMBIE);
	if(!isfinite(group->bbox_all.ur.x))
		group_compute_rtree_bbox(group);
	return &group->bbox_all;
}

const struct geom_rectangle* group_get_selected_rtree_bbox(struct go_group *group)
{
	ASSERT_GROUP(group);
	ASSERT(group->go.type != GOST_GROUP_TEMP && group->go.type != GOST_GROUP_ZOMBIE);
	if(!isfinite(group->bbox_all.ur.x))
		group_compute_rtree_bbox(group);
	return &group->bbox_sel;
}

const struct geom_rectangle* go_get_rtree_bbox(struct go *go)
{
	static struct geom_rectangle bbox_invalid = { { 1, 1 }, { 0, 0 } };
	ASSERT_GO_CACHE_INFO(go);
	if(go->type == GOT_GROUP)
		return group_get_rtree_bbox(go2group(go));
	struct rtree_bbox *rb = rtree_bbox_find(go);		
	return rb ? &rb->rtree.bbox : &bbox_invalid;
}

const struct geom_rectangle* tlo_get_rtree_bbox(struct obj_tlo *tlo)
{
	ASSERT_TLO_CACHE(tlo);
	return &tlo->rtree.root->bbox;
}

void expand_ref(struct obj_tlo *tlo)
{
	ASSERT_TLO(tlo);
	tlo->expand_ref_count++;
	DBG_EXPAND("  Increase expand reference: tlo %s at %p, new ref_count %d\n", obj_lookup_name(&tlo->obj), tlo, tlo->expand_ref_count);
	if(tlo->expand_ref_count > 1)
		return;

	geom_rtree_init(&tlo->rtree);

	char buf[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN(tlo, "Create rtree", buf)
		tsort_start(tlo);
		tsort_insert_group(tlo_get_tlg(tlo), tlo);
		tsort_dirty(&tlo_get_tlg(tlo)->go);
		tsort_call_recompute(tlo);
	TRANS_FAILED
		die(buf);
	TRANS_END
}	

void expand_unref(struct obj_tlo *tlo)
{
	ASSERT_TLO_CACHE(tlo);
	tlo->expand_ref_count--;
	DBG_EXPAND("  Decrease expand reference: tlo %s at %p, new ref_count %d\n", obj_lookup_name(&tlo->obj), tlo, tlo->expand_ref_count);
	if(tlo->expand_ref_count > 0)
		return;

	struct go_group *tlg = tlo_get_tlg(tlo), *owner;
	struct go *go;
	geom_rtree_clear(&tlo->rtree);
	WALK_GROUP_SMART(tlg, 1, go, owner, , rtree_bbox_delete(go), );
}

void go_cache_remove(struct go *go)
{
	ASSERT_GO(go);
	ASSERT_TLO(trans_tlo);
	if(!trans_tlo->expand_ref_count)
		return;
	struct rtree_bbox *rb = rtree_bbox_find(go);
	if(rb)
	{
		geom_rtree_delete(&trans_tlo->rtree, &rb->rtree);
		rtree_bbox_remove(rb);
	}
}

void kernel_cache_init(void)
{
	rtree_bbox_init();
}

void kernel_cache_finish(void)
{
	rtree_bbox_cleanup();
}

