/*
 * Kernel - GO linking functions
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "kernel/internal.h"

void go_insert(struct go *go, struct go_group *group, struct go *after)
{
#ifdef STUPID_GROUP
	clist_insert_after(&go->nmember, after ? &after->nmember : &group->lmember.head);
	go->owner = group;
#else
	if(after)
		geom_group_insert_after(go2geom(go), go2geom(after));
	else
		geom_group_add_head(&group->geom, go2geom(go));
#endif
	ASSERT(go_get_group(go) == group);
}

static inline void go_remove(struct go *go)
{
#ifdef STUPID_GROUP
	clist_remove(&go->nmember);
#else
	geom_group_remove(go2geom(go));
#endif
}

void go_relink_low(struct go *go, struct go_group *group, struct go *prev)
{
	ASSERT_GO(go);
	ASSERT_GROUP(group);
	ASSERT(!prev || (prev->kind == T_GO && go_get_group(prev) == group));
	ASSERT(!go_is_selected(go));
	ASSERT(go->type != GOT_GROUP || !group_is_go_descentant(go2group(go), &group->go));
	
	struct go_group *old_group = go_get_group(go);
	struct go *old_prev = group_get_prev_go(old_group, go), *g;
	struct go_tlg *old_tlg = go_get_root(go),
				 *new_tlg = go_get_root(&group->go);
	struct anchor *a;
	int i;

	// When go is moved from zombie to name-space it must obtained one reference. Otherwise, it has reference from temp.
	if(go_get_group(go) == &tlo_zombie->tlg->group)
		go_ref(go);

	go_remove(go);
	go_insert(go, group, prev);
	
	if(new_tlg->group.go.subtype == GOST_GROUP_TLG)
	{
		if(old_tlg->group.go.subtype == GOST_GROUP_TLG)
		{
			hook_call_go_relink(go, old_group, old_prev);
	
	// Paths need to be checked whether they are continuos after this relink.
			WALK_GO_RECURSIVE(go, g) 
				tsort_dirty(g);
		}
		else
		{
			WALK_GO_RECURSIVE(go, g)
			{
				WALK_GO_ANCHOR(g, a, i)
					clist_add_tail(&a->h->lanchor, &a->nanchor);
				tsort_dirty(g);
			}
			hook_call_go_link(go);
		}
	}	
	else if(old_tlg->group.go.subtype == GOST_GROUP_TLG)
	{
		hook_call_go_unlink(go, old_group, old_prev);
		WALK_GO_RECURSIVE(go, g)
		{
			go_cache_remove(g);
			WALK_GO_ANCHOR(g, a, i)
				clist_remove(&a->nanchor);
			tsort_remove(g, old_group, old_tlg->tlo);
			geom_item_expansion_release(go2geom(g));
		}
	}

	if(group == &tlo_zombie->tlg->group)
		go_unref(go);		// go was moved from name-space to zombie so it lost one reference
}

struct go* go_link_after(struct go *go, struct go_group *group, struct go *after)
{
	ASSERT(go_get_root(go) == trans_tlo->temp && go_get_tlo(&group->go) == trans_tlo);
	TSORT_NO_RECOMPUTE_BEGIN
		go_relink_low(go, group, after);
		TRANS_NEW_UNDO(TRANS_GO_LINK, go_link, go_ref(go), group, after);
	TSORT_NO_RECOMPUTE_END
	return go;
}

struct go* go_unlink(struct go *go)
{
	ASSERT(go_get_tlo(go) == trans_tlo);
	struct go_group *group = go_get_group(go);

	TSORT_NO_RECOMPUTE_BEGIN
		go_unselect(go);
		TRANS_NEW_UNDO(TRANS_GO_UNLINK, go_link, go_ref(go), group, group_get_prev_go(group, go));
		go_relink_low(go, &trans_tlo->temp->group, NULL);
	TSORT_NO_RECOMPUTE_END
	
	return go;
}

struct go* go_relink(struct go *go, struct go_group *group, struct go *after)
{
	ASSERT(go_get_root(go) == go_get_root(&group->go) && go_get_tlo(go) == trans_tlo);

	struct go_group *old_group = go_get_group(go);
	struct go *old_after = group_get_prev_go(old_group, go);
	uns selected = go_is_selected(go);

	TSORT_NO_RECOMPUTE_BEGIN
		if(selected)
			go_unselect(go);
		TRANS_NEW_UNDO(TRANS_GO_RELINK, go_link, go_ref(go), old_group, old_after);
		go_relink_low(go, group, after);
		if(selected)
			go_select(go);
	TSORT_NO_RECOMPUTE_END

	return go;
}


/*
 * Hanger and anchor linking/rehanging functions
 */

static void hanger_relink(struct hanger_smart *hs, struct go *go)
{
	clist_remove(&hs->nhanger);
	hs->go = go;
	clist_add_tail(go_get_lhanger(go), &hs->nhanger);
}

static void anchor_relink(struct anchor *a, struct hanger *h)
{
	ASSERT(hanger_get_go(h) != anchor_get_go(a));
	clist_remove(&a->nanchor);
	a->h = h;
	clist_add_tail(&h->lanchor, &a->nanchor);
}

struct hanger* hanger_new_mc(struct geom_point *gp)
{
	ASSERT_GO_TRANS(NULL);
	if(!geom_point_valid(gp))
		TRANS_FAIL(("Numeric error."), ("Function hanger_new_mc received invalid point [ %f, %f ].", gp->x, gp->y));
	struct hanger_smart* hs = xmalloc(sizeof(struct hanger_smart));

	hs->h.info = go_info[GOT_GROUP][GOST_GROUP_TLG].h_info + QT_POINT;
	clist_init(&hs->h.lanchor);
	hs->go = &trans_tlo->temp->group.go;
	clist_add_tail(&trans_tlo->temp->mc, &hs->nhanger);
	hs->ref_count = 0;
	hs->q.p = *gp;
	ASSERT_HANGER(hs);

	return &hs->h;
}

struct hanger* hanger_new_mc_copy(struct hanger *h)
{
	ASSERT_GO_TRANS(NULL);
	struct hanger_smart* hs = xmalloc(sizeof(struct hanger_smart));

	hs->h.info = go_info[GOT_GROUP][GOST_GROUP_TLG].h_info + h->info->quantity;
	clist_init(&hs->h.lanchor);
	hs->go = &trans_tlo->temp->group.go;
	clist_add_tail(&trans_tlo->temp->mc, &hs->nhanger);
	hs->ref_count = 0;

	switch(h->info->quantity)
	{
		case QT_POINT:
			hs->q.p = hanger_get_position(h);
			break;

		case QT_CURVE:
			hs->q.p = (struct geom_point) { NAN, NAN };
			break;

		default:
			die("Function hanger_new_mc_copy hasn't case for hanger type %d.\n", h->info->quantity);
	}
	
	ASSERT_HANGER(hs);
	return &hs->h;
}

// FIXME: Asi by stacil jen jeden krivkovy hanger v kazdem tlg.
struct hanger* hanger_new_mc_curve(void)
{
	ASSERT_GO_TRANS(NULL);
	struct hanger_smart* hs = xmalloc(sizeof(struct hanger_smart));

	hs->h.info = go_info[GOT_GROUP][GOST_GROUP_TLG].h_info + QT_CURVE;
	clist_init(&hs->h.lanchor);
	hs->go = &trans_tlo->temp->group.go;
	clist_add_tail(&trans_tlo->temp->mc, &hs->nhanger);
	hs->ref_count = 0;
	ASSERT_HANGER(hs);

	return &hs->h;
}

void anchor_rehang(struct anchor* a, struct hanger* h)
{
	struct go *go = anchor_get_go(a);
	ASSERT(go_can_work(go) && (go_can_work(hanger_get_go(h)) || hanger_get_go(h) == tlg2go(trans_tlo->temp)));

	if(a->h == h)
		return;

	if(h->info->quantity != a->info->quantity)
		TRANS_FAIL(("Incompatible anchor/hanger."), ("Quantity: anchor %s, hanger %s",
				string_quantity_type[a->info->quantity]->text, string_quantity_type[h->info->quantity]->text));

	TRANS_NEW_UNDO(TRANS_ANCHOR_REHANG, anchor_rehang, a->h, a); // Reference is moved from the anchor to undo.
	go_ref(go);		// Reference for undo.
	anchor_relink(a, hanger_ref(h));  // Reference for anchor.

	if(tsort_is_go_presorted(hanger_get_go(h)) && (trans_tlo->obj.flags & OF_TSORT_PRESORT))
	{
		tsort_unsort(trans_tlo);
		tsort_presort(trans_tlo);
	}
	
	TSORT_NO_RECOMPUTE_BEGIN
		tsort_dirty(go);
	TSORT_NO_RECOMPUTE_END
}

void anchor_rehang_mc_no_recompute(struct anchor* a, struct geom_point *gp)
{
	ASSERT(go_can_work(anchor_get_go(a)));
	TRANS_NEW_UNDO(TRANS_ANCHOR_REHANG, anchor_rehang, a->h, a);	// Reference of the old hanger is moved from the anchor to undo.
	go_ref(anchor_get_go(a));		// Reference for undo.
	anchor_relink(a, hanger_ref(hanger_new_mc(gp)));  // Reference for anchor.
}

void anchor_rehang_copy_no_recompute(struct anchor* a, struct hanger *h)
{
	ASSERT(go_can_work(anchor_get_go(a)));
	TRANS_NEW_UNDO(TRANS_ANCHOR_REHANG, anchor_rehang, a->h, a);	// Reference of the old hanger is moved from the anchor to undo.
	go_ref(anchor_get_go(a));		// Reference for undo.
	anchor_relink(a, hanger_ref(hanger_new_mc_copy(h)));  // Reference for anchor.
}

int hanger_mc_move(struct hanger* h, struct geom_point* gp)
{
	ASSERT(h && hanger_is_mc(h) && gp && isfinite(gp->x) && isfinite(gp->y) && go_can_work(hanger_get_go(h)));
	struct hanger_smart *mc = hanger2smart(h);

	if(IS_EQUAL_XY(mc->q.p, *gp))
		return 0;

	TSORT_NO_RECOMPUTE_BEGIN
	TRANS_NEW_UNDO(TRANS_MOUSE_CLICK_MOVE, mc_move, mc, mc->q.p);
	hanger_ref(h);		// Reference for undo.
	mc->q.p = *gp;
	tsort_dirty_hanger(h);
	TSORT_NO_RECOMPUTE_END

	return 1;
}



/*
 * Miscellaneous linking functions
 */

void group_relink_all_go(struct go_group *source, struct go_group *target, struct go *after)
{
	ASSERT_GROUP(source);
	struct go *go;
	TSORT_NO_RECOMPUTE_BEGIN
	  	while((go = group_get_last_go(source)))
			go_relink(go, target, after);
	TSORT_NO_RECOMPUTE_END
}

void group_relink_selected_go(struct go_group *source, struct go_group *target, struct go *after)
{
	ASSERT_GROUP(source);
	struct go *go, *tmp;
	
	TSORT_NO_RECOMPUTE_BEGIN
		for(go = group_get_last_go(source); go && (tmp = group_get_prev_go(source, go), 1); go = tmp)
			if(go_is_selected(go))
				go_relink(go, target, after);
	TSORT_NO_RECOMPUTE_END
}

void tlo_unlink_all_go(void)
{
	ASSERT(trans_present && trans_tlo && trans_tlo->obj.type == OT_TLO);
	struct go *go;
	
	TSORT_NO_RECOMPUTE_BEGIN
		while((go = group_get_first_go(&trans_tlo->tlg->group)))
			go_unlink(go);
	TSORT_NO_RECOMPUTE_END
}

void group_unlink_selected_go(struct go_group *group)
{
	ASSERT(go_can_work(&group->go));
	struct go *go;
	
	TSORT_NO_RECOMPUTE_BEGIN
		while((go = group_get_first_selected_go(group)))
			go_unlink(go);
	TSORT_NO_RECOMPUTE_END
}

void go_remove_rehang(struct go *go)
{
	ASSERT(go_can_work(go));
	struct go *g;
	struct hanger *h, *mc;
	struct anchor *a, *tmp;

	TSORT_NO_RECOMPUTE_BEGIN
		WALK_GO_RECURSIVE(go, g)
			WALK_GO_HANGER_BEGIN(g, h)
				mc = NULL;
				WALK_HANGER_ANCHOR_DELSAFE(h, a, tmp)
					if(!go_is_father(anchor_get_go(a), go))
					{
						if(!mc)
							mc = hanger_new_mc_copy(h);
						anchor_rehang(a, mc);
					}
		WALK_GO_HANGER_END
	go_unlink(go);
	TSORT_NO_RECOMPUTE_END
}

void group_remove_rehang_selected_go(struct go_group *group)
{
	ASSERT_GROUP(group);
	ASSERT(go_can_work(&group->go));
	struct go* go;
	
	TSORT_NO_RECOMPUTE_BEGIN
	while((go = group_get_first_selected_go(group)))
		go_remove_rehang(go);
	TSORT_NO_RECOMPUTE_END
}

uns go_remove_carefully(struct go *go)
{
	ASSERT(go_can_work(go));
	struct go *g, *dep;

	TSORT_NO_RECOMPUTE_BEGIN
		WALK_GO_RECURSIVE(go, g)
			WALK_GO_DEPENDENT_BEGIN(g, dep)
				if(!go_is_father(dep, go))
					return 0;
		WALK_GO_DEPENDENT_END
	go_unlink(go);
	TSORT_NO_RECOMPUTE_END

	return 1;
}



/*
 * Selection functions
 */

void go_select_low(struct go *go)
{
	ASSERT(go_can_work(go) && !go_is_selected(go));
	go->flags |= GOF_SELECTED;
	clist_add_tail(&go_get_group(go)->lselected, &go->nselected);
	tsort_dirty(&go_get_group(go)->go);
	hook_call_go_selected(go);
}

struct go* go_select(struct go *go)
{
	if(go_is_selected(go))
		return go;
	
	TSORT_NO_RECOMPUTE_BEGIN
		TRANS_NEW_UNDO(TRANS_GO_SELECT, go_select, go_ref(go));
		go_select_low(go);
	TSORT_NO_RECOMPUTE_END
		
	return go;
}

void group_select_all_go(struct go_group *group)
{
	ASSERT(go_can_work(&group->go));
	struct go* go;

	TSORT_NO_RECOMPUTE_BEGIN
		WALK_GROUP_GO(group, go)
			go_select(go);
	TSORT_NO_RECOMPUTE_END
}

void go_unselect_low(struct go *go)
{
	ASSERT(go_can_work(go) && go_is_selected(go));
	go->flags &= ~GOF_SELECTED;
	clist_remove(&go->nselected);
	tsort_dirty(&go_get_group(go)->go);
	hook_call_go_unselected(go);
}

struct go* go_unselect(struct go* go)
{
	if(!go_is_selected(go))
		return go;

	TSORT_NO_RECOMPUTE_BEGIN
		TRANS_NEW_UNDO(TRANS_GO_UNSELECT, go_select, go_ref(go));
		go_unselect_low(go);
	TSORT_NO_RECOMPUTE_END
		
	return go;
}

void group_unselect_all_go(struct go_group *group)
{
	ASSERT(go_can_work(&group->go));
	struct go* go;
	
	TSORT_NO_RECOMPUTE_BEGIN
		WALK_GROUP_GO(group, go)
			go_unselect(go);
	TSORT_NO_RECOMPUTE_END
}

void tlo_unselect_all_go(void)
{
	ASSERT_GO_TRANS(NULL);
	struct go* go;
	
	TSORT_NO_RECOMPUTE_BEGIN
		WALK_GO_RECURSIVE(tlg2go(trans_tlo->tlg), go)
			go_unselect(go);
	TSORT_NO_RECOMPUTE_END
}


/*
 * Miscellaneous
 */

struct go* go_recursive_next(const struct go *root, struct go *walk)
{
	ASSERT(root->kind == T_GO && walk->kind == T_GO);
	struct go* go;
	if(walk->type == GOT_GROUP && (go = group_get_first_go(go2group(walk))))
		return go;
	else
		for(; walk != root; walk = &go_get_group(walk)->go)
			if((go = group_get_next_go(go_get_group(walk), walk)))
				return go;
	return NULL;
}

struct go* go_recursive_prev(const struct go *root, struct go *walk)
{
	ASSERT(root->kind == T_GO && walk->kind == T_GO);
	struct go* go;
	if(walk->type == GOT_GROUP && (go = group_get_last_go(go2group(walk))))
		return go;
	else
		for(; walk != root; walk = &go_get_group(walk)->go)
			if((go = group_get_prev_go(go_get_group(walk), walk)))
				return go;
	return NULL;
}

uns go_has_dependent(struct go* go)
{
	struct hanger *h;
	WALK_GO_HANGER_BEGIN(go, h)
		if(hanger_get_first_anchor(h))
			return 1;
	WALK_GO_HANGER_END
	return 0;
}

uns go_get_count_dependents(struct go *go)
{
	ASSERT_GO(go);
	struct anchor *a;
	struct hanger *h;
	uns count = 0;
	WALK_GO_HANGER_BEGIN(go, h)
		WALK_HANGER_ANCHOR(h, a)
		 count++;
	WALK_GO_HANGER_END
	return count;
}

/**
 * Returns GO satisfied two conditions:
 * Is is (direct) member of given group.
 * It is given go or go's (undirect) owner.
 * Returns NULL if such GO not exists.
 */
struct go* go_get_owner_and_member(struct go* go, struct go_group *group)
{
	ASSERT_GO(go);
	ASSERT_GROUP(group);
	struct go_group *owner;
	for(; (owner = go_get_group(go)); go = &owner->go)
		if(owner == group)
			return go;
	return NULL;
}

struct obj_tlo* go_get_tlo(struct go *go)
{
	return go_get_root(go)->tlo;
}


/*
 * Inicialization and cleanup
 */

struct go_group* group_init(struct go_group *group, uns subtype)
{
	go_init(&group->go, GOT_GROUP, subtype);
	geom_group_create(&group->geom.group);
	clist_init(&group->lselected);

#ifdef STUPID_GROUP
	clist_init(&group->lmember);
#else
#endif

	return group;
}

#undef go_init
struct go* go_init(struct go *go, uns type, uns subtype, ...)
{
	uns i;
	struct go_info *gi = &go_info[type][subtype];
	va_list ap;
	struct hanger *h;
	struct anchor *a;

	o_init((struct o*)go, T_GO, type, subtype, gi->prop_virtual_list);
	ASSERT(!go->temp_data && !go->nfifo.next);

#define INSTANCE_INIT(_id) case GO_TYPE(_id): geom_instance_zero_init(go2geom(go), GEOM_CLASS(GO_GEOM(_id))); break;
	GO_TYPE_SWITCH(GO_FOR_ALL_ID(INSTANCE_INIT));	

	if(trans_tlo && trans_tlo->obj.type == OT_TLO)
		go_insert(go, &trans_tlo->temp->group, NULL);

	h = SKIP_OFFSET(struct hanger, go, gi->h_offset);
	for(i = 0; i < gi->h_count; i++, h++)
	{
		h->info = gi->h_info + i;
		clist_init(&h->lanchor);
		ASSERT_HANGER(h);
	}
	if(gi->hs_offset)
		clist_init(SKIP_OFFSET(clist, go, gi->hs_offset));

	va_start(ap, subtype);
	a = SKIP_OFFSET(struct anchor, go, gi->a_offset);
	for(i = 0; i < gi->a_count; i++, a++)
	{
		h = va_arg(ap, struct hanger*);
		a->info = gi->a_info + i;
		a->h = hanger_ref(h);
		ASSERT_ANCHOR(a);
	}
	ASSERT_KERNEL(va_arg(ap, uns) == END_OF_ARGS);
	va_end(ap);

	WALK_GO_ANCHOR(go, a, i)
		if(a->info->quantity != a->h->info->quantity)
			TRANS_FAIL(("Incompatible anchor/hanger."), ("Quantity: anchor %s, hanger %s",
						string_quantity_type[a->info->quantity]->text, string_quantity_type[a->h->info->quantity]->text));

	return go;
}

void go_finish(struct go *go)
{
	// If vrr failed in this assert then you may recompute your referencing. You probably call unref more times than ref.
	// The go's tlo must be the zombie otherwise this go has another reference.
	ASSERT_GO_SUBTYPE(go_get_group(go), GROUP, ZOMBIE);
	ASSERT(go->ref_count == 0);

	clist *l;
	struct hanger_smart *hs;
	int i;
	struct anchor *a;

	go_remove(go);

	WALK_GO_ANCHOR(go, a, i)
		hanger_unref(a->h);

	if((l = go_get_lhanger(go)))
		while((hs = go_lhanger_get_first_hanger_smart(l)))
		{
			ASSERT(hs->go == go);
			hanger_relink(hs, &tlo_zombie->tlg->group.go);
		}

#define DESTROY(_id) GEOM_INSTANCE_DESTROY2(&((GO_STRUCT(_id)*)go)->geom, GO_GEOM(_id))
#define DESTROY_CASE(_id) case GO_TYPE(_id): DESTROY(_id); break;		

	switch(go->type)
	{
	case GOT_GROUP:
		{
			struct go *g;
			struct go_group *group = go2group(go);
			while((g = group_get_first_go(group)))
			{
				go_remove(g);
				go_insert(g, &tlo_zombie->tlg->group, NULL);
				go_unref(g);
			}
			DESTROY(GROUP);
		}
		break;
	
	case GOT_TEXT:
		string_unref(((struct go_text *)go)->text);
		xfree(((struct go_text*)go)->transformation);
		DESTROY(TEXT);
		break;

	case GOT_TEX_TEXT:
		tex_process_unref(((struct go_tex_text*)go)->process);
		DESTROY(TEX);
		break;

	DESTROY_CASE(POINT)
	DESTROY_CASE(PARPOINT)
	DESTROY_CASE(INTERSECTION)
	DESTROY_CASE(DECPOINT)
	DESTROY_CASE(ARROW)
	DESTROY_CASE(SEGMENT)
	DESTROY_CASE(BEZIER)
	DESTROY_CASE(ELARC)
	
	default:
		die("Function go_finish hasn't label for the go type[%d] %s.\n", go->type, string_go_type[go->type]->text);
	}
	o_finish(go2o(go));
}

void go_info_finish(struct go_info *gi, uns count)
{
	uns i, j;
	for(i = 0; i < count; i++, gi++)
	{
		string_unref(gi->name);
		for(j = 0; j < gi->h_count; j++)
			string_unref(gi->h_info[j].name);
		for(j = 0; j < gi->a_count; j++)
			string_unref(gi->a_info[j].name);
	}
}



/*
 * Referencing functions
 */

#undef go_ref
struct go* go_ref(struct go *go DEBUG_REF_O_ARGS)
{
	ASSERT_GO(go);
	ASSERT(go->ref_count > 0);
	go->ref_count++;
	DBG_O("  Increase reference: go %s at %p, new ref_count %d, file %s, line %d, fce %s\n", go_lookup_name(go), go, go->ref_count, file, line, fce);
	return go;
}

#undef go_unref
void go_unref(struct go* go DEBUG_REF_O_ARGS)
{
	ASSERT_GO(go);
	ASSERT(go->ref_count > 0);
	go->ref_count--;
	DBG_O("  Decrease reference: go %s at %p, new ref_count %d, file %s, line %d, fce %s\n", go_lookup_name(go), go, go->ref_count, file, line, fce);

	if(go->ref_count == 0)
		go_finish(go);
}

#undef hanger_ref
struct hanger* hanger_ref(struct hanger *h DEBUG_REF_O_ARGS)
{
	if(!hanger_is_smart(h))
#ifndef DEBUG_REF_O
		go_ref(hanger_get_go(h));
#else
		go_ref(hanger_get_go(h), file, line, fce);
#endif

	else
	{
		struct hanger_smart *hs = hanger2smart(h);
		if(hs->ref_count == 0)
		{
			ASSERT(hanger_is_mc(h) && tlo_is_in_trans_stack(go2tlg(hs->go)->tlo));
		  	ASSERT_GO_SUBTYPE(hs->go, GROUP, TEMP);
			hanger_relink(hs, tlg2go(go2tlg(hs->go)->tlo->tlg));
		}
		hs->ref_count++;
		DBG_O("  Increase reference: hanger %p, new ref_count %d, file %s, line %d, fce %s\n", h, hs->ref_count, file, line, fce);
	}
	return h;
}

static void hanger_smart_finish(struct hanger_smart *hs)
{
	ASSERT(clist_empty(&hs->h.lanchor));
	clist_remove(&hs->nhanger);
	hanger_free(&hs->h);
}

#undef hanger_unref
void hanger_unref(struct hanger *h DEBUG_REF_O_ARGS)
{
	if(!hanger_is_smart(h))
	{
#ifndef DEBUG_REF_O
		go_unref(hanger_get_go(h));
#else
		go_unref(hanger_get_go(h), file, line, fce);
#endif
		return;
	}

	struct hanger_smart *hs = hanger2smart(h);
	ASSERT(hs->ref_count > 0);
	hs->ref_count--;
	DBG_O("  Decrease reference: hanger %p, new ref_count %d, file %s, line %d, fce %s\n", h, hs->ref_count, file, line, fce);
	if(hs->ref_count == 0)
	{
		if(trans_present && tlo_is_in_trans_stack(go2tlg(hs->go)->tlo))
		{
			ASSERT_GO_SUBTYPE(hs->go, GROUP, TLG);
			hanger_relink(hs, tlg2go(go2tlg(hs->go)->tlo->temp));
		}
		else
			hanger_smart_finish(hanger2smart(h));
	}
}

