#ifndef _KERNEL_OBJECT_H_
#define _KERNEL_OBJECT_H_

/** @file
 * Kernel - Object handling functions	
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */



/**********************************************
 *                                            *
 *   Verify macros and type-cast functions    *
 *                                            *
 **********************************************/


#define ASSERT_OBJ(_o) ASSERT((_o) != NULL && ((struct o*)(_o))->kind == T_OBJ && ((struct o*)(_o))->subtype == 0)
#define ASSERT_DOC(_o) ASSERT((_o) != NULL && ((struct o*)(_o))->kind == T_OBJ && ((struct o*)(_o))->type == OT_DOCUMENT)
#define ASSERT_TLO(_o) ASSERT((_o) != NULL && ((struct o*)(_o))->kind == T_OBJ && ((struct o*)(_o))->type == OT_TLO)

static inline struct obj_doc* obj2doc(struct obj* obj) { ASSERT_DOC(obj); return (struct obj_doc*)obj; }
static inline struct obj_tlo* obj2tlo(struct obj* obj) { ASSERT_TLO(obj); return (struct obj_tlo*)obj; }

static inline struct obj* o2obj(struct o* o) { ASSERT_OBJ(o); return (struct obj*)o; }
static inline struct o* obj2o(struct obj* o) { ASSERT_OBJ(o); return (struct o*)o; }

static inline struct obj* void2obj(void *o) { ASSERT_OBJ(o); return o; }
static inline struct obj_doc* void2doc(void *o) { ASSERT_DOC(o); return o; }
static inline struct obj_tlo* void2tlo(void *o) { ASSERT_TLO(o); return o; }

static inline struct obj* nson2obj(cnode *n) { return n ? void2obj(SKIP_BACK(struct obj, nson, n)) : NULL; }



/**********************************************
 *                                            *
 *    Movement functions and walk macros      *
 *                                            *
 **********************************************/


/*
 * Basic OBJ movement
 */
	
SCHEME static inline SO_obj obj_get_first_son(SO_obj obj) { return nson2obj(clist_head(&void2obj(obj)->lson)); }

SCHEME static inline SO_obj obj_get_next_son(SO_obj obj) { return nson2obj(clist_next(&void2obj(obj)->father->lson, &obj->nson)); }

SCHEME static inline SO_obj obj_get_last_son(SO_obj obj) { return nson2obj(clist_tail(&void2obj(obj)->lson)); }

SCHEME static inline SO_obj obj_get_prev_son(SO_obj obj) { return nson2obj(clist_prev(&void2obj(obj)->father->lson, &obj->nson)); }
	
#define WALK_OBJ_SON(_father, _walk) \
	for(cnode *_n = void2obj(_father)->lson.head.next; \
	_n != &void2obj(_father)->lson.head && (_walk = void2obj(SKIP_BACK(struct obj, nson, _n)), 1); \
	_n = _n->next)

/* Generic DFS WALK-macro
 * Calling commands _pre and _in and _post and condition _down is equivalent to following recursion.
 *
 *	void recursion(walk)
 *	{	
 *		if(!is_leaf(walk) && _down)
 *		{	
 *			_pre;
 *			WALK_GROUP_GO(walk, obj)
 *				recursion(obj);
 *			_post;
 *		}
 *		else 
 *			_in;
 *	}
 */		
#define WALK_OBJ_SMART(_root, _obj, _down, _pre, _in, _post) \
	do { \
		_obj = _root; \
		if(!obj_get_first_son(_obj)) do { _in; } while(0); else \
		for(;;) { \
			if(obj_get_first_son(_obj) && (_down)) \
			{	do { _pre; } while(0); _obj = obj_get_first_son(_obj); continue; } \
			do { _in; } while(0); \
NEXT:		if(obj_get_next_son(_obj)) \
			{ _obj = obj_get_next_son(_obj); continue; } \
			_obj = _obj->father; \
			do { _post; } while(0); \
			if(_obj == _root) break; \
			goto NEXT; \
		} \
	} while(0); 
	
/*
 * Type-specified OBJ movement
 */

#define WALK_UNIVERSE_DOC(_walk) \
	for(cnode *_n = obj_universum->lson.head.next; \
	_n != &obj_universum->lson.head && (_walk = obj2doc(SKIP_BACK(struct obj, nson, _n)), 1); \
	_n = _n->next)
	
#define WALK_DOC_TLO(_doc, _walk) \
	for(cnode *_n = void2doc(_doc)->obj.lson.head.next; \
	_n != &void2obj(_doc)->lson.head && (_walk = void2tlo(SKIP_BACK(struct obj, nson, _n)), 1); \
	_n = _n->next)

	
/*
 * Recursive OBJ movement
 */

SCHEME SO_obj obj_recursive_next(const SO_obj root, SO_obj walk);

#define WALK_OBJ_RECURSIVE(_root, _walk) \
	for(_walk = void2obj(_root); _walk; _walk = obj_recursive_next(_root, _walk))


/*
 * Miscellaneous OBJ movement
 */

SCHEME static inline SO_obj obj_get_nth_first_son(SO_obj obj, SUNS n) { return nson2obj(clist_head_nth(&void2obj(obj)->lson, n)); }

SCHEME static inline SO_obj obj_get_nth_last_son(SO_obj obj, SUNS n) { return nson2obj(clist_tail_nth(&void2obj(obj)->lson, n)); }

SCHEME static inline SINT obj_get_rank_son(SO_obj obj) { return obj ? clist_rank(&void2obj(obj->father)->lson, &void2obj(obj)->nson) : -1; }

SCHEME static inline SO_obj obj_get_root(SO_obj obj) { while(void2obj(obj)->father) obj = obj->father; return obj; }

SCHEME static inline SBOOL obj_is_linked(SO_obj obj) { return obj_get_root(obj) == obj_universum; }

SCHEME static inline SBOOL obj_is_ancestor(SO_obj anc, SO_obj des) { for(; des; des = des->father) if(void2obj(des) == void2obj(anc)) return 1; return 0; }


/*
 * Movement in selections
 */

SCHEME static inline SO_obj obj_get_first_selected(void)
	{ cnode *n = obj_lsel.head.next; return n != &obj_lsel.head ? SKIP_BACK(struct obj, nsel, n) : NULL; }

SCHEME static inline SO_obj obj_get_next_selected(SO_obj o)
	{ cnode *n = o->nsel.next; return n != &obj_lsel.head ? SKIP_BACK(struct obj, nsel, n) : NULL; }

#define WALK_OBJ_SELECTED(_walk) for(_walk = obj_get_first_selected(); _walk; _walk = obj_get_next_selected(_walk))


/*
 * Creating (named) object.
 */

SCHEME SO_obj_doc doc_new(void) SMETATRANS ;
SCHEME SO_obj_doc doc_new_named(SSTRINGC name) SMETATRANS ;

SCHEME SO_obj_tlo tlo_new(void) SMETATRANS ;
SCHEME SO_obj_tlo tlo_new_named(SSTRINGC name) SMETATRANS ;


/*
 * Link object.
 */

SCHEME SO_obj obj_link_after(SO_obj obj, SO_obj father, SO_obj after) SMETATRANS ;
SCHEME static inline SO_obj obj_link_head(SO_obj obj, SO_obj father) SMETATRANS
	{ return obj_link_after(obj, father, NULL); }
SCHEME static inline SO_obj obj_link_prev(SO_obj obj, SO_obj father, SO_obj prev) SMETATRANS
	{ return obj_link_after(obj, father, prev ? obj_get_prev_son(prev) : obj_get_last_son(father)); }
SCHEME static inline SO_obj obj_link_tail(SO_obj obj, SO_obj father) SMETATRANS
	{ return obj_link_after(obj, father, obj_get_last_son(father)); }
SCHEME SO_obj obj_unlink(SO_obj obj) SMETATRANS ;
SCHEME SVOID obj_unlink_selected(void) SMETATRANS ;
SCHEME SVOID obj_unlink_all_son(SO_obj obj) SMETATRANS ;
SCHEME SO_obj obj_relink_after(SO_obj obj, SO_obj father, SO_obj after) SMETATRANS ;



/*
 * Create named object and link it.
 */

SCHEME static inline SO_obj_doc doc_new_named_link(SSTRINGC name) SMETATRANS
	{ return obj2doc(obj_link_tail(&doc_new_named(name)->obj, obj_universum)); }

SCHEME static inline SO_obj_tlo tlo_new_named_link(SO_obj_doc doc, SSTRINGC name) SMETATRANS
	{ return obj2tlo(obj_link_tail(&tlo_new_named(name)->obj, &doc->obj)); }


/*
 * Selecting functions
 */

SCHEME SVOID obj_select(SO_obj obj) SMETATRANS ;
SCHEME SVOID obj_unselect(SO_obj obj) SMETATRANS ;
SCHEME SVOID obj_unselect_all(void) SMETATRANS ;

SCHEME static inline SUNS obj_is_selected(SO_obj obj)
	{ return (void2obj(obj)->flags & OF_SELECTED); }



/*************************
 *                       *
 *     Save and load     *
 *                       *
 *************************/


SCHEME SVOID doc_save(SO_obj_doc doc, SSTRINGC filename);
SCHEME SO_obj_doc doc_load(SSTRINGC filename);

SCHEME SVOID settings_save(SSTRINGC filename) SMETATRANS ;
SCHEME SVOID settings_load(SSTRINGC filename) SMETATRANS ;

SCHEME static inline SUNS tlo_is_saved(SO_obj_tlo tlo) { return void2tlo(tlo)->undo_saved == tlo->undo_present; }
SCHEME SVOID tlo_set_saved(SO_obj_tlo tlo);

SCHEME static inline SUNS doc_is_saved(SO_obj_doc doc) { return void2doc(doc)->saved; }
SCHEME SVOID doc_set_saved(SO_obj_doc doc);

SCHEME SUNS doc_is_all_saved(SO_obj_doc doc);
SCHEME SVOID doc_set_all_saved(SO_obj_doc doc);


	
/**********************************************
 *                                            *
 *        OBJ hook handling functions         *
 *                                            *
 **********************************************/


struct obj_hook* obj_hook_add(struct obj_hook *hook);
struct obj_hook* obj_hook_remove(struct obj_hook *hook);
struct obj_hook* obj_hook_remove_by_data(void *data);

static inline struct obj_hook* obj_hook_fill(
		struct obj_hook *h,
		void* data,
		void (*o_link) (struct o *obj, struct obj_hook* h),
		void (*o_unlink) (struct o *obj, struct o *father, struct o *prev, struct obj_hook* h),
		void (*obj_selected) (struct o *obj, struct obj_hook* h),
		void (*obj_unselected) (struct o *obj, struct obj_hook* h),
		void (*obj_prop_create) (struct o *obj, struct prop* p, struct obj_hook* h),
		void (*obj_prop_change) (struct o *obj, struct obj_hook* h),
		void (*obj_prop_delete) (struct o *obj, struct prop* p, struct obj_hook* h))
	{ h->data = data; h->obj_link = o_link; h->obj_unlink = o_unlink; h->obj_selected = obj_selected; h->obj_unselected = obj_unselected; h->obj_prop_create = obj_prop_create; h->obj_prop_change = obj_prop_change; h->obj_prop_delete = obj_prop_delete; return h; }



/*****************************
 *                           *
 *    Dumping functions      *
 *                           *
 *****************************/


SCHEME SVOID obj_dump_father(SO_obj obj);
SCHEME SVOID obj_dump_son(SO_obj obj);
SCHEME SVOID obj_dump_subtree(SO_obj root);
SCHEME SVOID obj_dump_flag(SO_obj root);
SCHEME SVOID obj_dump_hook(SO_obj obj);
SCHEME SVOID obj_dump_selected(void);
SCHEME SVOID obj_dump_all(SO_obj obj);


// miscelaneous

int tlo_get_exact_bbox_after_trans (struct obj_tlo *tlo, const struct geom_transform *t, struct geom_rectangle *bbox);

#endif
