#ifndef _KERNEL_PROPERTY_H_
#define _KERNEL_PROPERTY_H_

/** @file
 *
 * Kernel - Property and unit operation.
 * Functions for working with color and name of objects.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */


/*
 * Create a new property.
 */

void prop_create(struct o *o, string key, uns type, uns subtype, uns unit, prop_value value);


/*
 * Get value from given property.
 * Do not read value from structure prop. Use this function instead.
 */

prop_value prop_get_value(struct o *o, struct prop *p);


/*
 * Properties searching.
 * Each property is uniquely determined by its name
 */

struct prop* prop_find(struct o *o, string key);
static inline prop_value prop_lookup(struct o *o, string key, uns type, prop_value err)
	{ struct prop *p = prop_find(o, key); return (p && p->type == type) ? prop_get_value(o, p) : err; }
static inline prop_value prop_named_lookup(struct o *o, const char* key, uns type, prop_value err)
	{ string s = string_find(key); return s ? prop_lookup(o, s, type, err) : err; }

/* Shortcuts of function prop_lookup. */
static inline uns prop_lookup_uns(struct o *o, string key, uns err)
	{ return prop_lookup(o, key, PT_UNS, (prop_value){.u=err}).u; }
static inline real prop_lookup_real(struct o *o, string key, real err)
   { return prop_lookup(o, key, PT_REAL, (prop_value){.r=err}).r; }
static inline const char* prop_lookup_text(struct o *o, string key)
	{ string ret = prop_lookup(o, key, PT_STRING, (prop_value){.s=NULL}).s; return ret ? ret->text : NULL; }
real prop_lookup_real_unit(struct o *o, string key, real err);
static inline real prop_named_lookup_real_unit(struct o *o, const char* key, real err)
	{ string s = string_find(key); return s ? prop_lookup_real_unit(o, s, err) : err; }


/*
 * Change value of a property which allready exist.
 */

void prop_set(struct o *o, string key, prop_value value);

/* Shortcuts of function prop_set. */
static inline void prop_set_uns(struct o *o, string key, uns value)
	{ prop_set(o, key, (prop_value){.u=value}); }
static inline void prop_set_real(struct o *o, string key, real value)
	{ prop_set(o, key, (prop_value){.r=value}); }
static inline void prop_set_text(struct o *o, string key, const char* value)
	{ prop_set(o, key, (prop_value){.s=string_lookup(value)}); }
void prop_set_real_unit(struct o *o, string key, real value);
void prop_named_set_real_unit(struct o *o, const char* key, real value);


/*
 * Properties deletion
 * Property cannot be virtual.
 */

void prop_delete(struct o *o, struct prop* prop);
void prop_find_delete(struct o *o, string key);


/*
 * Movement PROP function
 */

static inline struct prop* prop_get_first(struct o *o)
	{ struct snode* n = o->prop.head.next; return n ? SKIP_BACK(struct prop, n, n) : NULL; }
static inline struct prop* prop_get_next(struct prop* prop)
	{ struct snode* n = prop->n.next; return n ? SKIP_BACK(struct prop, n, n) : NULL; }
#define WALK_PROP(_o, _walk) for(_walk = prop_get_first(_o); _walk; _walk = prop_get_next(_walk))
static inline uns prop_get_count(struct o *o)
	{ uns n = 0; struct prop* p; WALK_PROP(o, p) n++; return n; }


/*
 * Property unit functions
 */

uns unit_find(string name);
static inline uns unit_named_find(const char *name)
	{ string s = string_find(name); return s ? unit_find(s) : PROP_UNIT_MAX; }

uns prop_get_unit(struct o *o, struct prop* prop);
static inline real prop_get_unit_multiplier(struct o *o, struct prop *prop)
{ uns u = prop_get_unit(o, prop); return u == PROP_UNIT_MAX ? 1 : unit_array[u].multiplier; }

void prop_set_unit(struct o *o, struct prop* prop, uns unit);

static inline uns prop_get_quantity(struct prop *p)
	{ return prop_subtype2quantity[p->type][p->subtype]; }

/* Function for working with array of units. */
uns unit_set(string name, real mult, uns quantity);
void unit_default_set(uns quantity, uns index);
void unit_delete(uns index);
void unit_change_name(uns index, string name);


/*
 * Property name functions
 */

prop_value o_get_prop_name(struct o *o);
static inline string o_get_name(struct o *o)
	{ return o_get_prop_name(o).s; }
static inline const char* o_lookup_name(struct o *o)
	{ return o_get_prop_name(o).s->text; }
static inline string obj_get_name(struct obj *obj)
	{ return o_get_prop_name((struct o*)obj).s; }
static inline const char* obj_lookup_name(struct obj *obj)
	{ return o_get_prop_name((struct o*)obj).s->text; }
static inline string go_get_name(struct go *go)
	{ return o_get_prop_name((struct o*)go).s; }
static inline const char* go_lookup_name(struct go *go)
	{ return o_get_prop_name((struct o*)go).s->text; }

SCHEME static inline SVOID o_set_name(SO_o o, SSTRINGC str) SOTRANS(0)
	{ prop_set(o, string_name, (prop_value){ .s = string_lookup(str) }); }

SCHEME SO_o o_lookup(SSTRINGS name);


/*
 * Functions for working with color
 */

SCHEME static inline SUNS rgba_uns2uns(SUNS r, SUNS g, SUNS b, SUNS a)
	{ return (r << 24) | (g << 16) | (b << 8) | a; }
static inline void uns2rgba_uns(uns u, uns *r, uns *g, uns *b, uns *a)
	{ *r = (u>>24)&255; *g = (u>>16)&255; *b = (u>>8)&255; *a = u&255; }

static inline uns rgba_real2uns(real r, real g, real b, real a)
	{ ASSERT(r>=0 && r<=1 && g>=0 && g<=1 && b>=0 && b<=1 && a>=0 && a<=1); return (((uns)(r*255)) << 24) | (((uns)(g*255)) << 16) | (((uns)(b*255)) << 8) | ((uns)(a*255)); }
static inline void uns2rgba_real(uns u, real *r, real *g, real *b, real *a)
	{ *r = ((real)((u>>24)&255))/255; *g = ((real)((u>>16)&255))/255; *b = ((real)((u>>8)&255))/255; *a = ((real)(u&255))/255; }

SCHEME static inline SUNS uns2r_uns(SUNS u)
	{ return (u>>24)&255; }
SCHEME static inline SUNS uns2g_uns(SUNS u)
	{ return (u>>16)&255; }
SCHEME static inline SUNS uns2b_uns(SUNS u)
	{ return (u>>8)&255; }
SCHEME static inline SUNS uns2a_uns(SUNS u)
	{ return u&255; }

SCHEME static inline SUNS rgba_uns_set_r_uns(SUNS c, SUNS r)
	{ return (c & 0x00FFFFFF) | (r << 24); }
SCHEME static inline SUNS rgba_uns_set_g_uns(SUNS c, SUNS g)
	{ return (c & 0xFF00FFFF) | (g << 16); }
SCHEME static inline SUNS rgba_uns_set_b_uns(SUNS c, SUNS b)
	{ return (c & 0xFFFF00FF) | (b <<  8); }
SCHEME static inline SUNS rgba_uns_set_a_uns(SUNS c, SUNS a)
	{ return (c & 0xFFFFFF00) | a; }

/*
 * Stroke pattern properties
 */
struct stroke_pattern
{
	int n_items;
	char * string;
	real length[];
};

extern struct stroke_pattern * stroke_patterns[PSS_MAX];

/*
 * Functions for compare property value
 */

uns prop_value_compare(uns prop_type, prop_value v1, prop_value v2);


/*
 * Property unit hooks
 */

struct unit_hook* unit_hook_add(struct unit_hook *hook);
struct unit_hook* unit_hook_remove(struct unit_hook *hook);
struct unit_hook* unit_hook_remove_by_data(void *data);


/*
 * Dump property.
 */

char* prop_dump_value(char buf[PROP_DUMP_SIZE], struct prop* prop, prop_value value);
char* prop_dump(char buf[PROP_DUMP_SIZE], struct o *o, struct prop* prop);

#endif
