/*
 * Kernel - Settings
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "kernel/internal.h"
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <pwd.h>
#include "lib/cache.h"


#define SETTING(_var, _next, _fce) \
static struct prop_virtual settings_##_var; \
static prop_value settings_##_var##_get(struct o* o) \
	{ ASSERT(o2obj(o) == obj_universum); return (prop_value) { .u = _var }; } \
static void settings_##_var##_set(struct o* o, prop_value value) \
{ \
	ASSERT(o2obj(o) == obj_universum); \
	trans_new_undo_prop_virtual(o, &settings_##_var, (prop_value){ .u = _var }); \
	_var = value.u; \
	_fce(); \
	hook_call_prop_change(o); \
} \
static void settings_##_var##_set_low(struct o* o, prop_value value) \
{ \
	ASSERT(o2obj(o) == obj_universum); \
	_var = value.u; \
	_fce(); \
	hook_call_prop_change(o); \
} \
static uns settings_##_var##_get_unit(struct o* o UNUSED) { BUG(); } \
static void settings_##_var##_set_unit(struct o* o UNUSED, uns unit UNUSED) { BUG(); } \
static struct prop_virtual settings_##_var = \
	{ { { &_next.p.n }, NULL, {0}, PT_UNS, PTU_UNSPECIFIED, PTF_VIRTUAL | PTF_SAVE, 0 }, \
	UINT_MAX, settings_##_var##_get, settings_##_var##_set, settings_##_var##_set_low, settings_##_var##_get_unit, settings_##_var##_set_unit };
	


static uns max_history_volume, max_history_count,
	max_cache_memory, max_cache_count, min_cache_count;

static inline void history_limits_change(void)
{
	struct obj *obj;
	trans_undo_limit_control(tlo_universum);	
	WALK_OBJ_RECURSIVE(obj_universum, obj)
		if(obj->type == OT_TLO)
			trans_undo_limit_control(obj2tlo(obj));
}

static inline void cache_limits_change(void)
{
	cache_set_limits(&global_cache, max_cache_memory, min_cache_count, max_cache_count);
}

SETTING(max_history_volume, o_prop_name, history_limits_change)
SETTING(max_history_count, settings_max_history_volume, history_limits_change)
SETTING(max_cache_memory, settings_max_history_count, cache_limits_change)
SETTING(max_cache_count, settings_max_cache_memory, if(min_cache_count > max_cache_count) prop_set_uns(o, string_min_cache_count, max_cache_count); cache_limits_change)
SETTING(min_cache_count, settings_max_cache_count, if(min_cache_count > max_cache_count) prop_set_uns(o, string_max_cache_count, min_cache_count); cache_limits_change)

struct prop_virtual *tlo_universum_prop_virtual = &settings_min_cache_count;



static void load_settings(void)
{
	struct stat st;
	struct passwd * pasw = getpwuid( getuid() );
	if (!pasw)
	{ err("Cannot getpwuid:%s\n", strerror(errno)); return; }

	char filename[strlen(pasw->pw_dir) + strlen(VRR_DIR GLOBAL_SETTINGS_FILE) + 1];
	strcpy( filename, pasw->pw_dir );
        strcat( filename, VRR_DIR GLOBAL_SETTINGS_FILE );

	if (stat( filename, &st ))
	{
		if ( errno != ENOENT ) err("Cannot stat file %s: %s\n", filename, strerror(errno));
		return;
	}

	char buf[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( tlo_universum, "Load settings", buf )
		settings_load(filename);
	TRANS_FAILED
		err("Cannot read settings file %s: %s\n", filename, buf);
	TRANS_END
}

static void save_settings(void)
{
	struct stat st;
	struct passwd * pasw = getpwuid( getuid() );
	if (!pasw)
	{ err("Cannot getpwuid:%s\n", strerror(errno)); return; }

	char filename[strlen(pasw->pw_dir) + strlen(VRR_DIR GLOBAL_SETTINGS_FILE) + 1];
	strcpy( filename, pasw->pw_dir );
	strcat( filename, VRR_DIR );

	int res = stat(filename, &st);
	if (res && errno == ENOENT)
	{
		if (mkdir(filename, S_IRWXU))
		{ err("Cannot create dir %s: %s\n", filename, strerror(errno)); return; }
	} else {
		if (res == 0 && !S_ISDIR(st.st_mode))
		{ err("%s is not a directory\n", filename); return; }
	}
	strcat( filename, GLOBAL_SETTINGS_FILE );

	char buf[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( tlo_universum, "Save settings", buf )
		settings_save( filename );
	TRANS_FAILED
		err("Cannot write settings file %s: %s\n", filename, buf);
	TRANS_END
}

void settings_init(void)
{
	settings_max_history_volume.p.key = string_max_history_volume;
	settings_max_history_count.p.key = string_max_history_count;
	settings_max_cache_memory.p.key = string_max_cache_memory;
	settings_max_cache_count.p.key = string_max_cache_count;
	settings_min_cache_count.p.key = string_min_cache_count;

	max_history_volume = MAX_UNDO_HISTORY_VOLUME_KB;
	max_history_count = MAX_UNDO_HISTORY_COUNT;
	max_cache_memory = MAX_CACHE_MEMORY_LIMIT_KB;
	max_cache_count = MAX_CACHE_COUNT;
	min_cache_count = MIN_CACHE_COUNT;
	
	load_settings();
}

void settings_finish(void)
{
	save_settings();
}
