#ifndef _KERNEL_TYPES_H_
#define _KERNEL_TYPES_H_

/** @file
 *
 *  Kernel's structures and declaration of global variables.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */


/**********************************
 *                                *
 *            UNDO                *
 *                                *
 **********************************/


struct undo_gui
{
	cnode ng;			// Node for obj_undo
	string name_operation;
	clist lk;			// List of kernel operations in this gui command.
};



/*************************************
 *                                   *
 *      Non-graphics objects         *
 *                                   *
 *************************************/


/* First we define the greatgrandfather of all name-space objects */
#define STRUCT_O_MAGIC 		u8 kind; u8 type; u8 subtype; u8 _dummy; uns ref_count; struct slist prop; uns flags
struct o
{
	STRUCT_O_MAGIC;
};

/* Abstract non-graphical object */
struct obj
{
	STRUCT_O_MAGIC;
	struct obj* father;		// predecessor in name-space tree
	clist lson;
	cnode nson;			// brothers in NS tree are linked via this node
	cnode nsel;			// selected objects are linked via this node
};

/* OBJ TLO, top-level object, descendant of OBJ */
struct obj_tlo
{
	struct obj obj;
	
	struct undo_gui undo_start;				// undo history
	struct undo_gui* undo_present;		// pointer to current gui command in undo history
	struct undo_gui* undo_saved;
	uns kernel_count, gui_count; // total count of items in undo

	slist trans_hooks, go_hooks; // lists of hooks
	
	struct go_tlg *tlg; 		// top-level group
	struct go_tlg *temp;
	
	struct flist tsort; // tsort list
	
	struct geom_rtree rtree;	// rtree main data
	uns expand_ref_count;
		
	struct fmalloc_head  // Small allocators for
		*fprop, // property (struct prop)
		*fname, // property name (struct prop_o)
		*fuk, // item of undo kernel
		*frtree; // rtree node (struct rtree_bbox_bucket)
};

/* OBJ DOCUMENT, descendant of OBJ */
struct obj_doc
{
	struct obj obj;
	struct undo_kernel *saved_start, *saved_end;
	int saved;
};



/****************************************
 *                                      *
 *     Graphics object and groups       *
 *                                      *
 ****************************************/


/* Abstract graphical object */
struct go
{
	STRUCT_O_MAGIC;
#ifdef STUPID_GROUP
	cnode nmember;
	struct go_group *owner;
#else
#endif
	cnode nselected;		// Node for obj_tlo.lselected
	snode nfifo;			// Internal: for topological sort
	uintptr_t temp_data;  // Internal: data for topological sort and clipboard
};

struct go_group
{
	struct go go;
	struct geom_path geom;

#ifdef STUPID_GROUP
	clist lmember;
#else
#endif

	uns fill_color, stroke_color, stroke_cap, stroke_join, stroke_style;
	real stroke_width;
	u8 unit_stroke_width;

	clist lselected;
	struct geom_rectangle bbox_all, bbox_sel;
};

struct go_tlg
{
	struct go_group group;
	struct obj_tlo *tlo;
	clist mc;
};

struct go_path
{
	struct go_group group;
};

/**************************************
 *                                    *
 *   Objects dependence realization   *
 *                                    *
 **************************************/


typedef union quantity
{
	struct geom_point p;
	real r;
} quantity;

struct hanger
{
	struct hanger_info *info;
	clist lanchor;			// The list of anchors which are hanging on this hanger.
};

struct hanger_smart
{
	struct hanger h;
	struct go *go;			// The hanger's owners.
	cnode nhanger;
	uns ref_count;
	quantity q;
};

struct anchor
{
	struct anchor_info *info;
	cnode nanchor;			// Node for hanger.lanchor.
	struct hanger *h;		// This anchor is hanging on the hanger.
};




/****************************
 *                          *
 *    Property and unit     *
 *                          *
 ****************************/


typedef union prop_value
{
	string s;
	uns u;
	void* v;
	real r;
	struct geom_transform* t;
} prop_value;

struct prop
{
	struct snode n;
	string key;
	prop_value value;
	u8 type, subtype, flags, unit;
};

struct prop_unit_struct
{
	string name;
	real multiplier;
	uns quantity;
};



/******************************
 *                            *
 *           HOOKS            *
 *                            *
 ******************************/

/*
 * Structuries for visualization's hook
 */

struct changed_data_generic
{
	uns kind;
};

struct changed_data
{
	uns kind;
	const struct geom_rectangle *old_bbox;
	const struct geom_rectangle *new_bbox;
};

struct altered_data
{
	uns kind;
	const struct geom_rectangle *where;
};

struct transformed_data
{
	uns kind;
	const struct geom_rectangle *old_bbox;
	const struct geom_rectangle *new_bbox;
	const struct geom_transform2 *transformation;
};


struct obj_hook
{
	snode n;
	void* data;
	void (*obj_link) (struct o *o, struct obj_hook* h);		// A obj has been linked.
	void (*obj_unlink) (struct o *o, struct o *father, struct o *prev, struct obj_hook* h);	// The obj is going to be unlinked.
	void (*obj_selected) (struct o *o, struct obj_hook* h);	// The obj has been selected.
	void (*obj_unselected) (struct o *o, struct obj_hook* h);	// The obj has been unselected.
	void (*obj_prop_create) (struct o *o, struct prop* p, struct obj_hook* h);	// The property of obj was created.
	void (*obj_prop_change) (struct o *o, struct obj_hook* h);	// The property of obj was changed.
	void (*obj_prop_delete) (struct o *o, struct prop* p, struct obj_hook* h);	// The property of obj will be deleted.
};

struct go_hook
{
	snode n;
	void* data;
	void (*go_link) (struct o *o, struct go_hook* h);		// The go has been linked.
	void (*go_unlink) (struct o* o, struct o *father, struct o *prev, struct go_hook* h);		// The go is going to be unlinked.
	void (*go_relink) (struct o *o, struct go_group *old_group, struct go *old_prev, struct go_hook* h);		// The go has been relinked.
	void (*go_selected) (struct o *o, struct go_hook* h);		// The go has been selected.
	void (*go_unselected) (struct o *o, struct go_hook* h);	// The go has been unselected.
	void (*go_prop_create) (struct o *o, struct prop* p, struct go_hook* h);	// The property of go was created.
	void (*go_prop_change) (struct o *o, struct go_hook* h);	// The property of go was changed.
	void (*go_prop_delete) (struct o *o, struct prop* p, struct go_hook* h);	// The property of go will be deleted.
	void (*go_changed) (struct go *go, const struct changed_data_generic *d, struct go_hook* h);	// Visualization's hook: The go has been changed.
};

struct trans_hook
{
	snode n;
	void* data;
	void (*undo_new) (struct obj_tlo *tlo, struct undo_gui* ug, struct trans_hook* h);
	void (*undo_delete) (struct obj_tlo *tlo, struct undo_gui* ug, struct undo_gui *prev_item, struct trans_hook* h);
	void (*undo_update) (struct obj_tlo *tlo, struct undo_gui* ug, struct trans_hook* h);
	void (*undo) (struct obj_tlo* tlo, struct trans_hook* h);	// One operation undo.
	void (*redo) (struct obj_tlo *tlo, struct trans_hook* h);	// One operation redo.
	void (*saved) (struct obj * obj, struct trans_hook* h);		// The obj was set as saved
};

struct unit_hook
{
	snode n;
	void* data;
	void (*add) (uns slot, struct unit_hook *h);		// New unit was added into slot.
	void (*deleted) (uns slot, struct unit_hook *h);			// Unit was deleted.
	void (*changed) (uns slot, struct unit_hook *h);		// Value or quantity of unit was changed.
	void (*def) (uns quantity, struct unit_hook *h);		// Default unit of quantity was changed.
};



/******************************
 *                            *
 *   Description subsystem    *
 *                            *
 ******************************/


struct obj_info
{
	string name;
};

struct go_info
{
	string name;
	uns h_count; // Number of hangers in fixed-size array.
	int h_offset; // Offset of the first hanger and go. Zero iff go hasn't stupid hanger.
	int hs_offset; // Offset of clist (which contains smart hangers) and go. Zero iff go hasn't smart hanger.
	struct hanger_info *h_info; // Information of the first hanger.
	int h_curve; // Index of hanger reprezenting whole curve. -1 if go is not parametrizable.
	int h_start; // Index of hanger reprezenting start point of curve. -1 if go hasn't start.
	int h_end; // Index of hanger reprezenting end point of curve. -1 if go hasn't end.
	int h_center; // Index of hanger reprezenting center of curve. -1 if go hasn't center.
	uns a_count; // Number of anchors in fixed-size array.
	int a_offset; // Offset of first anchor and go.
	struct anchor_info *a_info; // Information of the first anchor.
	int a_start; // Index of anchor reprezenting start point of curve. -1 if go hasn't start.
	int a_end; // Index of anchor reprezenting end point of curve. -1 if go hasn't end.
	struct prop_virtual *prop_virtual_list;
};

struct hanger_info
{
	string name;
	uns quantity;
	int go_offset; // Offset of hanger and go. Zero iff hanger is smart.
	struct go_info *go_info;
	int q_offset; // Offset of hanger and value of quantity. Zero iff following function must be used instead.
	quantity (*get_quantity) (struct hanger *h); // Also zero iff type of hanger is QT_CURVE.
	uns linear;
};

struct anchor_info
{
	string name;
	uns quantity;
	int offset; // Offset of anchor and go.
	struct go_info *go_info;
	uns linear;
};



/*
 * Miscellaneous
 */

struct rtree_bbox
{
	struct geom_rtree_obj rtree;
	struct go *go;
};


/***************************************
 *                                     *
 *   Declaration of global variables   *
 *                                     *
 ***************************************/


extern struct obj *obj_universum; 		// Tree of all objects linked to the universum.
extern struct obj_tlo *tlo_universum; 		// Tree of all objects linked to the universum.
extern struct obj_tlo *tlo_clipboard;

extern struct obj *obj_zombie; 			// Forest of unused objects.
extern struct obj_tlo *tlo_zombie;
extern struct obj *obj_temp;			// Forest of temp (working) objects.

extern clist obj_lsel;				// List of selected objects.
extern uns obj_sel_count[OT_MAX];		// Counts of selected objects per each object type.

extern struct obj_info obj_info[];
extern struct go_info *go_info[], go_group_info[];

extern uns vrr_with_gui;

extern struct trans *trans_present;
extern struct obj_tlo *trans_tlo;

extern struct prop_unit_struct unit_array[];
extern uns prop_subtype2quantity[][PROP_SUBTYPE_MAX];
extern uns unit_default[];


#endif
