/*
 *	Sherlock Library -- Circular Linked Lists
 *
 *	(c) 2003 Martin Mares <mj@ucw.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU Lesser General Public License.
 */

#ifndef _SHERLOCK_CLISTS_H
#define _SHERLOCK_CLISTS_H

typedef struct cnode {
  struct cnode *next, *prev;
} cnode;

typedef struct clist {
  struct cnode head;
} clist;

static inline void *clist_head(clist *l)
{
  return (l->head.next != &l->head) ? l->head.next : NULL;
}

static inline void *clist_tail(clist *l)
{
  return (l->head.prev != &l->head) ? l->head.prev : NULL;
}

static inline void *clist_next(clist *l, cnode *n)
{
  return (n->next != &l->head) ? (void *) n->next : NULL;
}

static inline void *clist_prev(clist *l, cnode *n)
{
  return (n->prev != &l->head) ? (void *) n->prev : NULL;
}

static inline void *clist_head_nth(clist *l, uns i)
{
  cnode *n;
  for(n = l->head.next; i && n != &l->head; i--)
	  n = n->next;
  return n == &l->head ? NULL : n;
}

static inline void *clist_tail_nth(clist *l, uns i)
{
  cnode *n;
  for(n = l->head.prev; i && n != &l->head; i--)
    n = n->prev;
  return n == &l->head ? NULL : n;
}

static inline int clist_rank(clist *l, cnode *n)
{
  int i = -1;
  for(; n != &l->head; n = n->prev)
    i++;
  return i;
}

static inline int clist_empty(clist *l)
{
  return (l->head.next == &l->head);
}

#define CLIST_WALK(n,list) for(n=(void*)(list).head.next; (cnode*)(n) != &(list).head; n=(void*)((cnode*)(n))->next)
#define CLIST_WALK_DELSAFE(n,list,tmp) for(n=(void*)(list).head.next; tmp=(void*)((cnode*)(n))->next, (cnode*)(n) != &(list).head; n=(void*)tmp)

static inline uns clist_is_node_member(clist *l, cnode *n)
{
  cnode *cn;
  CLIST_WALK(cn, *l)
    if(n == cn)
      return 1;
  return 0;
}
	
static inline void clist_insert_after(cnode *what, cnode *after)
{
  cnode *before = after->next;
  what->next = before;
  what->prev = after;
  before->prev = what;
  after->next = what;
}

static inline void clist_insert_before(cnode *what, cnode *before)
{
  cnode *after = before->prev;
  what->next = before;
  what->prev = after;
  before->prev = what;
  after->next = what;
}

static inline void clist_add_tail(clist *l, cnode *n)
{
  clist_insert_before(n, &l->head);
}

static inline void clist_add_head(clist *l, cnode *n)
{
  clist_insert_after(n, &l->head);
}

static inline void clist_remove(cnode *n)
{
  cnode *before = n->prev;
  cnode *after = n->next;
  before->next = after;
  after->prev = before;
#ifdef DEBUG
  n->next = n->prev = NULL;
#endif
}

static inline void clist_init(clist *l)
{
  cnode *head = &l->head;
  head->next = head->prev = head;
}

#endif
