
/*
 * VRRLIB - Growing arrays
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

/*
 * You should define the following macros:
 *
 *   GARR_PREFIX(x)
 *   GARR_TYPE
 *
 * GARR_PREFIX defaults to garr_
 * GARR_TYPE defaults to byte
 *
 * Optional:
 *   GARR_ZERO			zero all allocated memory, this is default
 *   				at anonymous garr function, and disabled at named functions
 *   GARR_MUL_FACTOR xxx	array grow factor
 *
 *
 * Example:
 * -------
 *
 *   #define GARR_PREFIX(x) num_##x
 *   #define GARR_TYPE unsigned int
 *   #include "lib/garr.h"
 *
 *   garr array;	// garr is typedef'ed struct garr *
 *
 * This will create functions:
 *   num_init(struct garr *g, uns init_size)	// init must be called before first use of g
 *   num_new(uns init_size)			// allocate new garr & init
 *   num_cleanup(struct garr *g)		// should be called after finished using
 *   num_destroy(struct garr *g)		// cleanup & deallocate
 *   num_get(struct garr *g, uns i)		// access element of index i
 *
 *   array = num_new(1024);
 *   x = *num_get(array, 2000);
 *   *num_get(array, 3000) = y;
 *   num_destroy(array);
 *   ...
 *
 * Anonymous functions look like:
 *   garr_init(struct garr *g, uns init_size, uns elem_size);
 *   garr_new(uns init_size, uns elem_size);
 *   garr_cleanup(struct garr *g);
 *   garr_destroy(struct garr *g);
 *   garr_get(struct garr *g, uns i);
 */


#ifndef _LIB_GARR_H_
#define _LIB_GARR_H_

#include <strings.h>

typedef struct garr {
	uns size;
	uns elem_size;
	uns off;
	byte *data;
} *garr;
#endif

#ifndef GARR_MUL_FACTOR
#define GARR_MUL_FACTOR 2
#endif


#ifndef GARR_PREFIX
#ifndef GARR_DEFAULT
#define GARR_DEFAULT
static inline struct garr *garr_init(struct garr *g, uns init_size, uns elem_size)
{
	g->size = init_size;
	g->elem_size = elem_size;
	g->off = 0;
	g->data = xmalloc_zero(g->elem_size * g->size);
	return g;
}

static inline struct garr *garr_new(uns init_size, uns elem_size)
{
	struct garr *g = xmalloc(sizeof(struct garr));
	garr_init(g, init_size, elem_size);
	return g;
}

static inline void garr_cleanup(struct garr *g)
{
	xfree(g->data);
}

static inline void garr_destroy(struct garr *g)
{
	garr_cleanup(g);
	xfree(g);
}

static inline byte *garr_get(struct garr *g, uns i)
{
	i += g->off;
	if (unlikely(g->size <= i)) {
		uns old_size = g->size;
		while (g->size <= i)
			g->size *= GARR_MUL_FACTOR;
		g->data = xrealloc(g->data, g->size * g->elem_size);
		bzero(g->data + old_size*g->elem_size, g->elem_size*(g->size - old_size));
	}
	return g->data + i*g->elem_size;
}

static inline struct garr *garr_seek(struct garr *g, int disp)
{
	g->off += disp;
	return g;
}

static inline struct garr *garr_seekset(struct garr *g, uns off)
{
	g->off = off;
	return g;
}
#endif

#else

#ifndef GARR_TYPE
#define GARR_TYPE byte
#endif

static inline struct garr * GARR_PREFIX(init) (struct garr *g, uns init_size)
{
	g->size = init_size;
	g->elem_size = sizeof(GARR_TYPE);
	g->off = 0;
#ifdef GARR_ZERO
	g->data = xmalloc_zero(g->size * sizeof(GARR_TYPE));
#else
	g->data = xmalloc(g->size * sizeof(GARR_TYPE));
#endif
	return g;
}

static inline struct garr * GARR_PREFIX(new) (uns init_size)
{
	struct garr *g = xmalloc(sizeof(struct garr));
	GARR_PREFIX(init)(g, init_size);
	return g;
}

static inline void GARR_PREFIX(cleanup) (struct garr *g)
{
	xfree(g->data);
}

static inline void GARR_PREFIX(destroy) (struct garr *g)
{
	GARR_PREFIX(cleanup)(g);
	xfree(g);
}

static inline GARR_TYPE * GARR_PREFIX(get) (struct garr *g, uns i)
{
	i += g->off;
	if (unlikely(g->size <= i)) {
#ifdef GARR_ZERO
		uns old_size = g->size;
#endif
		while (g->size <= i)
			g->size *= GARR_MUL_FACTOR;
		g->data = xrealloc(g->data, g->size * g->elem_size);
#ifdef GARR_ZERO
		bzero(g->data + old_size*g->elem_size, g->elem_size*(g->size - old_size));
#endif
	}
	return (GARR_TYPE *) (g->data + i*g->elem_size);
}

static inline struct garr * GARR_PREFIX(seek) (struct garr *g, int off)
{
	g->off += off;
	return g;
}

static inline struct garr * GARR_PREFIX(seekset) (struct garr *g, uns off)
{
	g->off = off;
	return g;
}
#endif

#undef GARR_PREFIX
#undef GARR_TYPE
#undef GARR_MUL_FACTOR
#undef GARR_ZERO
