#undef LOCAL_DEBUG

#include "lib/lib.h"
#include "kernel/kernel.h"
#include "gui/main.h"
#include "gui/cmdmgr.h"
#include "gui/properties.h"
// this is needed only because of the VRR install icon path
#include "build/path.h"

// the ID obtained from the Command Structure
int my_menu_id = -1;
// a flag if an instance of this plugin was already stopped
int stopped;

// the mandatory plugin functions
u32 plugin_start(void);
void plugin_stop(void);

// two menu commands
void twist_beziers(void * subj);
void hello_world(void * subj);

// the kernel plugin function prototypes
void star(union plugin_arg *);
void ribbon(union plugin_arg *);

// the GO Factory states user for creation of a star
static struct of_state ofs_star_4, ofs_star_3, ofs_star_2, ofs_star_1;
static struct of_state ofs_ribbon_6, ofs_ribbon_5, ofs_ribbon_4, ofs_ribbon_3, ofs_ribbon_2, ofs_ribbon_1;

// the command which will be registered into the command structure manually
static struct cmd new_star_cmd = { .title = "New star", .description = "New star",
	.type = CT_FACTORY_OP, .flags = CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN,
	.spec = { .op = { &ofs_star_1 } }, .icon_path = LIB_PATH"star.png" };

static struct cmd new_ribbon_cmd = { .title = "New ribbon", .description = "New ribbon",
	.type = CT_FACTORY_OP, .flags = CL_VIEW_MENU | CL_VIEW_TBAR2 | CC_CHECKABLE | CS_VIS_EN,
	.spec = { .op = { &ofs_ribbon_1 } }, .icon_path = LIB_PATH"ribbon.png" };

//string identifiers for property values as input in our GO Factory states
static string string_stretch, string_apex_no;
static string string_start_color, string_end_color, string_pearl_no, string_start_radius, string_end_radius;

u32 plugin_start(void)
{
	if (my_menu_id != -1)
	{
		//the plugin has already been loaded, do not register anything more
		return 0;
	}

	my_menu_id = plugin_menu_register("Heaven");

	// set the property identifier strings for property values
	// as input in our GO Factory commands
	ofs_star_1.value.prop.key = string_apex_no = string_lookup_ref("number-of-apices");
	ofs_star_2.value.prop.key = string_stretch = string_lookup_ref("star-stretch");

	ofs_ribbon_2.value.prop.key = string_start_color = string_lookup_ref("start-color");
	ofs_ribbon_3.value.prop.key = string_end_color = string_lookup_ref("end-color");
	ofs_ribbon_4.value.prop.key = string_pearl_no = string_lookup_ref("pearl-no");
	ofs_ribbon_5.value.prop.key = string_start_radius = string_lookup_ref("start-radius");
	ofs_ribbon_6.value.prop.key = string_end_radius = string_lookup_ref("end-radius");
	// load the star and ribbon icons
	new_star_cmd.icon_pixbuf = icon_load(new_star_cmd.icon_path);
	new_ribbon_cmd.icon_pixbuf = icon_load(new_ribbon_cmd.icon_path);
	// add the manually created command into the Command Structure
	plugin_menu_command_register( my_menu_id, &new_star_cmd );
	plugin_menu_command_register( my_menu_id, &new_ribbon_cmd );

	// other menu functions registered conveniently
	plugin_menu_command_register_conv( my_menu_id, "Twist", "Twist a Bezier curve",
		VTM_BEZIER, -1, RT_GROUP, twist_beziers );
	plugin_menu_command_register_conv( my_menu_id, "Hello world", "Say hello world",
		VTM_ANYTHING, 0, RT_ANY, hello_world );

	// a kernel plugin function -- another interface for the star
	plugin_function_register("star", "craetes a star", PLUGIN_T_VOID, 4,
		PLUGIN_T_UNS, PLUGIN_T_REAL, PLUGIN_T_HANGER, PLUGIN_T_HANGER);

	plugin_function_register("ribbon", "creates a ribbon", PLUGIN_T_VOID, 6,
		PLUGIN_T_GO, PLUGIN_T_UNS, PLUGIN_T_UNS, PLUGIN_T_UNS, PLUGIN_T_REAL, PLUGIN_T_REAL );

	return 0;
}

void plugin_stop(void)
{
	if (stopped)
		return;

	plugin_menu_unregister(my_menu_id);

	string_unref(string_apex_no);
	string_unref(string_stretch);
	string_unref(string_start_color);
	string_unref(string_end_color);
	string_unref(string_pearl_no);
	string_unref(string_start_radius);
	string_unref(string_end_radius);

	stopped = 1;
}

// A simple function which works with Bezier curves.
void twist_beziers(void * subj)
{
	char buf[TRANS_ERR_SIZE];
	// this si called from the GUI menu, so we can suppose that
	// the command requirements are fulfilled and that there
	// is a selection or a context object
	// (which implies a context group and tlo)
	ASSERT_TLO(context.tlo);
	ASSERT_GROUP(context.group);
	TRANS_BEGIN_MAIN(context.tlo, "Twist", buf)
		struct go * b;
		struct anchor * a1, * a2;
		struct hanger * h1, * h2;
		// the "subj" may contain the context object
		// if there is no selection
		if (subj)
		{
			ASSERT_GO(subj);
			go_select(subj);
		}
		WALK_GROUP_SELECTED_GO(context.group, b)
		{
			if (b->type != GOT_BEZIER)
				trans_fail("A non-bezier selected.");
			// switch the first two anchor positions
			h1 = (a1 = go_get_anchor(b, 0))->h;
			h2 = (a2 = go_get_anchor(b, 1))->h;
			anchor_rehang( a1, h2 );
			anchor_rehang( a2, h1 );
		}
	TRANS_FAILED
		statusbar_msg(NULL, "Cannot twist: %s\n", buf);
	TRANS_END
}

// A very simple callback which just demonstrates
// the usage of GUI callbacks.
void hello_world(void * subj UNUSED)
{
	statusbar_msg(NULL, "HELLO WORLD! :)");
}

// GO Factory callbacks
static void star_check_apex_number( struct of_state * state );
static void star_check_stretch( struct of_state * state );
static void star_create_star( struct of_state * state );

// GO Factory states definition
static struct of_state ofs_star_4 =
	{ 0, OFIK_DPR, DIST_PASS_HANGER, 0, {.dpr = { DIST_PASS_POINT, {.point = {0.0,{0.0, 0.0}}} }},
	  "Choose an apex of the star:",
	  star_create_star, NULL, &ofs_star_3, &ofs_end, NULL, { NULL }};
static struct of_state ofs_star_3 =
	{ 0, OFIK_DPR, DIST_PASS_HANGER, 0, {.dpr = { DIST_PASS_POINT, {.point = {0.0,{0.0, 0.0}}} }},
	  "Choose the center point of the star:",
	  NULL, NULL, &ofs_star_2, &ofs_star_4, NULL, { NULL }};
static struct of_state ofs_star_2 =
	{ 0, OFIK_PROP, PT_REAL, PTR_POSITIVE, {.prop = {.value = {.r = 2.0}}}, "Enter the star stretch of the star:",
	  star_check_stretch, NULL, &ofs_star_1, &ofs_star_3, NULL, { NULL }};
static struct of_state ofs_star_1 =
	{ 0, OFIK_PROP, PT_UNS, PTU_UNSPECIFIED, {.prop = {.value = {.u = 5}}}, "Enter the number of apices of the star:",
	  star_check_apex_number, NULL, &ofs_start, &ofs_star_2, NULL, { NULL }};


static void star_check_apex_number( struct of_state * state )
{
	uns n = state->value.prop.value.u;
	if ( n < 3 || n > 100 )
		trans_fail("Invalid number of apices.");
}

static void star_check_stretch( struct of_state * state )
{
	real r = state->value.prop.value.r;
	if ( !isfinite(r) || r <= 1.0 || r > 10.0 )
		trans_fail("Invalid stretch.");
}

// A function called both by the GUI and kernel interface.
static struct go * star_internal( uns apex_no, real stretch, struct hanger * center, struct hanger * apex )
{
	if ( apex_no < 3 || apex_no > 100 )
                trans_fail("Invalid number of apices.");
	if ( !isfinite(stretch) || stretch <= 1.0 || stretch > 10.0 )
                trans_fail("Invalid stretch.");
	if ( !center )
		trans_fail("Center must be non-NULL.");
	if ( !apex )
		trans_fail("Apex must be non-NULL.");

	struct go_group * group = context.group;
	if (!context.group)
		trans_fail("No context group.");
	if (!context.tlo)
		trans_fail("No context tlo.");

	// the resulting path to be craeted and filled with the star
	struct go_path * path = NULL;

	char errs[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( context.tlo, "Star", errs)

		// First, create some temporary gos on which the star will hang.
		
		//the outer circle
		struct go_elarc * circ = go_elarc_new_xy1ecc( center, apex, 0.0, 0.0, GEOM_CONIC_ARC_POINT_ENTIRE );
		go_link_head(&circ->go, group);

		//a segment with a parametric point on it
		struct hanger * h = go_get_hanger_start( &circ->go );
		ASSERT(h);
		struct go_segment * seg = go_segment_new( center, h );
		go_link_head(&seg->go, group);

		struct go_parametric_point * pp = go_parametric_point_new( &seg->go, 1 / stretch );
		go_link_head(&pp->go, group);

		// another circle hanging on the center and the parametric point
		struct go_elarc * circ_inner = go_elarc_new_xy1ecc( center, go_get_hanger( &pp->go, 0 ),
				0.0, 0.0, GEOM_CONIC_ARC_POINT_ENTIRE );
		go_link_head(&circ_inner->go, group);
		prop_set_uns( go2o(&circ_inner->go), string_invisible, 1 );

		// parametric points for the apices
		struct go * apices_outer[apex_no];
		struct go * apices_inner[apex_no];

		real difference = 0.5 / apex_no; // the difference by which the inner points are shifted
		for (uns u = 0; u < apex_no; u++ )
		{
			apices_outer[u] = &go_parametric_point_new( &circ->go, ((real)u) / apex_no )->go;
			go_link_head(apices_outer[u], group);
	
			apices_inner[u] = &go_parametric_point_new( &circ_inner->go, difference + ((real)u) / apex_no )->go;
			go_link_head(apices_inner[u], group);
		}
	
		// the main path which contains the star
		path = go_path_new();
		go_link_head(&path->group.go, group);
	
		struct go * segment = NULL;
		struct hanger * h1, * h2;
		for (int i = apex_no - 1; i >= 0; i-- )
		{
			// an inward ray
			h1 = ((uns)i == apex_no - 1) ? go_get_hanger(apices_outer[0], 0) : go_get_hanger_end(segment);
			h2 = go_get_hanger(apices_inner[i], 0);
			ASSERT( h1 && h2 );
			go_link_tail( segment = &go_segment_new( h1, h2 )->go, &path->group );
	
			// an outward ray
			h1 = go_get_hanger_end(segment);
			h2 = go_get_hanger(apices_outer[i], 0);
			ASSERT( h1 && h2 );
			go_link_tail( segment = &go_segment_new( h1, h2 )->go, &path->group );
		}
		// remove the temporarily created gos
		for (uns u = 0; u < apex_no; u++)
		{
			go_remove_rehang(apices_outer[u]);
			go_remove_rehang(apices_inner[u]);
		}
		go_remove_rehang(&circ->go);
		go_remove_rehang(&seg->go);
		go_remove_rehang(&circ_inner->go);
		go_remove_rehang(&pp->go);

	TRANS_FAILED
		trans_fail(errs);
		path = NULL;
	TRANS_END

	return path ? &path->group.go : NULL;
}

// Go Factory callback which extracts the GO Factory input values
// and passes them to the internal function.
// After that, if the function was succesful, recycles
// the GUI style properties and sets the object to context.
static void star_create_star( struct of_state * state UNUSED )
{
	struct go * result = star_internal(
		ofs_star_1.value.prop.value.u, ofs_star_2.value.prop.value.r,
		ofs_star_3.value.dpr.data.hanger.hanger,
		ofs_star_4.value.dpr.data.hanger.hanger );

	if (result)
	{
		gui_prop_recycle( go2o(result) );
		change_context( context.w, go_get_group(result), go2o(result) );
	}
}

// The kernel interface.
void star(union plugin_arg * arg)
{
	uns apex_no = arg[0].t_uns;
	real stretch = arg[1].t_real;
	struct hanger * center = arg[2].ptr;
	struct hanger * apex = arg[3].ptr;

	star_internal( apex_no, stretch, center, apex );
}

/******************** RIBBON ************************/

static void ribbon_check_no( struct of_state * state );
static void ribbon_check_go( struct of_state * state );
static void ribbon_create_ribbon( struct of_state * state );

static struct of_state ofs_ribbon_6 =
	{ 0, OFIK_PROP, PT_REAL, PTR_POSITIVE, {.prop = {.value = {.r = 5.0 }}}, "Enter the end pearl radius:",
	  ribbon_create_ribbon, NULL, &ofs_ribbon_5, &ofs_end, NULL, { NULL }};
static struct of_state ofs_ribbon_5 =
	{ 0, OFIK_PROP, PT_REAL, PTR_POSITIVE, {.prop = {.value = {.r = 5.0 }}}, "Enter the start pearl radius:",
	  NULL, NULL, &ofs_ribbon_4, &ofs_ribbon_6, NULL, { NULL }};
static struct of_state ofs_ribbon_4 =
	{ 0, OFIK_PROP, PT_UNS, PTR_UNSPECIFIED, {.prop = {.value = {.u = 10}}}, "Enter the number of pearls:",
	  ribbon_check_no, NULL, &ofs_ribbon_3, &ofs_ribbon_5, NULL, { NULL }};
static struct of_state ofs_ribbon_3 =
	{ 0, OFIK_PROP, PT_UNS, PTU_COLOR, {.prop = {.value = {.u = 0xFF}}}, "Choose the end color:",
	  NULL, NULL, &ofs_ribbon_2, &ofs_ribbon_4, NULL, { NULL }};
static struct of_state ofs_ribbon_2 =
	{ 0, OFIK_PROP, PT_UNS, PTU_COLOR, {.prop = {.value = {.u = 0xFFFFFFFF}}}, "Choose the start color:",
	  NULL, NULL, &ofs_ribbon_1, &ofs_ribbon_3, NULL, { NULL }};
static struct of_state ofs_ribbon_1 =
	{ 0, OFIK_DPR, DIST_PASS_GO, 0, {.dpr = { DIST_PASS_GO, {.go = { .go = NULL }}}},
	  "Choose a curve for the ribbon",
	  ribbon_check_go, NULL, &ofs_start, &ofs_ribbon_2, NULL, { NULL }};

static void ribbon_check_go( struct of_state * state )
{
	struct go * go;
	if (!(go = state->value.dpr.data.go.go) || !go_get_hanger_curve(go))
		trans_fail("Not a curve go.");
}

static void ribbon_check_no( struct of_state * state )
{
	uns n = state->value.prop.value.u;
	if ( n < 2 || n > 1000 )
		trans_fail("Invalid number of pearls.");
}

#define GRAD_COL(_x1, _x2) (((int)_x2 - (int)_x1) * param + 0.5 );

// Create a ribbon: n decorator points on the given curve with
// gradually changing color and radius.
// The newly created gos are linked into a group.
static void ribbon_internal( struct go * curve, uns col1, uns col2, uns n, real radius1, real radius2 )
{
	if (!curve || !go_get_hanger_curve(curve))
		trans_fail("Not a curve go.");
	if ( n < 2 || n > 1000)
		trans_fail("Invalid number of pearls.");
	if (!context.group)
		trans_fail("No context group.");
	if (!context.tlo)
		trans_fail("No context tlo.");

	struct go_group * grp = NULL;
	char errs[TRANS_ERR_SIZE];
	uns r1, g1, b1, a1, r2, g2, b2, a2, r3, g3, b3, a3;
	uns2rgba_uns( col1, &r1, &g1, &b1, &a1 );
	uns2rgba_uns( col2, &r2, &g2, &b2, &a2 );

        TRANS_BEGIN_MAIN( context.tlo, "Pearl Ribbon", errs )

		grp = go_group_new();
		struct go * prev = go_get_owner_and_member( curve, context.group );
		if (prev)
			go_link_prev( &grp->go, context.group, prev );
		else
			go_link_head( &grp->go, context.group );

		struct go_decorator_point * pearl = NULL;
		struct go_parametric_point * pt = NULL;
		for (uns i = 0; i < n; i++)
		{
			real param = i * 1.0 / (n - 1);
			pt = go_parametric_point_new( curve, param );
			go_link_head(&pt->go, grp);
			pearl = go_decorator_point_new( go_get_hanger( &pt->go, 0 ), 0,
				radius1 + (radius2 - radius1) * param, 0.0 );
			go_link_head(&pearl->go, grp);

			r3 = r1 + GRAD_COL(r1, r2);
			g3 = g1 + GRAD_COL(g1, g2);
			b3 = b1 + GRAD_COL(b1, b2);
			a3 = a1 + GRAD_COL(a1, a2);
			DBG("Color = %u, %u, %u, %u\n", r3, g3, b3, a3 );
			prop_set_uns( go2o(&pearl->go), string_fill_color, rgba_uns2uns( r3, g3, b3, a3 ));
		}

	TRANS_FAILED
		trans_fail(errs);
	TRANS_END
}

// GO Factory wrapper for ribbon_internal.
static void ribbon_create_ribbon( struct of_state * state UNUSED )
{
	ribbon_internal(
			ofs_ribbon_1.value.dpr.data.go.go,
			ofs_ribbon_2.value.prop.value.u,
			ofs_ribbon_3.value.prop.value.u,
			ofs_ribbon_4.value.prop.value.u,
			ofs_ribbon_5.value.prop.value.r,
			ofs_ribbon_6.value.prop.value.r );
}

// Kernel wrapper for ribbon_internal.
void ribbon( union plugin_arg * arg )
{
	ribbon_internal( arg[0].ptr, arg[1].t_uns, arg[2].t_uns,
		arg[3].t_uns, arg[4].t_real, arg[5].t_real );
}
