/** @file
 *
 * Miscelaneous kernel functions.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */

#include "lib/lib.h"
#include <stdio.h>
#include <string.h>
#include <guile/gh.h>
#include "kernel/kernel.h"
#include "font/font.h"
#include "export/export.h"
#include "build/path.h"
#include "import/import.h"

#include "scheme/gl_trans.h"
#include "scheme/glt_common.h"
#include "scheme/glt_kernel.h"

#include "kernel/guileaccess.h"
char err_buf [TRANS_ERR_SIZE];

static uns direct_gprintf;

void scheme_module_load (const char *str)
{
  SCM m = scm_current_module ();
  scm_c_primitive_load (str);
  scm_set_current_module (m);
}

static UNUSED void set_direct_gprintf (uns val)
{ direct_gprintf = val; }

static UNUSED uns get_direct_gprintf (void)
{ return (direct_gprintf); }

static int guile_printf_int (const char *s, int bufsize, va_list args)
{
  byte buf[bufsize];
  int n = vsnprintf ((char *)buf, bufsize, s, args);
  if  ((n > -1) && (n < bufsize))
    scm_c_write (scm_current_output_port (), buf,  n);
  return (n);
}

static int guile_printf(const char *s, va_list args)
{
  int bufsize = 128;
  int n = guile_printf_int (s, bufsize, args);

  if (n >= bufsize)
    n = guile_printf_int (s, n + 1, args);
  else while (n < 0)
    {
      bufsize *= 2;
      n = guile_printf_int (s, bufsize, args);
    }

  return (n);
}

/* normal log, no other action */
static UNUSED int gprintf(const char *s, ...)
{
  va_list args;
  int i;

  va_start(args, s);
  if (direct_gprintf) i = vprintf (s, args);
  else i = guile_printf(s, args);
  va_end(args);

  return (i);
}

// This functions doesn't do any type checking

static inline void * scm2ptr (SCM number) { return (void *) scm_num2ulong (number, SCM_ARG1, "scm2ptr"); }
static inline SCM ptr2scm (void * ptr) { return scm_ulong2num ((unsigned long) ptr); }

static SCM prop2scm (struct o* o, struct prop *p)
{
  prop_value pv = prop_get_value (o, p);
  switch (p -> type)
    {
    case PT_UNS:
      return (out_SUNS (pv . u));
    case PT_POINTER:
      return (ptr2scm (pv . v));
    case PT_STRING:
      return (out_SSTRINGC (pv . s -> text));
    case PT_REAL:
      return (out_SREAL (pv . r));
    default:
      BUG();
    }
}

static SCM gl_property_get_int (SCM o, SCM name)
{
  SCM_ASSERT (is_o_proxy (o), o, SCM_ARG1, "property-get");
  SCM_ASSERT (scm_symbol_p (name), name, SCM_ARG2, "property-get");

  string s = string_find (SCM_SYMBOL_CHARS (name));
  if(!s) return (SCM_BOOL_F);
  struct prop *p = prop_find ((optr (o)), s);
  if (p == NULL) return (SCM_BOOL_F);
  else return (SCM_LIST4 (prop2scm (optr (o), p),
			  SCM_MAKINUM (p -> type),
			  SCM_MAKINUM (p -> subtype),
			  SCM_MAKINUM (p -> flags)));
}

/*
// static SCM gl_property_set_uns (SCM o, SCM name, SCM type, SCM value)
static SCM gl_property_set_uns (SCM o, SCM name, SCM value)
{
  SCM_ASSERT (is_o_proxy (o), o, SCM_ARG1, "property-set!");
  SCM_ASSERT (scm_symbol_p (name), name, SCM_ARG2, "property-set!");

  prop_value pv;
  pv . u = scm_num2uint (value, SCM_ARG3, "property-set!");

  TRANS_BEGIN (err_buf)
    prop_set (optr (o), string_lookup (SCM_SYMBOL_CHARS (name)), pv);
  TRANS_FAILED
    throw_transaction_failed (err_buf);
  TRANS_END;

  return (SCM_UNSPECIFIED);
}

static SCM gl_property_set_string (SCM o, SCM name, SCM value)
{
  SCM_ASSERT (is_o_proxy (o), o, SCM_ARG1, "property-set!");
  SCM_ASSERT (scm_symbol_p (name), name, SCM_ARG2, "property-set!");
  SCM_ASSERT (SCM_STRINGP (value), value, SCM_ARG3, "property-set!");
  ASSERT_O_TRANS_SCHEME("gl_property_set_string", optr(o));

  TRANS_BEGIN (err_buf)
    prop_set (optr (o), string_lookup (SCM_SYMBOL_CHARS (name)),
	      (prop_value){.s = string_lookup (SCM_STRING_CHARS (value))});
  TRANS_FAILED
    throw_transaction_failed (err_buf);
  TRANS_END;

  return (SCM_UNSPECIFIED);
}

static SCM gl_property_set_real (SCM o, SCM name, SCM value)
{
  SCM_ASSERT (is_o_proxy (o), o, SCM_ARG1, "property-set!");
  SCM_ASSERT (scm_symbol_p (name), name, SCM_ARG2, "property-set!");
  //  SCM_ASSERT (SCM_INUMP (type), type, SCM_ARG3, "property-set!");
  // no need for check type of value, because scm_num2float does it itself
  ASSERT_O_TRANS_SCHEME("gl_property_set_real", optr(o));

  prop_value pv;
  pv . r = scm_num2float (value, SCM_ARG3, "property-set!");

  TRANS_BEGIN (err_buf)
    prop_set (optr (o), string_lookup (SCM_SYMBOL_CHARS (name)), pv);
  TRANS_FAILED
    throw_transaction_failed (err_buf);
  TRANS_END;

  return (SCM_UNSPECIFIED);
}
*/
static SCM gl_decode_prop_transform (SCM xt)
{
  struct geom_transform *t = scm2ptr (xt);
  return (scm_list_n (out_SSYMBOLC (string_transform->text),
		      out_SREAL (t -> coef [0] [0]), out_SREAL (t -> coef [0] [1]),
		      out_SREAL (t -> coef [1] [0]), out_SREAL (t -> coef [1] [1]),
		      SCM_UNDEFINED));
}

static SCM gl_property_transform_set (SCM o, SCM name, SCM value)
{
  SCM_ASSERT (is_o_proxy (o), o, SCM_ARG1, "property-set!");
  SCM_ASSERT (scm_symbol_p (name), name, SCM_ARG2, "property-set!");
  SCM_ASSERT ((scm_ilength (value) == 4), name, SCM_ARGn, "property-set!");
  ASSERT_O_TRANS_SCHEME("gl_property_transform_set", optr(o));

  // FIXME : convert typearg errors (ilength, num2float) to misc-error "Bad property value definition"

  real ar [4];
  for (int i = 0; i < 4; i ++)
    {
      ar [i] = scm_num2float (SCM_CAR (value), SCM_ARGn, "property-set!");
      value = SCM_CDR (value);
    }

  prop_value pv;
  pv . t = xmalloc (sizeof (struct geom_transform));
  int err = geom_transform_set (pv . t, ar [0], ar [1], 0, ar [2], ar [3], 0);

  if (err)
    {
      xfree (pv . t);
      scm_misc_error ("property-set!", "invalid transformation matrix", SCM_EOL);
    }

  TRANS_BEGIN (err_buf)
    prop_set_forte (optr (o), SCM_SYMBOL_CHARS (name), PT_POINTER, PTP_TRANSFORM, 1, pv);
  TRANS_FAILED
    xfree (pv . t); // Should i?
    throw_transaction_failed (err_buf);
  TRANS_END;

  return (SCM_UNSPECIFIED);
}



/*
static SCM gl_property_get_subtype_i (SCM o, SCM name)
{
  SCM_ASSERT (is_o_proxy (o), o, SCM_ARG1, "property-get-subtype");
  SCM_ASSERT (scm_symbol_p (name), name, SCM_ARG2, "property-get-subtype");

  struct prop *p = prop_find (optr (o),
			      string_lookup (SCM_SYMBOL_CHARS (name)));
  if (p == NULL) return (SCM_BOOL_F);
  else return (scm_cons (SCM_MAKINUM (p -> type), SCM_MAKINUM (p -> subtype)));
}
*/

static SCM gl_property_for_each (SCM o, SCM cb)
{
  SCM_ASSERT (is_o_proxy (o), o, SCM_ARG1, "property-for-each");
  SCM_ASSERT (scm_procedure_p (cb), cb, SCM_ARG2, "property-for-each");

  struct prop *p;
  WALK_PROP(optr (o), p)
    {
      // trivial C code, no need for jump protection
      scm_apply_0 (cb, scm_list_5 (scm_str2symbol (p -> key -> text),
				   prop2scm (optr (o), p),
				   SCM_MAKINUM (p -> type),
				   SCM_MAKINUM (p -> subtype),
				   SCM_MAKINUM (p -> flags)));

      // FIXME str2symbol eats ref?
      /*
      SCM exception = SCM_BOOL_F;
      SCM thunk = scm_call_3 (make_thunk, cb, scm_str2symbol (p -> key -> text),
			      prop2scm (optr (o), p));
      call_guile_thunk (thunk, &exception);
      forward_exception (exception);
     */
    }
  return (SCM_UNSPECIFIED);
}

static SCM gl_proxy2ptr(SCM o)
{
  SCM_ASSERT (is_o_proxy (o), o, SCM_ARG1, "proxy2ptr");
  return (scm_ulong2num(SCM_SMOB_DATA(o)));
}

static void
assert_in (uns type, SCM value, uns argn, const char *fn_name)
{
  if (type & PLUGIN_T_O)
    assert_SO (value, argn, fn_name, -1, -1); // FIXME correct check
  else switch (type)
    {
    case PLUGIN_T_INT:
      assert_SINT (value, argn, fn_name);
      break;
    case PLUGIN_T_UNS:
      assert_SUNS (value, argn, fn_name);
      break;
    case PLUGIN_T_REAL:
      assert_SREAL (value, argn, fn_name);
      break;
    case PLUGIN_T_BOOL:
      assert_SBOOL (value, argn, fn_name);
      break;
    case PLUGIN_T_STRING_C:
      assert_SSTRINGC (value, argn, fn_name);
      break;
    case PLUGIN_T_SYMBOL_C:
      assert_SSYMBOLC (value, argn, fn_name);
      break;
    case PLUGIN_T_STRING_S:
      assert_SSTRINGS (value, argn, fn_name);
      break;
    case PLUGIN_T_SYMBOL_S:
      assert_SSYMBOLS (value, argn, fn_name);
      break;
    case PLUGIN_T_ANCHOR:
      assert_SANCHOR (value, argn, fn_name);
      break;
    case PLUGIN_T_HANGER:
      assert_SHANGER (value, argn, fn_name);
      break;
    default:
      BUG ();
    }
}

static union plugin_arg
convert_in (uns type, SCM value)
{
  volatile union plugin_arg retval;

  if (type & PLUGIN_T_O)
    retval . ptr = in_SO (value);
  else switch (type)
    {
    case PLUGIN_T_INT:
      retval . t_int = in_SINT (value);
      break;
    case PLUGIN_T_UNS:
      retval . t_uns = in_SUNS (value);
      break;
    case PLUGIN_T_REAL:
      retval . t_real = in_SREAL (value);
      break;
    case PLUGIN_T_BOOL:
      retval . t_uns = in_SBOOL (value);
      break;
    case PLUGIN_T_STRING_C:
      retval . t_cs = in_SSTRINGC (value);
      break;
    case PLUGIN_T_SYMBOL_C:
      retval . t_cs = in_SSYMBOLC (value);
      break;
      //    case PLUGIN_T_STRING_S:
      //      retval . ptr = in_SSTRINGS (value);
      //      break;
      //    case PLUGIN_T_SYMBOL_S:
      //      retval . ptr = in_SSYMBOLS (value);
      //      break;
    case PLUGIN_T_ANCHOR:
      retval . ptr = in_SANCHOR (value);
      break;
    case PLUGIN_T_HANGER:
      retval . ptr = in_SHANGER (value);
      break;
    default:
      BUG ();
    }

  return (retval);
}

static SCM
convert_out (uns type, union plugin_arg value)
{
  SCM retval;

  if (type & PLUGIN_T_O)
    retval = out_SO (value . ptr);
  else switch (type)
    {
    case PLUGIN_T_VOID:
      retval = SCM_UNSPECIFIED;
      break;
    case PLUGIN_T_INT:
      retval = out_SINT (value . t_int);
      break;
    case PLUGIN_T_UNS:
      retval = out_SUNS (value . t_uns);
      break;
    case PLUGIN_T_REAL:
      retval = out_SREAL (value . t_real);
      break;
    case PLUGIN_T_BOOL:
      retval = out_SBOOL (value . t_uns);
      break;
    case PLUGIN_T_STRING_C:
      retval = out_SSTRINGC (value . t_cs);
      break;
    case PLUGIN_T_SYMBOL_C:
      retval = out_SSYMBOLC (value . t_cs);
      break;
      // FIXME: check strings
      //    case PLUGIN_T_STRING_S:
      //      retval = out_SSTRINGS (value . ptr);
      //      break;
      // FIXME: check strings
      //    case PLUGIN_T_SYMBOL_S:
      //      retval = out_SSYMBOLS (value . ptr);
      //      break;
    case PLUGIN_T_ANCHOR:
      retval = out_SANCHOR (value . ptr);
      break;
    case PLUGIN_T_HANGER:
      retval = out_SHANGER (value . ptr);
      break;
    default:
      BUG ();
    }

  return (retval);
}

static SCM plugin_call (union plugin_arg (*fn) (union plugin_arg *), SCM args,
			const char *fn_name, uns ret_type, uns arg_count, uns *arg_types)
{
  uns i = 0;
  SCM l = args;

  for (; pair_c_p (l) && (i < arg_count); l = SCM_CDR (l), i ++)
    assert_in (arg_types [i], SCM_CAR (l), i + 1, fn_name);

  if ((i < arg_count) || (! SCM_NULLP (l)))
    scm_error_num_args_subr (fn_name);

  union plugin_arg data [arg_count];
  volatile SCM retval = SCM_UNSPECIFIED;

  TRANS_BEGIN_ANONYMOUS(err_buf)
    {
      for (i = 0, l = args; i < arg_count; l = SCM_CDR (l), i ++)
	data [i] = convert_in (arg_types [i], SCM_CAR (l));

      if (ret_type == PLUGIN_T_VOID)
	((void (*) (union plugin_arg *)) fn) (data);
      else retval = convert_out (ret_type, fn (data));
    }
  TRANS_FAILED
    throw_transaction_failed (err_buf);
  TRANS_END;

  return (retval);
}

static SCM plugin_call_fn;

static SCM gl_plugin_call (SCM id, SCM args)
{
  struct plugin_function *fn_dsc = (void *) scm_num2ulong (id, 1, "plugin-call");

  return (plugin_call (fn_dsc -> func, args, fn_dsc -> name, fn_dsc -> retval,
		       fn_dsc -> argc, fn_dsc -> prototype));
}

extern void register_plugin_fn (const char *name, unsigned long specific);
void register_plugin_fn (const char *name, unsigned long specific)
{ scm_c_define (name, scm_call_2 (curry, plugin_call_fn, scm_ulong2num (specific))); }



static void scheme_go_cb (struct go * go, void * cb_data)
{
  SCM scheme_cb =  *((SCM *) cb_data);
  SCM exception = SCM_BOOL_F;
  SCM sgo = out_SO (go);
  SCM thunk = scm_call_2 (make_thunk, scheme_cb, sgo);
  call_guile_thunk (thunk, &exception);
  //  scm_write (exception, SCM_UNDEFINED);
  propagate_exception (exception);
}

static SCM gl_tsort_tlo (SCM tlo, SCM cb)
{
  SCM_ASSERT (is_specific_proxy (tlo, T_OBJ, OT_TLO, -1), tlo, SCM_ARG1, "tsort-tlo");
  SCM_ASSERT (scm_procedure_p(cb), cb, SCM_ARG2, "tsort-tlo");
  ASSERT_GO_TRANS_SCHEME ("tsort-tlo");

  struct obj_tlo *tlo_c = (struct obj_tlo *) scm2o (tlo);
  struct go_group *group = tlo_get_tlg (tlo_c);

  TRANS_BEGIN (err_buf)
    tsort_start (tlo_c);
    tsort_insert_group (group, tlo_c);
    tsort_call (tlo_c, scheme_go_cb, &cb);
  TRANS_FAILED
    propagate_trans_fail (err_buf);
  TRANS_END;

  return (SCM_UNSPECIFIED);
}

static SCM
gl_unify_go_list (SCM ar0)
{
  struct obj_tlo *tlo = NULL;
  SCM lst = ar0;
  SCM retval = SCM_EOL;

  for (; pair_c_p (lst); lst = SCM_CDR (lst))
    {
      SCM g = SCM_CAR (lst);
      if (is_specific_proxy (g, T_GO, -1, -1))
	{
	  struct go * g2 = (struct go *) scm2o (g);

	  if (tlo == NULL)
	    {
	      tlo = go_get_tlo (g2);
	      tsort_start (tlo);
	    }

	  if (tlo == go_get_tlo (g2)) tsort_insert (g2, tlo);
	  else lst = SCM_BOOL_F;
	}
      else lst = SCM_BOOL_F;
    }

  if (tlo == NULL)
    return (SCM_EOL);
  else if (lst == SCM_EOL)
    {
      struct go * go;

      WALK_TSORT_GO (tlo, go)
	retval = scm_cons (o2scm ((struct o*) go), retval);

      tsort_presort_end (tlo);
      return (retval);
    }
  else
    {
      tsort_presort_end (tlo);
      SCM_ASSERT (0, ar0, SCM_ARG1, "unify-list");
      return (retval); // to supress warning
    }
}

static SCM
gl_go_get_anchors (SCM ar0)
{
  SCM_ASSERT (is_specific_proxy (ar0, T_GO, -1, -1), ar0, SCM_ARG1, "go-get-anchors");
  struct go *go = (struct go *) scm2o (ar0);
  SCM retval = SCM_EOL;
  struct anchor *a;
  int i;
  
  WALK_GO_ANCHOR (go, a, i)
    retval = scm_cons (anchor2scm (a), retval);
  
  return (retval);
}
static SCM
gl_go_get_hangers (SCM ar0)
{
  SCM_ASSERT (is_specific_proxy (ar0, T_GO, -1, -1), ar0, SCM_ARG1, "go-get-hangers");
  struct go *go = (struct go *) scm2o (ar0);
  SCM retval = SCM_EOL;
  struct hanger *h;
  
  WALK_GO_HANGER_BEGIN (go, h)
    retval = scm_cons (hanger2scm (h), retval);
  WALK_GO_HANGER_END;
  
  return (retval);
}

// SCM gl_finish(void) NONRET;
static SCM gl_finish(void)
{
  if(vrr_with_gui)
    return (SCM_UNSPECIFIED); // FIXME:
  else
    kernel_finish(),
  exit(0);
}


#include "obj/kernel/generated_bindings.h"

void doc_save (struct obj_doc * doc, const char * filename)
{
  SCM save_doc_fn = scm_variable_ref (scm_c_module_lookup (scm_c_resolve_module ("vrr save"), "save-doc"));
  call_guile_fn_from_c (save_doc_fn, scm_list_2 (o2scm ((struct o *)doc),
						 scm_makfrom0str (filename))); 
}

struct obj_doc * doc_load (const char * filename)
{
  SCM load_doc_fn = scm_variable_ref (scm_c_module_lookup (scm_c_resolve_module ("vrr load"), "load-doc"));
  return (struct obj_doc*) scm2o (call_guile_fn_from_c (load_doc_fn,
							scm_list_1 (scm_makfrom0str (filename))));
}

void settings_save (const char * filename)
{ 
   SCM save_doc_fn = scm_variable_ref (scm_c_module_lookup (scm_c_resolve_module ("vrr save"), "save-doc"));
   call_guile_fn_from_c (save_doc_fn, scm_list_2 (o2scm ((struct o *)obj_universum),
						  scm_makfrom0str (filename)));
}

void settings_load (const char * filename)
{ 
  SCM load_doc_fn = scm_variable_ref (scm_c_module_lookup (scm_c_resolve_module ("vrr load"), "load-doc"));
  call_guile_fn_from_c (load_doc_fn, scm_list_1 (scm_makfrom0str (filename)));
}

void
guilelink_start(void *data, int argc, char **argv, void (*start)(void *data, int argc, char **argv))
{
  scm_boot_guile(argc, argv, start, data);
}

void
gl_misc_bindings_init(void)
{
  // vars
  scm_c_define("universe", o2scm ((struct o *) obj_universum));
  scm_c_define("morgue", o2scm ((struct o *) obj_zombie));

  scm_c_export ("universe", "morgue", NULL);

  // low level
  scm_c_define_gsubr("property-get-int", 2, 1, 0, gl_property_get_int);
  scm_c_define_gsubr("property-transform-set", 3, 0, 0, gl_property_transform_set);
  scm_c_define_gsubr("decode-prop-transform", 1, 0, 0, gl_decode_prop_transform);
  scm_c_define_gsubr("property-for-each", 2, 0, 0, gl_property_for_each);
  scm_c_define_gsubr("proxy2ptr", 1, 0, 0, gl_proxy2ptr);
  scm_c_define_gsubr("tsort-tlo", 2, 0, 0, gl_tsort_tlo);
  scm_c_define_gsubr("finish", 0, 0, 0, gl_finish);

  scm_c_export ("property-get-int", "property-transform-set", "decode-prop-transform",
		"property-for-each", "tsort-tlo", "proxy2ptr", NULL);

  // high level
  plugin_call_fn = scm_c_define_gsubr ("plugin-call", 1, 0, 1, gl_plugin_call);
  scm_c_define_gsubr ("go-get-anchors", 1, 0, 0, gl_go_get_anchors);
  scm_c_define_gsubr ("go-get-hangers", 1, 0, 0, gl_go_get_hangers);
  scm_c_define_gsubr ("unify-go-list", 1, 0, 0, gl_unify_go_list);
  scm_c_export ("plugin-call", "go-get-anchors", "go-get-hangers", "unify-go-list", NULL);
}

static void
low_module_init (UNUSED void *data)
{
  gl_kernel_bindings_init ();
  gl_trans_bindings_init ();
  gl_misc_bindings_init ();
  kernel_generated_bindings_init ();
}


void
guilelink_kernel_init(void)
{
  glt_kernel_init ();

  scm_c_define_module ("vrr low", low_module_init, NULL);

  scheme_module_load (LIB_PATH "misc.scm");
  scheme_module_load (LIB_PATH "support.scm");
  scheme_module_load (LIB_PATH "property.scm");
  scheme_module_load (LIB_PATH "save.scm");
  scheme_module_load (LIB_PATH "load.scm");
  scheme_module_load (LIB_PATH "high.scm");

  gl_trans_init ();

  SCM high = scm_c_resolve_module ("vrr high");
  scm_c_call_with_current_module (high, glt_kernel_update_high, NULL);

  SCM misc = scm_c_resolve_module ("vrr misc");
  curry = scm_variable_ref (scm_c_module_lookup (misc, "curry"));
  make_thunk = scm_variable_ref (scm_c_module_lookup (misc, "make-thunk"));

  scm_c_use_module ("vrr high");
}
