/** @file
 * GUILE-link - interface functions for kernel data types.
 *
 * Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
 * This program is distributed under GNU General Public License, see COPYING.
 */


#include "lib/lib.h"
#include <stdio.h>
#include <string.h>
#include <guile/gh.h>
#include "kernel/kernel.h"
#include "build/path.h"

#include "scheme/gl_trans.h"
#include "scheme/glt_common.h"
#include "scheme/glt_kernel.h"

// htable for (symbol -> number) conversion 
SCM obj_type_numbers;
SCM go_type_numbers;
// SCM go_subtype_numbers;

// vector for (number -> symbol) conversion 
SCM *obj_type_symbols;
SCM *go_type_symbols;
// SCM **go_subtype_symbols;

SCM symbol_obj    = SCM_BOOL_F;
SCM symbol_go     = SCM_BOOL_F;
SCM symbol_hanger = SCM_BOOL_F;
SCM symbol_anchor = SCM_BOOL_F;
SCM symbol_transaction_failed = SCM_BOOL_F;
SCM symbol_condition_not_satisfied = SCM_BOOL_F;
SCM symbol_misc_error = SCM_BOOL_F;

typedef struct rst_item
{
  struct rst_item* next;
  SCM proxy;
  void* p;
} rst_item;


#define HT_PREFIX(x) rst_ ## x
#define HT_KEY void*
#define HT_KEY_POINTER p
#define HT_TYPE SCM
#define HT_STRUCT rst_item
#define HT_ALLOC(_struct, _key, _type) \
	(_struct) = xmalloc (sizeof (rst_item)); \
	(_struct)->p = _key; \
	(_struct)->proxy = _type;
#define HT_ONLY_ONE rst_oo
#include "lib/hashtable.h"

#define WALK_RST(_walk) for(rst_item **_index = rst_oo.table, *_walk = *_index ? *_index : rst_next(&_index); _index < rst_oo.table + rst_oo.size && _walk; _walk = _walk->next ? _walk->next : rst_next(&_index))

scm_t_bits o_tag;
scm_t_bits hanger_tag;
scm_t_bits anchor_tag;

int is_o_proxy(SCM proxy)
{ return SCM_SMOB_PREDICATE(o_tag, proxy); }

int is_hanger_proxy(SCM proxy)
{ return SCM_SMOB_PREDICATE(hanger_tag, proxy); }

int is_anchor_proxy(SCM proxy)
{ return SCM_SMOB_PREDICATE(anchor_tag, proxy); }


int is_specific_proxy(SCM proxy, int kind, int type, int subtype)
{
  if ((kind == T_HANGER) && (is_hanger_proxy (proxy)) && (type == -1))
    return (1);
  if ((kind == T_ANCHOR) && (is_anchor_proxy (proxy)) && (type == -1))
    return (1);
  else if (is_o_proxy (proxy))
    {
      struct o * obj = optr (proxy);
      return (((kind == -1) || (obj -> kind == kind)) &&
	      ((type == -1) || (obj -> type == type)) &&
	      ((subtype == -1) || (obj -> subtype == subtype)));
    }
  else return (0);
}

SCM vrr_exception_handler;

static size_t o_proxy_free (SCM proxy)
{
  rst_delete (optr (proxy));
  o_unref (optr (proxy));
  return (0);
}

static size_t hanger_proxy_free (SCM proxy)
{
  rst_delete( hptr (proxy));
  hanger_unref (hptr (proxy));
  return (0);
}

static size_t anchor_proxy_free (SCM proxy)
{
  rst_delete( aptr (proxy));
  go_unref (anchor_get_go (aptr (proxy)));
  return (0);
}

SCM o2scm(struct o *p)
{
  if (p == NULL) return (SCM_BOOL_F);

  struct rst_item *it = rst_get(p);
  if (it == NULL)
    {
      o_ref (p);
      SCM proxy;
      SCM_NEWSMOB (proxy, o_tag, p);
      rst_insert(p, proxy);
      return proxy;
    }
  else return it->proxy;
}

SCM hanger2scm(struct hanger *p)
{
  if (p == NULL) return (SCM_BOOL_F);

  struct rst_item *it = rst_get(p);
  if (it == NULL)
    {
      hanger_ref (p);
      SCM proxy;
      SCM_NEWSMOB (proxy, hanger_tag, p);
      rst_insert(p, proxy);
      return proxy;
    }
  else return it->proxy;
}

SCM anchor2scm(struct anchor *p)
{
  if (p == NULL) return (SCM_BOOL_F);

  struct rst_item *it = rst_get(p);
  if (it == NULL)
    {
      go_ref ( anchor_get_go(p));
      SCM proxy;
      SCM_NEWSMOB (proxy, anchor_tag, p);
      rst_insert(p, proxy);
      return proxy;
    }
  else return it->proxy;
}

static int
o_proxy_print (SCM smob, SCM port, UNUSED scm_print_state *pstate)
{
  struct o * obj = optr (smob);
  const char *name = o_lookup_name (obj);
  size_t l = (name ? strlen (name) : 0) + 14 + 32; // %p output from printf should be smaller than 32
  char buff [l]; 

  int rv = snprintf (buff, l, "#<vrr-o \"%s\" %p>", name, obj);
  
  if ((rv > -1) && (rv < (int)l))
    scm_puts (buff, port);
  else
    scm_puts ("#<vrr-o unknown>", port);

  return 1; // Success
}

void condition_not_satisfied(const char *name, const char *dsc)
{ scm_error_scm (symbol_condition_not_satisfied, scm_makfrom0str (name), scm_makfrom0str (dsc), SCM_EOL, SCM_BOOL_F); }

void misc_exception(const char *name, const char *dsc)
{ scm_error_scm (symbol_misc_error, scm_makfrom0str (name), scm_makfrom0str (dsc), SCM_EOL, SCM_BOOL_F); }

static SCM gl_proxy_get_kind (SCM proxy)
{
  if (is_o_proxy (proxy))
    {
      struct o * o = optr (proxy);
      if (o -> kind == T_OBJ) return (symbol_obj);
      else if (o -> kind == T_GO) return (symbol_go);
      else BUG();
    }
  else if (is_hanger_proxy (proxy)) return (symbol_hanger);
  else if (is_anchor_proxy (proxy)) return (symbol_anchor);
  else SCM_ASSERT (0, proxy, SCM_ARG1, "proxy-get-kind");
}

static SCM gl_proxy_get_type (SCM proxy)
{
  if (is_o_proxy (proxy))
    {
      struct o * o = optr (proxy);
      if (o -> kind == T_OBJ) return (obj_type_symbols [o -> type]);
      else if (o -> kind == T_GO) return (go_type_symbols [o -> type]);
      else BUG ();
    }
  else if (is_hanger_proxy (proxy)) return (SCM_BOOL_F);
  else if (is_anchor_proxy (proxy)) return (SCM_BOOL_F);
  else SCM_ASSERT (0, proxy, SCM_ARG1, "proxy-get-type");
}

static SCM gl_proxy_get_subtype (SCM proxy)
{
  if (is_o_proxy (proxy))
    {
      struct o * o = optr (proxy);
      if (o -> kind == T_OBJ) return (SCM_BOOL_F);
      else if (o -> kind == T_GO) return (out_SSYMBOLS (go_get_info ((struct go *) o) -> name));
      else BUG ();
    }
  else if (is_hanger_proxy (proxy)) return (SCM_BOOL_F);
  else if (is_anchor_proxy (proxy)) return (SCM_BOOL_F);
  else SCM_ASSERT (0, proxy, SCM_ARG1, "proxy-get-subtype");
}

static SCM gl_proxy_p (SCM proxy, SCM kind, SCM type, SCM subt)
{
  SCM_ASSERT (SCM_EQ_P (kind, SCM_UNDEFINED) || SCM_SYMBOLP (kind), kind,
	      SCM_ARG2, "proxy?");
  SCM_ASSERT (SCM_EQ_P (type, SCM_UNDEFINED) || SCM_SYMBOLP (type), type,
	      SCM_ARG3, "proxy?");
  SCM_ASSERT (SCM_EQ_P (subt, SCM_UNDEFINED) || SCM_SYMBOLP (subt), subt,
	      SCM_ARG4, "proxy?");

  return SCM_BOOL ((is_o_proxy (proxy) ||
		    is_hanger_proxy (proxy) ||
		    is_anchor_proxy (proxy)) &&
		   (SCM_EQ_P (kind, SCM_UNDEFINED) ||
		    SCM_EQ_P (kind, gl_proxy_get_kind (proxy))) &&
		   (SCM_EQ_P (type, SCM_UNDEFINED) ||
		    SCM_EQ_P (type, gl_proxy_get_type (proxy))) &&
		   (SCM_EQ_P (subt, SCM_UNDEFINED) ||
		    SCM_EQ_P (subt, gl_proxy_get_subtype (proxy))));
}

static void
string_array_convert (string * src, uns size, SCM ** dst, SCM * rev)
{
  // SCM sdst = scm_c_make_vector (size, SCM_BOOL_F);
    *rev = scm_c_make_hash_table (size * 2);
    *dst = xmalloc (size * sizeof (SCM));

    for (uns i = 0; i < size; i ++)
    {
	SCM symbol = (*dst) [i] = scm_str2symbol (src[i] -> text);
	scm_gc_protect_object (symbol);
	// scm_vector_set_x (sdst, SCM_MAKINUM (i), symbol);
	scm_hashq_set_x (*rev, symbol, SCM_MAKINUM (i));
    }

    scm_gc_protect_object (*rev);
}

/*
UNUSED static void
string_array_line (string * src, uns size, uns ln,
		   SCM ** dst, SCM sdst, SCM rev)
{
    dst [ln] = xmalloc (size * sizeof (SCM));
    SCM line = scm_c_make_vector (size, SCM_BOOL_F);
    scm_vector_set_x (sdst, SCM_MAKINUM (ln), line);

    for (uns i = 0; i < size; i ++)
    {
	SCM symbol = dst [ln][i] = scm_str2symbol (src[i] -> text);
	scm_vector_set_x (line, SCM_MAKINUM (i), symbol);
	scm_hashq_set_x (rev, symbol, scm_cons (SCM_MAKINUM (ln), SCM_MAKINUM (i)));
    }
}
*/


void
glt_kernel_finish (void)
{
  WALK_RST(walk)
    {
      if (is_o_proxy (walk -> proxy)) o_unref (optr (walk -> proxy));
      else if (is_hanger_proxy (walk -> proxy)) hanger_unref (hptr (walk -> proxy));
      else if (is_anchor_proxy (walk -> proxy)) go_unref (anchor_get_go (aptr (walk -> proxy)));
      else BUG ();
    }

  xfree (obj_type_symbols);
  xfree (go_type_symbols);
  //  xfree (go_subtype_symbols);

  rst_finish();
}

struct lst
{
  const char * name;
  uns length;
  struct lst * next;
};

struct lst *hs, *as;

void
glt_kernel_init_hanger (struct hanger_info * info)
{
  DBG("glt_kernel_init_hanger: %s\n", info->name->text);
  struct lst *n = xmalloc (sizeof (struct lst));
  n -> name = xstrdup (info -> name -> text);
  n -> length = info -> name -> length;
  n -> next = hs;
  hs = n;
}

void
glt_kernel_init_anchor (struct anchor_info * info)
{
  DBG("glt_kernel_init_anchor: %s\n", info->name->text);
  struct lst *n = xmalloc (sizeof (struct lst));
  n -> name = xstrdup (info -> name -> text);
  n -> length = info -> name -> length;
  n -> next = as;
  as = n;
}

// this function should be called with (vrr high) as actual module 
SCM
glt_kernel_update_high (UNUSED void *arg)
{
  struct lst * n;

  for (n = hs; n != NULL; n = n -> next)
    {
      char buff [n -> length * 2 + 80];
      sprintf (buff, "(define-public (h-%s go) (go-get-hanger-by-name go '%s))", n -> name, n -> name);
      scm_c_eval_string (buff);
    }

  for (n = hs; n != NULL; n = n -> next)
    {
      char buff [n -> length * 2 + 80];
      sprintf (buff, "(define-public (a-%s go) (go-get-anchor-by-name go '%s))", n -> name, n -> name);
      scm_c_eval_string (buff);
      sprintf (buff, "(define-public (i-%s go) (anchor-get-hanger (go-get-anchor-by-name go '%s)))", n -> name, n -> name);
      scm_c_eval_string (buff);
      sprintf (buff, "(define-public (set-i-%s! go h) (anchor-rehang (go-get-anchor-by-name go '%s) h))", n -> name, n -> name);
      scm_c_eval_string (buff);
    }
  
  return (SCM_UNSPECIFIED);
}

void
glt_kernel_init(void)
{
  rst_init(31);

  // Scheme smob definitions
  o_tag = scm_make_smob_type ("vrr-o", 0);
  scm_set_smob_free (o_tag, o_proxy_free);
  scm_set_smob_print (o_tag, o_proxy_print);

  hanger_tag = scm_make_smob_type ("vrr-hanger", 0);
  scm_set_smob_free (hanger_tag, hanger_proxy_free);

  anchor_tag = scm_make_smob_type ("vrr-anchor", 0);
  scm_set_smob_free (anchor_tag, anchor_proxy_free);

  // string-symbol-number tables
  string_array_convert (string_obj_type, OT_MAX,  &obj_type_symbols, &obj_type_numbers);
  string_array_convert (string_go_type,	 GOT_MAX, &go_type_symbols,  &go_type_numbers);

  //  go_subtype_symbols = xmalloc_zero (GOT_MAX * sizeof (SCM *));
  //  for (int i = 0; i < GOT_MAX; i ++) go_subtype_symbols [i] = go_type_symbols + i;
  //  SCM subtype_symbols_scm = scm_c_make_vector (GOT_MAX, SCM_BOOL_F);
  //  go_subtype_numbers = scm_c_make_hash_table (GOT_MAX * 8); // heuristic

  //  scm_gc_protect_object (subtype_symbols_scm);
  //  scm_gc_protect_object (go_subtype_numbers);

  // misc symbols
  #define DEF(VAR, STR) VAR = scm_str2symbol (STR); scm_gc_protect_object (VAR);
  DEF (symbol_obj,	"obj");
  DEF (symbol_go,	"go");
  DEF (symbol_hanger,	"hanger");
  DEF (symbol_anchor,	"anchor");
  DEF (symbol_transaction_failed,      "transaction-failed");
  DEF (symbol_condition_not_satisfied, "condition-not-satisfied");
  DEF (symbol_misc_error, "misc-error");
  #undef DEF
}



void
gl_kernel_bindings_init (void)
{
  scm_c_define_gsubr ("proxy?", 1, 3, 0, gl_proxy_p);
  scm_c_define_gsubr ("proxy-kind", 1, 0, 0, gl_proxy_get_kind);
  scm_c_define_gsubr ("proxy-type", 1, 0, 0, gl_proxy_get_type);
  scm_c_define_gsubr ("proxy-subtype", 1, 0, 0, gl_proxy_get_subtype);

  scm_c_export ("proxy?", "proxy-kind", "proxy-type", "proxy-subtype", NULL);
}

