(define-module (vrr property))

(use-modules (vrr misc))
(use-modules (vrr low))
(use-modules (vrr support))

(export get-property set-property! prop-decode)

;; actions: fills tables prop-table, prop-table2, prop-definition-table


;;; Property manipulation support
;;;
;;; Copyright (c) 2004-2005    The VRR Team <vrr@ucw.cz>
;;; This program is distributed under GNU General Public License, see COPYING.


(define prop-table (make-hash-table 31))

(define prop-table2 (make-vector CC_PT_MAX #f))
(vector-set! prop-table2 CC_PT_UNS    (make-vector CC_PTU_MAX #f))
(vector-set! prop-table2 CC_PT_POINTER (make-vector CC_PTP_MAX #f))
(vector-set! prop-table2 CC_PT_STRING (make-vector CC_PTS_MAX #f))
(vector-set! prop-table2 CC_PT_REAL   (make-vector CC_PTR_MAX #f))

(define (register-prop-subtype symbol type-number subtype-number decoder)
  (let ((obj (vector symbol type-number subtype-number decoder)))
    (hashq-set! prop-table symbol obj)
    (vector-set! (vector-ref prop-table2 type-number) subtype-number obj)))

(define (property-type-symbol t s)
  (pt-symbol (vector-ref (vector-ref prop-table2 t) s)))

(define (pt-symbol x) (vector-ref x 0))
(define (pt-type-number x) (vector-ref x 1))
(define (pt-subtype-number x) (vector-ref x 2))
(define (pt-decode x y) ((vector-ref x 3) y))


(define (prop-decode tn stn value)
  (pt-decode (vector-ref (vector-ref prop-table2 tn) stn) value))

(define (real-between? a b c)
  (and (real? a) (real? b) (real? c) (<= a b) (<= b c)))

(define (natural-between? a b c)
  (and (natural? a) (natural? b) (natural? c) (<= a b) (<= b c)))


(define prop-definition-table (make-hash-table 31))

(define (register-prop-definition name fn)
  (hashq-set! prop-definition-table name fn))

(define (property-specific-set obj prop name data)
  (let ((fn (hashq-ref prop-definition-table name)))
    (if fn
	(fn obj prop data)
	(error "Bad property value definition"))))


(define (rgb-list? a)
  (and (list? a) (= (length a) 3) (every (lambda (x) (natural-between? 0 x 255)) a)))

(define (rgba-list? a)
  (and (list? a) (= (length a) 4) (every (lambda (x) (natural-between? 0 x 255)) a)))

(define unpack-r uns2r-uns)
(define unpack-g uns2g-uns)
(define unpack-b uns2b-uns)
(define unpack-a uns2a-uns)

(define pack-rgba rgba-uns2uns)
(define (pack-rgb r g b)
  (pack-rgba r g b 255))


(define (natural-list? data)
  (and (pair? data) (natural? (car data)) (null? (cdr data))))

(define (property-natural-set obj prop data)
  (or (natural-list? data)
      (error "Bad property value definition"))
  (prop-uns-set-forte obj prop CC_PT_UNS CC_PTU_UNSPECIFIED #t (car data)))

(define (property-rgb-set obj prop data)
  (or (rgb-list? data)
      (error "Bad property value definition"))
  (prop-uns-set-forte obj prop CC_PT_UNS CC_PTU_COLOR #t (apply pack-rgb data)))

(define (property-rgba-set obj prop data)
  (or (rgba-list? data)
      (error "Bad property value definition"))
  (prop-uns-set-forte obj prop CC_PT_UNS CC_PTU_COLOR #t (apply pack-rgba data)))

(define (font-list? data)
  (and (pair? data) (string? (car data)) (string? (cadr data)) (null? (cddr data))))

(define (property-font-set obj prop data)
  (or (font-list? data)
      (error "Bad property value definition"))
  (prop-uns-set-forte obj prop CC_PT_UNS CC_PTU_FONT #t (font-load-by-family-style (car data) (cadr data))))

(define (property-boolean-set obj prop data)
  (or (boolean? data)
      (error "Bad property value definition"))
  (prop-uns-set-forte obj prop CC_PT_UNS CC_PTU_BOOLEAN #t (if data 1 0)))

(define (make-prop-set-1 ct cst code)
  (lambda (obj prop data)
    (or (and (pair? data) (code (car data)) (null? (cdr data)))
	(error "Bad property value definition"))
    (prop-uns-set-forte obj prop ct cst #t (code (car data)))))



(register-prop-definition 'natural property-natural-set)
(register-prop-definition 'rgb property-rgb-set)
(register-prop-definition 'rgba property-rgba-set)
(register-prop-definition 'font property-font-set)
(register-prop-definition 'transform property-transform-set)

(register-prop-definition 'conic-type-0 (make-prop-set-1 CC_PT_UNS CC_PTU_CONIC_TYPE_0P code-conic))
(register-prop-definition 'conic-type-1 (make-prop-set-1 CC_PT_UNS CC_PTU_CONIC_TYPE_1P code-conic))
(register-prop-definition 'conic-type-2 (make-prop-set-1 CC_PT_UNS CC_PTU_CONIC_TYPE_2P code-conic))
(register-prop-definition 'conic-type-3 (make-prop-set-1 CC_PT_UNS CC_PTU_CONIC_TYPE_3P code-conic))
(register-prop-definition 'cap-style (make-prop-set-1 CC_PT_UNS CC_PTU_CAP_STYLE code-cap-style))
(register-prop-definition 'join-style (make-prop-set-1 CC_PT_UNS CC_PTU_JOIN_STYLE code-join-style))
(register-prop-definition 'stroke-style (make-prop-set-1 CC_PT_UNS CC_PTU_STROKE_STYLE code-stroke-style))
(register-prop-definition 'alignment-x (make-prop-set-1 CC_PT_UNS CC_PTU_ALIGNMENT_X code-alignment-x))
(register-prop-definition 'alignment-y (make-prop-set-1 CC_PT_UNS CC_PTU_ALIGNMENT_Y code-alignment-y))
(register-prop-definition 'arrow-front (make-prop-set-1 CC_PT_UNS CC_PTU_ARROW_FRONT code-arrow-front))
(register-prop-definition 'arrow-back (make-prop-set-1 CC_PT_UNS CC_PTU_ARROW_BACK code-arrow-back))
(register-prop-definition 'arrow-alignment (make-prop-set-1 CC_PT_UNS CC_PTU_ARROW_ALIGN code-arrow-alignment))


(define (decode-prop-natural n)
  (list 'natural n))

(define (decode-prop-color n)
  (if (= (unpack-a n) 255)
      (list 'rgb (unpack-r n) (unpack-g n) (unpack-b n))
      (list 'rgba (unpack-r n) (unpack-g n) (unpack-b n) (unpack-a n))))

(define (decode-prop-font n)
  (list 'font (font-family n) (font-style n)))

(define (decode-prop-boolean n)
  (not (zero? n)))

(define (decode-prop-cap-style n)
  (list 'cap-style (decode-cap-style n)))

(define (decode-prop-join-style n)
  (list 'join-style (decode-join-style n)))

(define (decode-prop-stroke-style n)
  (list 'stroke-style (decode-stroke-style n)))

(define (decode-prop-alignment-x n)
  (list 'alignment-x (decode-alignment-x n)))

(define (decode-prop-alignment-y n)
  (list 'alignment-y (decode-alignment-y n)))

(define (decode-prop-arrow-front n)
  (list 'arrow-front (decode-arrow-front n)))

(define (decode-prop-arrow-back n)
  (list 'arrow-back (decode-arrow-back n)))

(define (decode-prop-arrow-alignment n)
  (list 'arrow-alignment (decode-arrow-alignment n)))

(define (decode-prop-conic-0 n)
  (list 'conic-type-0 (decode-conic n)))

(define (decode-prop-conic-1 n)
  (list 'conic-type-1 (decode-conic n)))

(define (decode-prop-conic-2 n)
  (list 'conic-type-2 (decode-conic n)))

(define (decode-prop-conic-3 n)
  (list 'conic-type-3 (decode-conic n)))

(register-prop-subtype 'natural  CC_PT_UNS CC_PTU_UNSPECIFIED decode-prop-natural)
(register-prop-subtype 'color    CC_PT_UNS CC_PTU_COLOR decode-prop-color)
(register-prop-subtype 'font     CC_PT_UNS CC_PTU_FONT decode-prop-font)
(register-prop-subtype 'conic-type-0 CC_PT_UNS CC_PTU_CONIC_TYPE_0P decode-prop-conic-0)
(register-prop-subtype 'conic-type-1 CC_PT_UNS CC_PTU_CONIC_TYPE_1P decode-prop-conic-1)
(register-prop-subtype 'conic-type-2 CC_PT_UNS CC_PTU_CONIC_TYPE_2P decode-prop-conic-2)
(register-prop-subtype 'conic-type-3 CC_PT_UNS CC_PTU_CONIC_TYPE_3P decode-prop-conic-3)
(register-prop-subtype 'boolean  CC_PT_UNS CC_PTU_BOOLEAN decode-prop-boolean)
(register-prop-subtype 'cap-style CC_PT_UNS CC_PTU_CAP_STYLE decode-prop-cap-style)
(register-prop-subtype 'join-style CC_PT_UNS CC_PTU_JOIN_STYLE decode-prop-join-style)
(register-prop-subtype 'stroke-style CC_PT_UNS CC_PTU_STROKE_STYLE decode-prop-stroke-style)
(register-prop-subtype 'alignment-x CC_PT_UNS CC_PTU_ALIGNMENT_X decode-prop-alignment-x)
(register-prop-subtype 'alignment-y CC_PT_UNS CC_PTU_ALIGNMENT_Y decode-prop-alignment-y)
(register-prop-subtype 'arrow-front CC_PT_UNS CC_PTU_ARROW_FRONT decode-prop-arrow-front)
(register-prop-subtype 'arrow-back CC_PT_UNS CC_PTU_ARROW_BACK decode-prop-arrow-back)
(register-prop-subtype 'arrow-alignment CC_PT_UNS CC_PTU_ARROW_ALIGN decode-prop-arrow-alignment)

(register-prop-subtype 'string   CC_PT_STRING CC_PTS_UNSPECIFIED id)
(register-prop-subtype 'filename CC_PT_STRING CC_PTS_FILE_NAME id)
(register-prop-subtype 'text     CC_PT_STRING CC_PTS_LARGE_TEXT id)

(register-prop-subtype 'transform CC_PT_POINTER CC_PTP_TRANSFORM decode-prop-transform)

(register-prop-subtype 'real CC_PT_REAL CC_PTR_UNSPECIFIED id)
(register-prop-subtype 'coordinate CC_PT_REAL CC_PTR_COORDINATE id)
(register-prop-subtype 'positivive CC_PT_REAL CC_PTR_POSITIVE id)
(register-prop-subtype 'length CC_PT_REAL CC_PTR_LENGTH id)
(register-prop-subtype 'angle-1pi CC_PT_REAL CC_PTR_ANGLE_PI id)
(register-prop-subtype 'angle-2pi CC_PT_REAL CC_PTR_ANGLE_2PI id)
(register-prop-subtype 'angle-4pi CC_PT_REAL CC_PTR_ANGLE_4PI id)
(register-prop-subtype 'reference CC_PT_REAL CC_PTR_REFERENCE id)




(define (get-property obj name . missing)
  (let ((st (property-get-int obj name)))
    (if st
	(prop-decode (cadr st) (caddr st) (car st))
	(if (null? missing) #f (car missing)))))

(define (set-property! obj name value)
  
  (cond
   ((real? value)
    (property-real-set obj name value))
   ((string? value)
    (property-string-set obj name value))
   ((boolean? value)
    (property-boolean-set obj name value))
   ((pair? value)
    (property-specific-set obj name (car value) (cdr value)))
   (else
    (typearg-error "set-property!" 3 value))))


(define-public (set-conic! go conic . rest)
  (or (proxy? go 'go 'elliptic-arc) (typearg-error "set-conic!" 1 go))
  (or (code-conic conic) (typearg-error "set-conic!" 2 conic))
  (property-uns-set go 'conic (code-conic conic))
  (case conic
    ((start-entire)
     (or (real? (car rest)) (typearg-error "set-conic!" 3 (car rest)))
     (or (null? (cdr rest)) (typearg-error "set-conic!" 4 (cadr rest)))
     (property-real-set go 'start (car rest)))
    ((start-dif)
     (or (real? (car rest)) (typearg-error "set-conic!" 3 (car rest)))
     (or (real? (cadr rest)) (typearg-error "set-conic!" 4 (cadr rest)))
     (or (null? (cddr rest)) (typearg-error "set-conic!" 5 (caddr rest)))
     (property-real-set go 'start (car rest))
     (property-real-set go 'dif (cadr rest)))
    ((point-dif)
     (or (real? (car rest)) (typearg-error "set-conic!" 3 (car rest)))
     (or (null? (cdr rest)) (typearg-error "set-conic!" 4 (cadr rest)))
     (property-real-set go 'dif (car rest)))
    (else
     (or (null? rest) (typearg-error "set-conic!" 3 (car rest))))))

(define-public (get-conic go)
  (or (proxy? go 'go 'elliptic-arc) (typearg-error "get-conic" 1 go))
  (let ((conic (cadr (get-property go 'conic))))
    (case conic
      ((start-entire)
       (list conic (get-property go 'start)))
      ((start-dif)
       (list conic (get-property go 'start) (get-property go 'dif)))
      ((point-dif)
       (list conic (get-property go 'dif)))
      (else
       (list conic)))))

(define (skip-head f)
  (if (pair? f)
      (cdr f)
      f))

(define-public (go-get-sources go)
  (let ((v (hashq-ref src-table (proxy-subtype go))))
    (and v (flatten (map (lambda (x) ; in v should be only numbers, getters or property names
			   (cond
			    ((number? x)
			     (if (negative? x)
				 (go-get-anchor go (- -1 x))
				 (go-get-anchor go x)))
			    ((symbol? x) (skip-head (get-property go x)))
			    ((procedure? x) (x go))
			    (else (error "BUG in go-get-sources"))))
			 (if (number? v) (iota v) v))))))

(define-public (go-get-source-dsc subtype)
  (let ((v (hashq-ref src-table subtype)))
    (and v (map (lambda (x) ; in v should be only numbers, getters or property names
		  (if (number? x)
		      (if (negative? x) 'hanger-c 'hanger)
		      'data))
		(if (number? v) (iota v) v)))))


